/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040401
#define SWIGRUBY

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#if __GNUC__ >= 7
#pragma GCC diagnostic push
#if defined(__cplusplus)
#pragma GCC diagnostic ignored "-Wregister"
#if __GNUC__ >= 10
#pragma GCC diagnostic ignored "-Wvolatile"
#if __GNUC__ >= 11
#pragma GCC diagnostic ignored "-Wdeprecated-enum-enum-conversion"
#endif
#endif
#endif
#endif

#include <ruby.h>
#include <ruby/version.h> /* For RUBY_API_VERSION_CODE */

#if __GNUC__ >= 7
#pragma GCC diagnostic pop
#endif

/* Ruby 1.9.1 has a "memoisation optimisation" when compiling with GCC which
 * breaks using rb_intern as an lvalue, as SWIG does.  We work around this
 * issue for now by disabling this.
 * https://sourceforge.net/tracker/?func=detail&aid=2859614&group_id=1645&atid=101645
 */
#ifdef rb_intern
# undef rb_intern
#endif

/* Remove global macros defined in Ruby's win32.h */
#ifdef write
# undef write
#endif
#ifdef read
# undef read
#endif
#ifdef bind
# undef bind
#endif
#ifdef close
# undef close
#endif
#ifdef connect
# undef connect
#endif


/* Ruby 1.7 defines NUM2LL(), LL2NUM() and ULL2NUM() macros */
#ifndef NUM2LL
#define NUM2LL(x) NUM2LONG((x))
#endif
#ifndef LL2NUM
#define LL2NUM(x) INT2NUM((long) (x))
#endif
#ifndef ULL2NUM
#define ULL2NUM(x) UINT2NUM((unsigned long) (x))
#endif

/* Ruby 1.7 doesn't (yet) define NUM2ULL() */
#ifndef NUM2ULL
#ifdef HAVE_LONG_LONG
#define NUM2ULL(x) rb_num2ull((x))
#else
#define NUM2ULL(x) NUM2ULONG(x)
#endif
#endif

/* RSTRING_LEN, etc are new in Ruby 1.9, but ->ptr and ->len no longer work */
/* Define these for older versions so we can just write code the new way */
#ifndef RSTRING_LEN
# define RSTRING_LEN(x) RSTRING(x)->len
#endif
#ifndef RSTRING_PTR
# define RSTRING_PTR(x) RSTRING(x)->ptr
#endif
#ifndef RSTRING_END
# define RSTRING_END(x) (RSTRING_PTR(x) + RSTRING_LEN(x))
#endif
#ifndef RARRAY_LEN
# define RARRAY_LEN(x) RARRAY(x)->len
#endif
#ifndef RARRAY_PTR
# define RARRAY_PTR(x) RARRAY(x)->ptr
#endif
#ifndef RFLOAT_VALUE
# define RFLOAT_VALUE(x) RFLOAT(x)->value
#endif
#ifndef DOUBLE2NUM
# define DOUBLE2NUM(x) rb_float_new(x)
#endif
#ifndef RHASH_TBL
# define RHASH_TBL(x) (RHASH(x)->tbl)
#endif
#ifndef RHASH_ITER_LEV
# define RHASH_ITER_LEV(x) (RHASH(x)->iter_lev)
#endif
#ifndef RHASH_IFNONE
# define RHASH_IFNONE(x) (RHASH(x)->ifnone)
#endif
#ifndef RHASH_SIZE
# define RHASH_SIZE(x) (RHASH(x)->tbl->num_entries)
#endif
#ifndef RHASH_EMPTY_P
# define RHASH_EMPTY_P(x) (RHASH_SIZE(x) == 0)
#endif
#ifndef RSTRUCT_LEN
# define RSTRUCT_LEN(x) RSTRUCT(x)->len
#endif
#ifndef RSTRUCT_PTR
# define RSTRUCT_PTR(x) RSTRUCT(x)->ptr
#endif
#ifndef RTYPEDDATA_P
# define RTYPEDDATA_P(x) (TYPE(x) != T_DATA)
#endif



/*
 * The following macros are used for providing the correct type of a
 * function pointer to the Ruby C API.
 *
 * Starting with Ruby 2.7 these macros act transparently due to Ruby's moving
 * moving away from ANYARGS and instead employing strict function signatures.
 *
 * Note: In case of C (not C++) the macros are transparent even before
 * Ruby 2.7 due to the fact that the Ruby C API used function declarators
 * with empty parentheses, which allows for an unspecified number of
 * arguments.
 *
 * PROTECTFUNC(f) is used for the function pointer argument of the Ruby
 * C API function rb_protect().
 *
 * VALUEFUNC(f) is used for the function pointer argument(s) of Ruby C API
 * functions like rb_define_method() and rb_define_singleton_method().
 *
 * VOIDFUNC(f) is used to typecast a C function that implements either
 * the "mark" or "free" stuff for a Ruby Data object, so that it can be
 * passed as an argument to Ruby C API functions like Data_Wrap_Struct()
 * and Data_Make_Struct().
 *
 * SWIG_RUBY_VOID_ANYARGS_FUNC(f) is used for the function pointer
 * argument(s) of Ruby C API functions like rb_define_virtual_variable().
 *
 * SWIG_RUBY_INT_ANYARGS_FUNC(f) is used for the function pointer
 * argument(s) of Ruby C API functions like st_foreach().
 */
#if defined(__cplusplus) && RUBY_API_VERSION_CODE < 20700
#  define PROTECTFUNC(f) ((VALUE (*)(VALUE)) f)
#  define VALUEFUNC(f) ((VALUE (*)(ANYARGS)) f)
#  define VOIDFUNC(f) ((RUBY_DATA_FUNC) f)
#  define SWIG_RUBY_VOID_ANYARGS_FUNC(f) ((void (*)(ANYARGS))(f))
#  define SWIG_RUBY_INT_ANYARGS_FUNC(f) ((int (*)(ANYARGS))(f))
#else
#  define PROTECTFUNC(f) (f)
#  define VALUEFUNC(f) (f)
#  define VOIDFUNC(f) (f)
#  define SWIG_RUBY_VOID_ANYARGS_FUNC(f) (f)
#  define SWIG_RUBY_INT_ANYARGS_FUNC(f) (f)
#endif

/* Don't use for expressions have side effect */
#ifndef RB_STRING_VALUE
#define RB_STRING_VALUE(s) (TYPE(s) == T_STRING ? (s) : (*(volatile VALUE *)&(s) = rb_str_to_str(s)))
#endif
#ifndef StringValue
#define StringValue(s) RB_STRING_VALUE(s)
#endif
#ifndef StringValuePtr
#define StringValuePtr(s) RSTRING_PTR(RB_STRING_VALUE(s))
#endif
#ifndef StringValueLen
#define StringValueLen(s) RSTRING_LEN(RB_STRING_VALUE(s))
#endif
#ifndef SafeStringValue
#define SafeStringValue(v) do {\
    StringValue(v);\
    rb_check_safe_str(v);\
} while (0)
#endif

#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
#define rb_define_alloc_func(klass, func) rb_define_singleton_method((klass), "new", VALUEFUNC((func)), -1)
#define rb_undef_alloc_func(klass) rb_undef_method(CLASS_OF((klass)), "new")
#endif

static VALUE _mSWIG = Qnil;

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */


/* Define some additional error types */
#define SWIG_ObjectPreviouslyDeletedError  -100


/* Define custom exceptions for errors that do not map to existing Ruby
   exceptions. */

SWIGINTERN VALUE 
getNullReferenceError(void) {
  static VALUE rb_eNullReferenceError = Qnil;
  if (rb_eNullReferenceError == Qnil) {
    rb_eNullReferenceError = rb_define_class("NullReferenceError", rb_eRuntimeError);
  }
  return rb_eNullReferenceError;
} 

SWIGINTERN VALUE 
getObjectPreviouslyDeletedError(void) {
  static VALUE rb_eObjectPreviouslyDeleted = Qnil;
  if (rb_eObjectPreviouslyDeleted == Qnil) {
    rb_eObjectPreviouslyDeleted = rb_define_class("ObjectPreviouslyDeleted", rb_eRuntimeError);
  }
  return rb_eObjectPreviouslyDeleted;
} 


SWIGINTERN VALUE
SWIG_Ruby_ErrorType(int SWIG_code) {
  VALUE type;
  switch (SWIG_code) {
  case SWIG_MemoryError:
    type = rb_eNoMemError;
    break;
  case SWIG_IOError:
    type = rb_eIOError;
    break;
  case SWIG_RuntimeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_IndexError:
    type = rb_eIndexError;
    break;
  case SWIG_TypeError:
    type = rb_eTypeError;
    break;
  case SWIG_DivisionByZero:
    type = rb_eZeroDivError;
    break;
  case SWIG_OverflowError:
    type = rb_eRangeError;
    break;
  case SWIG_SyntaxError:
    type = rb_eSyntaxError;
    break;
  case SWIG_ValueError:
    type = rb_eArgError;
    break;
  case SWIG_SystemError:
    type = rb_eFatal;
    break;
  case SWIG_AttributeError:
    type = rb_eRuntimeError;
    break;
  case SWIG_NullReferenceError:
    type = getNullReferenceError();
    break;
  case SWIG_ObjectPreviouslyDeletedError:
    type = getObjectPreviouslyDeletedError();
    break;
  case SWIG_UnknownError:
    type = rb_eRuntimeError;
    break;
  default:
    type = rb_eRuntimeError;
  }
  return type;
}


/* This function is called when a user inputs a wrong argument to
   a method.
 */
SWIGINTERN 
const char* Ruby_Format_TypeError( const char* msg,
				   const char* type, 
				   const char* name, 
				   const int argn,
				   VALUE input )
{
  char buf[128];
  VALUE str;
  VALUE asStr;
  if ( msg && *msg )
    {
      str = rb_str_new2(msg);
    }
  else
    {
      str = rb_str_new(NULL, 0);
    }

  str = rb_str_cat2( str, "Expected argument " );
  SWIG_snprintf( buf, sizeof( buf), "%d of type ", argn-1 );
  str = rb_str_cat2( str, buf );
  str = rb_str_cat2( str, type );
  str = rb_str_cat2( str, ", but got " );
  str = rb_str_cat2( str, rb_obj_classname(input) );
  str = rb_str_cat2( str, " " );
  asStr = rb_inspect(input);
  if ( RSTRING_LEN(asStr) > 30 )
    {
      str = rb_str_cat( str, StringValuePtr(asStr), 30 );
      str = rb_str_cat2( str, "..." );
    }
  else
    {
      str = rb_str_append( str, asStr );
    }

  if ( name )
    {
      str = rb_str_cat2( str, "\n\tin SWIG method '" );
      str = rb_str_cat2( str, name );
      str = rb_str_cat2( str, "'" );
    }

  return StringValuePtr( str );
}

/* This function is called when an overloaded method fails */
SWIGINTERN 
void Ruby_Format_OverloadedError(
				 const int argc,
				 const int maxargs,
				 const char* method, 
				 const char* prototypes 
				 )
{
  const char* msg = "Wrong # of arguments";
  if ( argc <= maxargs ) msg = "Wrong arguments";
  rb_raise(rb_eArgError,"%s for overloaded method '%s'.\n"  
	   "Possible C/C++ prototypes are:\n%s",
	   msg, method, prototypes);
}

/* -----------------------------------------------------------------------------
 * rubytracking.swg
 *
 * This file contains support for tracking mappings from
 * Ruby objects to C++ objects.  This functionality is needed
 * to implement mark functions for Ruby's mark and sweep
 * garbage collector.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(ST_DATA_T_DEFINED)
/* Needs to be explicitly included for Ruby 1.8 and earlier */
#include <st.h>
#endif

/* Ruby 1.8 actually assumes the first case. */
#if SIZEOF_VOIDP == SIZEOF_LONG
#  define SWIG2NUM(v) LONG2NUM((unsigned long)v)
#  define NUM2SWIG(x) (unsigned long)NUM2LONG(x)
#elif SIZEOF_VOIDP == SIZEOF_LONG_LONG
#  define SWIG2NUM(v) LL2NUM((unsigned long long)v)
#  define NUM2SWIG(x) (unsigned long long)NUM2LL(x)
#else
#  error sizeof(void*) is not the same as long or long long
#endif

/* Global hash table to store Trackings from C/C++
   structs to Ruby Objects.
*/
static st_table* swig_ruby_trackings = NULL;

static VALUE swig_ruby_trackings_count(ID id, VALUE *var) {
  return SWIG2NUM(swig_ruby_trackings->num_entries);
}


/* Setup a hash table to store Trackings */
SWIGRUNTIME void SWIG_RubyInitializeTrackings(void) {
  /* Create a hash table to store Trackings from C++
     objects to Ruby objects. */

  /* Try to see if some other .so has already created a
     tracking hash table, which we keep hidden in an instance var
     in the SWIG module.
     This is done to allow multiple DSOs to share the same
     tracking table.
  */
  VALUE trackings_value = Qnil;
  /* change the variable name so that we can mix modules
     compiled with older SWIG's - this used to be called "@__trackings__" */
  ID trackings_id = rb_intern( "@__safetrackings__" );
  VALUE verbose = rb_gv_get("VERBOSE");
  rb_gv_set("VERBOSE", Qfalse);
  trackings_value = rb_ivar_get( _mSWIG, trackings_id );
  rb_gv_set("VERBOSE", verbose);

  /* The trick here is that we have to store the hash table
  pointer in a Ruby variable. We do not want Ruby's GC to
  treat this pointer as a Ruby object, so we convert it to
  a Ruby numeric value. */
  if (trackings_value == Qnil) {
    /* No, it hasn't.  Create one ourselves */
    swig_ruby_trackings = st_init_numtable();
    rb_ivar_set( _mSWIG, trackings_id, SWIG2NUM(swig_ruby_trackings) );
  } else {
    swig_ruby_trackings = (st_table*)NUM2SWIG(trackings_value);
  }

  rb_define_virtual_variable("SWIG_TRACKINGS_COUNT",
                             VALUEFUNC(swig_ruby_trackings_count),
                             SWIG_RUBY_VOID_ANYARGS_FUNC((rb_gvar_setter_t*)NULL));
}

/* Add a Tracking from a C/C++ struct to a Ruby object */
SWIGRUNTIME void SWIG_RubyAddTracking(void* ptr, VALUE object) {
  /* Store the mapping to the global hash table. */
  st_insert(swig_ruby_trackings, (st_data_t)ptr, object);
}

/* Get the Ruby object that owns the specified C/C++ struct */
SWIGRUNTIME VALUE SWIG_RubyInstanceFor(void* ptr) {
  /* Now lookup the value stored in the global hash table */
  VALUE value;

  if (st_lookup(swig_ruby_trackings, (st_data_t)ptr, &value)) {
    return value;
  } else {
    return Qnil;
  }
}

/* Remove a Tracking from a C/C++ struct to a Ruby object.  It
   is very important to remove objects once they are destroyed
   since the same memory address may be reused later to create
   a new object. */
SWIGRUNTIME void SWIG_RubyRemoveTracking(void* ptr) {
  /* Delete the object from the hash table */
  st_delete(swig_ruby_trackings, (st_data_t *)&ptr, NULL);
}

/* This is a helper method that unlinks a Ruby object from its
   underlying C++ object.  This is needed if the lifetime of the
   Ruby object is longer than the C++ object. */
SWIGRUNTIME void SWIG_RubyUnlinkObjects(void* ptr) {
  VALUE object = SWIG_RubyInstanceFor(ptr);

  if (object != Qnil) {
    // object might have the T_ZOMBIE type, but that's just
    // because the GC has flagged it as such for a deferred
    // destruction. Until then, it's still a T_DATA object.
    DATA_PTR(object) = 0;
  }
}

/* This is a helper method that iterates over all the trackings
   passing the C++ object pointer and its related Ruby object
   to the passed callback function. */

/* Proxy method to abstract the internal trackings datatype */
static int swig_ruby_internal_iterate_callback(st_data_t ptr, st_data_t obj, st_data_t meth) {
  ((void (*) (void *, VALUE))meth)((void *)ptr, (VALUE)obj);
  return ST_CONTINUE;
}

SWIGRUNTIME void SWIG_RubyIterateTrackings( void(*meth)(void* ptr, VALUE obj) ) {
  st_foreach(swig_ruby_trackings,
             SWIG_RUBY_INT_ANYARGS_FUNC(swig_ruby_internal_iterate_callback),
             (st_data_t)meth);
}

#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * Ruby API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

SWIGINTERN VALUE
SWIG_Ruby_AppendOutput(VALUE target, VALUE o, int is_void) {
  if (NIL_P(target) && is_void) {
    target = o;
  } else {
    if (TYPE(target) != T_ARRAY) {
      VALUE o2 = target;
      target = rb_ary_new();
      rb_ary_push(target, o2);
    }
    rb_ary_push(target, o);
  }
  return target;
}

/* For ruby1.8.4 and earlier. */
#ifndef RUBY_INIT_STACK
   RUBY_EXTERN void Init_stack(VALUE* addr);
#  define RUBY_INIT_STACK \
   VALUE variable_in_this_stack_frame; \
   Init_stack(&variable_in_this_stack_frame);
#endif


#ifdef __cplusplus
}
#endif


/* -----------------------------------------------------------------------------
 * rubyrun.swg
 *
 * This file contains the runtime support for Ruby modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Ruby_ConvertPtrAndOwn(obj, pptr, type, flags, own)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Ruby_NewPointerObj(ptr, type, flags)
#define SWIG_AcquirePtr(ptr, own)                       SWIG_Ruby_AcquirePtr(ptr, own)
#define swig_owntype                                    swig_ruby_owntype

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty, flags)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Ruby_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Ruby_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Ruby_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer) 		SWIG_Ruby_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Ruby_ErrorType(code)               
#define SWIG_Error(code, msg)            		rb_raise(SWIG_Ruby_ErrorType(code), "%s", msg)
#define SWIG_fail                        		goto fail				 


/* Ruby-specific SWIG API */

#define SWIG_InitRuntime()                              SWIG_Ruby_InitRuntime()              
#define SWIG_define_class(ty)                        	SWIG_Ruby_define_class(ty)
#define SWIG_NewClassInstance(value, ty)             	SWIG_Ruby_NewClassInstance(value, ty)
#define SWIG_MangleStr(value)                        	SWIG_Ruby_MangleStr(value)		  
#define SWIG_CheckConvert(value, ty)                 	SWIG_Ruby_CheckConvert(value, ty)	  

#include "assert.h"

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  VALUE klass;
  VALUE mImpl;
  void  (*mark)(void *);
  void  (*destroy)(void *);
  int trackObjects;
} swig_class;


/* Global pointer used to keep some internal SWIG stuff */
static VALUE _cSWIG_Pointer = Qnil;
static VALUE swig_runtime_data_type_pointer = Qnil;

/* Global IDs used to keep some internal SWIG stuff */
static ID swig_arity_id = 0;
static ID swig_call_id  = 0;
static ID swig_lowerthan_id = 0;

/*
  If your swig extension is to be run within an embedded ruby and has
  director callbacks, you should set -DRUBY_EMBEDDED during compilation.  
  This will reset ruby's stack frame on each entry point from the main 
  program the first time a virtual director function is invoked (in a 
  non-recursive way).
  If this is not done, you run the risk of Ruby trashing the stack.
*/

#ifdef RUBY_EMBEDDED

#  define SWIG_INIT_STACK                            \
      if ( !swig_virtual_calls ) { RUBY_INIT_STACK } \
      ++swig_virtual_calls;
#  define SWIG_RELEASE_STACK --swig_virtual_calls;
#  define Ruby_DirectorTypeMismatchException(x) \
          rb_raise( rb_eTypeError, "%s", x ); return c_result;

      static unsigned int swig_virtual_calls = 0;

#else  /* normal non-embedded extension */

#  define SWIG_INIT_STACK
#  define SWIG_RELEASE_STACK
#  define Ruby_DirectorTypeMismatchException(x) \
          throw Swig::DirectorTypeMismatchException( x );

#endif  /* RUBY_EMBEDDED */


SWIGRUNTIME VALUE 
getExceptionClass(void) {
  static VALUE rubyExceptionClass = Qnil;
  if (rubyExceptionClass == Qnil) {
    rubyExceptionClass = rb_const_get(_mSWIG, rb_intern("Exception"));
  }
  return rubyExceptionClass;
} 

/* This code checks to see if the Ruby object being raised as part
   of an exception inherits from the Ruby class Exception.  If so,
   the object is simply returned.  If not, then a new Ruby exception
   object is created and that will be returned to Ruby.*/
SWIGRUNTIME VALUE
SWIG_Ruby_ExceptionType(swig_type_info *desc, VALUE obj) {
  VALUE exceptionClass = getExceptionClass();
  if (rb_obj_is_kind_of(obj, exceptionClass)) {
    return obj;
  } else {
    return rb_exc_new3(rb_eRuntimeError, rb_obj_as_string(obj));
  }
}

/* Initialize Ruby runtime support */
SWIGRUNTIME void
SWIG_Ruby_InitRuntime(void)
{
  if (_mSWIG == Qnil) {
    _mSWIG = rb_define_module("SWIG");
    swig_call_id  = rb_intern("call");
    swig_arity_id = rb_intern("arity");
    swig_lowerthan_id = rb_intern("<");
  }
}

/* Define Ruby class for C type */
SWIGRUNTIME void
SWIG_Ruby_define_class(swig_type_info *type)
{
  size_t klass_len = 4 + strlen(type->name) + 1;
  char *klass_name = (char *) malloc(klass_len);
  SWIG_snprintf(klass_name, klass_len, "TYPE%s", type->name);
  if (NIL_P(_cSWIG_Pointer)) {
    _cSWIG_Pointer = rb_define_class_under(_mSWIG, "Pointer", rb_cObject);
    rb_undef_method(CLASS_OF(_cSWIG_Pointer), "new");
  }
  rb_undef_alloc_func(rb_define_class_under(_mSWIG, klass_name, _cSWIG_Pointer));
  free((void *) klass_name);
}

/* Create a new pointer object */
SWIGRUNTIME VALUE
SWIG_Ruby_NewPointerObj(void *ptr, swig_type_info *type, int flags)
{
  int own =  flags & SWIG_POINTER_OWN; 
  int track;
  char *klass_name;
  swig_class *sklass;
  VALUE klass;
  VALUE obj;

  if (!ptr)
    return Qnil;

  assert(type);
  if (type->clientdata) {
    sklass = (swig_class *) type->clientdata;
		
    /* Are we tracking this class and have we already returned this Ruby object? */
    track = sklass->trackObjects;
    if (track) {
      obj = SWIG_RubyInstanceFor(ptr);

      /* Check the object's type and make sure it has the correct type.
        It might not in cases where methods do things like 
        downcast methods. */
      if (obj != Qnil) {
        VALUE value = rb_iv_get(obj, "@__swigtype__");
        const char* type_name = RSTRING_PTR(value);
				
        if (strcmp(type->name, type_name) == 0) {
          return obj;
        }
      }
    }

    /* Create a new Ruby object */
    obj = Data_Wrap_Struct(sklass->klass, VOIDFUNC(sklass->mark), 
			   ( own ? VOIDFUNC(sklass->destroy) : 
			     (track ? VOIDFUNC(SWIG_RubyRemoveTracking) : 0 )
			     ), ptr);

    /* If tracking is on for this class then track this object. */
    if (track) {
      SWIG_RubyAddTracking(ptr, obj);
    }
  } else {
    size_t klass_len = 4 + strlen(type->name) + 1;
    klass_name = (char *) malloc(klass_len);
    SWIG_snprintf(klass_name, klass_len, "TYPE%s", type->name);
    klass = rb_const_get(_mSWIG, rb_intern(klass_name));
    free((void *) klass_name);
    obj = Data_Wrap_Struct(klass, 0, 0, ptr);
  }
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));

  return obj;
}

/* Create a new class instance (always owned) */
SWIGRUNTIME VALUE
SWIG_Ruby_NewClassInstance(VALUE klass, swig_type_info *type)
{
  VALUE obj;
  swig_class *sklass = (swig_class *) type->clientdata;
  obj = Data_Wrap_Struct(klass, VOIDFUNC(sklass->mark), VOIDFUNC(sklass->destroy), 0);
  rb_iv_set(obj, "@__swigtype__", rb_str_new2(type->name));
  return obj;
}

/* Get type mangle from class name */
SWIGRUNTIMEINLINE char *
SWIG_Ruby_MangleStr(VALUE obj)
{
  VALUE stype = rb_iv_get(obj, "@__swigtype__");
  if (NIL_P(stype))
    return NULL;
  return StringValuePtr(stype);
}

/* Acquire a pointer value */
typedef struct {
  void (*datafree)(void *);
  int own;
} swig_ruby_owntype;

SWIGRUNTIME swig_ruby_owntype
SWIG_Ruby_AcquirePtr(VALUE obj, swig_ruby_owntype own) {
  swig_ruby_owntype oldown = {0, 0};
  if (TYPE(obj) == T_DATA && !RTYPEDDATA_P(obj)) {
    oldown.datafree = RDATA(obj)->dfree;
    RDATA(obj)->dfree = own.datafree;
  }
  return oldown;
}

/* Convert a pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPtrAndOwn(VALUE obj, void **ptr, swig_type_info *ty, int flags, swig_ruby_owntype *own)
{
  char *c;
  swig_cast_info *tc;
  void *vptr = 0;

  /* Grab the pointer */
  if (NIL_P(obj)) {
    if (ptr)
      *ptr = 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  } else {
    if (TYPE(obj) != T_DATA || (TYPE(obj) == T_DATA && RTYPEDDATA_P(obj))) {
      return SWIG_ERROR;
    }
    Data_Get_Struct(obj, void, vptr);
  }
  
  if (own) {
    own->datafree = RDATA(obj)->dfree;
    own->own = 0;
  }
    
  if (((flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE)) {
    if (!RDATA(obj)->dfree)
      return SWIG_ERROR_RELEASE_NOT_OWNED;
  }

  /* Check to see if the input object is giving up ownership
     of the underlying C struct or C++ object.  If so then we
     need to reset the destructor since the Ruby object no 
     longer owns the underlying C++ object.*/ 
  if (flags & SWIG_POINTER_DISOWN) {
    /* Is tracking on for this class? */
    int track = 0;
    if (ty && ty->clientdata) {
      swig_class *sklass = (swig_class *) ty->clientdata;
      track = sklass->trackObjects;
    }

    if (track) {
      /* We are tracking objects for this class.  Thus we change the destructor
       * to SWIG_RubyRemoveTracking.  This allows us to
       * remove the mapping from the C++ to Ruby object
       * when the Ruby object is garbage collected.  If we don't
       * do this, then it is possible we will return a reference 
       * to a Ruby object that no longer exists thereby crashing Ruby. */
      RDATA(obj)->dfree = SWIG_RubyRemoveTracking;
    } else {    
      RDATA(obj)->dfree = 0;
    }
  }

  if (flags & SWIG_POINTER_CLEAR) {
    DATA_PTR(obj) = 0;
  }

  /* Do type-checking if type info was provided */
  if (ty) {
    if (ty->clientdata) {
      if (rb_obj_is_kind_of(obj, ((swig_class *) (ty->clientdata))->klass)) {
        if (vptr == 0) {
          /* The object has already been deleted */
          return SWIG_ObjectPreviouslyDeletedError;
        }
      }
    }
    if ((c = SWIG_MangleStr(obj)) == NULL) {
      return SWIG_ERROR;
    }
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) {
      return SWIG_ERROR;
    } else {
      if (ptr) {
        if (tc->type == ty) {
          *ptr = vptr;
        } else {
          int newmemory = 0;
          *ptr = SWIG_TypeCast(tc, vptr, &newmemory);
          if (newmemory == SWIG_CAST_NEW_MEMORY) {
            assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
            if (own)
              own->own = own->own | SWIG_CAST_NEW_MEMORY;
          }
        }
      }
    }
  } else {
    if (ptr)
      *ptr = vptr;
  }
  
  return SWIG_OK;
}

/* Check convert */
SWIGRUNTIMEINLINE int
SWIG_Ruby_CheckConvert(VALUE obj, swig_type_info *ty)
{
  char *c = SWIG_MangleStr(obj);
  if (!c) return 0;
  return SWIG_TypeCheck(c,ty) != 0;
}

SWIGRUNTIME VALUE
SWIG_Ruby_NewPackedObj(void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(type->name)) > 1000) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r, ptr, sz);
  strcpy(r, type->name);
  return rb_str_new2(result);
}

/* Convert a packed pointer value */
SWIGRUNTIME int
SWIG_Ruby_ConvertPacked(VALUE obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c;

  if (TYPE(obj) != T_STRING) goto type_error;
  c = StringValuePtr(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') goto type_error;
  c++;
  c = SWIG_UnpackData(c, ptr, sz);
  if (ty) {
    tc = SWIG_TypeCheck(c, ty);
    if (!tc) goto type_error;
  }
  return SWIG_OK;

 type_error:
  return SWIG_ERROR;
}

SWIGRUNTIME swig_module_info *
SWIG_Ruby_GetModule(void *SWIGUNUSEDPARM(clientdata))
{
  VALUE pointer;
  swig_module_info *ret = 0;
  VALUE verbose = rb_gv_get("VERBOSE");

 /* temporarily disable warnings, since the pointer check causes warnings with 'ruby -w' */
  rb_gv_set("VERBOSE", Qfalse);
  
  /* first check if pointer already created */
  pointer = rb_gv_get("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME);
  if (pointer != Qnil) {
    Data_Get_Struct(pointer, swig_module_info, ret);
  }

  /* reinstate warnings */
  rb_gv_set("VERBOSE", verbose);
  return ret;
}

SWIGRUNTIME void 
SWIG_Ruby_SetModule(swig_module_info *pointer)
{
  /* register a new class */
  VALUE cl = rb_define_class("swig_runtime_data", rb_cObject);
  rb_undef_alloc_func(cl);
  /* create and store the structure pointer to a global variable */
  swig_runtime_data_type_pointer = Data_Wrap_Struct(cl, 0, 0, pointer);
  rb_define_readonly_variable("$swig_runtime_data_type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, &swig_runtime_data_type_pointer);
}

/* This function can be used to check whether a proc or method or similarly
   callable function has been passed.  Usually used in a %typecheck, like:

   %typecheck(c_callback_t, precedence=SWIG_TYPECHECK_POINTER) {
        $result = SWIG_Ruby_isCallable( $input );
   }
 */
SWIGINTERN
int SWIG_Ruby_isCallable( VALUE proc )
{
  if ( rb_respond_to( proc, swig_call_id ) )
    return 1;
  return 0;
}

/* This function can be used to check the arity (number of arguments)
   a proc or method can take.  Usually used in a %typecheck.
   Valid arities will be that equal to minimal or those < 0
   which indicate a variable number of parameters at the end.
 */
SWIGINTERN
int SWIG_Ruby_arity( VALUE proc, int minimal )
{
  if ( rb_respond_to( proc, swig_arity_id ) )
    {
      VALUE num = rb_funcall2( proc, swig_arity_id, 0, 0 );
      int arity = NUM2INT(num);
      if ( arity < 0 && (arity+1) < -minimal ) return 1;
      if ( arity == minimal ) return 1;
      return 1;
    }
  return 0;
}


#ifdef __cplusplus
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg);; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_FILE swig_types[0]
#define SWIGTYPE_p_WS_LASTERR_Code swig_types[1]
#define SWIGTYPE_p_WsEnumerateInfo swig_types[2]
#define SWIGTYPE_p_WsmanKnownStatusCode swig_types[3]
#define SWIGTYPE_p_WsmanMessage swig_types[4]
#define SWIGTYPE_p__WS_CONTEXT swig_types[5]
#define SWIGTYPE_p__WsManClient swig_types[6]
#define SWIGTYPE_p__WsManFault swig_types[7]
#define SWIGTYPE_p__WsManTransport swig_types[8]
#define SWIGTYPE_p__WsXmlDoc swig_types[9]
#define SWIGTYPE_p__WsmanStatus swig_types[10]
#define SWIGTYPE_p___Soap swig_types[11]
#define SWIGTYPE_p___SoapOp swig_types[12]
#define SWIGTYPE_p___WsEnumerateInfo swig_types[13]
#define SWIGTYPE_p___WsXmlAttr swig_types[14]
#define SWIGTYPE_p___WsXmlNode swig_types[15]
#define SWIGTYPE_p___WsXmlNs swig_types[16]
#define SWIGTYPE_p___WsmanFaultCodeTable swig_types[17]
#define SWIGTYPE_p___WsmanFaultCodeType swig_types[18]
#define SWIGTYPE_p___WsmanFaultDetailTable swig_types[19]
#define SWIGTYPE_p___WsmanFaultDetailType swig_types[20]
#define SWIGTYPE_p_char swig_types[21]
#define SWIGTYPE_p_client_opt_t swig_types[22]
#define SWIGTYPE_p_epr_t swig_types[23]
#define SWIGTYPE_p_filter_t swig_types[24]
#define SWIGTYPE_p_wsman_auth_request_func_t swig_types[25]
#define SWIGTYPE_p_wsman_auth_type_t swig_types[26]
static swig_type_info *swig_types[28];
static swig_module_info swig_module = {swig_types, 27, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    Init_Openwsman
#define SWIG_name    "Openwsman"

static VALUE mOpenwsman;

#define SWIG_RUBY_THREAD_BEGIN_BLOCK
#define SWIG_RUBY_THREAD_END_BLOCK


#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


/*
 * type definitions to keep the C code generic
 */
 
#if defined(SWIGPYTHON)
#define Target_Null_p(x) (x == Py_None)
#define Target_INCREF(x) Py_INCREF(x)
#define Target_DECREF(x) Py_DECREF(x)
#define Target_True Py_True
#define Target_False Py_False
#define Target_Null Py_None
#define Target_Type PyObject*
#define Target_Bool(x) PyBool_FromLong(x)
#define Target_Char16(x) PyInt_FromLong(x)
#define Target_Int(x) PyInt_FromLong(x)
#define Target_String(x) PyString_FromString(x)
#define Target_Real(x) Py_None
#define Target_Array() PyList_New(0)
#define Target_SizedArray(len) PyList_New(len)
#define Target_ListSet(x,n,y) PyList_SetItem(x,n,y)
#define Target_Append(x,y) PyList_Append(x,y)
#include <Python.h>
#define TARGET_THREAD_BEGIN_BLOCK SWIG_PYTHON_THREAD_BEGIN_BLOCK
#define TARGET_THREAD_END_BLOCK SWIG_PYTHON_THREAD_END_BLOCK
#define TARGET_THREAD_BEGIN_ALLOW SWIG_PYTHON_THREAD_BEGIN_ALLOW
#define TARGET_THREAD_END_ALLOW SWIG_PYTHON_THREAD_END_ALLOW
#endif

#if defined(SWIGRUBY)
#define Target_Null_p(x) NIL_P(x)
#define Target_INCREF(x) 
#define Target_DECREF(x) 
#define Target_True Qtrue
#define Target_False Qfalse
#define Target_Null Qnil
#define Target_Type VALUE
#define Target_Bool(x) ((x)?Qtrue:Qfalse)
#define Target_Char16(x) INT2FIX(x)
#define Target_Int(x) INT2FIX(x)
#define Target_String(x) rb_str_new2(x)
#define Target_Real(x) rb_float_new(x)
#define Target_Array() rb_ary_new()
#define Target_SizedArray(len) rb_ary_new2(len)
#define Target_ListSet(x,n,y) rb_ary_store(x,n,y)
#define Target_Append(x,y) rb_ary_push(x,y)
#define TARGET_THREAD_BEGIN_BLOCK do {} while(0)
#define TARGET_THREAD_END_BLOCK do {} while(0)
#define TARGET_THREAD_BEGIN_ALLOW do {} while(0)
#define TARGET_THREAD_END_ALLOW do {} while(0)
#include <ruby.h>
#if HAVE_RUBY_IO_H
#include <ruby/io.h> /* Ruby 1.9 style */
#else
#include <rubyio.h>
#endif
#if HAVE_RUBY_VERSION_H
#include <ruby/version.h>
#endif
#endif

#if defined(SWIGPERL)
#define TARGET_THREAD_BEGIN_BLOCK do {} while(0)
#define TARGET_THREAD_END_BLOCK do {} while(0)
#define TARGET_THREAD_BEGIN_ALLOW do {} while(0)
#define TARGET_THREAD_END_ALLOW do {} while(0)

SWIGINTERNINLINE SV *SWIG_From_long  SWIG_PERL_DECL_ARGS_1(long value);
SWIGINTERNINLINE SV *SWIG_FromCharPtr(const char *cptr);
SWIGINTERNINLINE SV *SWIG_From_double  SWIG_PERL_DECL_ARGS_1(double value);

#define Target_Null_p(x) (x == NULL)
#define Target_INCREF(x) 
#define Target_DECREF(x) 
#define Target_True (&PL_sv_yes)
#define Target_False (&PL_sv_no)
#define Target_Null NULL
#define Target_Type SV *
#define Target_Bool(x) (x)?Target_True:Target_False
#define Target_Char16(x) SWIG_From_long(x)
#define Target_Int(x) SWIG_From_long(x)
#define Target_String(x) SWIG_FromCharPtr(x)
#define Target_Real(x) SWIG_From_double(x)
#define Target_Array() (SV *)newAV()
#define Target_SizedArray(len) (SV *)newAV()
#define Target_ListSet(x,n,y) av_store((AV *)(x),n,y)
#define Target_Append(x,y) av_push(((AV *)(x)), y)
#include <perl.h>
#include <EXTERN.h>
#endif



#include <ruby.h>
#if HAVE_RUBY_IO_H
#include <ruby/io.h> /* Ruby 1.9 style */
#else
#include <rubyio.h>
#endif

#if HAVE_RUBY_THREAD_H /* New threading model */
#include <ruby/thread.h>
#endif
#if RUBY_VERSION > 18
  #if HAVE_RB_IO_T
    #define rb_fptr_t rb_io_t
  #else
    #define rb_fptr_t struct rb_io
  #endif
#else
  #define rb_fptr_t struct OpenFile
#endif


#if defined(SWIGPERL)
/* filter_t is defined in Perls CORE/perl.h */
#define filter_t filter_type

/* undef Perl's die and warn macros as libu is going to reuse this names */
#undef die
#undef warn
#endif /* SWIGPERL */

#include <u/libu.h>
#include <wsman-types.h>
#include <wsman-client.h>
#include <wsman-client-transport.h>
#include <wsman-api.h>
#include <wsman-xml-binding.h>
#include <wsman-xml.h>
#include <wsman-epr.h>
#include <wsman-filter.h>
#include <wsman-soap.h>
#include <wsman-soap-envelope.h>
#include <openwsman.h>

#if defined(SWIGRUBY)
#include <ruby/helpers.h>

SWIGEXPORT
/* Init_ for the %module, defined by Swig
 * :nodoc:
 */
void Init_Openwsman(void);

SWIGEXPORT
/* Init_ for the .so lib, called by Ruby
 * :nodoc:
 */
void Init__openwsman(void) {
  Init_Openwsman();
}
#endif
#if defined(SWIGPYTHON)
#include <python/helpers.h>
#endif
#if defined(SWIGJAVA)
#include <java/helpers.h>
#endif
#if defined(SWIGPERL)
#include <perl/helpers.h>
#endif

/* Provide WsManTransport definition so it can be used as
 * dedicated datatype in bindings.
 * :nodoc:ly, its aliased to WsManClient
 *
 */
struct _WsManTransport { };
typedef struct _WsManTransport WsManTransport;

static void set_debug(int dbg);

/*
 * Set openwsman debug level.
 * call-seq:
 *   Openwsman::debug = -1 # full debug
 *
 */
static void set_debug(int dbg) {
  static int init = 0;

  if (!init && dbg != 0) {
    init = 1;
    debug_add_handler( debug_message_handler, DEBUG_LEVEL_ALWAYS, NULL );
  }
  wsman_debug_set_level( dbg );
	
}

/* module-level methods */

static int get_debug(void);

/*
 * Return openwsman debug level.
 * call-seq:
 *   Openwsman::debug -> Integer
 *
 */
static int get_debug(void) {
  return (int)wsman_debug_get_level();
}

static WsXmlDocH create_soap_envelope(void);

/*
 * Create empty SOAP envelope
 * call-seq:
 *   Openwsman::create_soap_envelope -> XmlDoc
 *
 */
static WsXmlDocH create_soap_envelope() {
  return ws_xml_create_soap_envelope();
}

static WsXmlDocH create_doc_from_file(const char *filename, const char *encoding);

/*
 * Read XmlDoc from file
 * call-seq:
 *   Openwsman::create_doc_from_file("/path/to/file", "utf-8") -> XmlDoc
 *
 */
static WsXmlDocH create_doc_from_file(const char *filename, const char *encoding) {
  return xml_parser_file_to_doc( filename, encoding, 0);                 
}

static WsXmlDocH create_doc_from_string(const char *buf, const char *encoding);

/*
 * Read XmlDoc from string
 * call-seq:
 *   Openwsman::create_doc_from_string("<xml ...>", "utf-8") -> XmlDoc
 *
 */
static WsXmlDocH create_doc_from_string(const char *buf, const char *encoding) {
  return xml_parser_memory_to_doc( buf, strlen(buf), encoding, 0);
}

static char *uri_classname(const char *uri);
/*
 * get classname from resource URI
 * call-seq:
 *   Openwsman::uri_classname("http://sblim.sf.net/wbem/wscim/1/cim-schema/2/Linux_OperatingSystem") -> "Linux_OperatingSystem"
 *
 */
static char *uri_classname(const char *uri) {
  const char *lastslash = strrchr(uri,'/');
  if (lastslash) {
    return strdup(lastslash+1);
  }
  return NULL;
}

static const char *uri_prefix(const char *classname);
/*
 * Map classname (class schema) to resource uri prefix
 * call-seq:
 *   Openwsman::uri_prefix("Linux") -> "http://sblim.sf.net/wbem/wscim/1/cim-schema/2"
 *   Openwsman::uri_prefix("Win32") -> "http://schemas.microsoft.com/wbem/wsman/1/wmi"
 *
 */
static const char *uri_prefix(const char *classname) {
  static struct map {
    int len;
    const char *schema;
    const char *prefix;
  } mapping[] = {
    /* dmtf CIM */
    { 3, "CIM", "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2" },
    /* dmtf reserved */
    { 3, "PRS", "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2" },
    /* Microsoft WMI */
    { 5, "Win32", "http://schemas.microsoft.com/wbem/wsman/1/wmi" },
    /* openwbem.org */
    { 8, "OpenWBEM", "http://schema.openwbem.org/wbem/wscim/1/cim-schema/2" },
    /* sblim */
    { 5, "Linux", "http://sblim.sf.net/wbem/wscim/1/cim-schema/2" },
    /* omc-project */
    { 3, "OMC", "http://schema.omc-project.org/wbem/wscim/1/cim-schema/2" },
    /* pegasus.org */
    { 2, "PG", "http://schema.openpegasus.org/wbem/wscim/1/cim-schema/2" },
    /* Intel AMT */
    { 3, "AMT", "http://intel.com/wbem/wscim/1/amt-schema/1" },
    /* Intel */
    { 3, "IPS", "http://intel.com/wbem/wscim/1/ips-schema/1" },
    /* Sun */
    { 3, "Sun","http://schemas.sun.com/wbem/wscim/1/cim-schema/2" },
    /* Microsoft HyperV */
    { 4, "Msvm", "http://schemas.microsoft.com/wbem/wsman/1/wmi" },
    /* Dell DRAC */
    { 4, "DCIM", "http://schemas.dell.com/wbem/wscim/1/cim-schema/2" },
    /* Unisys */
    { 4, "SPAR", "http://schema.unisys.com/wbem/wscim/1/cim-schema/2" },
    /* Fujitsu */
    { 3, "SVS", "http://schemas.ts.fujitsu.com/wbem/wscim/1/cim-schema/2" },
    { 0, NULL, NULL }
  };
  const char *schema_end;
  struct map *map;
  int len;
  if (classname == NULL)
    return NULL;
  if (strcmp(classname, "*") == 0) {
    return "http://schemas.dmtf.org/wbem/wscim/1";
  }
  if ((strcmp(classname, "meta_class") == 0)
      ||(strncmp(classname, "__", 2) == 0)) {
    return "http://schemas.microsoft.com/wbem/wsman/1/wmi";
  }
  schema_end = strchr(classname, '_');
  if (schema_end == NULL)
    return NULL; /* Bad class name */
  len = schema_end - classname;
  map = mapping;
  while (map->len > 0) {
    if ((len == map->len)
        && (strncasecmp(classname, map->schema, map->len) == 0)) {
      return map->prefix;
    }
    ++map;
  }
  return NULL;
}

#if defined(SWIGRUBY)
static epr_t *my_epr_deserialize(WsXmlNodeH node);
/*:nodoc:*/
static epr_t *my_epr_deserialize(WsXmlNodeH node) {
  if (strcmp(WSA_EPR, ws_xml_get_node_local_name(node)) == 0) {
    /* Use node as-is if its already a WSA_EPR */
    return epr_deserialize(node, NULL, NULL, 1);
  }
  /* else search the WSA_EPR node */
  return epr_deserialize(node, XML_NS_ADDRESSING, WSA_EPR, 1);
}

/*:nodoc:*/
static VALUE kv_list_to_hash(list_t *list)
{
  VALUE v = Qnil;
  if (!list_isempty(list)) {
  v = rb_hash_new();
  lnode_t *node = list_first(list);
  while (node) {
    key_value_t *kv = (key_value_t *)node->list_data;
    if (kv->type == 0) {
      rb_hash_aset( v, makestring(kv->key), makestring(kv->v.text));
    }
    else {
      VALUE epr = SWIG_NewPointerObj((void*)kv->v.epr, SWIGTYPE_p_epr_t, 0);
      rb_hash_aset( v, makestring(kv->key), epr);
    }
    node = list_next(list, node);
    }
  }
  return v;
}
#endif

static char *epr_prefix(const char *uri);
/* Get prefix from a EPR uri
 * :nodoc:
 */
static char *epr_prefix(const char *uri) {
  char *classname = uri_classname(uri);
  const char *prefix = uri_prefix(classname);
  if (prefix) {
    const char *lastslash;
    if (strncmp(uri, prefix, strlen(prefix)) == 0) {
      return strdup(prefix);
    }
    lastslash = strrchr(uri, '/');
    if (lastslash) {
      return strndup(uri, lastslash-uri);
    }
  }
  return strdup(uri);
}



  typedef struct {
    WsManClient *client;
    client_opt_t *options;
    filter_t *filter;
    const char *resource_uri;
    epr_t *epr;
    const char *context;
    const char *identifier;
    const char *data;
    size_t size;
    const char *encoding;
    const char *method;
    WsXmlDocH method_args;
  } wsmc_action_args_t;

  WsXmlDocH
  ruby_enumerate_thread(wsmc_action_args_t *args) {
    return wsmc_action_enumerate(args->client, args->resource_uri, args->options, args->filter);
  }
  WsXmlDocH
  ruby_identify_thread(wsmc_action_args_t *args) {
    return wsmc_action_identify(args->client, args->options);
  }
  WsXmlDocH
  ruby_get_from_epr_thread(wsmc_action_args_t *args) {
    return wsmc_action_get_from_epr(args->client, args->epr, args->options);
  }
  WsXmlDocH
  ruby_delete_from_epr_thread(wsmc_action_args_t *args) {
    return wsmc_action_delete_from_epr(args->client, args->epr, args->options);
  }
  WsXmlDocH
  ruby_pull_thread(wsmc_action_args_t *args) {
    return wsmc_action_pull(args->client, args->resource_uri, args->options, args->filter, args->context);
  }
  WsXmlDocH
  ruby_create_fromtext_thread(wsmc_action_args_t *args) {
    return wsmc_action_create_fromtext(args->client, args->resource_uri, args->options, args->data, args->size, args->encoding);
  }
  WsXmlDocH
  ruby_put_fromtext_thread(wsmc_action_args_t *args) {
    return wsmc_action_put_fromtext(args->client, args->resource_uri, args->options, args->data, args->size, args->encoding);
  }
  WsXmlDocH
  ruby_release_thread(wsmc_action_args_t *args) {
    return wsmc_action_release(args->client, args->resource_uri, args->options, args->context);
  }
  WsXmlDocH
  ruby_get_thread(wsmc_action_args_t *args) {
    return wsmc_action_get(args->client, args->resource_uri, args->options);
  }
  WsXmlDocH
  ruby_delete_thread(wsmc_action_args_t *args) {
    return wsmc_action_delete(args->client, args->resource_uri, args->options);
  }
  WsXmlDocH
  ruby_invoke_thread(wsmc_action_args_t *args) { 
    return wsmc_action_invoke(args->client, args->resource_uri, args->options, args->method, args->method_args);
  }
  WsXmlDocH
  ruby_subscribe_thread(wsmc_action_args_t *args) {
    return wsmc_action_subscribe(args->client, args->resource_uri, args->options, args->filter);
  }
  WsXmlDocH
  ruby_unsubscribe_thread(wsmc_action_args_t *args) {
    return wsmc_action_unsubscribe(args->client, args->resource_uri, args->options, args->identifier);
  }
  WsXmlDocH
  ruby_renew_thread(wsmc_action_args_t *args) {
    return wsmc_action_renew(args->client, args->resource_uri, args->options, args->identifier);
  }


#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


  #define SWIG_From_long   LONG2NUM 


SWIGINTERNINLINE VALUE
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > LONG_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_NewPointerObj((char *)(carray), pchar_descriptor, 0) : Qnil;
    } else {
      return rb_str_new(carray, (long)(size));
    }
  } else {
    return Qnil;
  }
}


SWIGINTERNINLINE VALUE 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(VALUE obj, char** cptr, size_t* psize, int *alloc)
{
  if (TYPE(obj) == T_STRING) {
    char *cstr = StringValuePtr(obj); 
    size_t size = RSTRING_LEN(obj) + 1;
    if (cptr)  {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *)vptr;
	if (psize) *psize = vptr ? (strlen((char*)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }  
  return SWIG_TypeError;
}




SWIGINTERN struct _WsXmlDoc *new__WsXmlDoc(char const *name,char const *ns){
    return ws_xml_create_doc(ns, name);
  }
SWIGINTERN char *_WsXmlDoc_string(struct _WsXmlDoc *self){
    int size;
    char *buf;
    /* force utf-8 encoding since e.g. winrm sends utf-16 */
    ws_xml_dump_memory_node_tree_enc( ws_xml_get_doc_root(self), &buf, &size, "UTF-8" );
    return buf;
  }
SWIGINTERN char *_WsXmlDoc_encode(struct _WsXmlDoc *self,char const *encoding){
    int size;
    char *buf;
    ws_xml_dump_memory_enc( self, &buf, &size, encoding );
    return buf;
  }
SWIGINTERN void _WsXmlDoc_dump_file(struct _WsXmlDoc *self,FILE *fp){
    ws_xml_dump_doc( fp, self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_root(struct _WsXmlDoc *self){
    return ws_xml_get_doc_root( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_envelope(struct _WsXmlDoc *self){
    return ws_xml_get_soap_envelope( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_header(struct _WsXmlDoc *self){
    return ws_xml_get_soap_header( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_body(struct _WsXmlDoc *self){
    return ws_xml_get_soap_body( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_element(struct _WsXmlDoc *self,char const *name){
    return ws_xml_get_soap_element( self, name );
  }
SWIGINTERN char const *_WsXmlDoc_context(struct _WsXmlDoc *self){
    char *c = wsmc_get_enum_context( self );
    if (c) {
      if (*c)
        return c;
      u_free(c);
    }
    return NULL;
  }
SWIGINTERN WsXmlDocH _WsXmlDoc_generate_fault(struct _WsXmlDoc *self,WsmanStatus *s){
    return wsman_generate_fault( self, s->fault_code, s->fault_detail_code, s->fault_msg);
  }
SWIGINTERN int _WsXmlDoc_is_fault(struct _WsXmlDoc *self){
    return wsmc_check_for_fault( self );
  }
SWIGINTERN WsManFault *_WsXmlDoc_fault(struct _WsXmlDoc *self){
    WsManFault *f = NULL;
    if (wsmc_check_for_fault(self)) {
      f = (WsManFault *)calloc(1, sizeof(WsManFault));
      wsmc_get_fault_data(self, f);
    }
    return f;
  }
SWIGINTERN WsXmlDocH _WsXmlDoc_create_response_envelope(struct _WsXmlDoc *self,char const *action){
    return wsman_create_response_envelope(self, action);
  }
SWIGINTERN int _WsXmlDoc_is_end_of_sequence(struct _WsXmlDoc *self){
    return NULL != ws_xml_find_in_tree( ws_xml_get_soap_body( self ), XML_NS_ENUMERATION, WSENUM_END_OF_SEQUENCE, 1 );
  }
SWIGINTERN char *__WsXmlNode_string(struct __WsXmlNode *self){
    int size;
    char *buf;
    ws_xml_dump_memory_node_tree( self, &buf, &size );
    return buf;
  }
SWIGINTERN void __WsXmlNode_dump_file(struct __WsXmlNode *self,FILE *fp){
    ws_xml_dump_node_tree( fp, self );
  }
SWIGINTERN int __WsXmlNode_equal(struct __WsXmlNode *self,WsXmlNodeH n){ return self == n; }
SWIGINTERN char *__WsXmlNode_text(struct __WsXmlNode *self){
    return ws_xml_get_node_text( self );
  }
SWIGINTERN void __WsXmlNode_set_text(struct __WsXmlNode *self,char const *text){
    ws_xml_set_node_text( self, text );
  }
SWIGINTERN WsXmlDocH __WsXmlNode_doc(struct __WsXmlNode *self){
    return ws_xml_get_node_doc( self );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_parent(struct __WsXmlNode *self){
    return ws_xml_get_node_parent( self );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_child(struct __WsXmlNode *self){
    return xml_parser_get_first_child(self);
  }
SWIGINTERN char *__WsXmlNode_name(struct __WsXmlNode *self){
    return ws_xml_get_node_local_name( self );
  }
SWIGINTERN void __WsXmlNode_set_name(struct __WsXmlNode *self,char const *name){
    ws_xml_set_node_name( self, ws_xml_get_node_name_ns( self ), name );
  }
SWIGINTERN char *__WsXmlNode_ns(struct __WsXmlNode *self){
    return ws_xml_get_node_name_ns( self );
  }
SWIGINTERN void __WsXmlNode_set_ns(struct __WsXmlNode *self,char const *ns){
    ws_xml_set_ns( self, ns, ws_xml_get_node_name_ns_prefix(self) );
  }
SWIGINTERN char const *__WsXmlNode_prefix(struct __WsXmlNode *self){
    return ws_xml_get_node_name_ns_prefix(self);
  }
SWIGINTERN void __WsXmlNode_set_lang(struct __WsXmlNode *self,char const *lang){
    ws_xml_set_node_lang(self, lang);
  }

SWIGINTERN VALUE
SWIG_ruby_failed(VALUE SWIGUNUSEDPARM(arg1), VALUE SWIGUNUSEDPARM(arg2))
{
  return Qnil;
} 


SWIGINTERN VALUE SWIG_AUX_NUM2LONG(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  long *res = (long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2LONG(obj) : rb_big2long(obj);
  return obj;
}

SWIGINTERN int
SWIG_AsVal_long (VALUE obj, long* val)
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2LONG), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int (VALUE obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}

SWIGINTERN WsXmlNodeH __WsXmlNode_find(struct __WsXmlNode *self,char const *ns,char const *name,int recursive){
    return ws_xml_find_in_tree( self, ns, name, recursive );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_next(struct __WsXmlNode *self,int all){
    WsXmlNodeH next_node = xml_parser_get_next_child(self);
    if (next_node && !all) {
      const char *ns_uri = ws_xml_get_node_name_ns(self);
      const char *name = ws_xml_get_node_local_name(self);
      if (ws_xml_is_node_qname(next_node, ns_uri, name) == 0) {
        next_node = NULL;
      }
    }
    return next_node;
  }
SWIGINTERN int __WsXmlNode_size(struct __WsXmlNode *self,char const *name,char const *ns){
    return ws_xml_get_child_count_by_qname(self, ns, name);
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add__SWIG_0(struct __WsXmlNode *self,char const *ns,char const *name,char const *text){
    return ws_xml_add_child( self, ns, name, text );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add_before(struct __WsXmlNode *self,char const *ns,char const *name,char const *text){
    return ws_xml_add_prev_sibling( self, ns, name, text );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add__SWIG_1(struct __WsXmlNode *self,WsXmlNodeH node){
    ws_xml_duplicate_tree( self, node );
    return self;
  }
SWIGINTERN void __WsXmlNode_each(struct __WsXmlNode *self,char const *name,char const *ns){
    int i = 0;
    WsXmlNodeH node = self;
    int count = ws_xml_get_child_count_by_qname( node, ns, name );
    while ( i < count ) {
      rb_yield( SWIG_NewPointerObj((void*) ws_xml_get_child(node, i, ns, name), SWIGTYPE_p___WsXmlNode, 0));
      ++i;
    }
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_get__SWIG_0(struct __WsXmlNode *self,int i,char const *name,char const *ns){
    if (i < 0 || i >= ws_xml_get_child_count_by_qname(self,ns,name))
      return NULL;
    return ws_xml_get_child(self, i, ns, name);
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_get__SWIG_1(struct __WsXmlNode *self,char const *name,char const *ns){
    return ws_xml_get_child(self, 0, ns, name);
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr(struct __WsXmlNode *self,VALUE index,VALUE namespace){
    if (NIL_P(index)) { /* nil */
      return ws_xml_get_node_attr( self, 0 );
    } else if (FIXNUM_P(index)) { /* numeric */
      return ws_xml_get_node_attr( self, FIX2INT(index) );      
    } else { /* convert to string */
      const char *ns = NULL;
      const char *name = as_string(index);
      if (!NIL_P(namespace)) {
        ns = as_string(namespace);
      }
      return ws_xml_find_node_attr( self, ns, name );
    }
  }
SWIGINTERN int __WsXmlNode_attr_count(struct __WsXmlNode *self){
    return ws_xml_get_node_attr_count( self );
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr_find(struct __WsXmlNode *self,char const *ns,char const *name){
    return ws_xml_find_node_attr( self, ns, name );
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr_add(struct __WsXmlNode *self,char const *ns,char const *name,char const *value){
    return ws_xml_add_node_attr( self, ns, name, value );
  }
SWIGINTERN epr_t *__WsXmlNode_epr(struct __WsXmlNode *self,char const *ns,char const *epr_node_name,int embedded){
    return epr_deserialize(self, ns, epr_node_name, embedded);
  }
SWIGINTERN void __WsXmlNode_each_attr(struct __WsXmlNode *self){
    int i = 0;
    while ( i < ws_xml_get_node_attr_count( self ) ) {
      rb_yield( SWIG_NewPointerObj((void*) ws_xml_get_node_attr(self, i), SWIGTYPE_p___WsXmlAttr, 0));
      ++i;
    }
  }
SWIGINTERN char *__WsXmlAttr_name(struct __WsXmlAttr *self){
    return ws_xml_get_attr_name( self );
  }
SWIGINTERN char *__WsXmlAttr_ns(struct __WsXmlAttr *self){
    return ws_xml_get_attr_ns( self );
  }
SWIGINTERN char *__WsXmlAttr_value(struct __WsXmlAttr *self){
    return ws_xml_get_attr_value( self );
  }
SWIGINTERN void __WsXmlAttr_remove(struct __WsXmlAttr *self){
    ws_xml_remove_node_attr( self );
  }
SWIGINTERN epr_t *new_epr_t(VALUE uri,VALUE address,VALUE selectors){
    KLASS_DECL(SwigClassXmlNode,SWIGTYPE_p___WsXmlNode);
    KLASS_DECL(SwigClassXmlDoc,SWIGTYPE_p__WsXmlDoc);
    if (!NIL_P(address) || !NIL_P(selectors)) {
      const char *uri_s = as_string(uri);
      const char *address_s = as_string(address);
      return epr_create(uri_s, value2hash(NULL,selectors,1), address_s);
    }
    else if (CLASS_OF(uri) == KLASS_OF(SwigClassXmlNode)) {
      WsXmlNodeH node;
      SWIG_ConvertPtr(uri, (void **)&node, SWIGTYPE_p___WsXmlNode, 0);
      return my_epr_deserialize(node);
    }
    else if (CLASS_OF(uri) == KLASS_OF(SwigClassXmlDoc)) {
      WsXmlDocH doc;
      WsXmlNodeH node;
      SWIG_ConvertPtr(uri, (void **)&doc, SWIGTYPE_p__WsXmlDoc, 0);
      node = ws_xml_get_soap_body(doc);
      if (node == NULL)
        node = ws_xml_get_doc_root(doc);
      return my_epr_deserialize(node);
    }
    else {
      return epr_from_string(as_string(uri));
    }
  }
SWIGINTERN epr_t *epr_t_clone(epr_t *self,epr_t const *epr){
    return epr_copy(epr);
  }
SWIGINTERN void epr_t_add_selector(epr_t *self,char const *name,char const *text){
    epr_add_selector_text(self, name, text);
  }
SWIGINTERN int epr_t_serialize(epr_t *self,WsXmlNodeH node,char const *ns,char const *epr_node_name,int embedded){
    return epr_serialize(node, ns, epr_node_name, self, embedded);
  }
SWIGINTERN int epr_t_cmp(epr_t *self,epr_t const *epr2){
    return epr_cmp(self, epr2);
  }
SWIGINTERN char *epr_t_to_xml(epr_t *self,char const *ns,char const *epr_node_name){
    return epr_to_txt(self, ns?ns:"http://schemas.xmlsoap.org/ws/2004/08/addressing", epr_node_name?epr_node_name:"EndpointReference");
  }
SWIGINTERN char *epr_t_string(epr_t *self){
    return epr_to_string(self);
  }
SWIGINTERN int epr_t_selector_count(epr_t *self){
    return epr_selector_count(self);
  }
SWIGINTERN char *epr_t_resource_uri(epr_t *self){
    return epr_get_resource_uri(self);
  }
SWIGINTERN char *epr_t_selector(epr_t *self,VALUE v){
    const char *name = as_string(v);







    return wsman_epr_selector_by_name(self, name);
  }
SWIGINTERN VALUE epr_t_selector_names(epr_t *self){







    int i;
    Target_Type ary = Target_SizedArray(self->refparams.selectorset.count);
    key_value_t *p = self->refparams.selectorset.selectors;
    for (i = 0; i < self->refparams.selectorset.count; i++) {
      Target_ListSet(ary, i, SWIG_FromCharPtr(p->key));
      ++p;
    }



    return ary;

  }
SWIGINTERN void epr_t_each(epr_t *self){
    int i;
    key_value_t *p = NULL;
    VALUE value, ary;
    p = self->refparams.selectorset.selectors;
    for (i = 0; i < self->refparams.selectorset.count; i++) {
      ary = rb_ary_new2(2);
      rb_ary_store(ary, 0, SWIG_FromCharPtr(p->key));
      if (p->type == 0) {
        value = SWIG_FromCharPtr(p->v.text);
      } else {
        value = SWIG_NewPointerObj((void*) p->v.epr, SWIGTYPE_p_epr_t, 0);
      }
      rb_ary_store(ary, 1, value);
      rb_yield(ary);
      p++;
    }
  }
SWIGINTERN char *epr_t_classname(epr_t *self){
    return uri_classname(self->refparams.uri);
  }
SWIGINTERN char *epr_t_namespace(epr_t *self){
    char *classname;
    int classnamelen, namespacelen;
    const char *prefix;
    const char *uri;

    /* try to get namespace from selectors (WS-Management style) */
    char *ns = get_cimnamespace_from_selectorset(&(self->refparams.selectorset));
    if (ns) {
      return strdup(ns);
    }
    /* WMI style? - extract namespace from uri */

    uri = self->refparams.uri;
    prefix = epr_prefix(uri);
    if (prefix == NULL) {
      return NULL; /* bad classname in URI */
    }
    classname = uri_classname(uri);
    if (classname == NULL)
      return NULL; /* bad URI */
    classnamelen = strlen(classname);
    free(classname);
    namespacelen = strlen(uri) - classnamelen - strlen(prefix) - 2; /* drop enclosing slashes */
    if (namespacelen <= 0)
      return strdup("");
    /* copy after prefix slash (+ 1) */
    return strndup(uri + strlen(prefix) + 1, namespacelen);
  }
SWIGINTERN char *epr_t_prefix(epr_t *self){
    return epr_prefix(self->refparams.uri);
  }
SWIGINTERN filter_t *new_filter_t(void){
    return filter_initialize();
  }
SWIGINTERN int filter_t_associators(filter_t *self,VALUE epr_v,VALUE assocClass_v,VALUE resultClass_v,VALUE role_v,VALUE resultRole_v,VALUE resultProp_v,VALUE propNum_v){
    return associators_references( self, 0, epr_v, assocClass_v, resultClass_v, role_v, resultRole_v, resultProp_v, propNum_v);
  }
SWIGINTERN int filter_t_references(filter_t *self,VALUE epr_v,VALUE assocClass_v,VALUE resultClass_v,VALUE role_v,VALUE resultRole_v,VALUE resultProp_v,VALUE propNum_v){
    return associators_references( self, 1, epr_v, assocClass_v, resultClass_v, role_v, resultRole_v, resultProp_v, propNum_v);
  }
SWIGINTERN int filter_t_simple(filter_t *self,char const *dialect,char const *query){
    return filter_set_simple(self, dialect, query );
  }
SWIGINTERN int filter_t_xpath(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://www.w3.org/TR/1999/REC-xpath-19991116", query );
  }
SWIGINTERN int filter_t_cql(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://schemas.dmtf.org/wbem/cql/1/dsp0202.pdf", query );
  }
SWIGINTERN int filter_t_wql(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://schemas.microsoft.com/wbem/wsman/1/WQL", query );
  }

  /*
   * Build associators (type == 0) or references (type == 1) filter
   *
   */
  static int associators_references( void *filter, int type, VALUE epr_v, VALUE assocClass_v, VALUE resultClass_v, VALUE role_v, VALUE resultRole_v, VALUE resultProp_v, VALUE propNum_v)
  {
    epr_t *epr = NULL;
    const char *assocClass = as_string(assocClass_v);
    const char *resultClass = as_string(resultClass_v);
    const char *role = as_string(role_v);
    const char *resultRole = as_string(resultRole_v);
    char **resultProp = NULL;
    int propNum = 0;
    int res;
    KLASS_DECL(SwigClassEndPointReference,SWIGTYPE_p_epr_t);

    if (CLASS_OF(epr_v) == KLASS_OF(SwigClassEndPointReference)) {
      SWIG_ConvertPtr(epr_v, (void **)&epr, SWIGTYPE_p_epr_t, 0);
    }
    else {
      epr = epr_from_string(as_string(epr_v));
    }
    if (!NIL_P(resultProp_v)) {
      int i;
      if (TYPE(resultProp_v) != T_ARRAY) {
        SWIG_exception( SWIG_ArgError(resultProp_v), Ruby_Format_TypeError( "", "Array", "resultProp", 6, resultProp_v ) );
      }
      resultProp = (char **)calloc(RARRAY_LEN(resultProp_v), sizeof(char *));
      for (i = 0; i < RARRAY_LEN(resultProp_v); ++i) {
        resultProp[i] = (char *)as_string(rb_ary_entry(resultProp_v, i));
        ++propNum;
      }
    }
    res = filter_set_assoc((filter_t *)filter, epr, type, assocClass, resultClass, role, resultRole, resultProp, propNum);
    if (resultProp) free(resultProp);
    return res;
  }

SWIGINTERN int __WsEnumerateInfo_max_items(struct __WsEnumerateInfo *self){ return self->maxItems; }
SWIGINTERN void __WsEnumerateInfo_set_max_items(struct __WsEnumerateInfo *self,int mi){ self->maxItems = mi; }
SWIGINTERN int __WsEnumerateInfo_flags(struct __WsEnumerateInfo *self){
    return self->flags;
  }
SWIGINTERN char const *__WsEnumerateInfo_epr_to(struct __WsEnumerateInfo *self){
    return self->epr_to;
  }
SWIGINTERN char const *__WsEnumerateInfo_epr_uri(struct __WsEnumerateInfo *self){
    return self->epr_uri;
  }
SWIGINTERN char const *__WsEnumerateInfo_encoding(struct __WsEnumerateInfo *self){
    return self->encoding;
  }
SWIGINTERN filter_t const *__WsEnumerateInfo_filter(struct __WsEnumerateInfo *self){
    return self->filter;
  }
SWIGINTERN int __WsEnumerateInfo_index(struct __WsEnumerateInfo *self){
    return self->index;
  }
SWIGINTERN void __WsEnumerateInfo_set_index(struct __WsEnumerateInfo *self,int i){
    self->index = i;
  }
SWIGINTERN int __WsEnumerateInfo_total_items(struct __WsEnumerateInfo *self){
    return self->totalItems;
  }
SWIGINTERN void __WsEnumerateInfo_set_total_items(struct __WsEnumerateInfo *self,int i){
    self->totalItems = i;
  }
SWIGINTERN WsXmlDocH __WsEnumerateInfo_pull_result(struct __WsEnumerateInfo *self){
    return (WsXmlDocH)self->pullResultPtr;
  }
SWIGINTERN void __WsEnumerateInfo_set_pull_result(struct __WsEnumerateInfo *self,WsXmlDocH result){
    self->pullResultPtr = (void *)result;
  }
SWIGINTERN WsXmlDocH __SoapOp_indoc(struct __SoapOp *self){
    return soap_get_op_doc(self, 1);
  }
SWIGINTERN void __SoapOp_set_indoc(struct __SoapOp *self,WsXmlDocH doc){
    soap_set_op_doc( self, doc, 1 );
  }
SWIGINTERN WsXmlDocH __SoapOp_outdoc(struct __SoapOp *self){
    return soap_get_op_doc(self, 0);
  }
SWIGINTERN void __SoapOp_set_outdoc(struct __SoapOp *self,WsXmlDocH doc){
    soap_set_op_doc( self, doc, 0 );
  }
SWIGINTERN struct __Soap *__SoapOp_soap(struct __SoapOp *self){
    return soap_get_op_soap(self);
  }
SWIGINTERN WsmanMessage *__SoapOp_msg(struct __SoapOp *self){
    return wsman_get_msg_from_op(self);
  }
SWIGINTERN unsigned long __SoapOp_maxsize(struct __SoapOp *self){
    return wsman_get_maxsize_from_op(self);
  }

SWIGINTERNINLINE VALUE
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return ULONG2NUM(value); 
}

SWIGINTERN WsContextH __Soap_create_context(struct __Soap *self){
    return ws_create_context(self);
  }
SWIGINTERN WsContextH __Soap_context(struct __Soap *self){
    return ws_get_soap_context(self);
  }
SWIGINTERN WsContextH __Soap_create_ep_context(struct __Soap *self,WsXmlDocH doc){
    return ws_create_ep_context( self, doc );
  }
SWIGINTERN WsXmlDocH _WS_CONTEXT_indoc(struct _WS_CONTEXT *self){
    return self->indoc;
  }
SWIGINTERN struct __Soap *_WS_CONTEXT_runtime(struct _WS_CONTEXT *self){
    return ws_context_get_runtime(self);
  }

SWIGINTERN VALUE SWIG_AUX_NUM2ULONG(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  unsigned long *res = (unsigned long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2ULONG(obj) : rb_big2ulong(obj);
  return obj;
}

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (VALUE obj, unsigned long *val) 
{
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    unsigned long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2ULONG), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (rb_funcall(obj, swig_lowerthan_id, 1, INT2FIX(0)) != Qfalse)
        return SWIG_OverflowError;
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}

SWIGINTERN void _WS_CONTEXT_set_enumIdleTimeout(struct _WS_CONTEXT *self,unsigned long timeout){
    ws_set_context_enumIdleTimeout(self, timeout);
  }
SWIGINTERN char const *_WS_CONTEXT_classname(struct _WS_CONTEXT *self){
    return wsman_get_class_name(self);
  }
SWIGINTERN char const *_WS_CONTEXT_method(struct _WS_CONTEXT *self){
    return wsman_get_method_name(self);
  }
SWIGINTERN hash_t *_WS_CONTEXT_method_args(struct _WS_CONTEXT *self,char const *resource_uri){
    return wsman_get_method_args(self, resource_uri);
  }
SWIGINTERN int _WS_CONTEXT_max_elements(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_max_elements(self, doc);
  }
SWIGINTERN unsigned long _WS_CONTEXT_max_envelope_size(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_max_envelope_size(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_fragment_string(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_fragment_string(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_selector(struct _WS_CONTEXT *self,WsXmlDocH doc,char const *name,int index){
    return wsman_get_selector(self, doc, name, index);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors_from_epr(struct _WS_CONTEXT *self,WsXmlNodeH epr_node){
    return wsman_get_selectors_from_epr(self, epr_node);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_selector_list(self, doc);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors_from_filter(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_selector_list_from_filter(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_action(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_action(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_resource_uri(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_resource_uri(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_option_set(struct _WS_CONTEXT *self,WsXmlDocH doc,char const *op){
    return wsman_get_option_set(self, doc, op);
  }
SWIGINTERN int _WS_CONTEXT_parse_enum_request(struct _WS_CONTEXT *self,WsEnumerateInfo *enumInfo){
    WsmanStatus status;
    wsman_status_init(&status);
    return wsman_parse_enum_request( self, enumInfo, &status);
    /* FIXME: return status */
  }
SWIGINTERN struct _WsmanStatus *new__WsmanStatus(int code,int detail,char const *msg){
    WsmanStatus *s = (WsmanStatus *)malloc(sizeof(WsmanStatus));
    wsman_status_init(s);
    if (code)
      s->fault_code = code;
    if (msg)
      s->fault_msg = strdup(msg);
    if (detail < 0
        || detail > OWSMAN_SYSTEM_ERROR) {
      SWIG_exception( SWIG_ValueError, "Bad fault detail" );
    }
    else {
      s->fault_detail_code = detail;
    }



    return s;
  }
SWIGINTERN char const *_WsmanStatus_to_s(struct _WsmanStatus *self){
    return self->fault_msg;
  }
SWIGINTERN void _WsmanStatus_set_code(struct _WsmanStatus *self,int code){ self->fault_code = code; }
SWIGINTERN int _WsmanStatus_code(struct _WsmanStatus *self){
    return self->fault_code;
  }
SWIGINTERN void _WsmanStatus_set_detail(struct _WsmanStatus *self,int detail){
    if (detail < 0
        || detail > OWSMAN_SYSTEM_ERROR) {
      SWIG_exception( SWIG_ValueError, "Bad fault detail" );
    }
    else {
      self->fault_detail_code = detail;
    }




  }
SWIGINTERN int _WsmanStatus_detail(struct _WsmanStatus *self){
    return self->fault_detail_code;
  }
SWIGINTERN void _WsmanStatus_set_msg(struct _WsmanStatus *self,char const *msg){
    if (self->fault_msg)
      free(self->fault_msg);
    if (msg)
      self->fault_msg = strdup(msg);
    else
      self->fault_msg = NULL;
  }
SWIGINTERN char const *_WsmanStatus_msg(struct _WsmanStatus *self){
    return self->fault_msg;
  }
SWIGINTERN WsXmlDocH _WsmanStatus_generate_fault(struct _WsmanStatus *self,WsXmlDocH doc){
    return wsman_generate_fault( doc, self->fault_code, self->fault_detail_code, self->fault_msg);
  }
SWIGINTERN struct _WsManFault *new__WsManFault(WsXmlDocH doc){
    WsManFault *fault = wsmc_fault_new();
    wsmc_get_fault_data(doc, fault);
    return fault;
  }
SWIGINTERN char const *_WsManFault_code(struct _WsManFault *self){
    return self->code;
  }
SWIGINTERN char const *_WsManFault_subcode(struct _WsManFault *self){
    return self->subcode;
  }
SWIGINTERN char const *_WsManFault_reason(struct _WsManFault *self){
    return self->reason;
  }
SWIGINTERN char const *_WsManFault_detail(struct _WsManFault *self){
    return self->fault_detail;
  }
SWIGINTERN int _WsManTransport_is_auth_method(struct _WsManTransport *self,int method){
    return wsman_is_auth_method((WsManClient *)self, method);
  }
SWIGINTERN void _WsManTransport_close(struct _WsManTransport *self){
    wsman_transport_close_transport((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_agent(struct _WsManTransport *self,char const *agent){
    wsman_transport_set_agent((WsManClient *)self, agent);
  }
SWIGINTERN char *_WsManTransport_agent(struct _WsManTransport *self){
    return wsman_transport_get_agent ((WsManClient *)self);
  }
SWIGINTERN char *_WsManTransport_get_username(struct _WsManTransport *self){
    return wsman_transport_get_userName((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_username(struct _WsManTransport *self,char *user_name){
    wsman_transport_set_userName((WsManClient *)self, user_name);
  }
SWIGINTERN char *_WsManTransport_get_password(struct _WsManTransport *self){
    return wsman_transport_get_password((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_password(struct _WsManTransport *self,char *password){
    wsman_transport_set_password((WsManClient *)self, password);
  }
SWIGINTERN char *_WsManTransport_get_proxy_username(struct _WsManTransport *self){
    return wsman_transport_get_proxy_username((WsManClient *)self );
  }
SWIGINTERN void _WsManTransport_set_proxy_username(struct _WsManTransport *self,char *proxy_username){
    wsman_transport_set_proxy_username((WsManClient *)self, proxy_username );
  }
SWIGINTERN char *_WsManTransport_get_proxy_password(struct _WsManTransport *self){
    return wsman_transport_get_proxy_password((WsManClient *)self );
  }
SWIGINTERN void _WsManTransport_set_proxy_password(struct _WsManTransport *self,char *proxy_password){
    wsman_transport_set_proxy_password((WsManClient *)self, proxy_password );
  }
SWIGINTERN void _WsManTransport_set_auth_method(struct _WsManTransport *self,char const *am){
    wsman_transport_set_auth_method((WsManClient *)self, am);
  }
SWIGINTERN char *_WsManTransport_auth_method(struct _WsManTransport *self){
    return wsman_transport_get_auth_method ((WsManClient *)self);
  }
SWIGINTERN char const *_WsManTransport_auth_name(int auth){
    return wsmc_transport_get_auth_name(auth);
  }
SWIGINTERN int _WsManTransport_auth_value(struct _WsManTransport *self){
    return wsmc_transport_get_auth_value((WsManClient *)self);
  }
SWIGINTERN char *_WsManTransport_error_string(int err){
    return wsman_transport_get_last_error_string(err);
  }
SWIGINTERN void _WsManTransport_set_timeout(struct _WsManTransport *self,unsigned long timeout){
    wsman_transport_set_timeout((WsManClient *)self, timeout);
  }
SWIGINTERN unsigned long _WsManTransport_timeout(struct _WsManTransport *self){
    return wsman_transport_get_timeout((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_verify_peer(struct _WsManTransport *self,VALUE rvalue){
    unsigned int value;
    if ((rvalue == Qfalse) || (rvalue == Qnil)) {
      value = 0;
    }
    else if ((TYPE(rvalue) == T_FIXNUM) && (FIX2INT(rvalue) == 0)) {
      value = 0;
    }
    else {
      value = 1;
    }









    wsman_transport_set_verify_peer((WsManClient *)self, value);
  }
SWIGINTERN unsigned int _WsManTransport_verify_peer(struct _WsManTransport *self){
    return wsman_transport_get_verify_peer((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_verify_host(struct _WsManTransport *self,VALUE rvalue){
    unsigned int value;
    if ((rvalue == Qfalse) || (rvalue == Qnil)) {
      value = 0;
    }
    else if ((TYPE(rvalue) == T_FIXNUM) && (FIX2INT(rvalue) == 0)) {
      value = 0;
    }
    else {
      value = 1;
    }









    wsman_transport_set_verify_host((WsManClient *)self, value);
  }
SWIGINTERN unsigned int _WsManTransport_verify_host(struct _WsManTransport *self){
    return wsman_transport_get_verify_host((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_proxy(struct _WsManTransport *self,char const *proxy){
    wsman_transport_set_proxy((WsManClient *)self, proxy);
  }
SWIGINTERN char *_WsManTransport_proxy(struct _WsManTransport *self){
    return wsman_transport_get_proxy((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_proxyauth(struct _WsManTransport *self,char const *pauth){
    wsman_transport_set_proxyauth((WsManClient *)self, pauth);
  }
SWIGINTERN char *_WsManTransport_proxyauth(struct _WsManTransport *self){
    return wsman_transport_get_proxyauth((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_cainfo(struct _WsManTransport *self,char const *cainfo){
    wsman_transport_set_cainfo((WsManClient *)self, cainfo);
  }
SWIGINTERN char *_WsManTransport_cainfo(struct _WsManTransport *self){
    return wsman_transport_get_cainfo((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_certhumbprint(struct _WsManTransport *self,char const *arg){
    wsman_transport_set_certhumbprint((WsManClient *)self, arg);
  }
SWIGINTERN char *_WsManTransport_certhumbprint(struct _WsManTransport *self){
    return wsman_transport_get_certhumbprint((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_capath(struct _WsManTransport *self,char const *capath){
    wsman_transport_set_capath((WsManClient *)self, capath);
  }
SWIGINTERN char *_WsManTransport_capath(struct _WsManTransport *self){
    return wsman_transport_get_capath((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_caoid(struct _WsManTransport *self,char const *oid){
    wsman_transport_set_caoid((WsManClient *)self, oid);
  }
SWIGINTERN char *_WsManTransport_caoid(struct _WsManTransport *self){
    return wsman_transport_get_caoid((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_cert(struct _WsManTransport *self,char const *cert){
    wsman_transport_set_cert((WsManClient *)self, cert);
  }
SWIGINTERN char *_WsManTransport_cert(struct _WsManTransport *self){
    return wsman_transport_get_cert((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_key(struct _WsManTransport *self,char const *key){
    wsman_transport_set_key((WsManClient *)self, key);
  }
SWIGINTERN char *_WsManTransport_key(struct _WsManTransport *self){
    return wsman_transport_get_key((WsManClient *)self);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (VALUE obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE VALUE
SWIG_From_unsigned_SS_int  (unsigned int value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}

SWIGINTERN client_opt_t *new_client_opt_t(void){
    client_opt_t *options = wsmc_options_init();
    return options;
  }
SWIGINTERN void client_opt_t_set_dump_request(client_opt_t *self){
    wsmc_set_action_option(self, 0x0010 );
  }
SWIGINTERN void client_opt_t_clear_dump_request(client_opt_t *self){
    wsmc_clear_action_option(self, 0x0010 );
  }
SWIGINTERN void client_opt_t_set_flags(client_opt_t *self,unsigned long flags){
    wsmc_set_action_option(self, flags);
  }
SWIGINTERN unsigned long client_opt_t_get_flags(client_opt_t *self){
    return wsmc_get_action_option(self);
  }
SWIGINTERN void client_opt_t_clear_flags(client_opt_t *self,unsigned long flags){
    wsmc_clear_action_option(self, flags);
  }
SWIGINTERN void client_opt_t_reset_flags(client_opt_t *self){
    wsmc_clear_action_option(self, ~0x0000);
  }
SWIGINTERN void client_opt_t_set_max_envelope_size(client_opt_t *self,unsigned long size){
    self->max_envelope_size = size;
  }
SWIGINTERN unsigned long client_opt_t_get_max_envelope_size(client_opt_t *self){
    return self->max_envelope_size;
  }
SWIGINTERN void client_opt_t_set_max_elements(client_opt_t *self,int elements){
    self->max_elements = elements;
  }
SWIGINTERN int client_opt_t_get_max_elements(client_opt_t *self){
    return self->max_elements;
  }
SWIGINTERN void client_opt_t_set_timeout(client_opt_t *self,unsigned long timeout){
    self->timeout = timeout;
  }
SWIGINTERN unsigned long client_opt_t_get_timeout(client_opt_t *self){
    return self->timeout;
  }
SWIGINTERN void client_opt_t_set_fragment(client_opt_t *self,char *fragment){
    wsmc_set_fragment(fragment, self);
  }
SWIGINTERN char const *client_opt_t_get_fragment(client_opt_t *self){
    return self->fragment;
  }
SWIGINTERN void client_opt_t_set_cim_namespace(client_opt_t *self,char *cim_namespace){
    wsmc_set_cim_ns(cim_namespace, self);
  }
SWIGINTERN char const *client_opt_t_get_cim_namespace(client_opt_t *self){
    return self->cim_ns;
  }
SWIGINTERN void client_opt_t_set_reference(client_opt_t *self,char const *reference){
    wsmc_set_reference(reference, self);
  }
SWIGINTERN char const *client_opt_t_get_reference(client_opt_t *self){
    return self->reference;
  }
SWIGINTERN void client_opt_t_add_option(client_opt_t *self,VALUE k,VALUE v){
    const char *key = as_string(k);
    const char *value = as_string(v);




    wsmc_add_option(self, key, value);
  }
SWIGINTERN void client_opt_t_set_options(client_opt_t *self,VALUE hash){
    self->options = value2hash(NULL, hash, 0);
  }
SWIGINTERN VALUE client_opt_t_get_options(client_opt_t *self){
    return hash2value(self->options);
  }
SWIGINTERN void client_opt_t_add_selector(client_opt_t *self,VALUE k,VALUE v){
    const char *key = as_string(k);
    KLASS_DECL(SwigClassEndPointReference,SWIGTYPE_p_epr_t);
    if (CLASS_OF(v) == KLASS_OF(SwigClassEndPointReference)) {
      epr_t *epr;
      SWIG_ConvertPtr(v, (void **)&epr, SWIGTYPE_p_epr_t, 0);
      wsmc_add_selector_epr(self, key, epr);
    }
    else {
      const char *value = as_string(v);
      wsmc_add_selector(self, key, value);
    }
  }
SWIGINTERN VALUE client_opt_t_get_selectors(client_opt_t *self){
    return kv_list_to_hash(self->selectors);
  }
SWIGINTERN void client_opt_t_add_property(client_opt_t *self,VALUE k,VALUE v){
    const char *key = as_string(k);
    KLASS_DECL(SwigClassEndPointReference,SWIGTYPE_p_epr_t);

    if (CLASS_OF(v) == KLASS_OF(SwigClassEndPointReference)) {
      const epr_t *epr;
      SWIG_ConvertPtr(v, (void **)&epr, SWIGTYPE_p_epr_t, 0);
      wsmc_add_property_epr(self, key, epr);
    }
    else {
      const char *value = as_string(v);
      wsmc_add_property(self, key, value);
    }
  }
SWIGINTERN VALUE client_opt_t_get_properties(client_opt_t *self){
    return kv_list_to_hash(self->properties);
  }
SWIGINTERN void client_opt_t_set_delivery_uri(client_opt_t *self,char const *delivery_uri){
    wsmc_set_delivery_uri(delivery_uri, self);
  }
SWIGINTERN char const *client_opt_t_delivery_uri(client_opt_t *self){
    return self->delivery_uri;
  }

#include <float.h>


#include <math.h>


/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN VALUE SWIG_AUX_NUM2DBL(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  double *res = (double *)(args[1]);
  *res = NUM2DBL(obj); (void)type;
  return obj;
}

SWIGINTERN int
SWIG_AsVal_double (VALUE obj, double *val)
{
  VALUE type = TYPE(obj);
  if ((type == T_FLOAT) || (type == T_FIXNUM) || (type == T_BIGNUM)) {
    double v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2DBL), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_float (VALUE obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (float)(v);
    }
  }  
  return res;
}

SWIGINTERN void client_opt_t_set_sub_expiry(client_opt_t *self,float event_subscription_expire){
	wsmc_set_sub_expiry(event_subscription_expire, self);
  }
SWIGINTERN float client_opt_t_sub_expiry(client_opt_t *self){
    return self->expires;
  }

  #define SWIG_From_double   rb_float_new 


SWIGINTERNINLINE VALUE
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}

SWIGINTERN void client_opt_t_set_heartbeat_interval(client_opt_t *self,float heartbeat_interval){
	wsmc_set_heartbeat_interval(heartbeat_interval, self);
  }
SWIGINTERN float client_opt_t_heartbeat_interval(client_opt_t *self){
    return self->heartbeat_interval;
  }
SWIGINTERN void client_opt_t_set_delivery_mode(client_opt_t *self,unsigned int delivery_mode){
    if (delivery_mode > 3)
      SWIG_exception( SWIG_ValueError, "Bad delivery mode" );
	
    wsmc_set_delivery_mode(delivery_mode, self);




  }
SWIGINTERN int client_opt_t_delivery_mode(client_opt_t *self){
    return self->delivery_mode;
  }
SWIGINTERN void client_opt_t_set_delivery_security_mode(client_opt_t *self,unsigned int delivery_sec_mode){
    if (delivery_sec_mode > 10)
      SWIG_exception( SWIG_ValueError, "Bad delivery security mode" );
    wsmc_set_delivery_security_mode(delivery_sec_mode, self);




  }
SWIGINTERN int client_opt_t_delivery_sec_mode(client_opt_t *self){
    return self->delivery_sec_mode;
  }
SWIGINTERN void client_opt_t_set_locale(client_opt_t *self,char const *locale){
	wsmc_set_locale(self, locale);
  }
SWIGINTERN char const *client_opt_t_locale(client_opt_t *self){
    return self->locale;
  }
SWIGINTERN struct _WsManClient *new__WsManClient__SWIG_0(char const *uri){
    struct _WsManClient *client = wsmc_create_from_uri( uri );
    if (client == NULL)
      SWIG_exception( SWIG_ValueError, "Can't create Openwsman::Client from given URI" );



    return client;
  }
SWIGINTERN struct _WsManClient *new__WsManClient__SWIG_1(char const *hostname,int const port,char const *path,char const *scheme,char const *username,char const *password){
    struct _WsManClient *client = wsmc_create( hostname, port, path, scheme, username, password );
    if (client == NULL)
      SWIG_exception( SWIG_ValueError, "Can't create Openwsman::Client from given values" );



    return client;
  }
SWIGINTERN void _WsManClient_set_dumpfile(struct _WsManClient *self,FILE *f){
    wsmc_set_dumpfile( self, f );
  }
SWIGINTERN long _WsManClient_response_code(struct _WsManClient *self){
    return wsmc_get_response_code( self );
  }
SWIGINTERN char *_WsManClient_scheme(struct _WsManClient *self){
    return wsmc_get_scheme( self );
  }
SWIGINTERN char *_WsManClient_host(struct _WsManClient *self){
    return wsmc_get_hostname( self );
  }
SWIGINTERN int _WsManClient_port(struct _WsManClient *self){
    return wsmc_get_port( self );
  }
SWIGINTERN char *_WsManClient_path(struct _WsManClient *self){
    return wsmc_get_path( self );
  }
SWIGINTERN char *_WsManClient_user(struct _WsManClient *self){
    return wsmc_get_user( self );
  }
SWIGINTERN char *_WsManClient_password(struct _WsManClient *self){
    return wsmc_get_password( self );
  }
SWIGINTERN WsManTransport *_WsManClient_transport(struct _WsManClient *self){
    wsmc_transport_init(self, NULL);
    wsmc_transport_set_auth_request_func( self, auth_request_callback );

    return (WsManTransport *)self;
  }
SWIGINTERN int _WsManClient_send_request(struct _WsManClient *self,WsXmlDocH request){
    return wsman_send_request(self, request);
  }
SWIGINTERN WsXmlDocH _WsManClient_build_envelope_from_response(struct _WsManClient *self){
    return wsmc_build_envelope_from_response(self);
  }
SWIGINTERN char *_WsManClient_encoding(struct _WsManClient *self){
    return wsmc_get_encoding(self);
  }
SWIGINTERN void _WsManClient_set_encoding(struct _WsManClient *self,char const *encoding){
    wsmc_set_encoding(self, encoding);
  }
SWIGINTERN WsXmlDocH _WsManClient_identify(struct _WsManClient *self,client_opt_t *options){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_identify_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_get_from_epr(struct _WsManClient *self,client_opt_t *options,epr_t *epr){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.epr = epr;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_get_from_epr_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_delete_from_epr(struct _WsManClient *self,client_opt_t *options,epr_t *epr){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.epr = epr;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_delete_from_epr_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_enumerate(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char *resource_uri){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.filter = filter;
    args.resource_uri = resource_uri;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_enumerate_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_pull(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri,char const *context){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.filter = filter;
    args.resource_uri = resource_uri;
    args.context = context;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_pull_thread, &args, RUBY_UBF_IO, 0);






  }

#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN VALUE SWIG_AUX_NUM2ULL(VALUE arg)
{
  VALUE *args = (VALUE *)arg;
  VALUE obj = args[0];
  VALUE type = TYPE(obj);
  long long *res = (long long *)(args[1]);
  *res = type == T_FIXNUM ? NUM2ULL(obj) : rb_big2ull(obj);
  return obj;
}

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long (VALUE obj, unsigned long long *val)
{ 
  VALUE type = TYPE(obj);
  if ((type == T_FIXNUM) || (type == T_BIGNUM)) {
    unsigned long long v;
    VALUE a[2];
    a[0] = obj;
    a[1] = (VALUE)(&v);
    if (rb_rescue(VALUEFUNC(SWIG_AUX_NUM2ULL), (VALUE)a, VALUEFUNC(SWIG_ruby_failed), 0) != Qnil) {
      if (rb_funcall(obj, swig_lowerthan_id, 1, INT2FIX(0)) != Qfalse)
        return SWIG_OverflowError;
      if (val) *val = v;
      return SWIG_OK;
    }
  }
  return SWIG_TypeError;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_size_t (VALUE obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long (obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}

SWIGINTERN WsXmlDocH _WsManClient_create(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *data,size_t size,char const *encoding){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;
    args.data = data;
    args.size = size;
    args.encoding = encoding;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_create_fromtext_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_put(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *data,size_t size,char const *encoding){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;
    args.data = data;
    args.size = size;
    args.encoding = encoding;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_put_fromtext_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_release(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *context){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;
    args.context = context;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_release_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_get(struct _WsManClient *self,client_opt_t *options,char const *resource_uri){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_get_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_delete(struct _WsManClient *self,client_opt_t *options,char const *resource_uri){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_delete_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_invoke(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *method,WsXmlDocH data){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;
    args.method = method;
    args.method_args = data;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_invoke_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_subscribe(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.filter = filter;
    args.resource_uri = resource_uri;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_subscribe_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_unsubscribe(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri,char const *identifier){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.filter = filter;
    args.resource_uri = resource_uri;
    args.identifier = identifier;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_unsubscribe_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN WsXmlDocH _WsManClient_renew(struct _WsManClient *self,client_opt_t *options,char *resource_uri,char *identifier){

    wsmc_action_args_t args;
    args.client = self;
    args.options = options;
    args.resource_uri = resource_uri;
    args.identifier = identifier;

    return (WsXmlDocH)rb_thread_call_without_gvl((void * (*)(void *))ruby_renew_thread, &args, RUBY_UBF_IO, 0);






  }
SWIGINTERN char *_WsManClient_fault_string(struct _WsManClient *self){
    return wsmc_get_fault_string(self);
  }
SWIGINTERN int _WsManClient_last_error(struct _WsManClient *self){
     return wsmc_get_last_error(self);
   }
static swig_class SwigClassXmlNs;

static swig_class SwigClassXmlDoc;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_XmlDoc_allocate(VALUE self)
#else
_wrap_XmlDoc_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p__WsXmlDoc);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_XmlDoc(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = (char *)NULL ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  struct _WsXmlDoc *result = 0 ;
  
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","_WsXmlDoc", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","_WsXmlDoc", 2, argv[1] ));
    }
    arg2 = (char *)(buf2);
  }
  result = (struct _WsXmlDoc *)new__WsXmlDoc((char const *)arg1,(char const *)arg2);
  DATA_PTR(self) = result;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN void delete__WsXmlDoc(struct _WsXmlDoc *self){
    ws_xml_destroy_doc( self );
  }
SWIGINTERN void
free__WsXmlDoc(void *self) {
    struct _WsXmlDoc *arg1 = (struct _WsXmlDoc *)self;
    delete__WsXmlDoc(arg1);
}

SWIGINTERN VALUE
_wrap_XmlDoc_string(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","string", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (char *)_WsXmlDoc_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_encode(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  char *arg2 = (char *)"utf-8" ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","encode", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  if (argc > 0) {
    res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","encode", 2, argv[0] ));
    }
    arg2 = (char *)(buf2);
  }
  result = (char *)_WsXmlDoc_encode(arg1,(char const *)arg2);
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  free(result);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_dump_file(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","dump_file", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  {
    rb_fptr_t *fptr;
    
    Check_Type(argv[0], T_FILE);
    GetOpenFile(argv[0], fptr);
    /*rb_io_check_writable(fptr);*/
    
    arg2 = rb_io_stdio_file(fptr);
    
    
    
  }
  _WsXmlDoc_dump_file(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_root(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","root", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (WsXmlNodeH)_WsXmlDoc_root(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_envelope(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","envelope", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (WsXmlNodeH)_WsXmlDoc_envelope(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_header(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","header", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (WsXmlNodeH)_WsXmlDoc_header(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_body(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","body", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (WsXmlNodeH)_WsXmlDoc_body(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_element(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","element", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","element", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  result = (WsXmlNodeH)_WsXmlDoc_element(arg1,(char const *)arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_context(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","context", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (char *)_WsXmlDoc_context(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_generate_fault(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  WsmanStatus *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","generate_fault", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsmanStatus *","generate_fault", 2, argv[0] )); 
  }
  arg2 = (WsmanStatus *)(argp2);
  result = (WsXmlDocH)_WsXmlDoc_generate_fault(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_faultq___(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","is_fault", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (int)_WsXmlDoc_is_fault(arg1);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_fault(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsManFault *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","fault", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (WsManFault *)_WsXmlDoc_fault(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManFault, SWIG_POINTER_OWN |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_create_response_envelope(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  char *arg2 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","create_response_envelope", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  if (argc > 0) {
    res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","create_response_envelope", 2, argv[0] ));
    }
    arg2 = (char *)(buf2);
  }
  result = (WsXmlDocH)_WsXmlDoc_create_response_envelope(arg1,(char const *)arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlDoc_end_of_sequenceq___(int argc, VALUE *argv, VALUE self) {
  struct _WsXmlDoc *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsXmlDoc *","is_end_of_sequence", 1, self )); 
  }
  arg1 = (struct _WsXmlDoc *)(argp1);
  result = (int)_WsXmlDoc_is_end_of_sequence(arg1);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassXmlNode;

SWIGINTERN void delete___WsXmlNode(struct __WsXmlNode *self){
    ws_xml_unlink_node(self);
  }
SWIGINTERN void
free___WsXmlNode(void *self) {
    struct __WsXmlNode *arg1 = (struct __WsXmlNode *)self;
    delete___WsXmlNode(arg1);
}

SWIGINTERN VALUE
_wrap_XmlNode_string(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","string", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (char *)__WsXmlNode_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_dump_file(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","dump_file", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  {
    rb_fptr_t *fptr;
    
    Check_Type(argv[0], T_FILE);
    GetOpenFile(argv[0], fptr);
    /*rb_io_check_writable(fptr);*/
    
    arg2 = rb_io_stdio_file(fptr);
    
    
    
  }
  __WsXmlNode_dump_file(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_equal(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  WsXmlNodeH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","equal", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlNodeH","equal", 2, argv[0] )); 
  }
  arg2 = (WsXmlNodeH)(argp2);
  result = (int)__WsXmlNode_equal(arg1,arg2);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_text(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","text", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (char *)__WsXmlNode_text(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_texte___(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","set_text", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_text", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  __WsXmlNode_set_text(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_doc(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","doc", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (WsXmlDocH)__WsXmlNode_doc(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_parent(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","parent", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (WsXmlNodeH)__WsXmlNode_parent(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_child(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","child", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (WsXmlNodeH)__WsXmlNode_child(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_name(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","name", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (char *)__WsXmlNode_name(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_namee___(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","set_name", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_name", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  __WsXmlNode_set_name(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_ns(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","ns", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (char *)__WsXmlNode_ns(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_nse___(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","set_ns", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_ns", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  __WsXmlNode_set_ns(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_prefix(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","prefix", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (char *)__WsXmlNode_prefix(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_lange___(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","set_lang", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_lang", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  __WsXmlNode_set_lang(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


/*
  Document-method: Openwsman::XmlNode.find

  call-seq:
    find(ns, name, recursive=1) -> XmlNode

Find an element in the class.
*/
SWIGINTERN VALUE
_wrap_XmlNode_find(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  int arg4 = 1 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","find", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","find", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","find", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    ecode4 = SWIG_AsVal_int(argv[2], &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","find", 4, argv[2] ));
    } 
    arg4 = (int)(val4);
  }
  result = (WsXmlNodeH)__WsXmlNode_find(arg1,(char const *)arg2,(char const *)arg3,arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_next(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  int arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","next", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  if (argc > 0) {
    ecode2 = SWIG_AsVal_int(argv[0], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","next", 2, argv[0] ));
    } 
    arg2 = (int)(val2);
  }
  result = (WsXmlNodeH)__WsXmlNode_next(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


/*
  Document-method: Openwsman::XmlNode.size

  call-seq:
    size(name=nil, ns=nil) -> int

Size or Length of the XmlNode.
*/
SWIGINTERN VALUE
_wrap_XmlNode_size(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = (char *)NULL ;
  char *arg3 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","size", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  if (argc > 0) {
    res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","size", 2, argv[0] ));
    }
    arg2 = (char *)(buf2);
  }
  if (argc > 1) {
    res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","size", 3, argv[1] ));
    }
    arg3 = (char *)(buf3);
  }
  result = (int)__WsXmlNode_size(arg1,(char const *)arg2,(char const *)arg3);
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_add__SWIG_0(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","add", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","add", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","add", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","add", 4, argv[2] ));
    }
    arg4 = (char *)(buf4);
  }
  result = (WsXmlNodeH)__WsXmlNode_add__SWIG_0(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_add_before(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","add_before", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","add_before", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","add_before", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  if (argc > 2) {
    res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","add_before", 4, argv[2] ));
    }
    arg4 = (char *)(buf4);
  }
  result = (WsXmlNodeH)__WsXmlNode_add_before(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_add__SWIG_1(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  WsXmlNodeH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","add", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlNodeH","add", 2, argv[0] )); 
  }
  arg2 = (WsXmlNodeH)(argp2);
  result = (WsXmlNodeH)__WsXmlNode_add__SWIG_1(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE _wrap_XmlNode_add(int nargs, VALUE *args, VALUE self) {
  int argc;
  VALUE argv[5];
  int ii;
  
  argc = nargs + 1;
  argv[0] = self;
  if (argc > 5) SWIG_fail;
  for (ii = 1; (ii < argc); ++ii) {
    argv[ii] = args[ii-1];
  }
  if (argc == 2) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p___WsXmlNode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p___WsXmlNode, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_XmlNode_add__SWIG_1(nargs, args, self);
      }
    }
  }
  if ((argc >= 3) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p___WsXmlNode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          if (argc <= 3) {
            return _wrap_XmlNode_add__SWIG_0(nargs, args, self);
          }
          int res = SWIG_AsCharPtrAndSize(argv[3], 0, NULL, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_XmlNode_add__SWIG_0(nargs, args, self);
          }
        }
      }
    }
  }
  
fail:
  Ruby_Format_OverloadedError( argc, 5, "add", 
    "    WsXmlNodeH add(char const *ns, char const *name, char const *text)\n"
    "    WsXmlNodeH add(WsXmlNodeH node)\n");
  
  return Qnil;
}


/*
  Document-method: Openwsman::XmlNode.each

  call-seq:
    each(name=nil, ns=nil)

Iterate thru each element in the XmlNode.  A block must be provided.
*/
SWIGINTERN VALUE
_wrap_XmlNode_each(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = (char *)NULL ;
  char *arg3 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if ((argc < 0) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","each", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  if (argc > 0) {
    res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","each", 2, argv[0] ));
    }
    arg2 = (char *)(buf2);
  }
  if (argc > 1) {
    res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","each", 3, argv[1] ));
    }
    arg3 = (char *)(buf3);
  }
  __WsXmlNode_each(arg1,(char const *)arg2,(char const *)arg3);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_get__SWIG_0(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  int arg2 ;
  char *arg3 = (char *)NULL ;
  char *arg4 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","get", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","get", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  if (argc > 1) {
    res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","get", 3, argv[1] ));
    }
    arg3 = (char *)(buf3);
  }
  if (argc > 2) {
    res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","get", 4, argv[2] ));
    }
    arg4 = (char *)(buf4);
  }
  result = (WsXmlNodeH)__WsXmlNode_get__SWIG_0(arg1,arg2,(char const *)arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_get__SWIG_1(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  WsXmlNodeH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","get", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","get", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  if (argc > 1) {
    res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","get", 3, argv[1] ));
    }
    arg3 = (char *)(buf3);
  }
  result = (WsXmlNodeH)__WsXmlNode_get__SWIG_1(arg1,(char const *)arg2,(char const *)arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE _wrap_XmlNode_get(int nargs, VALUE *args, VALUE self) {
  int argc;
  VALUE argv[5];
  int ii;
  
  argc = nargs + 1;
  argv[0] = self;
  if (argc > 5) SWIG_fail;
  for (ii = 1; (ii < argc); ++ii) {
    argv[ii] = args[ii-1];
  }
  if ((argc >= 2) && (argc <= 4)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p___WsXmlNode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        if (argc <= 2) {
          return _wrap_XmlNode_get__SWIG_0(nargs, args, self);
        }
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          if (argc <= 3) {
            return _wrap_XmlNode_get__SWIG_0(nargs, args, self);
          }
          int res = SWIG_AsCharPtrAndSize(argv[3], 0, NULL, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_XmlNode_get__SWIG_0(nargs, args, self);
          }
        }
      }
    }
  }
  if ((argc >= 2) && (argc <= 3)) {
    int _v = 0;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p___WsXmlNode, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsCharPtrAndSize(argv[1], 0, NULL, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        if (argc <= 2) {
          return _wrap_XmlNode_get__SWIG_1(nargs, args, self);
        }
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_XmlNode_get__SWIG_1(nargs, args, self);
        }
      }
    }
  }
  
fail:
  Ruby_Format_OverloadedError( argc, 5, "get", 
    "    WsXmlNodeH get(int i, char const *name, char const *ns)\n"
    "    WsXmlNodeH get(char const *name, char const *ns)\n");
  
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_attr(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  VALUE arg2 = Qnil ;
  VALUE arg3 = Qnil ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlAttrH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","attr", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  if (argc > 0) {
    arg2 = argv[0];
  }
  if (argc > 1) {
    arg3 = argv[1];
  }
  result = (WsXmlAttrH)__WsXmlNode_attr(arg1,arg2,arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_attr_count(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","attr_count", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  result = (int)__WsXmlNode_attr_count(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_attr_find(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  WsXmlAttrH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","attr_find", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","attr_find", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","attr_find", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  result = (WsXmlAttrH)__WsXmlNode_attr_find(arg1,(char const *)arg2,(char const *)arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_attr_add(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlAttrH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","attr_add", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","attr_add", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","attr_add", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","attr_add", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  result = (WsXmlAttrH)__WsXmlNode_attr_add(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_epr(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  epr_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","epr", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","epr", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","epr", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","epr", 4, argv[2] ));
  } 
  arg4 = (int)(val4);
  result = (epr_t *)__WsXmlNode_epr(arg1,(char const *)arg2,(char const *)arg3,arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_epr_t, 0 |  0 );
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlNode_each_attr(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlNode *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlNode *","each_attr", 1, self )); 
  }
  arg1 = (struct __WsXmlNode *)(argp1);
  __WsXmlNode_each_attr(arg1);
  return Qnil;
fail:
  return Qnil;
}


static swig_class SwigClassXmlAttr;

SWIGINTERN VALUE
_wrap_XmlAttr_name(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlAttr *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlAttr *","name", 1, self )); 
  }
  arg1 = (struct __WsXmlAttr *)(argp1);
  result = (char *)__WsXmlAttr_name(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlAttr_ns(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlAttr *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlAttr *","ns", 1, self )); 
  }
  arg1 = (struct __WsXmlAttr *)(argp1);
  result = (char *)__WsXmlAttr_ns(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlAttr_value(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlAttr *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlAttr *","value", 1, self )); 
  }
  arg1 = (struct __WsXmlAttr *)(argp1);
  result = (char *)__WsXmlAttr_value(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_XmlAttr_remove(int argc, VALUE *argv, VALUE self) {
  struct __WsXmlAttr *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsXmlAttr *","remove", 1, self )); 
  }
  arg1 = (struct __WsXmlAttr *)(argp1);
  __WsXmlAttr_remove(arg1);
  return Qnil;
fail:
  return Qnil;
}


static swig_class SwigClassEndPointReference;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_EndPointReference_allocate(VALUE self)
#else
_wrap_EndPointReference_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_epr_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_EndPointReference(int argc, VALUE *argv, VALUE self) {
  VALUE arg1 = 0 ;
  VALUE arg2 = Qnil ;
  VALUE arg3 = Qnil ;
  epr_t *result = 0 ;
  
  if ((argc < 1) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  arg1 = argv[0];
  if (argc > 1) {
    arg2 = argv[1];
  }
  if (argc > 2) {
    arg3 = argv[2];
  }
  result = (epr_t *)new_epr_t(arg1,arg2,arg3);
  DATA_PTR(self) = result;
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete_epr_t(epr_t *self){
    epr_destroy( self );
  }
SWIGINTERN void
free_epr_t(void *self) {
    epr_t *arg1 = (epr_t *)self;
    delete_epr_t(arg1);
}

/*
  Document-method: Openwsman::EndPointReference.clone

  call-seq:
    clone(epr) -> EndPointReference

Create a duplicate of the class.
*/
SWIGINTERN VALUE
_wrap_EndPointReference_clone(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  epr_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  epr_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","clone", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "epr_t const *","clone", 2, argv[0] )); 
  }
  arg2 = (epr_t *)(argp2);
  result = (epr_t *)epr_t_clone(arg1,(epr_t const *)arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_epr_t, SWIG_POINTER_OWN |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_add_selector(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","add_selector", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","add_selector", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","add_selector", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  epr_t_add_selector(arg1,(char const *)arg2,(char const *)arg3);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_serialize(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  WsXmlNodeH arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int val5 ;
  int ecode5 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","serialize", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlNodeH","serialize", 2, argv[0] )); 
  }
  arg2 = (WsXmlNodeH)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","serialize", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","serialize", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_int(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "int","serialize", 5, argv[3] ));
  } 
  arg5 = (int)(val5);
  result = (int)epr_t_serialize(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
  vresult = SWIG_From_int((int)(result));
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_cmp(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  epr_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","cmp", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "epr_t const *","cmp", 2, argv[0] )); 
  }
  arg2 = (epr_t *)(argp2);
  result = (int)epr_t_cmp(arg1,(epr_t const *)arg2);
  vresult = (result == 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_to_xml(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  char *arg2 = (char *)NULL ;
  char *arg3 = (char *)NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","to_xml", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  if (argc > 0) {
    res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","to_xml", 2, argv[0] ));
    }
    arg2 = (char *)(buf2);
  }
  if (argc > 1) {
    res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","to_xml", 3, argv[1] ));
    }
    arg3 = (char *)(buf3);
  }
  result = (char *)epr_t_to_xml(arg1,(char const *)arg2,(char const *)arg3);
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_to_s(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","string", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (char *)epr_t_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_selector_count(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","selector_count", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (int)epr_t_selector_count(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_resource_uri(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","resource_uri", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (char *)epr_t_resource_uri(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_selector(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  VALUE arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","selector", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  arg2 = argv[0];
  result = (char *)epr_t_selector(arg1,arg2);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_selector_names(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","selector_names", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (VALUE)epr_t_selector_names(arg1);
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


/*
  Document-method: Openwsman::EndPointReference.each

  call-seq:
    each

Iterate thru each element in the EndPointReference.  A block must be provided.
*/
SWIGINTERN VALUE
_wrap_EndPointReference_each(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","each", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  epr_t_each(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_classname(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","classname", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (char *)epr_t_classname(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_namespace(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","namespace", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (char *)epr_t_namespace(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EndPointReference_prefix(int argc, VALUE *argv, VALUE self) {
  epr_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "epr_t *","prefix", 1, self )); 
  }
  arg1 = (epr_t *)(argp1);
  result = (char *)epr_t_prefix(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassFilter;

SWIGINTERN VALUE
_wrap_Filter_resultClass_set(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","resultClass", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","resultClass", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->resultClass);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->resultClass = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->resultClass = 0;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_resultClass_get(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","resultClass", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  result = (char *) ((arg1)->resultClass);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_assocClass_set(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","assocClass", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","assocClass", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  free((char*)arg1->assocClass);
  if (arg2) {
    size_t size = strlen((const char *)(arg2)) + 1;
    arg1->assocClass = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
  } else {
    arg1->assocClass = 0;
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_assocClass_get(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","assocClass", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  result = (char *) ((arg1)->assocClass);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_Filter_allocate(VALUE self)
#else
_wrap_Filter_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_filter_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_Filter(int argc, VALUE *argv, VALUE self) {
  filter_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  result = (filter_t *)new_filter_t();
  DATA_PTR(self) = result;
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete_filter_t(filter_t *self){
    filter_destroy( self );
  }
SWIGINTERN void
free_filter_t(void *self) {
    filter_t *arg1 = (filter_t *)self;
    delete_filter_t(arg1);
}

SWIGINTERN VALUE
_wrap_Filter_associators(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  VALUE arg2 = Qnil ;
  VALUE arg3 = Qnil ;
  VALUE arg4 = Qnil ;
  VALUE arg5 = Qnil ;
  VALUE arg6 = Qnil ;
  VALUE arg7 = Qnil ;
  VALUE arg8 = Qnil ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","associators", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  if (argc > 0) {
    arg2 = argv[0];
  }
  if (argc > 1) {
    arg3 = argv[1];
  }
  if (argc > 2) {
    arg4 = argv[2];
  }
  if (argc > 3) {
    arg5 = argv[3];
  }
  if (argc > 4) {
    arg6 = argv[4];
  }
  if (argc > 5) {
    arg7 = argv[5];
  }
  if (argc > 6) {
    arg8 = argv[6];
  }
  result = (int)filter_t_associators(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_references(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  VALUE arg2 = Qnil ;
  VALUE arg3 = Qnil ;
  VALUE arg4 = Qnil ;
  VALUE arg5 = Qnil ;
  VALUE arg6 = Qnil ;
  VALUE arg7 = Qnil ;
  VALUE arg8 = Qnil ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 7)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","references", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  if (argc > 0) {
    arg2 = argv[0];
  }
  if (argc > 1) {
    arg3 = argv[1];
  }
  if (argc > 2) {
    arg4 = argv[2];
  }
  if (argc > 3) {
    arg5 = argv[3];
  }
  if (argc > 4) {
    arg6 = argv[4];
  }
  if (argc > 5) {
    arg7 = argv[5];
  }
  if (argc > 6) {
    arg8 = argv[6];
  }
  result = (int)filter_t_references(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_simple(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","simple", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","simple", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","simple", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  result = (int)filter_t_simple(arg1,(char const *)arg2,(char const *)arg3);
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_xpath(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","xpath", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","xpath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  result = (int)filter_t_xpath(arg1,(char const *)arg2);
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_cql(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","cql", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","cql", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  result = (int)filter_t_cql(arg1,(char const *)arg2);
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Filter_wql(int argc, VALUE *argv, VALUE self) {
  filter_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "filter_t *","wql", 1, self )); 
  }
  arg1 = (filter_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wql", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  result = (int)filter_t_wql(arg1,(char const *)arg2);
  vresult = SWIG_From_int((int)(result));
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


static swig_class SwigClassEnumerateInfo;

SWIGINTERN void delete___WsEnumerateInfo(struct __WsEnumerateInfo *self){
  }
SWIGINTERN void
free___WsEnumerateInfo(void *self) {
    struct __WsEnumerateInfo *arg1 = (struct __WsEnumerateInfo *)self;
    delete___WsEnumerateInfo(arg1);
}

SWIGINTERN VALUE
_wrap_EnumerateInfo_max_items(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","max_items", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (int)__WsEnumerateInfo_max_items(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_max_itemse___(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","set_max_items", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_max_items", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  __WsEnumerateInfo_set_max_items(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_flags(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","flags", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (int)__WsEnumerateInfo_flags(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_epr_to(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","epr_to", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (char *)__WsEnumerateInfo_epr_to(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_epr_uri(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","epr_uri", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (char *)__WsEnumerateInfo_epr_uri(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_encoding(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","encoding", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (char *)__WsEnumerateInfo_encoding(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_filter(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  filter_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","filter", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (filter_t *)__WsEnumerateInfo_filter(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_t, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_index(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","index", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (int)__WsEnumerateInfo_index(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_indexe___(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","set_index", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_index", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  __WsEnumerateInfo_set_index(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_total_items(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","total_items", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (int)__WsEnumerateInfo_total_items(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_total_itemse___(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","set_total_items", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_total_items", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  __WsEnumerateInfo_set_total_items(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_pull_result(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","pull_result", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  result = (WsXmlDocH)__WsEnumerateInfo_pull_result(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_EnumerateInfo_pull_resulte___(int argc, VALUE *argv, VALUE self) {
  struct __WsEnumerateInfo *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __WsEnumerateInfo *","set_pull_result", 1, self )); 
  }
  arg1 = (struct __WsEnumerateInfo *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","set_pull_result", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  __WsEnumerateInfo_set_pull_result(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


static swig_class SwigClassSoapOp;

SWIGINTERN VALUE
_wrap_SoapOp_indoc(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","indoc", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  result = (WsXmlDocH)__SoapOp_indoc(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_indoce___(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","set_indoc", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","set_indoc", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  __SoapOp_set_indoc(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_outdoc(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","outdoc", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  result = (WsXmlDocH)__SoapOp_outdoc(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_outdoce___(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","set_outdoc", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","set_outdoc", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  __SoapOp_set_outdoc(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_soap(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct __Soap *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","soap", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  result = (struct __Soap *)__SoapOp_soap(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___Soap, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_msg(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsmanMessage *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","msg", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  result = (WsmanMessage *)__SoapOp_msg(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_WsmanMessage, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_SoapOp_maxsize(int argc, VALUE *argv, VALUE self) {
  struct __SoapOp *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __SoapOp *","maxsize", 1, self )); 
  }
  arg1 = (struct __SoapOp *)(argp1);
  result = (unsigned long)__SoapOp_maxsize(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassSoap;

SWIGINTERN void delete___Soap(struct __Soap *self){
    soap_destroy(self);
  }
SWIGINTERN void
free___Soap(void *self) {
    struct __Soap *arg1 = (struct __Soap *)self;
    delete___Soap(arg1);
}

SWIGINTERN VALUE
_wrap_Soap_create_context(int argc, VALUE *argv, VALUE self) {
  struct __Soap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsContextH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___Soap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __Soap *","create_context", 1, self )); 
  }
  arg1 = (struct __Soap *)(argp1);
  result = (WsContextH)__Soap_create_context(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Soap_context(int argc, VALUE *argv, VALUE self) {
  struct __Soap *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsContextH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___Soap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __Soap *","context", 1, self )); 
  }
  arg1 = (struct __Soap *)(argp1);
  result = (WsContextH)__Soap_context(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Soap_create_ep_context(int argc, VALUE *argv, VALUE self) {
  struct __Soap *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  WsContextH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p___Soap, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct __Soap *","create_ep_context", 1, self )); 
  }
  arg1 = (struct __Soap *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","create_ep_context", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  result = (WsContextH)__Soap_create_ep_context(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassContext;

SWIGINTERN void delete__WS_CONTEXT(struct _WS_CONTEXT *self){
    ws_destroy_context(self);
  }
SWIGINTERN void
free__WS_CONTEXT(void *self) {
    struct _WS_CONTEXT *arg1 = (struct _WS_CONTEXT *)self;
    delete__WS_CONTEXT(arg1);
}

SWIGINTERN VALUE
_wrap_Context_indoc(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","indoc", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  result = (WsXmlDocH)_WS_CONTEXT_indoc(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_runtime(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct __Soap *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","runtime", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  result = (struct __Soap *)_WS_CONTEXT_runtime(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___Soap, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_enum_idle_timeoute___(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","set_enumIdleTimeout", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","set_enumIdleTimeout", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  _WS_CONTEXT_set_enumIdleTimeout(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_classname(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","classname", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  result = (char *)_WS_CONTEXT_classname(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_method(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","method", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  result = (char *)_WS_CONTEXT_method(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_method_args(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","method_args", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","method_args", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  result = (hash_t *)_WS_CONTEXT_method_args(arg1,(char const *)arg2);
  {
    vresult = hash2value(result);
    
  }
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_max_elements(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","max_elements", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","max_elements", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (int)_WS_CONTEXT_max_elements(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_max_envelope_size(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","max_envelope_size", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","max_envelope_size", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (unsigned long)_WS_CONTEXT_max_envelope_size(arg1,arg2);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_fragment_string(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","fragment_string", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","fragment_string", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (char *)_WS_CONTEXT_fragment_string(arg1,arg2);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_selector(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  char *arg3 = 0 ;
  int arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int val4 ;
  int ecode4 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","selector", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","selector", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","selector", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  ecode4 = SWIG_AsVal_int(argv[2], &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), Ruby_Format_TypeError( "", "int","selector", 4, argv[2] ));
  } 
  arg4 = (int)(val4);
  result = (char *)_WS_CONTEXT_selector(arg1,arg2,(char const *)arg3,arg4);
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_selectors_from_epr(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlNodeH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","selectors_from_epr", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlNodeH","selectors_from_epr", 2, argv[0] )); 
  }
  arg2 = (WsXmlNodeH)(argp2);
  result = (hash_t *)_WS_CONTEXT_selectors_from_epr(arg1,arg2);
  {
    vresult = hash2value(result);
    
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_selectors(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","selectors", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","selectors", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (hash_t *)_WS_CONTEXT_selectors(arg1,arg2);
  {
    vresult = hash2value(result);
    
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_selectors_from_filter(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  hash_t *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","selectors_from_filter", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","selectors_from_filter", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (hash_t *)_WS_CONTEXT_selectors_from_filter(arg1,arg2);
  {
    vresult = hash2value(result);
    
  }
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_action(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","action", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","action", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (char *)_WS_CONTEXT_action(arg1,arg2);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_resource_uri(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","resource_uri", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  if (argc > 0) {
    res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","resource_uri", 2, argv[0] )); 
    }
    arg2 = (WsXmlDocH)(argp2);
  }
  result = (char *)_WS_CONTEXT_resource_uri(arg1,arg2);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_option_set(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","option_set", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","option_set", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","option_set", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  result = (char *)_WS_CONTEXT_option_set(arg1,arg2,(char const *)arg3);
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Context_parse_enum_request(int argc, VALUE *argv, VALUE self) {
  struct _WS_CONTEXT *arg1 = 0 ;
  WsEnumerateInfo *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WS_CONTEXT *","parse_enum_request", 1, self )); 
  }
  arg1 = (struct _WS_CONTEXT *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_WsEnumerateInfo, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsEnumerateInfo *","parse_enum_request", 2, argv[0] )); 
  }
  arg2 = (WsEnumerateInfo *)(argp2);
  result = (int)_WS_CONTEXT_parse_enum_request(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassStatus;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_Status_allocate(VALUE self)
#else
_wrap_Status_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p__WsmanStatus);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_Status(int argc, VALUE *argv, VALUE self) {
  int arg1 = 0 ;
  int arg2 = 0 ;
  char *arg3 = (char *)NULL ;
  int val1 ;
  int ecode1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  struct _WsmanStatus *result = 0 ;
  
  if ((argc < 0) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  if (argc > 0) {
    ecode1 = SWIG_AsVal_int(argv[0], &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","_WsmanStatus", 1, argv[0] ));
    } 
    arg1 = (int)(val1);
  }
  if (argc > 1) {
    ecode2 = SWIG_AsVal_int(argv[1], &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","_WsmanStatus", 2, argv[1] ));
    } 
    arg2 = (int)(val2);
  }
  if (argc > 2) {
    res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","_WsmanStatus", 3, argv[2] ));
    }
    arg3 = (char *)(buf3);
  }
  result = (struct _WsmanStatus *)new__WsmanStatus(arg1,arg2,(char const *)arg3);
  DATA_PTR(self) = result;
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return self;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN void delete__WsmanStatus(struct _WsmanStatus *self){
    if (self->fault_msg) free(self->fault_msg);
    free(self);
  }
SWIGINTERN void
free__WsmanStatus(void *self) {
    struct _WsmanStatus *arg1 = (struct _WsmanStatus *)self;
    delete__WsmanStatus(arg1);
}

/*
  Document-method: Openwsman::Status.to_s

  call-seq:
    to_s -> char const *

Convert class to a String representation.
*/
SWIGINTERN VALUE
_wrap_Status_to_s(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","to_s", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  result = (char *)_WsmanStatus_to_s(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_codee___(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","set_code", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_code", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  _WsmanStatus_set_code(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_code(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","code", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  result = (int)_WsmanStatus_code(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_detaile___(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","set_detail", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_detail", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  _WsmanStatus_set_detail(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_detail(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","detail", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  result = (int)_WsmanStatus_detail(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_msge___(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","set_msg", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_msg", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsmanStatus_set_msg(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_msg(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","msg", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  result = (char *)_WsmanStatus_msg(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Status_generate_fault(int argc, VALUE *argv, VALUE self) {
  struct _WsmanStatus *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsmanStatus *","generate_fault", 1, self )); 
  }
  arg1 = (struct _WsmanStatus *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","generate_fault", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  result = (WsXmlDocH)_WsmanStatus_generate_fault(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassFault;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_Fault_allocate(VALUE self)
#else
_wrap_Fault_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p__WsManFault);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_Fault(int argc, VALUE *argv, VALUE self) {
  WsXmlDocH arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  struct _WsManFault *result = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsXmlDocH","_WsManFault", 1, argv[0] )); 
  }
  arg1 = (WsXmlDocH)(argp1);
  result = (struct _WsManFault *)new__WsManFault(arg1);
  DATA_PTR(self) = result;
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete__WsManFault(struct _WsManFault *self){
    wsmc_fault_destroy(self);
  }
SWIGINTERN void
free__WsManFault(void *self) {
    struct _WsManFault *arg1 = (struct _WsManFault *)self;
    delete__WsManFault(arg1);
}

SWIGINTERN VALUE
_wrap_Fault_code(int argc, VALUE *argv, VALUE self) {
  struct _WsManFault *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManFault *","code", 1, self )); 
  }
  arg1 = (struct _WsManFault *)(argp1);
  result = (char *)_WsManFault_code(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Fault_subcode(int argc, VALUE *argv, VALUE self) {
  struct _WsManFault *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManFault *","subcode", 1, self )); 
  }
  arg1 = (struct _WsManFault *)(argp1);
  result = (char *)_WsManFault_subcode(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Fault_reason(int argc, VALUE *argv, VALUE self) {
  struct _WsManFault *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManFault *","reason", 1, self )); 
  }
  arg1 = (struct _WsManFault *)(argp1);
  result = (char *)_WsManFault_reason(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Fault_detail(int argc, VALUE *argv, VALUE self) {
  struct _WsManFault *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManFault *","detail", 1, self )); 
  }
  arg1 = (struct _WsManFault *)(argp1);
  result = (char *)_WsManFault_detail(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassTransport;

SWIGINTERN VALUE
_wrap_Transport_auth_methodq___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","is_auth_method", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","is_auth_method", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  result = (int)_WsManTransport_is_auth_method(arg1,arg2);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_close(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","close", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  _WsManTransport_close(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_agente___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_agent", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_agent", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_agent(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_agent(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","agent", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_agent(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_username(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","get_username", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_get_username(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_usernamee___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_username", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_username", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_username(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_password(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","get_password", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_get_password(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_passworde___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_password", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_password", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_password(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxy_username(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","get_proxy_username", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_get_proxy_username(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxy_usernamee___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_proxy_username", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_proxy_username", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_proxy_username(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxy_password(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","get_proxy_password", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_get_proxy_password(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxy_passworde___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_proxy_password", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_proxy_password", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_proxy_password(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_auth_methode___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_auth_method", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_auth_method", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_auth_method(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_auth_method(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","auth_method", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_auth_method(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_auth_name(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","_WsManTransport_auth_name", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  result = (char *)_WsManTransport_auth_name(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_auth_value(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","auth_value", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (int)_WsManTransport_auth_value(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_error_string(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","_WsManTransport_error_string", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  result = (char *)_WsManTransport_error_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_timeoute___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_timeout", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","set_timeout", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  _WsManTransport_set_timeout(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_timeout(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","timeout", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (unsigned long)_WsManTransport_timeout(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_verify_peere___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  VALUE arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_verify_peer", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  arg2 = argv[0];
  _WsManTransport_set_verify_peer(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_verify_peerq___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","verify_peer", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (unsigned int)_WsManTransport_verify_peer(arg1);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_verify_hoste___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  VALUE arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_verify_host", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  arg2 = argv[0];
  _WsManTransport_set_verify_host(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_verify_hostq___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","verify_host", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (unsigned int)_WsManTransport_verify_host(arg1);
  vresult = (result != 0) ? Qtrue : Qfalse;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxye___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_proxy", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_proxy", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_proxy(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxy(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","proxy", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_proxy(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxyauthe___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_proxyauth", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_proxyauth", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_proxyauth(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_proxyauth(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","proxyauth", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_proxyauth(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_cainfoe___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_cainfo", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_cainfo", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_cainfo(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_cainfo(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","cainfo", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_cainfo(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_certhumbprinte___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_certhumbprint", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_certhumbprint", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_certhumbprint(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_certhumbprint(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","certhumbprint", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_certhumbprint(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_capathe___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_capath", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_capath", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_capath(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_capath(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","capath", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_capath(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_caoide___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_caoid", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_caoid", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_caoid(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_caoid(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","caoid", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_caoid(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_certe___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_cert", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_cert", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_cert(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_cert(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","cert", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_cert(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_keye___(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","set_key", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_key", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManTransport_set_key(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Transport_key(int argc, VALUE *argv, VALUE self) {
  struct _WsManTransport *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManTransport *","key", 1, self )); 
  }
  arg1 = (struct _WsManTransport *)(argp1);
  result = (char *)_WsManTransport_key(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_send_request(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_send_request", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","wsman_send_request", 2, argv[1] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  result = (int)wsman_send_request(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsmc_transport_set_auth_request_func(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  wsman_auth_request_func_t arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 ;
  int res2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsmc_transport_set_auth_request_func", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  {
    res2 = SWIG_ConvertPtr(argv[1], &argp2, SWIGTYPE_p_wsman_auth_request_func_t,  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "wsman_auth_request_func_t","wsmc_transport_set_auth_request_func", 2, argv[1] )); 
    }  
    if (!argp2) {
      SWIG_exception_fail(SWIG_NullReferenceError, Ruby_Format_TypeError("invalid null reference ", "wsman_auth_request_func_t","wsmc_transport_set_auth_request_func", 2, argv[1]));
    } else {
      arg2 = *((wsman_auth_request_func_t *)(argp2));
    }
  }
  wsmc_transport_set_auth_request_func(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_is_auth_method(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_is_auth_method", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","wsman_is_auth_method", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  result = (int)wsman_is_auth_method(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsmc_transport_init(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsmc_transport_init", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[1],SWIG_as_voidptrptr(&arg2), 0, 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "void *","wsmc_transport_init", 2, argv[1] )); 
  }
  result = (int)wsmc_transport_init(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_close_transport(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_close_transport", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  wsman_transport_close_transport(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsmc_transport_fini(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsmc_transport_fini", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  wsmc_transport_fini(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_agent(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_agent", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_agent", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_agent(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_agent(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_agent", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_agent(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_user_name(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_userName", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_userName", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_userName(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_user_name(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_userName", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_userName(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_password(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_password", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_password", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_password(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_password(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_password", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_password(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_proxy_username(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_proxy_username", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_proxy_username", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_proxy_username(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_proxy_username(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_proxy_username", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_proxy_username(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_proxy_password(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_proxy_password", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_proxy_password", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_proxy_password(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_proxy_password(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_proxy_password", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_proxy_password(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_auth_method(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_auth_method", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_auth_method", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_auth_method(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_auth_method(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_auth_method", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_auth_method(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsmc_transport_get_auth_name(int argc, VALUE *argv, VALUE self) {
  wsman_auth_type_t arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    res1 = SWIG_ConvertPtr(argv[0], &argp1, SWIGTYPE_p_wsman_auth_type_t,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "wsman_auth_type_t","wsmc_transport_get_auth_name", 1, argv[0] )); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, Ruby_Format_TypeError("invalid null reference ", "wsman_auth_type_t","wsmc_transport_get_auth_name", 1, argv[0]));
    } else {
      arg1 = *((wsman_auth_type_t *)(argp1));
    }
  }
  result = (char *)wsmc_transport_get_auth_name(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsmc_transport_get_auth_value(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  wsman_auth_type_t result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsmc_transport_get_auth_value", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = wsmc_transport_get_auth_value(arg1);
  vresult = SWIG_NewPointerObj((wsman_auth_type_t *)memcpy((wsman_auth_type_t *)calloc(1,sizeof(wsman_auth_type_t)),&result,sizeof(wsman_auth_type_t)), SWIGTYPE_p_wsman_auth_type_t, SWIG_POINTER_OWN |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_last_error_string(int argc, VALUE *argv, VALUE self) {
  WS_LASTERR_Code arg1 ;
  void *argp1 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  {
    res1 = SWIG_ConvertPtr(argv[0], &argp1, SWIGTYPE_p_WS_LASTERR_Code,  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WS_LASTERR_Code","wsman_transport_get_last_error_string", 1, argv[0] )); 
    }  
    if (!argp1) {
      SWIG_exception_fail(SWIG_NullReferenceError, Ruby_Format_TypeError("invalid null reference ", "WS_LASTERR_Code","wsman_transport_get_last_error_string", 1, argv[0]));
    } else {
      arg1 = *((WS_LASTERR_Code *)(argp1));
    }
  }
  result = (char *)wsman_transport_get_last_error_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_timeout(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_timeout", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","wsman_transport_set_timeout", 2, argv[1] ));
  } 
  arg2 = (unsigned long)(val2);
  wsman_transport_set_timeout(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_timeout(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_timeout", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (unsigned long)wsman_transport_get_timeout(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_verify_peer(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_verify_peer", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned int","wsman_transport_set_verify_peer", 2, argv[1] ));
  } 
  arg2 = (unsigned int)(val2);
  wsman_transport_set_verify_peer(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_verify_peer(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_verify_peer", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (unsigned int)wsman_transport_get_verify_peer(arg1);
  vresult = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_verify_host(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_verify_host", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned int","wsman_transport_set_verify_host", 2, argv[1] ));
  } 
  arg2 = (unsigned int)(val2);
  wsman_transport_set_verify_host(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_verify_host(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_verify_host", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (unsigned int)wsman_transport_get_verify_host(arg1);
  vresult = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_crlcheck(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_crlcheck", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned int","wsman_transport_set_crlcheck", 2, argv[1] ));
  } 
  arg2 = (unsigned int)(val2);
  wsman_transport_set_crlcheck(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_crlcheck(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_crlcheck", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (unsigned int)wsman_transport_get_crlcheck(arg1);
  vresult = SWIG_From_unsigned_SS_int((unsigned int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_crlfile(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_crlfile", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_crlfile", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_crlfile(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_crlfile(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_crlfile", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_crlfile(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_proxy(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_proxy", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_proxy", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_proxy(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_proxy(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_proxy", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_proxy(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_proxyauth(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_proxyauth", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_proxyauth", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_proxyauth(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_proxyauth(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_proxyauth", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_proxyauth(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_cainfo(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_cainfo", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_cainfo", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_cainfo(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_cainfo(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_cainfo", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_cainfo(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_certhumbprint(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_certhumbprint", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_certhumbprint", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_certhumbprint(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_certhumbprint(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_certhumbprint", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_certhumbprint(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_capath(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_capath", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_capath", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_capath(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_capath(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_capath", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_capath(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_caoid(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_caoid", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_caoid", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_caoid(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_caoid(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_caoid", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_caoid(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_cert(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_cert", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_cert", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_cert(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_cert(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_cert", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_cert(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_set_key(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_set_key", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","wsman_transport_set_key", 2, argv[1] ));
  }
  arg2 = (char *)(buf2);
  wsman_transport_set_key(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_wsman_transport_get_key(int argc, VALUE *argv, VALUE self) {
  WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(argv[0], &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "WsManClient *","wsman_transport_get_key", 1, argv[0] )); 
  }
  arg1 = (WsManClient *)(argp1);
  result = (char *)wsman_transport_get_key(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassClientOptions;

SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_ClientOptions_allocate(VALUE self)
#else
_wrap_ClientOptions_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p_client_opt_t);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_ClientOptions(int argc, VALUE *argv, VALUE self) {
  client_opt_t *result = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  result = (client_opt_t *)new_client_opt_t();
  DATA_PTR(self) = result;
  return self;
fail:
  return Qnil;
}


SWIGINTERN void delete_client_opt_t(client_opt_t *self){
    wsmc_options_destroy( self );
  }
SWIGINTERN void
free_client_opt_t(void *self) {
    client_opt_t *arg1 = (client_opt_t *)self;
    delete_client_opt_t(arg1);
}

SWIGINTERN VALUE
_wrap_ClientOptions_set_dump_request(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_dump_request", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  client_opt_t_set_dump_request(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_clear_dump_request(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","clear_dump_request", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  client_opt_t_clear_dump_request(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_flagse___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_flags", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","set_flags", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  client_opt_t_set_flags(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_flags(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_flags", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (unsigned long)client_opt_t_get_flags(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_clear_flags(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","clear_flags", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","clear_flags", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  client_opt_t_clear_flags(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_reset_flags(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","reset_flags", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  client_opt_t_reset_flags(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_max_envelope_sizee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_max_envelope_size", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","set_max_envelope_size", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  client_opt_t_set_max_envelope_size(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_max_envelope_size(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_max_envelope_size", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (unsigned long)client_opt_t_get_max_envelope_size(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_max_elementse___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_max_elements", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","set_max_elements", 2, argv[0] ));
  } 
  arg2 = (int)(val2);
  client_opt_t_set_max_elements(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_max_elements(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_max_elements", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (int)client_opt_t_get_max_elements(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_timeoute___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned long arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_timeout", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned long","set_timeout", 2, argv[0] ));
  } 
  arg2 = (unsigned long)(val2);
  client_opt_t_set_timeout(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_timeout(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_timeout", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (unsigned long)client_opt_t_get_timeout(arg1);
  vresult = SWIG_From_unsigned_SS_long((unsigned long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_fragmente___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_fragment", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_fragment", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  client_opt_t_set_fragment(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_fragment(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_fragment", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (char *)client_opt_t_get_fragment(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_cim_namespacee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_cim_namespace", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char *","set_cim_namespace", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  client_opt_t_set_cim_namespace(arg1,arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_cim_namespace(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_cim_namespace", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (char *)client_opt_t_get_cim_namespace(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_referencee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_reference", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_reference", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  client_opt_t_set_reference(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_reference(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_reference", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (char *)client_opt_t_get_reference(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_add_option(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  VALUE arg2 = 0 ;
  VALUE arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","add_option", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  arg2 = argv[0];
  arg3 = argv[1];
  client_opt_t_add_option(arg1,arg2,arg3);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_optionse___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  VALUE arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_options", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  arg2 = argv[0];
  client_opt_t_set_options(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_options(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_options", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (VALUE)client_opt_t_get_options(arg1);
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_add_selector(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  VALUE arg2 = 0 ;
  VALUE arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","add_selector", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  arg2 = argv[0];
  arg3 = argv[1];
  client_opt_t_add_selector(arg1,arg2,arg3);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_selectors(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_selectors", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (VALUE)client_opt_t_get_selectors(arg1);
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_add_property(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  VALUE arg2 = 0 ;
  VALUE arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","add_property", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  arg2 = argv[0];
  arg3 = argv[1];
  client_opt_t_add_property(arg1,arg2,arg3);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_properties(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  VALUE result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","get_properties", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (VALUE)client_opt_t_get_properties(arg1);
  vresult = result;
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_urie___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_delivery_uri", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_delivery_uri", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  client_opt_t_set_delivery_uri(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_uri(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","delivery_uri", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (char *)client_opt_t_delivery_uri(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_sub_expirye___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_sub_expiry", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_float(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "float","set_sub_expiry", 2, argv[0] ));
  } 
  arg2 = (float)(val2);
  client_opt_t_set_sub_expiry(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_sub_expiry(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","sub_expiry", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (float)client_opt_t_sub_expiry(arg1);
  vresult = SWIG_From_float((float)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_heartbeat_intervale___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  float arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_heartbeat_interval", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_float(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "float","set_heartbeat_interval", 2, argv[0] ));
  } 
  arg2 = (float)(val2);
  client_opt_t_set_heartbeat_interval(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_heartbeat_interval(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","heartbeat_interval", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (float)client_opt_t_heartbeat_interval(arg1);
  vresult = SWIG_From_float((float)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_modee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_delivery_mode", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned int","set_delivery_mode", 2, argv[0] ));
  } 
  arg2 = (unsigned int)(val2);
  client_opt_t_set_delivery_mode(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_mode(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","delivery_mode", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (int)client_opt_t_delivery_mode(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_security_modee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_delivery_security_mode", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(argv[0], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "unsigned int","set_delivery_security_mode", 2, argv[0] ));
  } 
  arg2 = (unsigned int)(val2);
  client_opt_t_set_delivery_security_mode(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_delivery_sec_mode(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","delivery_sec_mode", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (int)client_opt_t_delivery_sec_mode(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_localee___(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","set_locale", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_locale", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  client_opt_t_set_locale(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_ClientOptions_locale(int argc, VALUE *argv, VALUE self) {
  client_opt_t *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "client_opt_t *","locale", 1, self )); 
  }
  arg1 = (client_opt_t *)(argp1);
  result = (char *)client_opt_t_locale(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


static swig_class SwigClassClient;

SWIGINTERN VALUE
_wrap_new_Client__SWIG_0(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  struct _WsManClient *result = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","_WsManClient", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  result = (struct _WsManClient *)new__WsManClient__SWIG_0((char const *)arg1);
  DATA_PTR(self) = result;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}


SWIGINTERN VALUE
#ifdef HAVE_RB_DEFINE_ALLOC_FUNC
_wrap_Client_allocate(VALUE self)
#else
_wrap_Client_allocate(int argc, VALUE *argv, VALUE self)
#endif
{
  VALUE vresult = SWIG_NewClassInstance(self, SWIGTYPE_p__WsManClient);
#ifndef HAVE_RB_DEFINE_ALLOC_FUNC
  rb_obj_call_init(vresult, argc, argv);
#endif
  return vresult;
}


SWIGINTERN VALUE
_wrap_new_Client__SWIG_1(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int arg2 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  char *arg6 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int val2 ;
  int ecode2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  struct _WsManClient *result = 0 ;
  
  if ((argc < 6) || (argc > 6)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 6)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","_WsManClient", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  ecode2 = SWIG_AsVal_int(argv[1], &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), Ruby_Format_TypeError( "", "int","_WsManClient", 2, argv[1] ));
  } 
  arg2 = (int)(val2);
  res3 = SWIG_AsCharPtrAndSize(argv[2], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","_WsManClient", 3, argv[2] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[3], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","_WsManClient", 4, argv[3] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[4], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","_WsManClient", 5, argv[4] ));
  }
  arg5 = (char *)(buf5);
  res6 = SWIG_AsCharPtrAndSize(argv[5], &buf6, NULL, &alloc6);
  if (!SWIG_IsOK(res6)) {
    SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","_WsManClient", 6, argv[5] ));
  }
  arg6 = (char *)(buf6);
  result = (struct _WsManClient *)new__WsManClient__SWIG_1((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6);
  DATA_PTR(self) = result;
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return self;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
}


SWIGINTERN VALUE _wrap_new_Client(int nargs, VALUE *args, VALUE self) {
  int argc;
  VALUE argv[6];
  int ii;
  
  argc = nargs;
  if (argc > 6) SWIG_fail;
  for (ii = 0; (ii < argc); ++ii) {
    argv[ii] = args[ii];
  }
  if (argc == 1) {
    int _v = 0;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Client__SWIG_0(nargs, args, self);
    }
  }
  if (argc == 6) {
    int _v = 0;
    int res = SWIG_AsCharPtrAndSize(argv[0], 0, NULL, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsCharPtrAndSize(argv[2], 0, NULL, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = SWIG_AsCharPtrAndSize(argv[3], 0, NULL, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = SWIG_AsCharPtrAndSize(argv[4], 0, NULL, 0);
            _v = SWIG_CheckState(res);
            if (_v) {
              int res = SWIG_AsCharPtrAndSize(argv[5], 0, NULL, 0);
              _v = SWIG_CheckState(res);
              if (_v) {
                return _wrap_new_Client__SWIG_1(nargs, args, self);
              }
            }
          }
        }
      }
    }
  }
  
fail:
  Ruby_Format_OverloadedError( argc, 6, "Client.new", 
    "    Client.new(char const *uri)\n"
    "    Client.new(char const *hostname, int const port, char const *path, char const *scheme, char const *username, char const *password)\n");
  
  return Qnil;
}


SWIGINTERN void delete__WsManClient(struct _WsManClient *self){
    wsmc_release( self );
  }
SWIGINTERN void
free__WsManClient(void *self) {
    struct _WsManClient *arg1 = (struct _WsManClient *)self;
    delete__WsManClient(arg1);
}

SWIGINTERN VALUE
_wrap_Client_dumpfilee___(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  FILE *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","set_dumpfile", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  {
    rb_fptr_t *fptr;
    
    Check_Type(argv[0], T_FILE);
    GetOpenFile(argv[0], fptr);
    /*rb_io_check_writable(fptr);*/
    
    arg2 = rb_io_stdio_file(fptr);
    
    
    
  }
  _WsManClient_set_dumpfile(arg1,arg2);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_response_code(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  long result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","response_code", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (long)_WsManClient_response_code(arg1);
  vresult = SWIG_From_long((long)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_scheme(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","scheme", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_scheme(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_host(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","host", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_host(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_port(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","port", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (int)_WsManClient_port(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_path(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","path", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_path(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_user(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","user", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_user(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_password(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","password", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_password(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  free(result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_transport(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsManTransport *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","transport", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (WsManTransport *)_WsManClient_transport(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManTransport, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_send_request(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  WsXmlDocH arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","send_request", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "WsXmlDocH","send_request", 2, argv[0] )); 
  }
  arg2 = (WsXmlDocH)(argp2);
  result = (int)_WsManClient_send_request(arg1,arg2);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_build_envelope_from_response(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","build_envelope_from_response", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (WsXmlDocH)_WsManClient_build_envelope_from_response(arg1);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_encoding(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","encoding", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_encoding(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_encodinge___(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  char *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","set_encoding", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_AsCharPtrAndSize(argv[0], &buf2, NULL, &alloc2);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","set_encoding", 2, argv[0] ));
  }
  arg2 = (char *)(buf2);
  _WsManClient_set_encoding(arg1,(char const *)arg2);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
fail:
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_identify(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","identify", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","identify", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  result = (WsXmlDocH)_WsManClient_identify(arg1,arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_get_from_epr(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  epr_t *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","get_from_epr", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","get_from_epr", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "epr_t *","get_from_epr", 3, argv[1] )); 
  }
  arg3 = (epr_t *)(argp3);
  result = (WsXmlDocH)_WsManClient_get_from_epr(arg1,arg2,arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_delete_from_epr(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  epr_t *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","delete_from_epr", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","delete_from_epr", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_epr_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "epr_t *","delete_from_epr", 3, argv[1] )); 
  }
  arg3 = (epr_t *)(argp3);
  result = (WsXmlDocH)_WsManClient_delete_from_epr(arg1,arg2,arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_enumerate(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  filter_t *arg3 = 0 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","enumerate", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","enumerate", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "filter_t *","enumerate", 3, argv[1] )); 
  }
  arg3 = (filter_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char *","enumerate", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  result = (WsXmlDocH)_WsManClient_enumerate(arg1,arg2,arg3,arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_pull(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  filter_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","pull", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","pull", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "filter_t *","pull", 3, argv[1] )); 
  }
  arg3 = (filter_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","pull", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","pull", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  result = (WsXmlDocH)_WsManClient_pull(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_create(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  size_t arg5 ;
  char *arg6 = (char *)"utf-8" ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  size_t val5 ;
  int ecode5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","create", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","create", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","create", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","create", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_size_t(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "size_t","create", 5, argv[3] ));
  } 
  arg5 = (size_t)(val5);
  if (argc > 4) {
    res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","create", 6, argv[4] ));
    }
    arg6 = (char *)(buf6);
  }
  result = (WsXmlDocH)_WsManClient_create(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_put(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  size_t arg5 ;
  char *arg6 = (char *)"utf-8" ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  size_t val5 ;
  int ecode5 = 0 ;
  int res6 ;
  char *buf6 = 0 ;
  int alloc6 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 5)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","put", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","put", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","put", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","put", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  ecode5 = SWIG_AsVal_size_t(argv[3], &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), Ruby_Format_TypeError( "", "size_t","put", 5, argv[3] ));
  } 
  arg5 = (size_t)(val5);
  if (argc > 4) {
    res6 = SWIG_AsCharPtrAndSize(argv[4], &buf6, NULL, &alloc6);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), Ruby_Format_TypeError( "", "char const *","put", 6, argv[4] ));
    }
    arg6 = (char *)(buf6);
  }
  result = (WsXmlDocH)_WsManClient_put(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_release(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","release", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","release", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","release", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","release", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  result = (WsXmlDocH)_WsManClient_release(arg1,arg2,(char const *)arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_get(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","get", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","get", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","get", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  result = (WsXmlDocH)_WsManClient_get(arg1,arg2,(char const *)arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_delete(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 2) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 2)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","delete", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","delete", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","delete", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  result = (WsXmlDocH)_WsManClient_delete(arg1,arg2,(char const *)arg3);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_invoke(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  WsXmlDocH arg5 = NULL ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","invoke", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","invoke", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char const *","invoke", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","invoke", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  if (argc > 3) {
    res5 = SWIG_ConvertPtr(argv[3], &argp5,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "WsXmlDocH","invoke", 5, argv[3] )); 
    }
    arg5 = (WsXmlDocH)(argp5);
  }
  result = (WsXmlDocH)_WsManClient_invoke(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_subscribe(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  filter_t *arg3 = 0 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","subscribe", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","subscribe", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "filter_t *","subscribe", 3, argv[1] )); 
  }
  arg3 = (filter_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","subscribe", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  result = (WsXmlDocH)_WsManClient_subscribe(arg1,arg2,arg3,(char const *)arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_unsubscribe(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  filter_t *arg3 = 0 ;
  char *arg4 = 0 ;
  char *arg5 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  int res5 ;
  char *buf5 = 0 ;
  int alloc5 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 4) || (argc > 4)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 4)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","unsubscribe", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","unsubscribe", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_ConvertPtr(argv[1], &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "filter_t *","unsubscribe", 3, argv[1] )); 
  }
  arg3 = (filter_t *)(argp3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char const *","unsubscribe", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  res5 = SWIG_AsCharPtrAndSize(argv[3], &buf5, NULL, &alloc5);
  if (!SWIG_IsOK(res5)) {
    SWIG_exception_fail(SWIG_ArgError(res5), Ruby_Format_TypeError( "", "char const *","unsubscribe", 5, argv[3] ));
  }
  arg5 = (char *)(buf5);
  result = (WsXmlDocH)_WsManClient_unsubscribe(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return vresult;
fail:
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_renew(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  client_opt_t *arg2 = 0 ;
  char *arg3 = 0 ;
  char *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 ;
  char *buf3 = 0 ;
  int alloc3 = 0 ;
  int res4 ;
  char *buf4 = 0 ;
  int alloc4 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 3) || (argc > 3)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 3)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","renew", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  res2 = SWIG_ConvertPtr(argv[0], &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "client_opt_t *","renew", 2, argv[0] )); 
  }
  arg2 = (client_opt_t *)(argp2);
  res3 = SWIG_AsCharPtrAndSize(argv[1], &buf3, NULL, &alloc3);
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), Ruby_Format_TypeError( "", "char *","renew", 3, argv[1] ));
  }
  arg3 = (char *)(buf3);
  res4 = SWIG_AsCharPtrAndSize(argv[2], &buf4, NULL, &alloc4);
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), Ruby_Format_TypeError( "", "char *","renew", 4, argv[2] ));
  }
  arg4 = (char *)(buf4);
  result = (WsXmlDocH)_WsManClient_renew(arg1,arg2,arg3,arg4);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return vresult;
fail:
  if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_fault_string(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","fault_string", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (char *)_WsManClient_fault_string(arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_Client_last_error(int argc, VALUE *argv, VALUE self) {
  struct _WsManClient *arg1 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  res1 = SWIG_ConvertPtr(self, &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "struct _WsManClient *","last_error", 1, self )); 
  }
  arg1 = (struct _WsManClient *)(argp1);
  result = (int)_WsManClient_last_error(arg1);
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_debuge___(int argc, VALUE *argv, VALUE self) {
  int arg1 ;
  int val1 ;
  int ecode1 = 0 ;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  ecode1 = SWIG_AsVal_int(argv[0], &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), Ruby_Format_TypeError( "", "int","set_debug", 1, argv[0] ));
  } 
  arg1 = (int)(val1);
  set_debug(arg1);
  return Qnil;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_debug(int argc, VALUE *argv, VALUE self) {
  int result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  result = (int)get_debug();
  vresult = SWIG_From_int((int)(result));
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_create_soap_envelope(int argc, VALUE *argv, VALUE self) {
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 0) || (argc > 0)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 0)",argc); SWIG_fail;
  }
  result = (WsXmlDocH)create_soap_envelope();
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  return vresult;
fail:
  return Qnil;
}


SWIGINTERN VALUE
_wrap_create_doc_from_file(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = (char *)"UTF-8" ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","create_doc_from_file", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","create_doc_from_file", 2, argv[1] ));
    }
    arg2 = (char *)(buf2);
  }
  result = (WsXmlDocH)create_doc_from_file((char const *)arg1,(char const *)arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_create_doc_from_string(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  char *arg2 = (char *)"UTF-8" ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  int res2 ;
  char *buf2 = 0 ;
  int alloc2 = 0 ;
  WsXmlDocH result;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 2)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","create_doc_from_string", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  if (argc > 1) {
    res2 = SWIG_AsCharPtrAndSize(argv[1], &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), Ruby_Format_TypeError( "", "char const *","create_doc_from_string", 2, argv[1] ));
    }
    arg2 = (char *)(buf2);
  }
  result = (WsXmlDocH)create_doc_from_string((char const *)arg1,(char const *)arg2);
  vresult = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 |  0 );
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  return Qnil;
}


SWIGINTERN VALUE
_wrap_uri_prefix(int argc, VALUE *argv, VALUE self) {
  char *arg1 = 0 ;
  int res1 ;
  char *buf1 = 0 ;
  int alloc1 = 0 ;
  char *result = 0 ;
  VALUE vresult = Qnil;
  
  if ((argc < 1) || (argc > 1)) {
    rb_raise(rb_eArgError, "wrong # of arguments(%d for 1)",argc); SWIG_fail;
  }
  res1 = SWIG_AsCharPtrAndSize(argv[0], &buf1, NULL, &alloc1);
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), Ruby_Format_TypeError( "", "char const *","uri_prefix", 1, argv[0] ));
  }
  arg1 = (char *)(buf1);
  result = (char *)uri_prefix((char const *)arg1);
  vresult = SWIG_FromCharPtr((const char *)result);
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return vresult;
fail:
  if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  return Qnil;
}



/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_FILE = {"_p_FILE", "FILE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WS_LASTERR_Code = {"_p_WS_LASTERR_Code", "WS_LASTERR_Code *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsEnumerateInfo = {"_p_WsEnumerateInfo", "WsEnumerateInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsmanKnownStatusCode = {"_p_WsmanKnownStatusCode", "WsmanKnownStatusCode *|enum WsmanKnownStatusCode *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsmanMessage = {"_p_WsmanMessage", "WsmanMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WS_CONTEXT = {"_p__WS_CONTEXT", "WsContextH|struct _WS_CONTEXT *|_WS_CONTEXT *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WsManClient = {"_p__WsManClient", "WsManClient *|struct _WsManClient *|_WsManClient *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WsManFault = {"_p__WsManFault", "WsManFault *|struct _WsManFault *|_WsManFault *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WsManTransport = {"_p__WsManTransport", "WsManTransport *|struct _WsManTransport *|_WsManTransport *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WsXmlDoc = {"_p__WsXmlDoc", "WsXmlDocH|struct _WsXmlDoc *|_WsXmlDoc *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WsmanStatus = {"_p__WsmanStatus", "WsmanStatus *|struct _WsmanStatus *|_WsmanStatus *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___Soap = {"_p___Soap", "struct __Soap *|__Soap *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___SoapOp = {"_p___SoapOp", "struct __SoapOp *|__SoapOp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsEnumerateInfo = {"_p___WsEnumerateInfo", "struct __WsEnumerateInfo *|__WsEnumerateInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsXmlAttr = {"_p___WsXmlAttr", "WsXmlAttrH|struct __WsXmlAttr *|__WsXmlAttr *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsXmlNode = {"_p___WsXmlNode", "WsXmlNodeH|struct __WsXmlNode *|__WsXmlNode *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsXmlNs = {"_p___WsXmlNs", "__WsXmlNs *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultCodeTable = {"_p___WsmanFaultCodeTable", "WsmanFaultCodeTable *|struct __WsmanFaultCodeTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultCodeType = {"_p___WsmanFaultCodeType", "WsmanFaultCodeType *|enum __WsmanFaultCodeType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultDetailTable = {"_p___WsmanFaultDetailTable", "WsmanFaultDetailTable *|struct __WsmanFaultDetailTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultDetailType = {"_p___WsmanFaultDetailType", "WsmanFaultDetailType *|enum __WsmanFaultDetailType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_client_opt_t = {"_p_client_opt_t", "client_opt_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_epr_t = {"_p_epr_t", "epr_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_filter_t = {"_p_filter_t", "filter_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wsman_auth_request_func_t = {"_p_wsman_auth_request_func_t", "wsman_auth_request_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wsman_auth_type_t = {"_p_wsman_auth_type_t", "wsman_auth_type_t *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_FILE,
  &_swigt__p_WS_LASTERR_Code,
  &_swigt__p_WsEnumerateInfo,
  &_swigt__p_WsmanKnownStatusCode,
  &_swigt__p_WsmanMessage,
  &_swigt__p__WS_CONTEXT,
  &_swigt__p__WsManClient,
  &_swigt__p__WsManFault,
  &_swigt__p__WsManTransport,
  &_swigt__p__WsXmlDoc,
  &_swigt__p__WsmanStatus,
  &_swigt__p___Soap,
  &_swigt__p___SoapOp,
  &_swigt__p___WsEnumerateInfo,
  &_swigt__p___WsXmlAttr,
  &_swigt__p___WsXmlNode,
  &_swigt__p___WsXmlNs,
  &_swigt__p___WsmanFaultCodeTable,
  &_swigt__p___WsmanFaultCodeType,
  &_swigt__p___WsmanFaultDetailTable,
  &_swigt__p___WsmanFaultDetailType,
  &_swigt__p_char,
  &_swigt__p_client_opt_t,
  &_swigt__p_epr_t,
  &_swigt__p_filter_t,
  &_swigt__p_wsman_auth_request_func_t,
  &_swigt__p_wsman_auth_type_t,
};

static swig_cast_info _swigc__p_FILE[] = {  {&_swigt__p_FILE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WS_LASTERR_Code[] = {  {&_swigt__p_WS_LASTERR_Code, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsEnumerateInfo[] = {  {&_swigt__p_WsEnumerateInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsmanKnownStatusCode[] = {  {&_swigt__p_WsmanKnownStatusCode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsmanMessage[] = {  {&_swigt__p_WsmanMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WS_CONTEXT[] = {  {&_swigt__p__WS_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManClient[] = {  {&_swigt__p__WsManClient, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManFault[] = {  {&_swigt__p__WsManFault, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManTransport[] = {  {&_swigt__p__WsManTransport, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsXmlDoc[] = {  {&_swigt__p__WsXmlDoc, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsmanStatus[] = {  {&_swigt__p__WsmanStatus, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___Soap[] = {  {&_swigt__p___Soap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___SoapOp[] = {  {&_swigt__p___SoapOp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsEnumerateInfo[] = {  {&_swigt__p___WsEnumerateInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlAttr[] = {  {&_swigt__p___WsXmlAttr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlNode[] = {  {&_swigt__p___WsXmlNode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlNs[] = {  {&_swigt__p___WsXmlNs, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultCodeTable[] = {  {&_swigt__p___WsmanFaultCodeTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultCodeType[] = {  {&_swigt__p___WsmanFaultCodeType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultDetailTable[] = {  {&_swigt__p___WsmanFaultDetailTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultDetailType[] = {  {&_swigt__p___WsmanFaultDetailType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_client_opt_t[] = {  {&_swigt__p_client_opt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_epr_t[] = {  {&_swigt__p_epr_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_filter_t[] = {  {&_swigt__p_filter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wsman_auth_request_func_t[] = {  {&_swigt__p_wsman_auth_request_func_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wsman_auth_type_t[] = {  {&_swigt__p_wsman_auth_type_t, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_FILE,
  _swigc__p_WS_LASTERR_Code,
  _swigc__p_WsEnumerateInfo,
  _swigc__p_WsmanKnownStatusCode,
  _swigc__p_WsmanMessage,
  _swigc__p__WS_CONTEXT,
  _swigc__p__WsManClient,
  _swigc__p__WsManFault,
  _swigc__p__WsManTransport,
  _swigc__p__WsXmlDoc,
  _swigc__p__WsmanStatus,
  _swigc__p___Soap,
  _swigc__p___SoapOp,
  _swigc__p___WsEnumerateInfo,
  _swigc__p___WsXmlAttr,
  _swigc__p___WsXmlNode,
  _swigc__p___WsXmlNs,
  _swigc__p___WsmanFaultCodeTable,
  _swigc__p___WsmanFaultCodeType,
  _swigc__p___WsmanFaultDetailTable,
  _swigc__p___WsmanFaultDetailType,
  _swigc__p_char,
  _swigc__p_client_opt_t,
  _swigc__p_epr_t,
  _swigc__p_filter_t,
  _swigc__p_wsman_auth_request_func_t,
  _swigc__p_wsman_auth_type_t,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
    return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
    return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;

  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }

  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);

    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }

  /* When multiple interpreters are used, a module could have already been initialized in
     a different interpreter, but not yet have a pointer in this interpreter.
     In this case, we do not want to continue adding types... everything should be
     set up already */
  if (init == 0) return;

  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;

#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif

    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
	type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }

    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {

      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {

          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }

      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;

        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {

            /* Move cast to itself to the first entry in the array */

            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;

        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }

    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;

        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);

        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }

      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;

      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */

        SWIG_CastHashSort(first, num_mapped);
      }

      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }

    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;

#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
  printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;

  if (init_run) return;
  init_run = 1;

  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
              SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{ /* c-mode */
#endif
}
#endif

/*

*/
#ifdef __cplusplus
extern "C"
#endif
SWIGEXPORT void Init_Openwsman(void) {
  size_t i;
  
  SWIG_InitRuntime();
  mOpenwsman = rb_define_module("Openwsman");
  
  SWIG_InitializeModule(0);
  for (i = 0; i < swig_module.size; i++) {
    SWIG_define_class(swig_module.types[i]);
  }
  
  SWIG_RubyInitializeTrackings();
  rb_define_const(mOpenwsman, "OPENWSMAN_MAJOR", SWIG_From_int((int)(2)));
  rb_define_const(mOpenwsman, "OPENWSMAN_MINOR", SWIG_From_int((int)(8)));
  rb_define_const(mOpenwsman, "OPENWSMAN_PATCH", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "OPENWSMAN_VERSION", SWIG_FromCharPtr("2.8.1"));
  
  SwigClassXmlNs.klass = rb_define_class_under(mOpenwsman, "XmlNs", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlNs, (void *) &SwigClassXmlNs);
  rb_undef_alloc_func(SwigClassXmlNs.klass);
  SwigClassXmlNs.mark = 0;
  SwigClassXmlNs.trackObjects = 0;
  
  SwigClassXmlDoc.klass = rb_define_class_under(mOpenwsman, "XmlDoc", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WsXmlDoc, (void *) &SwigClassXmlDoc);
  rb_define_alloc_func(SwigClassXmlDoc.klass, _wrap_XmlDoc_allocate);
  rb_define_method(SwigClassXmlDoc.klass, "initialize", _wrap_new_XmlDoc, -1);
  rb_define_method(SwigClassXmlDoc.klass, "string", _wrap_XmlDoc_string, -1);
  rb_define_alias(SwigClassXmlDoc.klass, "to_xml", "string");
  rb_define_method(SwigClassXmlDoc.klass, "encode", _wrap_XmlDoc_encode, -1);
  rb_define_alias(SwigClassXmlDoc.klass, "to_s", "encode");
  rb_define_method(SwigClassXmlDoc.klass, "dump_file", _wrap_XmlDoc_dump_file, -1);
  rb_define_method(SwigClassXmlDoc.klass, "root", _wrap_XmlDoc_root, -1);
  rb_define_method(SwigClassXmlDoc.klass, "envelope", _wrap_XmlDoc_envelope, -1);
  rb_define_method(SwigClassXmlDoc.klass, "header", _wrap_XmlDoc_header, -1);
  rb_define_method(SwigClassXmlDoc.klass, "body", _wrap_XmlDoc_body, -1);
  rb_define_method(SwigClassXmlDoc.klass, "element", _wrap_XmlDoc_element, -1);
  rb_define_method(SwigClassXmlDoc.klass, "context", _wrap_XmlDoc_context, -1);
  rb_define_method(SwigClassXmlDoc.klass, "generate_fault", _wrap_XmlDoc_generate_fault, -1);
  rb_define_method(SwigClassXmlDoc.klass, "fault?", _wrap_XmlDoc_faultq___, -1);
  rb_define_method(SwigClassXmlDoc.klass, "fault", _wrap_XmlDoc_fault, -1);
  rb_define_method(SwigClassXmlDoc.klass, "create_response_envelope", _wrap_XmlDoc_create_response_envelope, -1);
  rb_define_method(SwigClassXmlDoc.klass, "end_of_sequence?", _wrap_XmlDoc_end_of_sequenceq___, -1);
  SwigClassXmlDoc.mark = 0;
  SwigClassXmlDoc.destroy = (void (*)(void *)) free__WsXmlDoc;
  SwigClassXmlDoc.trackObjects = 0;
  
  SwigClassXmlNode.klass = rb_define_class_under(mOpenwsman, "XmlNode", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlNode, (void *) &SwigClassXmlNode);
  rb_undef_alloc_func(SwigClassXmlNode.klass);
  rb_define_method(SwigClassXmlNode.klass, "string", _wrap_XmlNode_string, -1);
  rb_define_alias(SwigClassXmlNode.klass, "to_xml", "string");
  rb_define_method(SwigClassXmlNode.klass, "dump_file", _wrap_XmlNode_dump_file, -1);
  rb_define_method(SwigClassXmlNode.klass, "equal", _wrap_XmlNode_equal, -1);
  rb_define_alias(SwigClassXmlNode.klass, "==", "equal");
  rb_define_method(SwigClassXmlNode.klass, "text", _wrap_XmlNode_text, -1);
  rb_define_alias(SwigClassXmlNode.klass, "to_s", "text");
  rb_define_method(SwigClassXmlNode.klass, "text=", _wrap_XmlNode_texte___, -1);
  rb_define_method(SwigClassXmlNode.klass, "doc", _wrap_XmlNode_doc, -1);
  rb_define_method(SwigClassXmlNode.klass, "parent", _wrap_XmlNode_parent, -1);
  rb_define_method(SwigClassXmlNode.klass, "child", _wrap_XmlNode_child, -1);
  rb_define_alias(SwigClassXmlNode.klass, "first", "child");
  rb_define_method(SwigClassXmlNode.klass, "name", _wrap_XmlNode_name, -1);
  rb_define_method(SwigClassXmlNode.klass, "name=", _wrap_XmlNode_namee___, -1);
  rb_define_method(SwigClassXmlNode.klass, "ns", _wrap_XmlNode_ns, -1);
  rb_define_method(SwigClassXmlNode.klass, "ns=", _wrap_XmlNode_nse___, -1);
  rb_define_method(SwigClassXmlNode.klass, "prefix", _wrap_XmlNode_prefix, -1);
  rb_define_method(SwigClassXmlNode.klass, "lang=", _wrap_XmlNode_lange___, -1);
  rb_define_method(SwigClassXmlNode.klass, "find", _wrap_XmlNode_find, -1);
  rb_define_method(SwigClassXmlNode.klass, "next", _wrap_XmlNode_next, -1);
  rb_define_method(SwigClassXmlNode.klass, "size", _wrap_XmlNode_size, -1);
  rb_define_method(SwigClassXmlNode.klass, "add_before", _wrap_XmlNode_add_before, -1);
  rb_define_method(SwigClassXmlNode.klass, "add", _wrap_XmlNode_add, -1);
  rb_define_alias(SwigClassXmlNode.klass, "<<", "add");
  rb_define_method(SwigClassXmlNode.klass, "each", _wrap_XmlNode_each, -1);
  rb_define_method(SwigClassXmlNode.klass, "get", _wrap_XmlNode_get, -1);
  rb_define_alias(SwigClassXmlNode.klass, "[]", "get");
  rb_define_method(SwigClassXmlNode.klass, "attr", _wrap_XmlNode_attr, -1);
  rb_define_method(SwigClassXmlNode.klass, "attr_count", _wrap_XmlNode_attr_count, -1);
  rb_define_method(SwigClassXmlNode.klass, "attr_find", _wrap_XmlNode_attr_find, -1);
  rb_define_method(SwigClassXmlNode.klass, "attr_add", _wrap_XmlNode_attr_add, -1);
  rb_define_method(SwigClassXmlNode.klass, "epr", _wrap_XmlNode_epr, -1);
  rb_define_method(SwigClassXmlNode.klass, "each_attr", _wrap_XmlNode_each_attr, -1);
  SwigClassXmlNode.mark = 0;
  SwigClassXmlNode.destroy = (void (*)(void *)) free___WsXmlNode;
  SwigClassXmlNode.trackObjects = 0;
  
  SwigClassXmlAttr.klass = rb_define_class_under(mOpenwsman, "XmlAttr", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlAttr, (void *) &SwigClassXmlAttr);
  rb_undef_alloc_func(SwigClassXmlAttr.klass);
  rb_define_method(SwigClassXmlAttr.klass, "name", _wrap_XmlAttr_name, -1);
  rb_define_method(SwigClassXmlAttr.klass, "ns", _wrap_XmlAttr_ns, -1);
  rb_define_method(SwigClassXmlAttr.klass, "value", _wrap_XmlAttr_value, -1);
  rb_define_alias(SwigClassXmlAttr.klass, "to_s", "value");
  rb_define_method(SwigClassXmlAttr.klass, "remove", _wrap_XmlAttr_remove, -1);
  SwigClassXmlAttr.mark = 0;
  SwigClassXmlAttr.trackObjects = 0;
  rb_define_const(mOpenwsman, "XML_NS_SOAP_1_1", SWIG_FromCharPtr("http://schemas.xmlsoap.org/soap/envelope"));
  rb_define_const(mOpenwsman, "XML_NS_SOAP_1_2", SWIG_FromCharPtr("http://www.w3.org/2003/05/soap-envelope"));
  rb_define_const(mOpenwsman, "XML_NS_XML_NAMESPACES", SWIG_FromCharPtr("http://www.w3.org/XML/1998/namespace"));
  rb_define_const(mOpenwsman, "XML_NS_ADDRESSING", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing"));
  rb_define_const(mOpenwsman, "XML_NS_DISCOVERY", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/10/discovery"));
  rb_define_const(mOpenwsman, "XML_NS_EVENTING", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing"));
  rb_define_const(mOpenwsman, "XML_NS_ENUMERATION", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration"));
  rb_define_const(mOpenwsman, "XML_NS_TRANSFER", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer"));
  rb_define_const(mOpenwsman, "XML_NS_XML_SCHEMA", SWIG_FromCharPtr("http://www.w3.org/2001/XMLSchema"));
  rb_define_const(mOpenwsman, "XML_NS_SCHEMA_INSTANCE", SWIG_FromCharPtr("http://www.w3.org/2001/XMLSchema-instance"));
  rb_define_const(mOpenwsman, "XML_NS_POLICY", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/policy"));
  rb_define_const(mOpenwsman, "XML_NS_TRUST", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2005/02/trust"));
  rb_define_const(mOpenwsman, "XML_NS_SE", SWIG_FromCharPtr("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"));
  rb_define_const(mOpenwsman, "XML_NS_SCHEMA_INSTANCE_PREFIX", SWIG_FromCharPtr("xsi"));
  rb_define_const(mOpenwsman, "XML_NS_SCHEMA_INSTANCE_NIL", SWIG_FromCharPtr("nil"));
  rb_define_const(mOpenwsman, "XML_NS_OPENWSMAN", SWIG_FromCharPtr("http://schema.openwsman.org/2006/openwsman"));
  rb_define_const(mOpenwsman, "XML_NS_CIM_SCHEMA", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/common"));
  rb_define_const(mOpenwsman, "XML_NS_CIM_CLASS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2"));
  rb_define_const(mOpenwsman, "XML_NS_CIM_BINDING", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd"));
  rb_define_const(mOpenwsman, "XML_NS_CIM_INTRINSIC", SWIG_FromCharPtr("http://schemas.openwsman.org/wbem/wscim/1/intrinsic"));
  rb_define_const(mOpenwsman, "XML_NS_WS_MAN", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"));
  rb_define_const(mOpenwsman, "XML_NS_WSMAN_FAULT_DETAIL", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/faultDetail"));
  rb_define_const(mOpenwsman, "XML_NS_WS_MAN_CAT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2005/06/wsmancat"));
  rb_define_const(mOpenwsman, "XML_NS_WSMAN_ID", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/identity/1/wsmanidentity.xsd"));
  rb_define_const(mOpenwsman, "SOAP1_2_CONTENT_TYPE", SWIG_FromCharPtr("application/soap+xml;charset=UTF-8"));
  rb_define_const(mOpenwsman, "SOAP_CONTENT_TYPE", SWIG_FromCharPtr("application/soap+xml"));
  rb_define_const(mOpenwsman, "CIMXML_CONTENT_TYPE", SWIG_FromCharPtr("application/xml"));
  rb_define_const(mOpenwsman, "SOAP_SKIP_DEF_FILTERS", SWIG_From_int((int)(0x01)));
  rb_define_const(mOpenwsman, "SOAP_ACTION_PREFIX", SWIG_From_int((int)(0x02)));
  rb_define_const(mOpenwsman, "SOAP_ONE_WAY_OP", SWIG_From_int((int)(0x04)));
  rb_define_const(mOpenwsman, "SOAP_NO_RESP_OP", SWIG_From_int((int)(0x08)));
  rb_define_const(mOpenwsman, "SOAP_DONT_KEEP_INDOC", SWIG_From_int((int)(0x10)));
  rb_define_const(mOpenwsman, "SOAP_CLIENT_RESPONSE", SWIG_From_int((int)(0x20)));
  rb_define_const(mOpenwsman, "SOAP_CUSTOM_DISPATCHER", SWIG_From_int((int)(0x40)));
  rb_define_const(mOpenwsman, "SOAP_IDENTIFY_DISPATCH", SWIG_From_int((int)(0x80)));
  rb_define_const(mOpenwsman, "WSMID_IDENTIFY", SWIG_FromCharPtr("Identify"));
  rb_define_const(mOpenwsman, "WSMID_IDENTIFY_RESPONSE", SWIG_FromCharPtr("IdentifyResponse"));
  rb_define_const(mOpenwsman, "WSMID_PROTOCOL_VERSION", SWIG_FromCharPtr("ProtocolVersion"));
  rb_define_const(mOpenwsman, "WSMID_PRODUCT_VENDOR", SWIG_FromCharPtr("ProductVendor"));
  rb_define_const(mOpenwsman, "WSMID_PRODUCT_VERSION", SWIG_FromCharPtr("ProductVersion"));
  rb_define_const(mOpenwsman, "XML_SCHEMA_NIL", SWIG_FromCharPtr("nil"));
  rb_define_const(mOpenwsman, "WSA_TO_ANONYMOUS", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"));
  rb_define_const(mOpenwsman, "WSA_MESSAGE_ID", SWIG_FromCharPtr("MessageID"));
  rb_define_const(mOpenwsman, "WSA_ADDRESS", SWIG_FromCharPtr("Address"));
  rb_define_const(mOpenwsman, "WSA_EPR", SWIG_FromCharPtr("EndpointReference"));
  rb_define_const(mOpenwsman, "WSA_ACTION", SWIG_FromCharPtr("Action"));
  rb_define_const(mOpenwsman, "WSA_RELATES_TO", SWIG_FromCharPtr("RelatesTo"));
  rb_define_const(mOpenwsman, "WSA_TO", SWIG_FromCharPtr("To"));
  rb_define_const(mOpenwsman, "WSA_REPLY_TO", SWIG_FromCharPtr("ReplyTo"));
  rb_define_const(mOpenwsman, "WSA_FROM", SWIG_FromCharPtr("From"));
  rb_define_const(mOpenwsman, "WSA_FAULT_TO", SWIG_FromCharPtr("FaultTo"));
  rb_define_const(mOpenwsman, "WSA_REFERENCE_PROPERTIES", SWIG_FromCharPtr("ReferenceProperties"));
  rb_define_const(mOpenwsman, "WSA_REFERENCE_PARAMETERS", SWIG_FromCharPtr("ReferenceParameters"));
  rb_define_const(mOpenwsman, "WSA_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing/fault"));
  rb_define_const(mOpenwsman, "SOAP_ENVELOPE", SWIG_FromCharPtr("Envelope"));
  rb_define_const(mOpenwsman, "SOAP_HEADER", SWIG_FromCharPtr("Header"));
  rb_define_const(mOpenwsman, "SOAP_BODY", SWIG_FromCharPtr("Body"));
  rb_define_const(mOpenwsman, "SOAP_FAULT", SWIG_FromCharPtr("Fault"));
  rb_define_const(mOpenwsman, "SOAP_CODE", SWIG_FromCharPtr("Code"));
  rb_define_const(mOpenwsman, "SOAP_VALUE", SWIG_FromCharPtr("Value"));
  rb_define_const(mOpenwsman, "SOAP_SUBCODE", SWIG_FromCharPtr("Subcode"));
  rb_define_const(mOpenwsman, "SOAP_REASON", SWIG_FromCharPtr("Reason"));
  rb_define_const(mOpenwsman, "SOAP_TEXT", SWIG_FromCharPtr("Text"));
  rb_define_const(mOpenwsman, "SOAP_LANG", SWIG_FromCharPtr("lang"));
  rb_define_const(mOpenwsman, "SOAP_DETAIL", SWIG_FromCharPtr("Detail"));
  rb_define_const(mOpenwsman, "SOAP_FAULT_DETAIL", SWIG_FromCharPtr("FaultDetail"));
  rb_define_const(mOpenwsman, "SOAP_MUST_UNDERSTAND", SWIG_FromCharPtr("mustUnderstand"));
  rb_define_const(mOpenwsman, "SOAP_VERSION_MISMATCH", SWIG_FromCharPtr("VersionMismatch"));
  rb_define_const(mOpenwsman, "SOAP_UPGRADE", SWIG_FromCharPtr("Upgrade"));
  rb_define_const(mOpenwsman, "SOAP_SUPPORTED_ENVELOPE", SWIG_FromCharPtr("SupportedEnvelope"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_GET", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Get"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_GETRESPONSE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/GetResponse"));
  rb_define_const(mOpenwsman, "TRANSFER_GET", SWIG_FromCharPtr("Get"));
  rb_define_const(mOpenwsman, "TRANSFER_GET_RESP", SWIG_FromCharPtr("GetResponse"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_PUT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Put"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_PUTRESPONSE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/PutResponse"));
  rb_define_const(mOpenwsman, "TRANSFER_PUT", SWIG_FromCharPtr("Put"));
  rb_define_const(mOpenwsman, "TRANSFER_PUT_RESP", SWIG_FromCharPtr("PutResponse"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_CREATE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Create"));
  rb_define_const(mOpenwsman, "TRANSFER_CREATE", SWIG_FromCharPtr("Create"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_DELETE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete"));
  rb_define_const(mOpenwsman, "TRANSFER_DELETE", SWIG_FromCharPtr("Delete"));
  rb_define_const(mOpenwsman, "TRANSFER_ACTION_DELETERESPONSE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/DeleteResponse"));
  rb_define_const(mOpenwsman, "TRANSFER_DELETE_RESP", SWIG_FromCharPtr("DeleteResponse"));
  rb_define_const(mOpenwsman, "WSXF_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/fault"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_ENUMERATE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_ENUMERATERESPONSE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/EnumerateResponse"));
  rb_define_const(mOpenwsman, "WSENUM_ENUMERATE", SWIG_FromCharPtr("Enumerate"));
  rb_define_const(mOpenwsman, "WSENUM_ENUMERATE_RESP", SWIG_FromCharPtr("EnumerateResponse"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_RELEASE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Release"));
  rb_define_const(mOpenwsman, "WSENUM_RELEASE", SWIG_FromCharPtr("Release"));
  rb_define_const(mOpenwsman, "WSENUM_RELEASE_RESP", SWIG_FromCharPtr("ReleaseResponse"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_PULL", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull"));
  rb_define_const(mOpenwsman, "WSENUM_PULL", SWIG_FromCharPtr("Pull"));
  rb_define_const(mOpenwsman, "WSENUM_PULL_RESP", SWIG_FromCharPtr("PullResponse"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_RENEW", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Renew"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_GETSTATUS", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/GetStatus"));
  rb_define_const(mOpenwsman, "ENUM_ACTION_ENUMEND", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/EnumerationEnd"));
  rb_define_const(mOpenwsman, "EVT_ACTION_SUBSCRIBE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Subscribe"));
  rb_define_const(mOpenwsman, "WSEVENT_SUBSCRIBE", SWIG_FromCharPtr("Subscribe"));
  rb_define_const(mOpenwsman, "WSEVENT_SUBSCRIBE_RESP", SWIG_FromCharPtr("SubscribeResponse"));
  rb_define_const(mOpenwsman, "EVT_ACTION_PULL", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Pull"));
  rb_define_const(mOpenwsman, "EVT_ACTION_GETSTATUS", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/GetStaus"));
  rb_define_const(mOpenwsman, "EVT_ACTION_UNSUBSCRIBE", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Unsubscribe"));
  rb_define_const(mOpenwsman, "WSEVENT_UNSUBSCRIBE", SWIG_FromCharPtr("Unsubscribe"));
  rb_define_const(mOpenwsman, "WSEVENT_UNSUBSCRIBE_RESP", SWIG_FromCharPtr("UnsubscribeResponse"));
  rb_define_const(mOpenwsman, "EVT_ACTION_SUBEND", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/SubscriptionEnd"));
  rb_define_const(mOpenwsman, "EVT_ACTION_RENEW", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Renew"));
  rb_define_const(mOpenwsman, "WSEVENT_RENEW", SWIG_FromCharPtr("Renew"));
  rb_define_const(mOpenwsman, "WSEVENT_RENEW_RESP", SWIG_FromCharPtr("RenewResponse"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_EVENTS", SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Events"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_HEARTBEAT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Heartbeat"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_DROPPEDEVENTS", SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/DroppedEvents"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_ACK", SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Ack"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_EVENT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Event"));
  rb_define_const(mOpenwsman, "WSENUM_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/fault"));
  rb_define_const(mOpenwsman, "WSENUM_RENEW", SWIG_FromCharPtr("Renew"));
  rb_define_const(mOpenwsman, "WSENUM_RENEW_RESP", SWIG_FromCharPtr("RenewResponse"));
  rb_define_const(mOpenwsman, "WSENUM_GET_STATUS", SWIG_FromCharPtr("GetStatus"));
  rb_define_const(mOpenwsman, "WSENUM_GET_STATUS_RESP", SWIG_FromCharPtr("GetStatusResponse"));
  rb_define_const(mOpenwsman, "WSMAN_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/fault"));
  rb_define_const(mOpenwsman, "WSENUM_END_TO", SWIG_FromCharPtr("EndTo"));
  rb_define_const(mOpenwsman, "WSENUM_EXPIRES", SWIG_FromCharPtr("Expires"));
  rb_define_const(mOpenwsman, "WSENUM_FILTER", SWIG_FromCharPtr("Filter"));
  rb_define_const(mOpenwsman, "WSENUM_DIALECT", SWIG_FromCharPtr("Dialect"));
  rb_define_const(mOpenwsman, "WSENUM_ENUMERATION_CONTEXT", SWIG_FromCharPtr("EnumerationContext"));
  rb_define_const(mOpenwsman, "WSENUM_MAX_TIME", SWIG_FromCharPtr("MaxTime"));
  rb_define_const(mOpenwsman, "WSENUM_MAX_ELEMENTS", SWIG_FromCharPtr("MaxElements"));
  rb_define_const(mOpenwsman, "WSENUM_MAX_CHARACTERS", SWIG_FromCharPtr("MaxCharacters"));
  rb_define_const(mOpenwsman, "WSENUM_ITEMS", SWIG_FromCharPtr("Items"));
  rb_define_const(mOpenwsman, "WSENUM_END_OF_SEQUENCE", SWIG_FromCharPtr("EndOfSequence"));
  rb_define_const(mOpenwsman, "WSENUM_ENUMERATION_END", SWIG_FromCharPtr("EnumerationEnd"));
  rb_define_const(mOpenwsman, "WSENUM_REASON", SWIG_FromCharPtr("Reason"));
  rb_define_const(mOpenwsman, "WSENUM_CODE", SWIG_FromCharPtr("Code"));
  rb_define_const(mOpenwsman, "WSENUM_SOURCE_SHUTTING_DOWN", SWIG_FromCharPtr("SourceShuttingDown"));
  rb_define_const(mOpenwsman, "WSENUM_SOURCE_CANCELING", SWIG_FromCharPtr("SourceCanceling"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY", SWIG_FromCharPtr("Delivery"));
  rb_define_const(mOpenwsman, "WSEVENT_NOTIFY_TO", SWIG_FromCharPtr("NotifyTo"));
  rb_define_const(mOpenwsman, "WSEVENT_ENDTO", SWIG_FromCharPtr("EndTo"));
  rb_define_const(mOpenwsman, "WSEVENT_EXPIRES", SWIG_FromCharPtr("Expires"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY_MODE", SWIG_FromCharPtr("Mode"));
  rb_define_const(mOpenwsman, "WSEVENT_SUBSCRIPTION_MANAGER", SWIG_FromCharPtr("SubscriptionManager"));
  rb_define_const(mOpenwsman, "WSEVENT_IDENTIFIER", SWIG_FromCharPtr("Identifier"));
  rb_define_const(mOpenwsman, "WSEVENT_FILTER", SWIG_FromCharPtr("Filter"));
  rb_define_const(mOpenwsman, "WSEVENT_DIALECT", SWIG_FromCharPtr("Dialect"));
  rb_define_const(mOpenwsman, "WSEVENT_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/fault"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY_MODE_PUSH", SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/DeliveryModes/Push"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY_MODE_PUSHWITHACK", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/PushWithAck"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY_MODE_EVENTS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Events"));
  rb_define_const(mOpenwsman, "WSEVENT_DELIVERY_MODE_PULL", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Pull"));
  rb_define_const(mOpenwsman, "WSMB_ASSOCIATED_INSTANCES", SWIG_FromCharPtr("AssociatedInstances"));
  rb_define_const(mOpenwsman, "WSMB_ASSOCIATION_INSTANCES", SWIG_FromCharPtr("AssociationInstances"));
  rb_define_const(mOpenwsman, "WSMB_OBJECT", SWIG_FromCharPtr("Object"));
  rb_define_const(mOpenwsman, "WSMB_ASSOCIATION_CLASS_NAME", SWIG_FromCharPtr("AssociationClassName"));
  rb_define_const(mOpenwsman, "WSMB_RESULT_CLASS_NAME", SWIG_FromCharPtr("ResultClassName"));
  rb_define_const(mOpenwsman, "WSMB_ROLE", SWIG_FromCharPtr("Role"));
  rb_define_const(mOpenwsman, "WSMB_RESULT_ROLE", SWIG_FromCharPtr("ResultRole"));
  rb_define_const(mOpenwsman, "WSMB_INCLUDE_RESULT_PROPERTY", SWIG_FromCharPtr("IncludeResultProperty"));
  rb_define_const(mOpenwsman, "CIM_RESOURCE_NS_PREFIX", SWIG_FromCharPtr("p"));
  rb_define_const(mOpenwsman, "WSM_SYSTEM", SWIG_FromCharPtr("System"));
  rb_define_const(mOpenwsman, "WSM_LOCALE", SWIG_FromCharPtr("Locale"));
  rb_define_const(mOpenwsman, "WSM_RESOURCE_URI", SWIG_FromCharPtr("ResourceURI"));
  rb_define_const(mOpenwsman, "WSM_SELECTOR_SET", SWIG_FromCharPtr("SelectorSet"));
  rb_define_const(mOpenwsman, "WSM_SELECTOR", SWIG_FromCharPtr("Selector"));
  rb_define_const(mOpenwsman, "WSM_NAME", SWIG_FromCharPtr("Name"));
  rb_define_const(mOpenwsman, "WSM_REQUEST_TOTAL", SWIG_FromCharPtr("RequestTotalItemsCountEstimate"));
  rb_define_const(mOpenwsman, "WSM_TOTAL_ESTIMATE", SWIG_FromCharPtr("TotalItemsCountEstimate"));
  rb_define_const(mOpenwsman, "WSM_OPTIMIZE_ENUM", SWIG_FromCharPtr("OptimizeEnumeration"));
  rb_define_const(mOpenwsman, "WSM_MAX_ELEMENTS", SWIG_FromCharPtr("MaxElements"));
  rb_define_const(mOpenwsman, "WSM_ENUM_EPR", SWIG_FromCharPtr("EnumerateEPR"));
  rb_define_const(mOpenwsman, "WSM_ENUM_OBJ_AND_EPR", SWIG_FromCharPtr("EnumerateObjectAndEPR"));
  rb_define_const(mOpenwsman, "WSM_ENUM_MODE", SWIG_FromCharPtr("EnumerationMode"));
  rb_define_const(mOpenwsman, "WSM_ITEM", SWIG_FromCharPtr("Item"));
  rb_define_const(mOpenwsman, "WSM_FRAGMENT_TRANSFER", SWIG_FromCharPtr("FragmentTransfer"));
  rb_define_const(mOpenwsman, "WSM_XML_FRAGMENT", SWIG_FromCharPtr("XmlFragment"));
  rb_define_const(mOpenwsman, "WSM_OPTION_SET", SWIG_FromCharPtr("OptionSet"));
  rb_define_const(mOpenwsman, "WSM_OPTION", SWIG_FromCharPtr("Option"));
  rb_define_const(mOpenwsman, "WSM_TOTAL", SWIG_FromCharPtr("Total"));
  rb_define_const(mOpenwsman, "WSM_HEARTBEATS", SWIG_FromCharPtr("Heartbeats"));
  rb_define_const(mOpenwsman, "WSM_EVENTS", SWIG_FromCharPtr("Events"));
  rb_define_const(mOpenwsman, "WSM_ACTION", SWIG_FromCharPtr("Action"));
  rb_define_const(mOpenwsman, "WSM_EVENT", SWIG_FromCharPtr("Event"));
  rb_define_const(mOpenwsman, "WSM_ACKREQUESTED", SWIG_FromCharPtr("AckRequested"));
  rb_define_const(mOpenwsman, "WSM_MAX_ENVELOPE_SIZE", SWIG_FromCharPtr("MaxEnvelopeSize"));
  rb_define_const(mOpenwsman, "WSM_OPERATION_TIMEOUT", SWIG_FromCharPtr("OperationTimeout"));
  rb_define_const(mOpenwsman, "WSM_FAULT_SUBCODE", SWIG_FromCharPtr("FaultSubCode"));
  rb_define_const(mOpenwsman, "WSM_FILTER", SWIG_FromCharPtr("Filter"));
  rb_define_const(mOpenwsman, "WSM_DIALECT", SWIG_FromCharPtr("Dialect"));
  rb_define_const(mOpenwsman, "WSM_CONTENTCODING", SWIG_FromCharPtr("ContentEncoding"));
  rb_define_const(mOpenwsman, "WSM_CONNECTIONRETRY", SWIG_FromCharPtr("ConnectionRetry"));
  rb_define_const(mOpenwsman, "WSM_SENDBOOKMARKS", SWIG_FromCharPtr("SendBookmarks"));
  rb_define_const(mOpenwsman, "WSM_BOOKMARK", SWIG_FromCharPtr("Bookmark"));
  rb_define_const(mOpenwsman, "WSM_DROPPEDEVENTS", SWIG_FromCharPtr("DroppedEvents"));
  rb_define_const(mOpenwsman, "WSM_AUTH", SWIG_FromCharPtr("Auth"));
  rb_define_const(mOpenwsman, "WSM_PROFILE", SWIG_FromCharPtr("Profile"));
  rb_define_const(mOpenwsman, "WSM_CERTIFICATETHUMBPRINT", SWIG_FromCharPtr("CertificateThumbprint"));
  rb_define_const(mOpenwsman, "WSM_DEFAULTBOOKMARK", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/bookmark/earliest"));
  rb_define_const(mOpenwsman, "WXF_RESOURCE_CREATED", SWIG_FromCharPtr("ResourceCreated"));
  rb_define_const(mOpenwsman, "WSMB_POLYMORPHISM_MODE", SWIG_FromCharPtr("PolymorphismMode"));
  rb_define_const(mOpenwsman, "WSMB_INCLUDE_SUBCLASS_PROP", SWIG_FromCharPtr("IncludeSubClassProperties"));
  rb_define_const(mOpenwsman, "WSMB_EXCLUDE_SUBCLASS_PROP", SWIG_FromCharPtr("ExcludeSubClassProperties"));
  rb_define_const(mOpenwsman, "WSMB_NONE", SWIG_FromCharPtr("None"));
  rb_define_const(mOpenwsman, "WSMB_DERIVED_REPRESENTATION", SWIG_FromCharPtr("DerivedRepresentation"));
  rb_define_const(mOpenwsman, "WSMB_ACTION_FAULT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding/fault"));
  rb_define_const(mOpenwsman, "WSMB_SHOW_EXTENSION", SWIG_FromCharPtr("ShowExtensions"));
  rb_define_const(mOpenwsman, "WSMB_EXCLUDE_NIL_PROPS", SWIG_FromCharPtr("ExcludeNilProperties"));
  rb_define_const(mOpenwsman, "WSMANCAT_RESOURCE", SWIG_FromCharPtr("Resource"));
  rb_define_const(mOpenwsman, "WSMANCAT_RESOURCE_URI", SWIG_FromCharPtr("ResourceUri"));
  rb_define_const(mOpenwsman, "WSMANCAT_VERSION", SWIG_FromCharPtr("Version"));
  rb_define_const(mOpenwsman, "WSMANCAT_NOTES", SWIG_FromCharPtr("Notes"));
  rb_define_const(mOpenwsman, "WSMANCAT_VENDOR", SWIG_FromCharPtr("Vendor"));
  rb_define_const(mOpenwsman, "WSMANCAT_DISPLAY_NAME", SWIG_FromCharPtr("DisplayName"));
  rb_define_const(mOpenwsman, "WSMANCAT_KEYWORDS", SWIG_FromCharPtr("Keywords"));
  rb_define_const(mOpenwsman, "WSMANCAT_ACCESS", SWIG_FromCharPtr("Access"));
  rb_define_const(mOpenwsman, "WSMANCAT_RELATIONSHIPS", SWIG_FromCharPtr("Relationsships"));
  rb_define_const(mOpenwsman, "WSMANCAT_COMPLIANCE", SWIG_FromCharPtr("Compliance"));
  rb_define_const(mOpenwsman, "WSMANCAT_OPERATION", SWIG_FromCharPtr("Operation"));
  rb_define_const(mOpenwsman, "WSMANCAT_SELECTOR_SET", SWIG_FromCharPtr("SelectorSet"));
  rb_define_const(mOpenwsman, "WSMANCAT_SELECTOR", SWIG_FromCharPtr("Selector"));
  rb_define_const(mOpenwsman, "WSMANCAT_OPTION_SET", SWIG_FromCharPtr("OptionSet"));
  rb_define_const(mOpenwsman, "WSMANCAT_ACTION", SWIG_FromCharPtr("Action"));
  rb_define_const(mOpenwsman, "WSMANCAT_SELECTOR_SET_REF", SWIG_FromCharPtr("SelectorSetRef"));
  rb_define_const(mOpenwsman, "WSMANCAT_LOCATION", SWIG_FromCharPtr("Location"));
  rb_define_const(mOpenwsman, "WSMANCAT_NAME", SWIG_FromCharPtr("Name"));
  rb_define_const(mOpenwsman, "WSMANCAT_TYPE", SWIG_FromCharPtr("Type"));
  rb_define_const(mOpenwsman, "WSM_SELECTOR_FILTER_DIALECT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/SelectorFilter"));
  rb_define_const(mOpenwsman, "WSM_WQL_FILTER_DIALECT", SWIG_FromCharPtr("http://schemas.microsoft.com/wbem/wsman/1/WQL"));
  rb_define_const(mOpenwsman, "WSM_XPATH_FILTER_DIALECT", SWIG_FromCharPtr("http://www.w3.org/TR/1999/REC-xpath-19991116"));
  rb_define_const(mOpenwsman, "WSM_CQL_FILTER_DIALECT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/cql/1/dsp0202.pdf"));
  rb_define_const(mOpenwsman, "WSM_ASSOCIATION_FILTER_DIALECT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding/associationFilter"));
  rb_define_const(mOpenwsman, "WSM_XPATH_EVENTROOT_FILTER", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/filter/eventRootXPath"));
  rb_define_const(mOpenwsman, "WSFW_RESPONSE_STR", SWIG_FromCharPtr("Response"));
  rb_define_const(mOpenwsman, "WSFW_INDOC", SWIG_FromCharPtr("indoc"));
  rb_define_const(mOpenwsman, "WSFW_ENUM_PREFIX", SWIG_FromCharPtr("_en."));
  rb_define_const(mOpenwsman, "CIM_NAMESPACE_SELECTOR", SWIG_FromCharPtr("__cimnamespace"));
  rb_define_const(mOpenwsman, "CIM_ALL_AVAILABLE_CLASSES", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/*"));
  rb_define_const(mOpenwsman, "XML_NS_CIM_ALL_CLASS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1"));
  rb_define_const(mOpenwsman, "CIM_ACTION_ENUMERATE_CLASS_NAMES", SWIG_FromCharPtr("EnumerateClassNames"));
  rb_define_const(mOpenwsman, "CIM_ACTION_ENUMERATE_INSTANCE_NAMES", SWIG_FromCharPtr("EnumerateInstanceNames"));
  rb_define_const(mOpenwsman, "CIM_ACTION_ENUMERATE_CLASSES", SWIG_FromCharPtr("EnumerateClasses"));
  rb_define_const(mOpenwsman, "CIM_ACTION_GET_CLASS", SWIG_FromCharPtr("GetClass"));
  rb_define_const(mOpenwsman, "CIM_ACTION_DELETE_CLASS", SWIG_FromCharPtr("DeleteClass"));
  rb_define_const(mOpenwsman, "WST_ISSUEDTOKENS", SWIG_FromCharPtr("IssuedTokens"));
  rb_define_const(mOpenwsman, "WST_REQUESTSECURITYTOKENRESPONSE", SWIG_FromCharPtr("RequestSecurityTokenResponse"));
  rb_define_const(mOpenwsman, "WST_TOKENTYPE", SWIG_FromCharPtr("TokenType"));
  rb_define_const(mOpenwsman, "WST_REQUESTEDSECURITYTOKEN", SWIG_FromCharPtr("RequestedSecurityToken"));
  rb_define_const(mOpenwsman, "WSSE_USERNAMETOKEN", SWIG_FromCharPtr("UsernameToken"));
  rb_define_const(mOpenwsman, "WSSE_USERNAME", SWIG_FromCharPtr("Username"));
  rb_define_const(mOpenwsman, "WSSE_PASSWORD", SWIG_FromCharPtr("Password"));
  rb_define_const(mOpenwsman, "WSP_APPLIESTO", SWIG_FromCharPtr("AppliesTo"));
  rb_define_const(mOpenwsman, "WST_USERNAMETOKEN", SWIG_FromCharPtr("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#UsernameToken"));
  rb_define_const(mOpenwsman, "WST_CERTIFICATETHUMBPRINT", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/token/certificateThumbprint"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTP_BASIC", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/basic"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTP_DIGEST", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/digest"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_BASIC", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/basic"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_DIGEST", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/digest"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_BASIC", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/basic"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_DIGEST", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/digest"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_SPNEGO_KERBEROS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/spnego-kerberos"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_SPNEGO_KERBEROS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/spnego-kerberos"));
  rb_define_const(mOpenwsman, "WSMAN_SECURITY_PROFILE_HTTP_SPNEGO_KERBEROS", SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/spnego-kerberos"));
  rb_define_const(mOpenwsman, "FLAG_NONE", SWIG_From_int((int)(0x0000)));
  rb_define_const(mOpenwsman, "FLAG_ENUMERATION_COUNT_ESTIMATION", SWIG_From_int((int)(0x0001)));
  rb_define_const(mOpenwsman, "FLAG_ENUMERATION_OPTIMIZATION", SWIG_From_int((int)(0x0002)));
  rb_define_const(mOpenwsman, "FLAG_ENUMERATION_ENUM_EPR", SWIG_From_int((int)(0x0004)));
  rb_define_const(mOpenwsman, "FLAG_ENUMERATION_ENUM_OBJ_AND_EPR", SWIG_From_int((int)(0x0008)));
  rb_define_const(mOpenwsman, "FLAG_DUMP_REQUEST", SWIG_From_int((int)(0x0010)));
  rb_define_const(mOpenwsman, "FLAG_INCLUDESUBCLASSPROPERTIES", SWIG_From_int((int)(0x0020)));
  rb_define_const(mOpenwsman, "FLAG_EXCLUDESUBCLASSPROPERTIES", SWIG_From_int((int)(0x0040)));
  rb_define_const(mOpenwsman, "FLAG_POLYMORPHISM_NONE", SWIG_From_int((int)(0x0080)));
  rb_define_const(mOpenwsman, "FLAG_MUND_MAX_ESIZE", SWIG_From_int((int)(0x0100)));
  rb_define_const(mOpenwsman, "FLAG_MUND_LOCALE", SWIG_From_int((int)(0x0200)));
  rb_define_const(mOpenwsman, "FLAG_MUND_OPTIONSET", SWIG_From_int((int)(0x0400)));
  rb_define_const(mOpenwsman, "FLAG_MUND_FRAGMENT", SWIG_From_int((int)(0x0800)));
  rb_define_const(mOpenwsman, "FLAG_CIM_EXTENSIONS", SWIG_From_int((int)(0x1000)));
  rb_define_const(mOpenwsman, "FLAG_CIM_REFERENCES", SWIG_From_int((int)(0x2000)));
  rb_define_const(mOpenwsman, "FLAG_CIM_ASSOCIATORS", SWIG_From_int((int)(0x4000)));
  rb_define_const(mOpenwsman, "FLAG_EVENT_SENDBOOKMARK", SWIG_From_int((int)(0X8000)));
  rb_define_const(mOpenwsman, "FLAG_CIM_SCHEMA_OPT", SWIG_From_int((int)(0X10000)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_PUSH", SWIG_From_int((int)(0)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_PUSHWITHACK", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_EVENTS", SWIG_From_int((int)(2)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_PULL", SWIG_From_int((int)(3)));
  rb_define_const(mOpenwsman, "WS_NO_AUTH", SWIG_From_int((int)(0)));
  rb_define_const(mOpenwsman, "WS_BASIC_AUTH", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "WS_DIGEST_AUTH", SWIG_From_int((int)(2)));
  rb_define_const(mOpenwsman, "WS_PASS_AUTH", SWIG_From_int((int)(3)));
  rb_define_const(mOpenwsman, "WS_NTLM_AUTH", SWIG_From_int((int)(4)));
  rb_define_const(mOpenwsman, "WS_GSSNEGOTIATE_AUTH", SWIG_From_int((int)(5)));
  rb_define_const(mOpenwsman, "WS_MAX_AUTH", SWIG_From_int((int)(6)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_AUTO", SWIG_From_int((int)(0)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTP_BASIC", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTP_DIGEST", SWIG_From_int((int)(2)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_BASIC", SWIG_From_int((int)(3)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_DIGEST", SWIG_From_int((int)(4)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL", SWIG_From_int((int)(5)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_BASIC", SWIG_From_int((int)(6)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_DIGEST", SWIG_From_int((int)(7)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_SPNEGO_KERBEROS", SWIG_From_int((int)(8)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_SPNEGO_KERBEROS", SWIG_From_int((int)(9)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_SEC_HTTP_SPNEGO_KERBEROS", SWIG_From_int((int)(10)));
  
  SwigClassEndPointReference.klass = rb_define_class_under(mOpenwsman, "EndPointReference", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_epr_t, (void *) &SwigClassEndPointReference);
  rb_define_alloc_func(SwigClassEndPointReference.klass, _wrap_EndPointReference_allocate);
  rb_define_method(SwigClassEndPointReference.klass, "initialize", _wrap_new_EndPointReference, -1);
  rb_define_method(SwigClassEndPointReference.klass, "clone", _wrap_EndPointReference_clone, -1);
  rb_define_method(SwigClassEndPointReference.klass, "add_selector", _wrap_EndPointReference_add_selector, -1);
  rb_define_method(SwigClassEndPointReference.klass, "serialize", _wrap_EndPointReference_serialize, -1);
  rb_define_method(SwigClassEndPointReference.klass, "cmp", _wrap_EndPointReference_cmp, -1);
  rb_define_alias(SwigClassEndPointReference.klass, "==", "cmp");
  rb_define_method(SwigClassEndPointReference.klass, "to_xml", _wrap_EndPointReference_to_xml, -1);
  rb_define_method(SwigClassEndPointReference.klass, "to_s", _wrap_EndPointReference_to_s, -1);
  rb_define_method(SwigClassEndPointReference.klass, "selector_count", _wrap_EndPointReference_selector_count, -1);
  rb_define_method(SwigClassEndPointReference.klass, "resource_uri", _wrap_EndPointReference_resource_uri, -1);
  rb_define_method(SwigClassEndPointReference.klass, "selector", _wrap_EndPointReference_selector, -1);
  rb_define_method(SwigClassEndPointReference.klass, "selector_names", _wrap_EndPointReference_selector_names, -1);
  rb_define_method(SwigClassEndPointReference.klass, "each", _wrap_EndPointReference_each, -1);
  rb_define_method(SwigClassEndPointReference.klass, "classname", _wrap_EndPointReference_classname, -1);
  rb_define_method(SwigClassEndPointReference.klass, "namespace", _wrap_EndPointReference_namespace, -1);
  rb_define_method(SwigClassEndPointReference.klass, "prefix", _wrap_EndPointReference_prefix, -1);
  SwigClassEndPointReference.mark = 0;
  SwigClassEndPointReference.destroy = (void (*)(void *)) free_epr_t;
  SwigClassEndPointReference.trackObjects = 0;
  
  SwigClassFilter.klass = rb_define_class_under(mOpenwsman, "Filter", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_filter_t, (void *) &SwigClassFilter);
  rb_define_alloc_func(SwigClassFilter.klass, _wrap_Filter_allocate);
  rb_define_method(SwigClassFilter.klass, "initialize", _wrap_new_Filter, -1);
  rb_define_method(SwigClassFilter.klass, "resultClass=", _wrap_Filter_resultClass_set, -1);
  rb_define_method(SwigClassFilter.klass, "resultClass", _wrap_Filter_resultClass_get, -1);
  rb_define_method(SwigClassFilter.klass, "assocClass=", _wrap_Filter_assocClass_set, -1);
  rb_define_method(SwigClassFilter.klass, "assocClass", _wrap_Filter_assocClass_get, -1);
  rb_define_method(SwigClassFilter.klass, "associators", _wrap_Filter_associators, -1);
  rb_define_method(SwigClassFilter.klass, "references", _wrap_Filter_references, -1);
  rb_define_method(SwigClassFilter.klass, "simple", _wrap_Filter_simple, -1);
  rb_define_method(SwigClassFilter.klass, "xpath", _wrap_Filter_xpath, -1);
  rb_define_method(SwigClassFilter.klass, "cql", _wrap_Filter_cql, -1);
  rb_define_method(SwigClassFilter.klass, "wql", _wrap_Filter_wql, -1);
  SwigClassFilter.mark = 0;
  SwigClassFilter.destroy = (void (*)(void *)) free_filter_t;
  SwigClassFilter.trackObjects = 0;
  rb_define_const(mOpenwsman, "FAULT_SENDER_CODE", SWIG_FromCharPtr("Sender"));
  rb_define_const(mOpenwsman, "FAULT_MUSTUNDERSTAND_CODE", SWIG_FromCharPtr("MustUnderstand"));
  rb_define_const(mOpenwsman, "FAULT_RECEIVER_CODE", SWIG_FromCharPtr("Receiver"));
  rb_define_const(mOpenwsman, "FAULT_SENDER_CODE_NS", SWIG_FromCharPtr("s:Sender"));
  rb_define_const(mOpenwsman, "FAULT_RECEIVER_CODE_NS", SWIG_FromCharPtr("s:Receiver"));
  rb_define_const(mOpenwsman, "FAULT_MUSTUNDERSTAND_CODE_NS", SWIG_FromCharPtr("s:MustUnderstand"));
  rb_define_const(mOpenwsman, "WSMAN_RC_OK", SWIG_From_int((int)(WSMAN_RC_OK)));
  rb_define_const(mOpenwsman, "WSMAN_ACCESS_DENIED", SWIG_From_int((int)(WSMAN_ACCESS_DENIED)));
  rb_define_const(mOpenwsman, "WSA_ACTION_NOT_SUPPORTED", SWIG_From_int((int)(WSA_ACTION_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSMAN_ALREADY_EXISTS", SWIG_From_int((int)(WSMAN_ALREADY_EXISTS)));
  rb_define_const(mOpenwsman, "WSEN_CANNOT_PROCESS_FILTER", SWIG_From_int((int)(WSEN_CANNOT_PROCESS_FILTER)));
  rb_define_const(mOpenwsman, "WSMAN_CANNOT_PROCESS_FILTER", SWIG_From_int((int)(WSMAN_CANNOT_PROCESS_FILTER)));
  rb_define_const(mOpenwsman, "WSMAN_CONCURRENCY", SWIG_From_int((int)(WSMAN_CONCURRENCY)));
  rb_define_const(mOpenwsman, "WSE_DELIVERY_MODE_REQUESTED_UNAVAILABLE", SWIG_From_int((int)(WSE_DELIVERY_MODE_REQUESTED_UNAVAILABLE)));
  rb_define_const(mOpenwsman, "WSMAN_DELIVERY_REFUSED", SWIG_From_int((int)(WSMAN_DELIVERY_REFUSED)));
  rb_define_const(mOpenwsman, "WSA_DESTINATION_UNREACHABLE", SWIG_From_int((int)(WSA_DESTINATION_UNREACHABLE)));
  rb_define_const(mOpenwsman, "WSMAN_ENCODING_LIMIT", SWIG_From_int((int)(WSMAN_ENCODING_LIMIT)));
  rb_define_const(mOpenwsman, "WSA_ENDPOINT_UNAVAILABLE", SWIG_From_int((int)(WSA_ENDPOINT_UNAVAILABLE)));
  rb_define_const(mOpenwsman, "WSMAN_EVENT_DELIVER_TO_UNUSABLE", SWIG_From_int((int)(WSMAN_EVENT_DELIVER_TO_UNUSABLE)));
  rb_define_const(mOpenwsman, "WSE_EVENT_SOURCE_UNABLE_TO_PROCESS", SWIG_From_int((int)(WSE_EVENT_SOURCE_UNABLE_TO_PROCESS)));
  rb_define_const(mOpenwsman, "WSEN_FILTER_DIALECT_REQUESTED_UNAVAILABLE", SWIG_From_int((int)(WSEN_FILTER_DIALECT_REQUESTED_UNAVAILABLE)));
  rb_define_const(mOpenwsman, "WSE_FILTERING_NOT_SUPPORTED", SWIG_From_int((int)(WSE_FILTERING_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSEN_FILTERING_NOT_SUPPORTED", SWIG_From_int((int)(WSEN_FILTERING_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSE_FILTERING_REQUESTED_UNAVAILABLE", SWIG_From_int((int)(WSE_FILTERING_REQUESTED_UNAVAILABLE)));
  rb_define_const(mOpenwsman, "WSMAN_FRAGMENT_DIALECT_NOT_SUPPORTED", SWIG_From_int((int)(WSMAN_FRAGMENT_DIALECT_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSMAN_INTERNAL_ERROR", SWIG_From_int((int)(WSMAN_INTERNAL_ERROR)));
  rb_define_const(mOpenwsman, "WSMAN_INVALID_BOOKMARK", SWIG_From_int((int)(WSMAN_INVALID_BOOKMARK)));
  rb_define_const(mOpenwsman, "WSEN_INVALID_ENUMERATION_CONTEXT", SWIG_From_int((int)(WSEN_INVALID_ENUMERATION_CONTEXT)));
  rb_define_const(mOpenwsman, "WSE_INVALID_EXPIRATION_TIME", SWIG_From_int((int)(WSE_INVALID_EXPIRATION_TIME)));
  rb_define_const(mOpenwsman, "WSEN_INVALID_EXPIRATION_TIME", SWIG_From_int((int)(WSEN_INVALID_EXPIRATION_TIME)));
  rb_define_const(mOpenwsman, "WSE_INVALID_MESSAGE", SWIG_From_int((int)(WSE_INVALID_MESSAGE)));
  rb_define_const(mOpenwsman, "WSA_INVALID_MESSAGE_INFORMATION_HEADER", SWIG_From_int((int)(WSA_INVALID_MESSAGE_INFORMATION_HEADER)));
  rb_define_const(mOpenwsman, "WSMAN_INVALID_OPTIONS", SWIG_From_int((int)(WSMAN_INVALID_OPTIONS)));
  rb_define_const(mOpenwsman, "WSMAN_INVALID_PARAMETER", SWIG_From_int((int)(WSMAN_INVALID_PARAMETER)));
  rb_define_const(mOpenwsman, "WXF_INVALID_REPRESENTATION", SWIG_From_int((int)(WXF_INVALID_REPRESENTATION)));
  rb_define_const(mOpenwsman, "WSMAN_INVALID_SELECTORS", SWIG_From_int((int)(WSMAN_INVALID_SELECTORS)));
  rb_define_const(mOpenwsman, "WSA_MESSAGE_INFORMATION_HEADER_REQUIRED", SWIG_From_int((int)(WSA_MESSAGE_INFORMATION_HEADER_REQUIRED)));
  rb_define_const(mOpenwsman, "WSMAN_NO_ACK", SWIG_From_int((int)(WSMAN_NO_ACK)));
  rb_define_const(mOpenwsman, "WSMAN_QUOTA_LIMIT", SWIG_From_int((int)(WSMAN_QUOTA_LIMIT)));
  rb_define_const(mOpenwsman, "WSMAN_SCHEMA_VALIDATION_ERROR", SWIG_From_int((int)(WSMAN_SCHEMA_VALIDATION_ERROR)));
  rb_define_const(mOpenwsman, "WSEN_TIMED_OUT", SWIG_From_int((int)(WSEN_TIMED_OUT)));
  rb_define_const(mOpenwsman, "WSMAN_TIMED_OUT", SWIG_From_int((int)(WSMAN_TIMED_OUT)));
  rb_define_const(mOpenwsman, "WSE_UNABLE_TO_RENEW", SWIG_From_int((int)(WSE_UNABLE_TO_RENEW)));
  rb_define_const(mOpenwsman, "WSE_UNSUPPORTED_EXPIRATION_TYPE", SWIG_From_int((int)(WSE_UNSUPPORTED_EXPIRATION_TYPE)));
  rb_define_const(mOpenwsman, "WSEN_UNSUPPORTED_EXPIRATION_TYPE", SWIG_From_int((int)(WSEN_UNSUPPORTED_EXPIRATION_TYPE)));
  rb_define_const(mOpenwsman, "WSMAN_UNSUPPORTED_FEATURE", SWIG_From_int((int)(WSMAN_UNSUPPORTED_FEATURE)));
  rb_define_const(mOpenwsman, "SOAP_FAULT_VERSION_MISMATCH", SWIG_From_int((int)(SOAP_FAULT_VERSION_MISMATCH)));
  rb_define_const(mOpenwsman, "SOAP_FAULT_MUSTUNDERSTAND", SWIG_From_int((int)(SOAP_FAULT_MUSTUNDERSTAND)));
  rb_define_const(mOpenwsman, "WSMB_POLYMORPHISM_MODE_NOT_SUPPORTED", SWIG_From_int((int)(WSMB_POLYMORPHISM_MODE_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSMAN_UNKNOWN", SWIG_From_int((int)(WSMAN_UNKNOWN)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_OK", SWIG_From_int((int)(WSMAN_DETAIL_OK)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ACK", SWIG_From_int((int)(WSMAN_DETAIL_ACK)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ACTION_MISMATCH", SWIG_From_int((int)(WSMAN_DETAIL_ACTION_MISMATCH)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ALREADY_EXISTS", SWIG_From_int((int)(WSMAN_DETAIL_ALREADY_EXISTS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_AMBIGUOUS_SELECTORS", SWIG_From_int((int)(WSMAN_DETAIL_AMBIGUOUS_SELECTORS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ASYNCHRONOUS_REQUEST", SWIG_From_int((int)(WSMAN_DETAIL_ASYNCHRONOUS_REQUEST)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ADDRESSING_MODE", SWIG_From_int((int)(WSMAN_DETAIL_ADDRESSING_MODE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_AUTHERIZATION_MODE", SWIG_From_int((int)(WSMAN_DETAIL_AUTHERIZATION_MODE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_BOOKMARKS", SWIG_From_int((int)(WSMAN_DETAIL_BOOKMARKS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_CHARECHTER_SET", SWIG_From_int((int)(WSMAN_DETAIL_CHARECHTER_SET)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_DELIVERY_RETRIES", SWIG_From_int((int)(WSMAN_DETAIL_DELIVERY_RETRIES)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_DUPLICATE_SELECTORS", SWIG_From_int((int)(WSMAN_DETAIL_DUPLICATE_SELECTORS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ENCODING_TYPE", SWIG_From_int((int)(WSMAN_DETAIL_ENCODING_TYPE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_ENUMERATION_MODE", SWIG_From_int((int)(WSMAN_DETAIL_ENUMERATION_MODE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_EXPIRATION_TIME", SWIG_From_int((int)(WSMAN_DETAIL_EXPIRATION_TIME)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_EXPIRED", SWIG_From_int((int)(WSMAN_DETAIL_EXPIRED)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_FILTERING_REQUIRED", SWIG_From_int((int)(WSMAN_DETAIL_FILTERING_REQUIRED)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_FORMAT_MISMATCH", SWIG_From_int((int)(WSMAN_DETAIL_FORMAT_MISMATCH)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_FORMAT_SECURITY_TOKEN", SWIG_From_int((int)(WSMAN_DETAIL_FORMAT_SECURITY_TOKEN)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_FRAGMENT_LEVEL_ACCESS", SWIG_From_int((int)(WSMAN_DETAIL_FRAGMENT_LEVEL_ACCESS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_HEARTBEATS", SWIG_From_int((int)(WSMAN_DETAIL_HEARTBEATS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INSECURE_ADDRESS", SWIG_From_int((int)(WSMAN_DETAIL_INSECURE_ADDRESS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INSUFFICIENT_SELECTORS", SWIG_From_int((int)(WSMAN_DETAIL_INSUFFICIENT_SELECTORS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID", SWIG_From_int((int)(WSMAN_DETAIL_INVALID)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_ADDRESS", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_ADDRESS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_FORMAT", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_FORMAT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_FRAGMENT", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_FRAGMENT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_NAME", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_NAME)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_NAMESPACE", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_NAMESPACE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_RESOURCEURI", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_RESOURCEURI)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_SELECTOR_ASSIGNMENT", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_SELECTOR_ASSIGNMENT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_SYSTEM", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_SYSTEM)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_TIMEOUT", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_TIMEOUT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_VALUE", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_VALUE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_INVALID_VALUES", SWIG_From_int((int)(WSMAN_DETAIL_INVALID_VALUES)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_LOCALE", SWIG_From_int((int)(WSMAN_DETAIL_LOCALE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MAX_ELEMENTS", SWIG_From_int((int)(WSMAN_DETAIL_MAX_ELEMENTS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MAX_ENVELOPE_POLICY", SWIG_From_int((int)(WSMAN_DETAIL_MAX_ENVELOPE_POLICY)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MAX_ENVELOPE_SIZE", SWIG_From_int((int)(WSMAN_DETAIL_MAX_ENVELOPE_SIZE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MAX_TIME", SWIG_From_int((int)(WSMAN_DETAIL_MAX_TIME)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MINIMUM_ENVELOPE_LIMIT", SWIG_From_int((int)(WSMAN_DETAIL_MINIMUM_ENVELOPE_LIMIT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_MISSING_VALUES", SWIG_From_int((int)(WSMAN_DETAIL_MISSING_VALUES)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_NOT_SUPPORTED", SWIG_From_int((int)(WSMAN_DETAIL_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_OPERATION_TIMEOUT", SWIG_From_int((int)(WSMAN_DETAIL_OPERATION_TIMEOUT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_OPTION_LIMIT", SWIG_From_int((int)(WSMAN_DETAIL_OPTION_LIMIT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_OPTION_SET", SWIG_From_int((int)(WSMAN_DETAIL_OPTION_SET)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_READ_ONLY", SWIG_From_int((int)(WSMAN_DETAIL_READ_ONLY)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_RESOURCE_OFFLINE", SWIG_From_int((int)(WSMAN_DETAIL_RESOURCE_OFFLINE)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_RENAME", SWIG_From_int((int)(WSMAN_DETAIL_RENAME)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_SELECTOR_LIMIT", SWIG_From_int((int)(WSMAN_DETAIL_SELECTOR_LIMIT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_SERVICE_ENVELOPE_LIMIT", SWIG_From_int((int)(WSMAN_DETAIL_SERVICE_ENVELOPE_LIMIT)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_TARGET_ALREADY_EXISTS", SWIG_From_int((int)(WSMAN_DETAIL_TARGET_ALREADY_EXISTS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_TYPE_MISMATCH", SWIG_From_int((int)(WSMAN_DETAIL_TYPE_MISMATCH)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_UNEXPECTED_SELECTORS", SWIG_From_int((int)(WSMAN_DETAIL_UNEXPECTED_SELECTORS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_UNREPORTABLE_SUCCESS", SWIG_From_int((int)(WSMAN_DETAIL_UNREPORTABLE_SUCCESS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_UNUSABLE_ADDRESS", SWIG_From_int((int)(WSMAN_DETAIL_UNUSABLE_ADDRESS)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_URI_LIMIT_EXCEEDED", SWIG_From_int((int)(WSMAN_DETAIL_URI_LIMIT_EXCEEDED)));
  rb_define_const(mOpenwsman, "WSMAN_DETAIL_WHITESPACE", SWIG_From_int((int)(WSMAN_DETAIL_WHITESPACE)));
  rb_define_const(mOpenwsman, "WSA_DETAIL_DUPLICATE_MESSAGE_ID", SWIG_From_int((int)(WSA_DETAIL_DUPLICATE_MESSAGE_ID)));
  rb_define_const(mOpenwsman, "SOAP_DETAIL_HEADER_NOT_UNDERSTOOD", SWIG_From_int((int)(SOAP_DETAIL_HEADER_NOT_UNDERSTOOD)));
  rb_define_const(mOpenwsman, "WST_DETAIL_UNSUPPORTED_TOKENTYPE", SWIG_From_int((int)(WST_DETAIL_UNSUPPORTED_TOKENTYPE)));
  rb_define_const(mOpenwsman, "WSP_DETAIL_INVALID_EPR", SWIG_From_int((int)(WSP_DETAIL_INVALID_EPR)));
  rb_define_const(mOpenwsman, "OWSMAN_DETAIL_ENDPOINT_ERROR", SWIG_From_int((int)(OWSMAN_DETAIL_ENDPOINT_ERROR)));
  rb_define_const(mOpenwsman, "OWSMAN_NO_DETAILS", SWIG_From_int((int)(OWSMAN_NO_DETAILS)));
  rb_define_const(mOpenwsman, "OWSMAN_SYSTEM_ERROR", SWIG_From_int((int)(OWSMAN_SYSTEM_ERROR)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NONE", SWIG_From_int((int)(WSMAN_STATUS_NONE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CANCELLED", SWIG_From_int((int)(WSMAN_STATUS_CANCELLED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CANT_RESOLVE", SWIG_From_int((int)(WSMAN_STATUS_CANT_RESOLVE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CANT_RESOLVE_PROXY", SWIG_From_int((int)(WSMAN_STATUS_CANT_RESOLVE_PROXY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CANT_CONNECT", SWIG_From_int((int)(WSMAN_STATUS_CANT_CONNECT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CANT_CONNECT_PROXY", SWIG_From_int((int)(WSMAN_STATUS_CANT_CONNECT_PROXY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_SSL_FAILED", SWIG_From_int((int)(WSMAN_STATUS_SSL_FAILED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_IO_ERROR", SWIG_From_int((int)(WSMAN_STATUS_IO_ERROR)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_MALFORMED", SWIG_From_int((int)(WSMAN_STATUS_MALFORMED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_TRY_AGAIN", SWIG_From_int((int)(WSMAN_STATUS_TRY_AGAIN)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CONTINUE", SWIG_From_int((int)(WSMAN_STATUS_CONTINUE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_SWITCHING_PROTOCOLS", SWIG_From_int((int)(WSMAN_STATUS_SWITCHING_PROTOCOLS)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PROCESSING", SWIG_From_int((int)(WSMAN_STATUS_PROCESSING)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_OK", SWIG_From_int((int)(WSMAN_STATUS_OK)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CREATED", SWIG_From_int((int)(WSMAN_STATUS_CREATED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_ACCEPTED", SWIG_From_int((int)(WSMAN_STATUS_ACCEPTED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NON_AUTHORITATIVE", SWIG_From_int((int)(WSMAN_STATUS_NON_AUTHORITATIVE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NO_CONTENT", SWIG_From_int((int)(WSMAN_STATUS_NO_CONTENT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_RESET_CONTENT", SWIG_From_int((int)(WSMAN_STATUS_RESET_CONTENT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PARTIAL_CONTENT", SWIG_From_int((int)(WSMAN_STATUS_PARTIAL_CONTENT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_MULTI_STATUS", SWIG_From_int((int)(WSMAN_STATUS_MULTI_STATUS)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_MULTIPLE_CHOICES", SWIG_From_int((int)(WSMAN_STATUS_MULTIPLE_CHOICES)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_MOVED_PERMANENTLY", SWIG_From_int((int)(WSMAN_STATUS_MOVED_PERMANENTLY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_FOUND", SWIG_From_int((int)(WSMAN_STATUS_FOUND)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_MOVED_TEMPORARILY", SWIG_From_int((int)(WSMAN_STATUS_MOVED_TEMPORARILY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_SEE_OTHER", SWIG_From_int((int)(WSMAN_STATUS_SEE_OTHER)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_MODIFIED", SWIG_From_int((int)(WSMAN_STATUS_NOT_MODIFIED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_USE_PROXY", SWIG_From_int((int)(WSMAN_STATUS_USE_PROXY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL", SWIG_From_int((int)(WSMAN_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_TEMPORARY_REDIRECT", SWIG_From_int((int)(WSMAN_STATUS_TEMPORARY_REDIRECT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_BAD_REQUEST", SWIG_From_int((int)(WSMAN_STATUS_BAD_REQUEST)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_UNAUTHORIZED", SWIG_From_int((int)(WSMAN_STATUS_UNAUTHORIZED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PAYMENT_REQUIRED", SWIG_From_int((int)(WSMAN_STATUS_PAYMENT_REQUIRED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_FORBIDDEN", SWIG_From_int((int)(WSMAN_STATUS_FORBIDDEN)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_FOUND", SWIG_From_int((int)(WSMAN_STATUS_NOT_FOUND)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_METHOD_NOT_ALLOWED", SWIG_From_int((int)(WSMAN_STATUS_METHOD_NOT_ALLOWED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_ACCEPTABLE", SWIG_From_int((int)(WSMAN_STATUS_NOT_ACCEPTABLE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PROXY_AUTHENTICATION_REQUIRED", SWIG_From_int((int)(WSMAN_STATUS_PROXY_AUTHENTICATION_REQUIRED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PROXY_UNAUTHORIZED", SWIG_From_int((int)(WSMAN_STATUS_PROXY_UNAUTHORIZED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_REQUEST_TIMEOUT", SWIG_From_int((int)(WSMAN_STATUS_REQUEST_TIMEOUT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_CONFLICT", SWIG_From_int((int)(WSMAN_STATUS_CONFLICT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_GONE", SWIG_From_int((int)(WSMAN_STATUS_GONE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_LENGTH_REQUIRED", SWIG_From_int((int)(WSMAN_STATUS_LENGTH_REQUIRED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_PRECONDITION_FAILED", SWIG_From_int((int)(WSMAN_STATUS_PRECONDITION_FAILED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_REQUEST_ENTITY_TOO_LARGE", SWIG_From_int((int)(WSMAN_STATUS_REQUEST_ENTITY_TOO_LARGE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_REQUEST_URI_TOO_LONG", SWIG_From_int((int)(WSMAN_STATUS_REQUEST_URI_TOO_LONG)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_UNSUPPORTED_MEDIA_TYPE", SWIG_From_int((int)(WSMAN_STATUS_UNSUPPORTED_MEDIA_TYPE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE", SWIG_From_int((int)(WSMAN_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_INVALID_RANGE", SWIG_From_int((int)(WSMAN_STATUS_INVALID_RANGE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_EXPECTATION_FAILED", SWIG_From_int((int)(WSMAN_STATUS_EXPECTATION_FAILED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_UNPROCESSABLE_ENTITY", SWIG_From_int((int)(WSMAN_STATUS_UNPROCESSABLE_ENTITY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_LOCKED", SWIG_From_int((int)(WSMAN_STATUS_LOCKED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_FAILED_DEPENDENCY", SWIG_From_int((int)(WSMAN_STATUS_FAILED_DEPENDENCY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_INTERNAL_SERVER_ERROR", SWIG_From_int((int)(WSMAN_STATUS_INTERNAL_SERVER_ERROR)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_IMPLEMENTED", SWIG_From_int((int)(WSMAN_STATUS_NOT_IMPLEMENTED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_BAD_GATEWAY", SWIG_From_int((int)(WSMAN_STATUS_BAD_GATEWAY)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_SERVICE_UNAVAILABLE", SWIG_From_int((int)(WSMAN_STATUS_SERVICE_UNAVAILABLE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_GATEWAY_TIMEOUT", SWIG_From_int((int)(WSMAN_STATUS_GATEWAY_TIMEOUT)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_HTTP_VERSION_NOT_SUPPORTED", SWIG_From_int((int)(WSMAN_STATUS_HTTP_VERSION_NOT_SUPPORTED)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_INSUFFICIENT_STORAGE", SWIG_From_int((int)(WSMAN_STATUS_INSUFFICIENT_STORAGE)));
  rb_define_const(mOpenwsman, "WSMAN_STATUS_NOT_EXTENDED", SWIG_From_int((int)(WSMAN_STATUS_NOT_EXTENDED)));
  
  SwigClassEnumerateInfo.klass = rb_define_class_under(mOpenwsman, "EnumerateInfo", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___WsEnumerateInfo, (void *) &SwigClassEnumerateInfo);
  rb_undef_alloc_func(SwigClassEnumerateInfo.klass);
  rb_define_method(SwigClassEnumerateInfo.klass, "max_items", _wrap_EnumerateInfo_max_items, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "max_items=", _wrap_EnumerateInfo_max_itemse___, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "flags", _wrap_EnumerateInfo_flags, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "epr_to", _wrap_EnumerateInfo_epr_to, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "epr_uri", _wrap_EnumerateInfo_epr_uri, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "encoding", _wrap_EnumerateInfo_encoding, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "filter", _wrap_EnumerateInfo_filter, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "index", _wrap_EnumerateInfo_index, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "index=", _wrap_EnumerateInfo_indexe___, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "total_items", _wrap_EnumerateInfo_total_items, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "total_items=", _wrap_EnumerateInfo_total_itemse___, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "pull_result", _wrap_EnumerateInfo_pull_result, -1);
  rb_define_method(SwigClassEnumerateInfo.klass, "pull_result=", _wrap_EnumerateInfo_pull_resulte___, -1);
  SwigClassEnumerateInfo.mark = 0;
  SwigClassEnumerateInfo.destroy = (void (*)(void *)) free___WsEnumerateInfo;
  SwigClassEnumerateInfo.trackObjects = 0;
  
  SwigClassSoapOp.klass = rb_define_class_under(mOpenwsman, "SoapOp", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___SoapOp, (void *) &SwigClassSoapOp);
  rb_undef_alloc_func(SwigClassSoapOp.klass);
  rb_define_method(SwigClassSoapOp.klass, "indoc", _wrap_SoapOp_indoc, -1);
  rb_define_method(SwigClassSoapOp.klass, "indoc=", _wrap_SoapOp_indoce___, -1);
  rb_define_method(SwigClassSoapOp.klass, "outdoc", _wrap_SoapOp_outdoc, -1);
  rb_define_method(SwigClassSoapOp.klass, "outdoc=", _wrap_SoapOp_outdoce___, -1);
  rb_define_method(SwigClassSoapOp.klass, "soap", _wrap_SoapOp_soap, -1);
  rb_define_method(SwigClassSoapOp.klass, "msg", _wrap_SoapOp_msg, -1);
  rb_define_method(SwigClassSoapOp.klass, "maxsize", _wrap_SoapOp_maxsize, -1);
  SwigClassSoapOp.mark = 0;
  SwigClassSoapOp.trackObjects = 0;
  
  SwigClassSoap.klass = rb_define_class_under(mOpenwsman, "Soap", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p___Soap, (void *) &SwigClassSoap);
  rb_undef_alloc_func(SwigClassSoap.klass);
  rb_define_method(SwigClassSoap.klass, "create_context", _wrap_Soap_create_context, -1);
  rb_define_method(SwigClassSoap.klass, "context", _wrap_Soap_context, -1);
  rb_define_method(SwigClassSoap.klass, "create_ep_context", _wrap_Soap_create_ep_context, -1);
  SwigClassSoap.mark = 0;
  SwigClassSoap.destroy = (void (*)(void *)) free___Soap;
  SwigClassSoap.trackObjects = 0;
  
  SwigClassContext.klass = rb_define_class_under(mOpenwsman, "Context", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WS_CONTEXT, (void *) &SwigClassContext);
  rb_undef_alloc_func(SwigClassContext.klass);
  rb_define_method(SwigClassContext.klass, "indoc", _wrap_Context_indoc, -1);
  rb_define_method(SwigClassContext.klass, "runtime", _wrap_Context_runtime, -1);
  rb_define_method(SwigClassContext.klass, "enum_idle_timeout=", _wrap_Context_enum_idle_timeoute___, -1);
  rb_define_method(SwigClassContext.klass, "classname", _wrap_Context_classname, -1);
  rb_define_method(SwigClassContext.klass, "method", _wrap_Context_method, -1);
  rb_define_method(SwigClassContext.klass, "method_args", _wrap_Context_method_args, -1);
  rb_define_method(SwigClassContext.klass, "max_elements", _wrap_Context_max_elements, -1);
  rb_define_method(SwigClassContext.klass, "max_envelope_size", _wrap_Context_max_envelope_size, -1);
  rb_define_method(SwigClassContext.klass, "fragment_string", _wrap_Context_fragment_string, -1);
  rb_define_method(SwigClassContext.klass, "selector", _wrap_Context_selector, -1);
  rb_define_method(SwigClassContext.klass, "selectors_from_epr", _wrap_Context_selectors_from_epr, -1);
  rb_define_method(SwigClassContext.klass, "selectors", _wrap_Context_selectors, -1);
  rb_define_method(SwigClassContext.klass, "selectors_from_filter", _wrap_Context_selectors_from_filter, -1);
  rb_define_method(SwigClassContext.klass, "action", _wrap_Context_action, -1);
  rb_define_method(SwigClassContext.klass, "resource_uri", _wrap_Context_resource_uri, -1);
  rb_define_method(SwigClassContext.klass, "option_set", _wrap_Context_option_set, -1);
  rb_define_method(SwigClassContext.klass, "parse_enum_request", _wrap_Context_parse_enum_request, -1);
  SwigClassContext.mark = 0;
  SwigClassContext.destroy = (void (*)(void *)) free__WS_CONTEXT;
  SwigClassContext.trackObjects = 0;
  
  SwigClassStatus.klass = rb_define_class_under(mOpenwsman, "Status", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WsmanStatus, (void *) &SwigClassStatus);
  rb_define_alloc_func(SwigClassStatus.klass, _wrap_Status_allocate);
  rb_define_method(SwigClassStatus.klass, "initialize", _wrap_new_Status, -1);
  rb_define_method(SwigClassStatus.klass, "to_s", _wrap_Status_to_s, -1);
  rb_define_method(SwigClassStatus.klass, "code=", _wrap_Status_codee___, -1);
  rb_define_method(SwigClassStatus.klass, "code", _wrap_Status_code, -1);
  rb_define_method(SwigClassStatus.klass, "detail=", _wrap_Status_detaile___, -1);
  rb_define_method(SwigClassStatus.klass, "detail", _wrap_Status_detail, -1);
  rb_define_method(SwigClassStatus.klass, "msg=", _wrap_Status_msge___, -1);
  rb_define_method(SwigClassStatus.klass, "msg", _wrap_Status_msg, -1);
  rb_define_method(SwigClassStatus.klass, "generate_fault", _wrap_Status_generate_fault, -1);
  SwigClassStatus.mark = 0;
  SwigClassStatus.destroy = (void (*)(void *)) free__WsmanStatus;
  SwigClassStatus.trackObjects = 0;
  
  SwigClassFault.klass = rb_define_class_under(mOpenwsman, "Fault", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WsManFault, (void *) &SwigClassFault);
  rb_define_alloc_func(SwigClassFault.klass, _wrap_Fault_allocate);
  rb_define_method(SwigClassFault.klass, "initialize", _wrap_new_Fault, -1);
  rb_define_method(SwigClassFault.klass, "code", _wrap_Fault_code, -1);
  rb_define_method(SwigClassFault.klass, "subcode", _wrap_Fault_subcode, -1);
  rb_define_method(SwigClassFault.klass, "reason", _wrap_Fault_reason, -1);
  rb_define_method(SwigClassFault.klass, "detail", _wrap_Fault_detail, -1);
  SwigClassFault.mark = 0;
  SwigClassFault.destroy = (void (*)(void *)) free__WsManFault;
  SwigClassFault.trackObjects = 0;
  rb_define_const(mOpenwsman, "NO_AUTH", SWIG_From_int((int)(0)));
  rb_define_const(mOpenwsman, "BASIC_AUTH", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "DIGEST_AUTH", SWIG_From_int((int)(2)));
  rb_define_const(mOpenwsman, "PASS_AUTH", SWIG_From_int((int)(3)));
  rb_define_const(mOpenwsman, "NTLM_AUTH", SWIG_From_int((int)(4)));
  rb_define_const(mOpenwsman, "GSSNEGOTIATE_AUTH", SWIG_From_int((int)(5)));
  rb_define_const(mOpenwsman, "NO_AUTH_STR", SWIG_FromCharPtr("No Auth"));
  rb_define_const(mOpenwsman, "BASIC_AUTH_STR", SWIG_FromCharPtr("Basic"));
  rb_define_const(mOpenwsman, "DIGEST_AUTH_STR", SWIG_FromCharPtr("Digest"));
  rb_define_const(mOpenwsman, "PASS_AUTH_STR", SWIG_FromCharPtr("Passport"));
  rb_define_const(mOpenwsman, "NTLM_AUTH_STR", SWIG_FromCharPtr("NTLM"));
  rb_define_const(mOpenwsman, "GSSNEGOTIATE_AUTH_STR", SWIG_FromCharPtr("GSS-Negotiate"));
  
  SwigClassTransport.klass = rb_define_class_under(mOpenwsman, "Transport", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WsManTransport, (void *) &SwigClassTransport);
  rb_undef_alloc_func(SwigClassTransport.klass);
  rb_define_method(SwigClassTransport.klass, "auth_method?", _wrap_Transport_auth_methodq___, -1);
  rb_define_method(SwigClassTransport.klass, "close", _wrap_Transport_close, -1);
  rb_define_method(SwigClassTransport.klass, "agent=", _wrap_Transport_agente___, -1);
  rb_define_method(SwigClassTransport.klass, "agent", _wrap_Transport_agent, -1);
  rb_define_method(SwigClassTransport.klass, "username", _wrap_Transport_username, -1);
  rb_define_method(SwigClassTransport.klass, "username=", _wrap_Transport_usernamee___, -1);
  rb_define_method(SwigClassTransport.klass, "password", _wrap_Transport_password, -1);
  rb_define_method(SwigClassTransport.klass, "password=", _wrap_Transport_passworde___, -1);
  rb_define_method(SwigClassTransport.klass, "proxy_username", _wrap_Transport_proxy_username, -1);
  rb_define_method(SwigClassTransport.klass, "proxy_username=", _wrap_Transport_proxy_usernamee___, -1);
  rb_define_method(SwigClassTransport.klass, "proxy_password", _wrap_Transport_proxy_password, -1);
  rb_define_method(SwigClassTransport.klass, "proxy_password=", _wrap_Transport_proxy_passworde___, -1);
  rb_define_method(SwigClassTransport.klass, "auth_method=", _wrap_Transport_auth_methode___, -1);
  rb_define_method(SwigClassTransport.klass, "auth_method", _wrap_Transport_auth_method, -1);
  rb_define_singleton_method(SwigClassTransport.klass, "auth_name", _wrap_Transport_auth_name, -1);
  rb_define_method(SwigClassTransport.klass, "auth_value", _wrap_Transport_auth_value, -1);
  rb_define_singleton_method(SwigClassTransport.klass, "error_string", _wrap_Transport_error_string, -1);
  rb_define_method(SwigClassTransport.klass, "timeout=", _wrap_Transport_timeoute___, -1);
  rb_define_method(SwigClassTransport.klass, "timeout", _wrap_Transport_timeout, -1);
  rb_define_method(SwigClassTransport.klass, "verify_peer=", _wrap_Transport_verify_peere___, -1);
  rb_define_method(SwigClassTransport.klass, "verify_peer?", _wrap_Transport_verify_peerq___, -1);
  rb_define_method(SwigClassTransport.klass, "verify_host=", _wrap_Transport_verify_hoste___, -1);
  rb_define_method(SwigClassTransport.klass, "verify_host?", _wrap_Transport_verify_hostq___, -1);
  rb_define_method(SwigClassTransport.klass, "proxy=", _wrap_Transport_proxye___, -1);
  rb_define_method(SwigClassTransport.klass, "proxy", _wrap_Transport_proxy, -1);
  rb_define_method(SwigClassTransport.klass, "proxyauth=", _wrap_Transport_proxyauthe___, -1);
  rb_define_method(SwigClassTransport.klass, "proxyauth", _wrap_Transport_proxyauth, -1);
  rb_define_method(SwigClassTransport.klass, "cainfo=", _wrap_Transport_cainfoe___, -1);
  rb_define_method(SwigClassTransport.klass, "cainfo", _wrap_Transport_cainfo, -1);
  rb_define_method(SwigClassTransport.klass, "certhumbprint=", _wrap_Transport_certhumbprinte___, -1);
  rb_define_method(SwigClassTransport.klass, "certhumbprint", _wrap_Transport_certhumbprint, -1);
  rb_define_method(SwigClassTransport.klass, "capath=", _wrap_Transport_capathe___, -1);
  rb_define_method(SwigClassTransport.klass, "capath", _wrap_Transport_capath, -1);
  rb_define_method(SwigClassTransport.klass, "caoid=", _wrap_Transport_caoide___, -1);
  rb_define_method(SwigClassTransport.klass, "caoid", _wrap_Transport_caoid, -1);
  rb_define_method(SwigClassTransport.klass, "cert=", _wrap_Transport_certe___, -1);
  rb_define_method(SwigClassTransport.klass, "cert", _wrap_Transport_cert, -1);
  rb_define_method(SwigClassTransport.klass, "key=", _wrap_Transport_keye___, -1);
  rb_define_method(SwigClassTransport.klass, "key", _wrap_Transport_key, -1);
  SwigClassTransport.mark = 0;
  SwigClassTransport.trackObjects = 0;
  rb_define_const(mOpenwsman, "TRUE", SWIG_From_int((int)(1)));
  rb_define_const(mOpenwsman, "FALSE", SWIG_From_int((int)(0)));
  rb_define_const(mOpenwsman, "DEFAULT_USER_AGENT", SWIG_FromCharPtr("Openwsman"));
  rb_define_module_function(mOpenwsman, "wsman_send_request", _wrap_wsman_send_request, -1);
  rb_define_module_function(mOpenwsman, "wsmc_transport_set_auth_request_func", _wrap_wsmc_transport_set_auth_request_func, -1);
  rb_define_module_function(mOpenwsman, "wsman_is_auth_method", _wrap_wsman_is_auth_method, -1);
  rb_define_module_function(mOpenwsman, "wsmc_transport_init", _wrap_wsmc_transport_init, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_close_transport", _wrap_wsman_transport_close_transport, -1);
  rb_define_module_function(mOpenwsman, "wsmc_transport_fini", _wrap_wsmc_transport_fini, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_agent", _wrap_wsman_transport_set_agent, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_agent", _wrap_wsman_transport_get_agent, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_user_name", _wrap_wsman_transport_set_user_name, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_user_name", _wrap_wsman_transport_get_user_name, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_password", _wrap_wsman_transport_set_password, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_password", _wrap_wsman_transport_get_password, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_proxy_username", _wrap_wsman_transport_set_proxy_username, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_proxy_username", _wrap_wsman_transport_get_proxy_username, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_proxy_password", _wrap_wsman_transport_set_proxy_password, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_proxy_password", _wrap_wsman_transport_get_proxy_password, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_auth_method", _wrap_wsman_transport_set_auth_method, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_auth_method", _wrap_wsman_transport_get_auth_method, -1);
  rb_define_module_function(mOpenwsman, "wsmc_transport_get_auth_name", _wrap_wsmc_transport_get_auth_name, -1);
  rb_define_module_function(mOpenwsman, "wsmc_transport_get_auth_value", _wrap_wsmc_transport_get_auth_value, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_last_error_string", _wrap_wsman_transport_get_last_error_string, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_timeout", _wrap_wsman_transport_set_timeout, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_timeout", _wrap_wsman_transport_get_timeout, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_verify_peer", _wrap_wsman_transport_set_verify_peer, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_verify_peer", _wrap_wsman_transport_get_verify_peer, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_verify_host", _wrap_wsman_transport_set_verify_host, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_verify_host", _wrap_wsman_transport_get_verify_host, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_crlcheck", _wrap_wsman_transport_set_crlcheck, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_crlcheck", _wrap_wsman_transport_get_crlcheck, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_crlfile", _wrap_wsman_transport_set_crlfile, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_crlfile", _wrap_wsman_transport_get_crlfile, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_proxy", _wrap_wsman_transport_set_proxy, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_proxy", _wrap_wsman_transport_get_proxy, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_proxyauth", _wrap_wsman_transport_set_proxyauth, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_proxyauth", _wrap_wsman_transport_get_proxyauth, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_cainfo", _wrap_wsman_transport_set_cainfo, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_cainfo", _wrap_wsman_transport_get_cainfo, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_certhumbprint", _wrap_wsman_transport_set_certhumbprint, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_certhumbprint", _wrap_wsman_transport_get_certhumbprint, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_capath", _wrap_wsman_transport_set_capath, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_capath", _wrap_wsman_transport_get_capath, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_caoid", _wrap_wsman_transport_set_caoid, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_caoid", _wrap_wsman_transport_get_caoid, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_cert", _wrap_wsman_transport_set_cert, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_cert", _wrap_wsman_transport_get_cert, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_set_key", _wrap_wsman_transport_set_key, -1);
  rb_define_module_function(mOpenwsman, "wsman_transport_get_key", _wrap_wsman_transport_get_key, -1);
  
  SwigClassClientOptions.klass = rb_define_class_under(mOpenwsman, "ClientOptions", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p_client_opt_t, (void *) &SwigClassClientOptions);
  rb_define_alloc_func(SwigClassClientOptions.klass, _wrap_ClientOptions_allocate);
  rb_define_method(SwigClassClientOptions.klass, "initialize", _wrap_new_ClientOptions, -1);
  rb_define_method(SwigClassClientOptions.klass, "set_dump_request", _wrap_ClientOptions_set_dump_request, -1);
  rb_define_method(SwigClassClientOptions.klass, "clear_dump_request", _wrap_ClientOptions_clear_dump_request, -1);
  rb_define_method(SwigClassClientOptions.klass, "flags=", _wrap_ClientOptions_flagse___, -1);
  rb_define_method(SwigClassClientOptions.klass, "flags", _wrap_ClientOptions_flags, -1);
  rb_define_method(SwigClassClientOptions.klass, "clear_flags", _wrap_ClientOptions_clear_flags, -1);
  rb_define_method(SwigClassClientOptions.klass, "reset_flags", _wrap_ClientOptions_reset_flags, -1);
  rb_define_method(SwigClassClientOptions.klass, "max_envelope_size=", _wrap_ClientOptions_max_envelope_sizee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "max_envelope_size", _wrap_ClientOptions_max_envelope_size, -1);
  rb_define_method(SwigClassClientOptions.klass, "max_elements=", _wrap_ClientOptions_max_elementse___, -1);
  rb_define_method(SwigClassClientOptions.klass, "max_elements", _wrap_ClientOptions_max_elements, -1);
  rb_define_method(SwigClassClientOptions.klass, "timeout=", _wrap_ClientOptions_timeoute___, -1);
  rb_define_method(SwigClassClientOptions.klass, "timeout", _wrap_ClientOptions_timeout, -1);
  rb_define_method(SwigClassClientOptions.klass, "fragment=", _wrap_ClientOptions_fragmente___, -1);
  rb_define_method(SwigClassClientOptions.klass, "fragment", _wrap_ClientOptions_fragment, -1);
  rb_define_method(SwigClassClientOptions.klass, "cim_namespace=", _wrap_ClientOptions_cim_namespacee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "cim_namespace", _wrap_ClientOptions_cim_namespace, -1);
  rb_define_method(SwigClassClientOptions.klass, "reference=", _wrap_ClientOptions_referencee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "reference", _wrap_ClientOptions_reference, -1);
  rb_define_method(SwigClassClientOptions.klass, "add_option", _wrap_ClientOptions_add_option, -1);
  rb_define_method(SwigClassClientOptions.klass, "options=", _wrap_ClientOptions_optionse___, -1);
  rb_define_method(SwigClassClientOptions.klass, "options", _wrap_ClientOptions_options, -1);
  rb_define_method(SwigClassClientOptions.klass, "add_selector", _wrap_ClientOptions_add_selector, -1);
  rb_define_method(SwigClassClientOptions.klass, "selectors", _wrap_ClientOptions_selectors, -1);
  rb_define_method(SwigClassClientOptions.klass, "add_property", _wrap_ClientOptions_add_property, -1);
  rb_define_method(SwigClassClientOptions.klass, "properties", _wrap_ClientOptions_properties, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_uri=", _wrap_ClientOptions_delivery_urie___, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_uri", _wrap_ClientOptions_delivery_uri, -1);
  rb_define_method(SwigClassClientOptions.klass, "sub_expiry=", _wrap_ClientOptions_sub_expirye___, -1);
  rb_define_method(SwigClassClientOptions.klass, "sub_expiry", _wrap_ClientOptions_sub_expiry, -1);
  rb_define_method(SwigClassClientOptions.klass, "heartbeat_interval=", _wrap_ClientOptions_heartbeat_intervale___, -1);
  rb_define_method(SwigClassClientOptions.klass, "heartbeat_interval", _wrap_ClientOptions_heartbeat_interval, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_mode=", _wrap_ClientOptions_delivery_modee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_mode", _wrap_ClientOptions_delivery_mode, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_security_mode=", _wrap_ClientOptions_delivery_security_modee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "delivery_sec_mode", _wrap_ClientOptions_delivery_sec_mode, -1);
  rb_define_method(SwigClassClientOptions.klass, "locale=", _wrap_ClientOptions_localee___, -1);
  rb_define_method(SwigClassClientOptions.klass, "locale", _wrap_ClientOptions_locale, -1);
  SwigClassClientOptions.mark = 0;
  SwigClassClientOptions.destroy = (void (*)(void *)) free_client_opt_t;
  SwigClassClientOptions.trackObjects = 0;
  
  SwigClassClient.klass = rb_define_class_under(mOpenwsman, "Client", rb_cObject);
  SWIG_TypeClientData(SWIGTYPE_p__WsManClient, (void *) &SwigClassClient);
  rb_define_alloc_func(SwigClassClient.klass, _wrap_Client_allocate);
  rb_define_method(SwigClassClient.klass, "initialize", _wrap_new_Client, -1);
  rb_define_method(SwigClassClient.klass, "dumpfile=", _wrap_Client_dumpfilee___, -1);
  rb_define_method(SwigClassClient.klass, "response_code", _wrap_Client_response_code, -1);
  rb_define_method(SwigClassClient.klass, "scheme", _wrap_Client_scheme, -1);
  rb_define_method(SwigClassClient.klass, "host", _wrap_Client_host, -1);
  rb_define_method(SwigClassClient.klass, "port", _wrap_Client_port, -1);
  rb_define_method(SwigClassClient.klass, "path", _wrap_Client_path, -1);
  rb_define_method(SwigClassClient.klass, "user", _wrap_Client_user, -1);
  rb_define_method(SwigClassClient.klass, "password", _wrap_Client_password, -1);
  rb_define_method(SwigClassClient.klass, "transport", _wrap_Client_transport, -1);
  rb_define_method(SwigClassClient.klass, "send_request", _wrap_Client_send_request, -1);
  rb_define_method(SwigClassClient.klass, "build_envelope_from_response", _wrap_Client_build_envelope_from_response, -1);
  rb_define_method(SwigClassClient.klass, "encoding", _wrap_Client_encoding, -1);
  rb_define_method(SwigClassClient.klass, "encoding=", _wrap_Client_encodinge___, -1);
  rb_define_method(SwigClassClient.klass, "identify", _wrap_Client_identify, -1);
  rb_define_method(SwigClassClient.klass, "get_from_epr", _wrap_Client_get_from_epr, -1);
  rb_define_method(SwigClassClient.klass, "delete_from_epr", _wrap_Client_delete_from_epr, -1);
  rb_define_method(SwigClassClient.klass, "enumerate", _wrap_Client_enumerate, -1);
  rb_define_method(SwigClassClient.klass, "pull", _wrap_Client_pull, -1);
  rb_define_method(SwigClassClient.klass, "create", _wrap_Client_create, -1);
  rb_define_method(SwigClassClient.klass, "put", _wrap_Client_put, -1);
  rb_define_method(SwigClassClient.klass, "release", _wrap_Client_release, -1);
  rb_define_method(SwigClassClient.klass, "get", _wrap_Client_get, -1);
  rb_define_method(SwigClassClient.klass, "delete", _wrap_Client_delete, -1);
  rb_define_method(SwigClassClient.klass, "invoke", _wrap_Client_invoke, -1);
  rb_define_method(SwigClassClient.klass, "subscribe", _wrap_Client_subscribe, -1);
  rb_define_method(SwigClassClient.klass, "unsubscribe", _wrap_Client_unsubscribe, -1);
  rb_define_method(SwigClassClient.klass, "renew", _wrap_Client_renew, -1);
  rb_define_method(SwigClassClient.klass, "fault_string", _wrap_Client_fault_string, -1);
  rb_define_method(SwigClassClient.klass, "last_error", _wrap_Client_last_error, -1);
  SwigClassClient.mark = 0;
  SwigClassClient.destroy = (void (*)(void *)) free__WsManClient;
  SwigClassClient.trackObjects = 0;
  rb_define_module_function(mOpenwsman, "debug=", _wrap_debuge___, -1);
  rb_define_module_function(mOpenwsman, "debug", _wrap_debug, -1);
  rb_define_module_function(mOpenwsman, "create_soap_envelope", _wrap_create_soap_envelope, -1);
  rb_define_module_function(mOpenwsman, "create_doc_from_file", _wrap_create_doc_from_file, -1);
  rb_define_module_function(mOpenwsman, "create_doc_from_string", _wrap_create_doc_from_string, -1);
  rb_define_module_function(mOpenwsman, "uri_prefix", _wrap_uri_prefix, -1);
}

