/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (https://www.swig.org).
 * Version 4.4.1
 *
 * Do not make changes to this file unless you know what you are doing - modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */


#define SWIG_VERSION 0x040401
#define SWIGPERL
#define SWIG_CASTRANK_MODE

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC) && (__SUNPRO_CC <= 0x560)
#  define SWIGTEMPLATEDISAMBIGUATOR template
# elif defined(__HP_aCC)
/* Needed even with `aCC -AA' when `aCC -V' reports HP ANSI C++ B3910B A.03.55 */
/* If we find a maximum version that requires this, the test would be __HP_aCC <= 35500 for A.03.55 */
#  define SWIGTEMPLATEDISAMBIGUATOR template
# else
#  define SWIGTEMPLATEDISAMBIGUATOR
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__))
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__))
# else
#   define SWIGUNUSED
# endif
#endif

#ifndef SWIG_MSC_UNSUPPRESS_4505
# if defined(_MSC_VER)
#   pragma warning(disable : 4505) /* unreferenced local function has been removed */
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if defined(__GNUC__)
#  if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#    ifndef GCC_HASCLASSVISIBILITY
#      define GCC_HASCLASSVISIBILITY
#    endif
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif

/* Deal with Microsoft's attempt at deprecating methods in the standard C++ library */
#if !defined(SWIG_NO_SCL_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_SCL_SECURE_NO_DEPRECATE)
# define _SCL_SECURE_NO_DEPRECATE
#endif

/* Deal with Apple's deprecated 'AssertMacros.h' from Carbon-framework */
#if defined(__APPLE__) && !defined(__ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES)
# define __ASSERT_MACROS_DEFINE_VERSIONS_WITHOUT_UNDERSCORES 0
#endif

/* Intel's compiler complains if a variable which was never initialised is
 * cast to void, which is a common idiom which we use to indicate that we
 * are aware a variable isn't used.  So we just silence that warning.
 * See: https://github.com/swig/swig/issues/192 for more discussion.
 */
#ifdef __INTEL_COMPILER
# pragma warning disable 592
#endif

#if defined(__cplusplus) && __cplusplus >=201103L
# define SWIG_NOEXCEPT noexcept
#else
# define SWIG_NOEXCEPT throw()
#endif 

/* -----------------------------------------------------------------------------
 * swigcompat.swg
 *
 * Macros to provide support compatibility with older C and C++ standards.
 *
 * Note that SWIG expects __cplusplus to be defined to the appropriate C++ standard.
 * MSVC users are urged to check and examine the /Zc:__cplusplus compiler option.
 * See https://learn.microsoft.com/en-us/cpp/build/reference/zc-cplusplus.
 * ----------------------------------------------------------------------------- */

/* C99 and C++11 should provide snprintf, but define SWIG_NO_SNPRINTF
 * if you're missing it.
 */
#if ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L) || \
     (defined __cplusplus && __cplusplus >= 201103L) || \
     defined SWIG_HAVE_SNPRINTF) && \
    !defined SWIG_NO_SNPRINTF
# define SWIG_snprintf(O,S,F,A) snprintf(O,S,F,A)
# define SWIG_snprintf2(O,S,F,A,B) snprintf(O,S,F,A,B)
#else
/* Fallback versions ignore the buffer size, but most of our uses either have a
 * fixed maximum possible size or dynamically allocate a buffer that's large
 * enough.
 */
# define SWIG_snprintf(O,S,F,A) sprintf(O,F,A)
# define SWIG_snprintf2(O,S,F,A,B) sprintf(O,F,A,B)
#endif

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic C API SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "5"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the SWIG runtime code.
  In 99.9% of the cases, SWIG just needs to declare them as 'static'.

  But only do this if strictly necessary, ie, if you have problems
  with your compiler or suchlike.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1
#define SWIG_CAST_NEW_MEMORY       0x2
#define SWIG_POINTER_NO_NULL       0x4
#define SWIG_POINTER_CLEAR         0x8
#define SWIG_POINTER_RELEASE       (SWIG_POINTER_CLEAR | SWIG_POINTER_DISOWN)

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/*
   Flags/methods for returning states.

   The SWIG conversion methods, as ConvertPtr, return an integer
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).

   Use the following macros/flags to set or process the returning
   states.

   In old versions of SWIG, code such as the following was usually written:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   which is the same really, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }

   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   also requires SWIG_ConvertPtr to return new result values, such as

      int SWIG_ConvertPtr(obj, ptr,...) {
        if (<obj is ok>) {
          if (<need new object>) {
            *ptr = <ptr to new allocated object>;
            return SWIG_NEWOBJ;
          } else {
            *ptr = <ptr to old object>;
            return SWIG_OLDOBJ;
          }
        } else {
          return SWIG_BADOBJ;
        }
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   SWIG errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows returning the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call

      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()
*/

#define SWIG_OK                    (0)
/* Runtime errors are < 0 */
#define SWIG_ERROR                 (-1)
/* Errors in range -1 to -99 are in swigerrors.swg (errors for all languages including those not using the runtime) */
/* Errors in range -100 to -199 are language specific errors defined in *errors.swg */
/* Errors < -200 are generic runtime specific errors */
#define SWIG_ERROR_RELEASE_NOT_OWNED (-200)

#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporary objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del object mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))

/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) {
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) {
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0;
}
#else /* no cast-rank mode */
#  define SWIG_AddCast(r) (r)
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif


#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *, int *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store information on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next array of casts | pointer to cast hashed by value */
  unsigned int            value;		/* index of the last valid element in the array | typename hash value */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/*
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (int)((l1 - f1) - (l2 - f2));
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCmp(const char *nb, const char *tb) {
  int equiv = 1;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (equiv != 0 && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = SWIG_TypeNameComp(nb, ne, tb, te);
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  return SWIG_TypeCmp(nb, tb) == 0 ? 1 : 0;
}

/*
 * Hash function for type name strings, based on maRushPrime1Hash (http://amsoftware.narod.ru/algo2.html)
 */
SWIGRUNTIME unsigned int SWIG_Hash(const char *str, unsigned int len) {
  const unsigned char *data = (const unsigned char *)str;
  unsigned int hash = len, i = 0, k;
  int rem = (int)len;

  while (rem >= (int)sizeof(unsigned int)) {
    memcpy(&k, data, sizeof(unsigned int));
    k += i++;
    hash ^= k;
    hash *= 171717;
    data += sizeof(unsigned int);
    rem -= sizeof(unsigned int);
  }

  switch (rem) {
    case 3: k = (unsigned int)(data[2]) << 16;
            k |= (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 2: k = (unsigned int)(data[1]) << 8;
            k |= (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
    case 1: k = (unsigned int)(data[0]);
            k += i++;
            hash ^= k;
            hash *= 171717;
            break;
  }
  return hash;
}

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  static const unsigned int scan_threshold = 4;
  if (ty) {
    swig_cast_info *head = ty->cast;
    unsigned int hash_value = 0;
    int hashed = 0;

    while (head) {

      if (strcmp(head->type->name, c) == 0) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;
        int search = 1;

        if (!hashed) {
          if (head->value < scan_threshold) {
            for (iter = first; iter <= last; iter++) {
              if (strcmp(iter->type->name, c) == 0) {
                return iter;
              }
            }
            search = 0;
          } else {
            hashed = 1;
            hash_value = SWIG_Hash(c, (unsigned int)strlen(c));
          }
        }

        if (search) {
          /* Binary search over sorted <'next'|'value'> pairs */
          do {
            iter = first + ((last - first) >> 1);
            if (iter->value < hash_value) {
              first = iter + 1;
            } else if (iter->value == hash_value) {

              if (strcmp(iter->next->type->name, c) == 0) {
                return iter->next;
              }

              /* Hash collision check */
              for (last = iter + 1; last->next && last->value == hash_value; last++) {
                if (strcmp(last->next->type->name, c) == 0) {
                  return last->next;
                }
              }
              for (first = iter - 1; first != head && first->value == hash_value; first--) {
                if (strcmp(first->next->type->name, c) == 0) {
                  return first->next;
                }
              }
              break;
            } else
              last = iter - 1;
          } while (first <= last);
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Check the type by type address
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(const swig_type_info *from, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head = ty->cast;
    while (head) {
      if (head->type == from) {
        return head;
      }

      if (head->value) {
        swig_cast_info *iter;
        swig_cast_info *last = head + head->value;
        swig_cast_info *first = head + 1;

        /* Binary search over sorted array of casts */
        do {
          iter = first + ((last - first) >> 1);
          if (iter->type < from) {
            first = iter + 1;
          } else if (iter->type == from) {
            return iter;
          } else
            last = iter - 1;
        } while (first <= last);
      }
      head = head->next;
    }
  }
  return 0;
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr, int *newmemory) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr, newmemory);
}

/*
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  Choose the last
     name. It should be the most specific; a fully resolved name
     but not necessarily with default template parameters expanded. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/*
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *head = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;

  while (head) {
    swig_cast_info *cast;
    for (cast = head; (cast - head) <= head->value; cast++) {
      if (!cast->converter) {
        swig_type_info *tc = cast->type;
        if (!tc->clientdata) {
          SWIG_TypeClientData(tc, clientdata);
        }
      }
    }
    head = head->next;
  }
}

SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}

/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start,
                            swig_module_info *end,
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      size_t l = 0;
      size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	size_t i = (l + r) >> 1;
	const char *iname = iter->types[i]->name;
	if (iname) {
	  int compare = strcmp(name, iname);
	  if (compare == 0) {
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).

  We start searching at module start, and finish searching when start == end.
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start,
                     swig_module_info *end,
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }

  /* neither found a match */
  return 0;
}

/*
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  const unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/*
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  unsigned char *u = (unsigned char *) ptr;
  const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    char d = *(c++);
    unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = (unsigned char)((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = (unsigned char)((d - ('a'-10)) << 4);
    else
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (unsigned char)(d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (unsigned char)(d - ('a'-10));
    else
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/*
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/* SWIG Errors applicable to all language modules, values are reserved from -1 to -99 */
#define  SWIG_UnknownError    	   -1
#define  SWIG_IOError        	   -2
#define  SWIG_RuntimeError   	   -3
#define  SWIG_IndexError     	   -4
#define  SWIG_TypeError      	   -5
#define  SWIG_DivisionByZero 	   -6
#define  SWIG_OverflowError  	   -7
#define  SWIG_SyntaxError    	   -8
#define  SWIG_ValueError     	   -9
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12
#define  SWIG_NullReferenceError   -13


#ifdef __cplusplus
/* Needed on some windows machines---since MS plays funny games with the header files under C++ */
#include <math.h>
#include <stdlib.h>
extern "C" {
#endif

#if __GNUC__ >= 10
#if defined(__cplusplus)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wvolatile"
#endif
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#if __GNUC__ >= 10
#if defined(__cplusplus)
#pragma GCC diagnostic pop
#endif
#endif

/* PERL_REVISION was added in Perl 5.6. */
#if !defined PERL_REVISION || (PERL_REVISION-0 == 5 && PERL_VERSION-0 < 8)
# error SWIG requires Perl >= 5.8.0
#endif

#if defined(WIN32) && defined(PERL_OBJECT) && !defined(PerlIO_exportFILE)
#define PerlIO_exportFILE(fh,fl) (FILE*)(fh)
#endif

#ifndef SvIOK_UV
# define SvIOK_UV(sv)       (SvIOK(sv) && (SvUVX(sv) == SvIVX(sv)))
#endif

#ifndef SvUOK
# define SvUOK(sv)           SvIOK_UV(sv)
#endif

#ifndef IVSIZE
#  ifdef LONGSIZE
#    define IVSIZE LONGSIZE
#  else
#    define IVSIZE 4 /* A bold guess, but the best we can make. */
#  endif
#endif

#ifndef INT2PTR
#  if (IVSIZE == PTRSIZE) && (UVSIZE == PTRSIZE)
#    define PTRV                  UV
#    define INT2PTR(any,d)        (any)(d)
#  else
#    if PTRSIZE == LONGSIZE
#      define PTRV                unsigned long
#    else
#      define PTRV                unsigned
#    endif
#    define INT2PTR(any,d)        (any)(PTRV)(d)
#  endif

#  define NUM2PTR(any,d)  (any)(PTRV)(d)
#  define PTR2IV(p)       INT2PTR(IV,p)
#  define PTR2UV(p)       INT2PTR(UV,p)
#  define PTR2NV(p)       NUM2PTR(NV,p)

#  if PTRSIZE == LONGSIZE
#    define PTR2ul(p)     (unsigned long)(p)
#  else
#    define PTR2ul(p)     INT2PTR(unsigned long,p)
#  endif
#endif /* !INT2PTR */

#ifndef SvPV_nolen
# define SvPV_nolen(x) SvPV(x,PL_na)
#endif

#ifndef get_sv
#  define get_sv perl_get_sv
#endif

#ifndef ERRSV
#  define ERRSV get_sv("@",FALSE)
#endif

#ifndef pTHX_
#define pTHX_
#endif   

#include <string.h>
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGINTERN const char*
SWIG_Perl_ErrorType(int code) {
  switch(code) {
  case SWIG_MemoryError:
    return "MemoryError";
  case SWIG_IOError:
    return "IOError";
  case SWIG_RuntimeError:
    return "RuntimeError";
  case SWIG_IndexError:
    return "IndexError";
  case SWIG_TypeError:
    return "TypeError";
  case SWIG_DivisionByZero:
    return "ZeroDivisionError";
  case SWIG_OverflowError:
    return "OverflowError";
  case SWIG_SyntaxError:
    return "SyntaxError";
  case SWIG_ValueError:
    return "ValueError";
  case SWIG_SystemError:
    return "SystemError";
  case SWIG_AttributeError:
    return "AttributeError";
  case SWIG_NullReferenceError:
    return "NullReferenceError";
  default:
    return "RuntimeError";
  }
}


/* -----------------------------------------------------------------------------
 * perlrun.swg
 *
 * This file contains the runtime support for Perl modules
 * and includes code for managing global variables and pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

#define SWIG_PERL_OBJECT_DECL
#define SWIG_PERL_OBJECT_CALL

/* Common SWIG API */

/* for raw pointers */
#define SWIG_ConvertPtr(obj, pp, type, flags)           SWIG_Perl_ConvertPtr(SWIG_PERL_OBJECT_CALL obj, pp, type, flags)
#define SWIG_ConvertPtrAndOwn(obj, pp, type, flags,own) SWIG_Perl_ConvertPtrAndOwn(SWIG_PERL_OBJECT_CALL obj, pp, type, flags, own)
#define SWIG_NewPointerObj(p, type, flags)              SWIG_Perl_NewPointerObj(SWIG_PERL_OBJECT_CALL p, type, flags)
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Perl_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, p, s, type)             SWIG_Perl_ConvertPacked(SWIG_PERL_OBJECT_CALL obj, p, s, type)
#define SWIG_NewPackedObj(p, s, type)	                SWIG_Perl_NewPackedObj(SWIG_PERL_OBJECT_CALL p, s, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_ConvertPtr(obj, pptr, type, 0)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Perl_GetModule(clientdata)
#define SWIG_SetModule(clientdata, pointer)             SWIG_Perl_SetModule(pointer)


/* Error manipulation */

#define SWIG_ErrorType(code)                            SWIG_Perl_ErrorType(code)               
#define SWIG_Error(code, msg)            		sv_setpvf(get_sv("@", GV_ADD), "%s %s", SWIG_ErrorType(code), msg)
#define SWIG_fail                        		goto fail						    

/* Perl-specific SWIG API */

#define SWIG_MakePtr(sv, ptr, type, flags)              SWIG_Perl_MakePtr(SWIG_PERL_OBJECT_CALL sv, ptr, type, flags)
#define SWIG_MakePackedObj(sv, p, s, type)	        SWIG_Perl_MakePackedObj(SWIG_PERL_OBJECT_CALL sv, p, s, type)
#define SWIG_SetError(str)                              SWIG_Error(SWIG_RuntimeError, str)


#define SWIG_PERL_DECL_ARGS_1(arg1)                     (SWIG_PERL_OBJECT_DECL arg1)
#define SWIG_PERL_CALL_ARGS_1(arg1)                     (SWIG_PERL_OBJECT_CALL arg1)
#define SWIG_PERL_DECL_ARGS_2(arg1, arg2)               (SWIG_PERL_OBJECT_DECL arg1, arg2)
#define SWIG_PERL_CALL_ARGS_2(arg1, arg2)               (SWIG_PERL_OBJECT_CALL arg1, arg2)

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

/* For backward compatibility only */
#define SWIG_POINTER_EXCEPTION  0

#ifdef __cplusplus
extern "C" {
#endif

#define SWIG_OWNER   SWIG_POINTER_OWN
#define SWIG_SHADOW  SWIG_OWNER << 1

#define SWIG_MAYBE_PERL_OBJECT SWIG_PERL_OBJECT_DECL

/* SWIG Perl macros */

/* Macro to declare an XS function */
#ifndef XSPROTO
#   define XSPROTO(name) void name(pTHX_ CV* cv)
#endif

/* Macro to call an XS function */
#ifndef MULTIPLICITY
#  define SWIG_CALLXS(_name) _name(cv)
#else
#  define SWIG_CALLXS(_name) _name(PERL_GET_THX, cv)
#endif

#define MAGIC_PPERL
#define SWIGCLASS_STATIC static SWIGUNUSED

#ifndef MULTIPLICITY
#define SWIG_MAGIC(a,b) (SV *a, MAGIC *b)

#ifdef __cplusplus
extern "C" {
#endif
typedef int (*SwigMagicFunc)(SV *, MAGIC *);
#ifdef __cplusplus
}
#endif

#else /* MULTIPLICITY */

#define SWIG_MAGIC(a,b) (struct interpreter *interp, SV *a, MAGIC *b)

#ifdef __cplusplus
extern "C" {
#endif
typedef int (*SwigMagicFunc)(struct interpreter *, SV *, MAGIC *);
#ifdef __cplusplus
}
#endif

#endif /* MULTIPLICITY */

static void SWIGUNUSED SWIG_croak_null(void)
{
  SV *err = get_sv("@", GV_ADD);
  if (sv_isobject(err))
    croak(0);
  else
    croak("%s", SvPV_nolen(err));
}


/* 
   Define how strict is the cast between strings and integers/doubles
   when overloading between these types occurs.
   
   The default is making it as strict as possible by using SWIG_AddCast
   when needed.
   
   You can use -DSWIG_PERL_NO_STRICT_STR2NUM at compilation time to
   disable the SWIG_AddCast, making the casting between string and
   numbers less strict.

   In the end, we try to solve the overloading between strings and
   numerical types in the more natural way, but if you can avoid it,
   well, avoid it using %rename, for example.
*/
#ifndef SWIG_PERL_NO_STRICT_STR2NUM
# ifndef SWIG_PERL_STRICT_STR2NUM
#  define SWIG_PERL_STRICT_STR2NUM
# endif
#endif
#ifdef SWIG_PERL_STRICT_STR2NUM
/* string takes precedence */
#define SWIG_Str2NumCast(x) SWIG_AddCast(x)  
#else
/* number takes precedence */
#define SWIG_Str2NumCast(x) x
#endif



#include <stdlib.h>

SWIGRUNTIME const char *
SWIG_Perl_TypeProxyName(const swig_type_info *type) {
  if (type) {
    if (type->clientdata != NULL) {
      return (const char *) type->clientdata;
    }
    else {
      return type->name;
    }
  }
  return NULL;
}

/* SWIG_TypeCheck without binary search by hash value */
SWIGRUNTIME swig_cast_info *
SWIG_TypeProxyCheck(const char *c, swig_type_info *ty) {
  if (ty) {
    swig_cast_info *head, *cast;
    head = ty->cast;
    while (head) {
      for (cast = head; (cast - head) <= head->value; cast++) {
        if (strcmp(SWIG_Perl_TypeProxyName(cast->type), c) == 0) {
          return cast;
        }
      }
      head = head->next;
    }
  }
  return 0;
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Perl_AcquirePtr(SWIG_MAYBE_PERL_OBJECT SV *SWIGUNUSEDPARM(sv), int SWIGUNUSEDPARM(own)) {
  /* TODO */
  return 0;
}

/* Function for getting a pointer value */

SWIGRUNTIME int
SWIG_Perl_ConvertPtrAndOwn(SWIG_MAYBE_PERL_OBJECT SV *sv, void **ptr, swig_type_info *_t, int flags, int *own) {
  swig_cast_info *tc;
  void *voidptr = (void *)0;
  SV *tsv = 0;
  int check_owned_pointer_release = (flags & SWIG_POINTER_RELEASE) == SWIG_POINTER_RELEASE;

  if (own)
    *own = 0;

  /* If magical, apply more magic */
  if (SvGMAGICAL(sv))
    mg_get(sv);

  /* Check to see if this is an object */
  if (sv_isobject(sv)) {
    IV tmp = 0;
    tsv = (SV*) SvRV(sv);
    if ((SvTYPE(tsv) == SVt_PVHV)) {
      MAGIC *mg;
      if (SvMAGICAL(tsv)) {
        mg = mg_find(tsv,'P');
        if (mg) {
          sv = mg->mg_obj;
          if (sv_isobject(sv)) {
	    tsv = (SV*)SvRV(sv);
            tmp = SvIV(tsv);
          }
        }
      } else {
        return SWIG_ERROR;
      }
    } else {
      tmp = SvIV(tsv);
    }
    voidptr = INT2PTR(void *,tmp);
  } else if (! SvOK(sv)) {            /* Check for undef */
    *(ptr) = (void *) 0;
    return (flags & SWIG_POINTER_NO_NULL) ? SWIG_NullReferenceError : SWIG_OK;
  } else if (SvTYPE(sv) == SVt_RV) {  /* Check for NULL pointer */
    if (!SvROK(sv)) {
      /* In Perl 5.12 and later, SVt_RV == SVt_IV, so sv could be a valid integer value.  */
      if (SvIOK(sv)) {
        return SWIG_ERROR;
      } else {
        /* NULL pointer (reference to undef). */
        *(ptr) = (void *) 0;
        return SWIG_OK;
      }
    } else {
      return SWIG_ERROR;
    }
  } else {                            /* Don't know what it is */
    return SWIG_ERROR;
  }
  if (_t) {
    /* Now see if the types match */
    char *_c = HvNAME(SvSTASH(SvRV(sv)));
    tc = SWIG_TypeProxyCheck(_c,_t);
#ifdef SWIG_DIRECTORS
    if (!tc && !sv_derived_from(sv,SWIG_Perl_TypeProxyName(_t))) {
#else
    if (!tc) {
#endif
      return SWIG_ERROR;
    }
    {
      int newmemory = 0;
      *ptr = SWIG_TypeCast(tc,voidptr,&newmemory);
      if (newmemory == SWIG_CAST_NEW_MEMORY) {
        assert(own); /* badly formed typemap which will lead to a memory leak - it must set and use own to delete *ptr */
        if (own)
          *own = *own | SWIG_CAST_NEW_MEMORY;
      }
    }
  } else {
    *ptr = voidptr;
  }

  /* 
   *  DISOWN implementation: we need a perl guru to check this one.
   */
  if (tsv && ((flags & SWIG_POINTER_DISOWN) || check_owned_pointer_release)) {
    /* 
     *  almost copy paste code from below SWIG_POINTER_OWN setting
     */
    SV *obj = sv;
    HV *stash = SvSTASH(SvRV(obj));
    GV *gv = *(GV**)hv_fetch(stash, "OWNER", 5, TRUE);
    int owned = 0;
    if (isGV(gv)) {
      HV *hv = GvHVn(gv);
      /*
       * To set ownership (see below), a newSViv(1) entry is added. 
       * Hence, to remove ownership, we delete the entry.
       */
      if (hv_exists_ent(hv, obj, 0)) {
        owned = 1;
        if (flags & SWIG_POINTER_DISOWN) {
          hv_delete_ent(hv, obj, 0, 0);
        }
      }
    }
    if (check_owned_pointer_release && !owned) {
      return SWIG_ERROR_RELEASE_NOT_OWNED;
    }
  }

  if (tsv && (flags & SWIG_POINTER_CLEAR)) {
    SvIV_set(tsv, 0);
  }

  return SWIG_OK;
}

SWIGRUNTIME int
SWIG_Perl_ConvertPtr(SWIG_MAYBE_PERL_OBJECT SV *sv, void **ptr, swig_type_info *_t, int flags) {
  return SWIG_Perl_ConvertPtrAndOwn(sv, ptr, _t, flags, 0);
}

SWIGRUNTIME void
SWIG_Perl_MakePtr(SWIG_MAYBE_PERL_OBJECT SV *sv, void *ptr, swig_type_info *t, int flags) {
  if (ptr && (flags & (SWIG_SHADOW | SWIG_POINTER_OWN))) {
    SV *self;
    SV *obj=newSV(0);
    HV *hash=newHV();
    HV *stash;
    sv_setref_pv(obj, SWIG_Perl_TypeProxyName(t), ptr);
    stash=SvSTASH(SvRV(obj));
    if (flags & SWIG_POINTER_OWN) {
      HV *hv;
      GV *gv = *(GV**)hv_fetch(stash, "OWNER", 5, TRUE);
      if (!isGV(gv))
        gv_init(gv, stash, "OWNER", 5, FALSE);
      hv=GvHVn(gv);
      hv_store_ent(hv, obj, newSViv(1), 0);
    }
    sv_magic((SV *)hash, (SV *)obj, 'P', Nullch, 0);
    SvREFCNT_dec(obj);
    self=newRV_noinc((SV *)hash);
    sv_setsv(sv, self);
    SvREFCNT_dec((SV *)self);
    sv_bless(sv, stash);
  }
  else {
    sv_setref_pv(sv, SWIG_Perl_TypeProxyName(t), ptr);
  }
}

SWIGRUNTIMEINLINE SV *
SWIG_Perl_NewPointerObj(SWIG_MAYBE_PERL_OBJECT void *ptr, swig_type_info *t, int flags) {
  SV *result = sv_newmortal();
  SWIG_MakePtr(result, ptr, t, flags);
  return result;
}

SWIGRUNTIME void
SWIG_Perl_MakePackedObj(SWIG_MAYBE_PERL_OBJECT SV *sv, void *ptr, int sz, swig_type_info *type) {
  char result[1024];
  char *r = result;
  if ((2*sz + 1 + strlen(SWIG_Perl_TypeProxyName(type))) > 1000) return;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  strcpy(r,SWIG_Perl_TypeProxyName(type));
  sv_setpv(sv, result);
}

SWIGRUNTIME SV *
SWIG_Perl_NewPackedObj(SWIG_MAYBE_PERL_OBJECT void *ptr, int sz, swig_type_info *type) {
  SV *result = sv_newmortal();
  SWIG_Perl_MakePackedObj(result, ptr, sz, type);
  return result;
}

/* Convert a packed pointer value */
SWIGRUNTIME int
SWIG_Perl_ConvertPacked(SWIG_MAYBE_PERL_OBJECT SV *obj, void *ptr, int sz, swig_type_info *ty) {
  swig_cast_info *tc;
  const char  *c = 0;

  if ((!obj) || (!SvOK(obj))) return SWIG_ERROR;
  c = SvPV_nolen(obj);
  /* Pointer values must start with leading underscore */
  if (*c != '_') return SWIG_ERROR;
  c++;
  c = SWIG_UnpackData(c,ptr,sz);
  if (ty) {
    tc = SWIG_TypeCheck(c,ty);
    if (!tc) return SWIG_ERROR;
  }
  return SWIG_OK;
}


/* Macros for low-level exception handling */
#define SWIG_croak(x)    { SWIG_Error(SWIG_RuntimeError, x); SWIG_fail; }


typedef XSPROTO(SwigPerlWrapper);
typedef SwigPerlWrapper *SwigPerlWrapperPtr;

/* Structure for command table */
typedef struct {
  const char         *name;
  SwigPerlWrapperPtr  wrapper;
} swig_command_info;

/* Information for constant table */

#define SWIG_INT     1
#define SWIG_FLOAT   2
#define SWIG_STRING  3
#define SWIG_POINTER 4
#define SWIG_BINARY  5

/* Constant information structure */
typedef struct swig_constant_info {
    int              type;
    const char      *name;
    long             lvalue;
    double           dvalue;
    void            *pvalue;
    swig_type_info **ptype;
} swig_constant_info;


/* Structure for variable table */
typedef struct {
  const char   *name;
  SwigMagicFunc   set;
  SwigMagicFunc   get;
  swig_type_info  **type;
} swig_variable_info;

/* Magic variable code */
#ifdef __cplusplus
#  define swig_create_magic(s,a,b,c) _swig_create_magic(s,const_cast<char*>(a),b,c)
#else
#  define swig_create_magic(s,a,b,c) _swig_create_magic(s,(char*)(a),b,c)
#endif
#ifndef MULTIPLICITY
SWIGRUNTIME void _swig_create_magic(SV *sv, char *name, int (*set)(SV *, MAGIC *), int (*get)(SV *,MAGIC *))
#else
SWIGRUNTIME void _swig_create_magic(SV *sv, char *name, int (*set)(struct interpreter*, SV *, MAGIC *), int (*get)(struct interpreter*, SV *,MAGIC *))
#endif
{
  MAGIC *mg;
  sv_magic(sv,sv,'U',name,(I32)strlen(name));
  mg = mg_find(sv,'U');
  mg->mg_virtual = (MGVTBL *) malloc(sizeof(MGVTBL));
  mg->mg_virtual->svt_get = (SwigMagicFunc) get;
  mg->mg_virtual->svt_set = (SwigMagicFunc) set;
  mg->mg_virtual->svt_len = 0;
  mg->mg_virtual->svt_clear = 0;
  mg->mg_virtual->svt_free = 0;
}


SWIGRUNTIME swig_module_info *
SWIG_Perl_GetModule(void *SWIGUNUSEDPARM(clientdata)) {
  static void *type_pointer = (void *)0;
  SV *pointer;

  /* first check if pointer already created */
  if (!type_pointer) {
    pointer = get_sv("swig_runtime_data::type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, FALSE | GV_ADDMULTI);
    if (pointer && SvOK(pointer)) {
      type_pointer = INT2PTR(swig_type_info **, SvIV(pointer));
    }
  }

  return (swig_module_info *) type_pointer;
}

SWIGRUNTIME void
SWIG_Perl_SetModule(swig_module_info *module) {
  SV *pointer;

  /* create a new pointer */
  pointer = get_sv("swig_runtime_data::type_pointer" SWIG_RUNTIME_VERSION SWIG_TYPE_TABLE_NAME, TRUE | GV_ADDMULTI);
  sv_setiv(pointer, PTR2IV(module));
}

#ifdef __cplusplus
}
#endif

/* Workaround perl5 global namespace pollution. Note that undefining library
 * functions like fopen will not solve the problem on all platforms as fopen
 * might be a macro on Windows but not necessarily on other operating systems. */
#ifdef do_open
  #undef do_open
#endif
#ifdef do_close
  #undef do_close
#endif
#ifdef do_exec
  #undef do_exec
#endif
#ifdef scalar
  #undef scalar
#endif
#ifdef list
  #undef list
#endif
#ifdef apply
  #undef apply
#endif
#ifdef convert
  #undef convert
#endif
#ifdef Error
  #undef Error
#endif
#ifdef form
  #undef form
#endif
#ifdef vform
  #undef vform
#endif
#ifdef LABEL
  #undef LABEL
#endif
#ifdef METHOD
  #undef METHOD
#endif
#ifdef Move
  #undef Move
#endif
#ifdef yylex
  #undef yylex
#endif
#ifdef yyparse
  #undef yyparse
#endif
#ifdef yyerror
  #undef yyerror
#endif
#ifdef invert
  #undef invert
#endif
#ifdef ref
  #undef ref
#endif
#ifdef read
  #undef read
#endif
#ifdef write
  #undef write
#endif
#ifdef eof
  #undef eof
#endif
#ifdef close
  #undef close
#endif
#ifdef rewind
  #undef rewind
#endif
#ifdef free
  #undef free
#endif
#ifdef malloc
  #undef malloc
#endif
#ifdef calloc
  #undef calloc
#endif
#ifdef Stat
  #undef Stat
#endif
#ifdef check
  #undef check
#endif
#ifdef seekdir
  #undef seekdir
#endif
#ifdef open
  #undef open
#endif
#ifdef readdir
  #undef readdir
#endif
#ifdef bind
  #undef bind
#endif
#ifdef access
  #undef access
#endif
#ifdef stat
  #undef stat
#endif
#ifdef seed
  #undef seed
#endif

#ifdef bool
  /* Leave if macro is from C99 stdbool.h */
  #ifndef __bool_true_false_are_defined
    #undef bool
  #endif
#endif




#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) do { if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } } while (0) 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_AV swig_types[0]
#define SWIGTYPE_p_FILE swig_types[1]
#define SWIGTYPE_p_WS_LASTERR_Code swig_types[2]
#define SWIGTYPE_p_WsEnumerateInfo swig_types[3]
#define SWIGTYPE_p_WsmanKnownStatusCode swig_types[4]
#define SWIGTYPE_p_WsmanMessage swig_types[5]
#define SWIGTYPE_p__WS_CONTEXT swig_types[6]
#define SWIGTYPE_p__WsManClient swig_types[7]
#define SWIGTYPE_p__WsManFault swig_types[8]
#define SWIGTYPE_p__WsManTransport swig_types[9]
#define SWIGTYPE_p__WsXmlDoc swig_types[10]
#define SWIGTYPE_p__WsmanStatus swig_types[11]
#define SWIGTYPE_p___Soap swig_types[12]
#define SWIGTYPE_p___SoapOp swig_types[13]
#define SWIGTYPE_p___WsEnumerateInfo swig_types[14]
#define SWIGTYPE_p___WsXmlAttr swig_types[15]
#define SWIGTYPE_p___WsXmlNode swig_types[16]
#define SWIGTYPE_p___WsXmlNs swig_types[17]
#define SWIGTYPE_p___WsmanFaultCodeTable swig_types[18]
#define SWIGTYPE_p___WsmanFaultCodeType swig_types[19]
#define SWIGTYPE_p___WsmanFaultDetailTable swig_types[20]
#define SWIGTYPE_p___WsmanFaultDetailType swig_types[21]
#define SWIGTYPE_p_char swig_types[22]
#define SWIGTYPE_p_client_opt_t swig_types[23]
#define SWIGTYPE_p_epr_t swig_types[24]
#define SWIGTYPE_p_filter_t swig_types[25]
#define SWIGTYPE_p_p_char swig_types[26]
#define SWIGTYPE_p_wsman_auth_request_func_t swig_types[27]
#define SWIGTYPE_p_wsman_auth_type_t swig_types[28]
static swig_type_info *swig_types[30];
static swig_module_info swig_module = {swig_types, 29, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#define SWIG_init    boot_openwsman

#define SWIG_name   "openwsmanc::boot_openwsman"
#define SWIG_prefix "openwsmanc::"

#define SWIG_as_voidptr(a) (void *)((const void *)(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),(void**)(a)) 


#ifdef __cplusplus
extern "C"
#endif
#ifndef MULTIPLICITY
SWIGEXPORT void SWIG_init (CV* cv);
#else
SWIGEXPORT void SWIG_init (pTHXo_ CV* cv);
#endif


/*
 * type definitions to keep the C code generic
 */
 
#if defined(SWIGPYTHON)
#define Target_Null_p(x) (x == Py_None)
#define Target_INCREF(x) Py_INCREF(x)
#define Target_DECREF(x) Py_DECREF(x)
#define Target_True Py_True
#define Target_False Py_False
#define Target_Null Py_None
#define Target_Type PyObject*
#define Target_Bool(x) PyBool_FromLong(x)
#define Target_Char16(x) PyInt_FromLong(x)
#define Target_Int(x) PyInt_FromLong(x)
#define Target_String(x) PyString_FromString(x)
#define Target_Real(x) Py_None
#define Target_Array() PyList_New(0)
#define Target_SizedArray(len) PyList_New(len)
#define Target_ListSet(x,n,y) PyList_SetItem(x,n,y)
#define Target_Append(x,y) PyList_Append(x,y)
#include <Python.h>
#define TARGET_THREAD_BEGIN_BLOCK SWIG_PYTHON_THREAD_BEGIN_BLOCK
#define TARGET_THREAD_END_BLOCK SWIG_PYTHON_THREAD_END_BLOCK
#define TARGET_THREAD_BEGIN_ALLOW SWIG_PYTHON_THREAD_BEGIN_ALLOW
#define TARGET_THREAD_END_ALLOW SWIG_PYTHON_THREAD_END_ALLOW
#endif

#if defined(SWIGRUBY)
#define Target_Null_p(x) NIL_P(x)
#define Target_INCREF(x) 
#define Target_DECREF(x) 
#define Target_True Qtrue
#define Target_False Qfalse
#define Target_Null Qnil
#define Target_Type VALUE
#define Target_Bool(x) ((x)?Qtrue:Qfalse)
#define Target_Char16(x) INT2FIX(x)
#define Target_Int(x) INT2FIX(x)
#define Target_String(x) rb_str_new2(x)
#define Target_Real(x) rb_float_new(x)
#define Target_Array() rb_ary_new()
#define Target_SizedArray(len) rb_ary_new2(len)
#define Target_ListSet(x,n,y) rb_ary_store(x,n,y)
#define Target_Append(x,y) rb_ary_push(x,y)
#define TARGET_THREAD_BEGIN_BLOCK do {} while(0)
#define TARGET_THREAD_END_BLOCK do {} while(0)
#define TARGET_THREAD_BEGIN_ALLOW do {} while(0)
#define TARGET_THREAD_END_ALLOW do {} while(0)
#include <ruby.h>
#if HAVE_RUBY_IO_H
#include <ruby/io.h> /* Ruby 1.9 style */
#else
#include <rubyio.h>
#endif
#if HAVE_RUBY_VERSION_H
#include <ruby/version.h>
#endif
#endif

#if defined(SWIGPERL)
#define TARGET_THREAD_BEGIN_BLOCK do {} while(0)
#define TARGET_THREAD_END_BLOCK do {} while(0)
#define TARGET_THREAD_BEGIN_ALLOW do {} while(0)
#define TARGET_THREAD_END_ALLOW do {} while(0)

SWIGINTERNINLINE SV *SWIG_From_long  SWIG_PERL_DECL_ARGS_1(long value);
SWIGINTERNINLINE SV *SWIG_FromCharPtr(const char *cptr);
SWIGINTERNINLINE SV *SWIG_From_double  SWIG_PERL_DECL_ARGS_1(double value);

#define Target_Null_p(x) (x == NULL)
#define Target_INCREF(x) 
#define Target_DECREF(x) 
#define Target_True (&PL_sv_yes)
#define Target_False (&PL_sv_no)
#define Target_Null NULL
#define Target_Type SV *
#define Target_Bool(x) (x)?Target_True:Target_False
#define Target_Char16(x) SWIG_From_long(x)
#define Target_Int(x) SWIG_From_long(x)
#define Target_String(x) SWIG_FromCharPtr(x)
#define Target_Real(x) SWIG_From_double(x)
#define Target_Array() (SV *)newAV()
#define Target_SizedArray(len) (SV *)newAV()
#define Target_ListSet(x,n,y) av_store((AV *)(x),n,y)
#define Target_Append(x,y) av_push(((AV *)(x)), y)
#include <perl.h>
#include <EXTERN.h>
#endif



#if defined(SWIGPERL)
/* filter_t is defined in Perls CORE/perl.h */
#define filter_t filter_type

/* undef Perl's die and warn macros as libu is going to reuse this names */
#undef die
#undef warn
#endif /* SWIGPERL */

#include <u/libu.h>
#include <wsman-types.h>
#include <wsman-client.h>
#include <wsman-client-transport.h>
#include <wsman-api.h>
#include <wsman-xml-binding.h>
#include <wsman-xml.h>
#include <wsman-epr.h>
#include <wsman-filter.h>
#include <wsman-soap.h>
#include <wsman-soap-envelope.h>
#include <openwsman.h>

#if defined(SWIGRUBY)
#include <ruby/helpers.h>

SWIGEXPORT
/* Init_ for the %module, defined by Swig
 * :nodoc:
 */
void Init_Openwsman(void);

SWIGEXPORT
/* Init_ for the .so lib, called by Ruby
 * :nodoc:
 */
void Init__openwsman(void) {
  Init_Openwsman();
}
#endif
#if defined(SWIGPYTHON)
#include <python/helpers.h>
#endif
#if defined(SWIGJAVA)
#include <java/helpers.h>
#endif
#if defined(SWIGPERL)
#include <perl/helpers.h>
#endif

/* Provide WsManTransport definition so it can be used as
 * dedicated datatype in bindings.
 * :nodoc:ly, its aliased to WsManClient
 *
 */
struct _WsManTransport { };
typedef struct _WsManTransport WsManTransport;

static void set_debug(int dbg);

/*
 * Set openwsman debug level.
 * call-seq:
 *   Openwsman::debug = -1 # full debug
 *
 */
static void set_debug(int dbg) {
  static int init = 0;

  if (!init && dbg != 0) {
    init = 1;
    debug_add_handler( debug_message_handler, DEBUG_LEVEL_ALWAYS, NULL );
  }
  wsman_debug_set_level( dbg );
	
}

/* module-level methods */

static int get_debug(void);

/*
 * Return openwsman debug level.
 * call-seq:
 *   Openwsman::debug -> Integer
 *
 */
static int get_debug(void) {
  return (int)wsman_debug_get_level();
}

static WsXmlDocH create_soap_envelope(void);

/*
 * Create empty SOAP envelope
 * call-seq:
 *   Openwsman::create_soap_envelope -> XmlDoc
 *
 */
static WsXmlDocH create_soap_envelope() {
  return ws_xml_create_soap_envelope();
}

static WsXmlDocH create_doc_from_file(const char *filename, const char *encoding);

/*
 * Read XmlDoc from file
 * call-seq:
 *   Openwsman::create_doc_from_file("/path/to/file", "utf-8") -> XmlDoc
 *
 */
static WsXmlDocH create_doc_from_file(const char *filename, const char *encoding) {
  return xml_parser_file_to_doc( filename, encoding, 0);                 
}

static WsXmlDocH create_doc_from_string(const char *buf, const char *encoding);

/*
 * Read XmlDoc from string
 * call-seq:
 *   Openwsman::create_doc_from_string("<xml ...>", "utf-8") -> XmlDoc
 *
 */
static WsXmlDocH create_doc_from_string(const char *buf, const char *encoding) {
  return xml_parser_memory_to_doc( buf, strlen(buf), encoding, 0);
}

static char *uri_classname(const char *uri);
/*
 * get classname from resource URI
 * call-seq:
 *   Openwsman::uri_classname("http://sblim.sf.net/wbem/wscim/1/cim-schema/2/Linux_OperatingSystem") -> "Linux_OperatingSystem"
 *
 */
static char *uri_classname(const char *uri) {
  const char *lastslash = strrchr(uri,'/');
  if (lastslash) {
    return strdup(lastslash+1);
  }
  return NULL;
}

static const char *uri_prefix(const char *classname);
/*
 * Map classname (class schema) to resource uri prefix
 * call-seq:
 *   Openwsman::uri_prefix("Linux") -> "http://sblim.sf.net/wbem/wscim/1/cim-schema/2"
 *   Openwsman::uri_prefix("Win32") -> "http://schemas.microsoft.com/wbem/wsman/1/wmi"
 *
 */
static const char *uri_prefix(const char *classname) {
  static struct map {
    int len;
    const char *schema;
    const char *prefix;
  } mapping[] = {
    /* dmtf CIM */
    { 3, "CIM", "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2" },
    /* dmtf reserved */
    { 3, "PRS", "http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2" },
    /* Microsoft WMI */
    { 5, "Win32", "http://schemas.microsoft.com/wbem/wsman/1/wmi" },
    /* openwbem.org */
    { 8, "OpenWBEM", "http://schema.openwbem.org/wbem/wscim/1/cim-schema/2" },
    /* sblim */
    { 5, "Linux", "http://sblim.sf.net/wbem/wscim/1/cim-schema/2" },
    /* omc-project */
    { 3, "OMC", "http://schema.omc-project.org/wbem/wscim/1/cim-schema/2" },
    /* pegasus.org */
    { 2, "PG", "http://schema.openpegasus.org/wbem/wscim/1/cim-schema/2" },
    /* Intel AMT */
    { 3, "AMT", "http://intel.com/wbem/wscim/1/amt-schema/1" },
    /* Intel */
    { 3, "IPS", "http://intel.com/wbem/wscim/1/ips-schema/1" },
    /* Sun */
    { 3, "Sun","http://schemas.sun.com/wbem/wscim/1/cim-schema/2" },
    /* Microsoft HyperV */
    { 4, "Msvm", "http://schemas.microsoft.com/wbem/wsman/1/wmi" },
    /* Dell DRAC */
    { 4, "DCIM", "http://schemas.dell.com/wbem/wscim/1/cim-schema/2" },
    /* Unisys */
    { 4, "SPAR", "http://schema.unisys.com/wbem/wscim/1/cim-schema/2" },
    /* Fujitsu */
    { 3, "SVS", "http://schemas.ts.fujitsu.com/wbem/wscim/1/cim-schema/2" },
    { 0, NULL, NULL }
  };
  const char *schema_end;
  struct map *map;
  int len;
  if (classname == NULL)
    return NULL;
  if (strcmp(classname, "*") == 0) {
    return "http://schemas.dmtf.org/wbem/wscim/1";
  }
  if ((strcmp(classname, "meta_class") == 0)
      ||(strncmp(classname, "__", 2) == 0)) {
    return "http://schemas.microsoft.com/wbem/wsman/1/wmi";
  }
  schema_end = strchr(classname, '_');
  if (schema_end == NULL)
    return NULL; /* Bad class name */
  len = schema_end - classname;
  map = mapping;
  while (map->len > 0) {
    if ((len == map->len)
        && (strncasecmp(classname, map->schema, map->len) == 0)) {
      return map->prefix;
    }
    ++map;
  }
  return NULL;
}

#if defined(SWIGRUBY)
static epr_t *my_epr_deserialize(WsXmlNodeH node);
/*:nodoc:*/
static epr_t *my_epr_deserialize(WsXmlNodeH node) {
  if (strcmp(WSA_EPR, ws_xml_get_node_local_name(node)) == 0) {
    /* Use node as-is if its already a WSA_EPR */
    return epr_deserialize(node, NULL, NULL, 1);
  }
  /* else search the WSA_EPR node */
  return epr_deserialize(node, XML_NS_ADDRESSING, WSA_EPR, 1);
}

/*:nodoc:*/
static VALUE kv_list_to_hash(list_t *list)
{
  VALUE v = Qnil;
  if (!list_isempty(list)) {
  v = rb_hash_new();
  lnode_t *node = list_first(list);
  while (node) {
    key_value_t *kv = (key_value_t *)node->list_data;
    if (kv->type == 0) {
      rb_hash_aset( v, makestring(kv->key), makestring(kv->v.text));
    }
    else {
      VALUE epr = SWIG_NewPointerObj((void*)kv->v.epr, SWIGTYPE_p_epr_t, 0);
      rb_hash_aset( v, makestring(kv->key), epr);
    }
    node = list_next(list, node);
    }
  }
  return v;
}
#endif

static char *epr_prefix(const char *uri);
/* Get prefix from a EPR uri
 * :nodoc:
 */
static char *epr_prefix(const char *uri) {
  char *classname = uri_classname(uri);
  const char *prefix = uri_prefix(classname);
  if (prefix) {
    const char *lastslash;
    if (strncmp(uri, prefix, strlen(prefix)) == 0) {
      return strdup(prefix);
    }
    lastslash = strrchr(uri, '/');
    if (lastslash) {
      return strndup(uri, lastslash-uri);
    }
  }
  return strdup(uri);
}



SWIGINTERNINLINE SV *
SWIG_From_long  SWIG_PERL_DECL_ARGS_1(long value)
{
  SV *sv;
  if (IVSIZE >= sizeof(value) || (value >= IV_MIN && value <= IV_MAX))
    sv = newSViv(value);
  else
    sv = newSVpvf("%ld", value);
  return sv_2mortal(sv);
}


SWIGINTERNINLINE SV *
SWIG_From_int  SWIG_PERL_DECL_ARGS_1(int value)
{    
  return SWIG_From_long  SWIG_PERL_CALL_ARGS_1(value);
}


SWIGINTERNINLINE SV *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  SV *obj = sv_newmortal();
  if (carray) {
    sv_setpvn(obj, carray, size);
  } else {
    sv_setsv(obj, &PL_sv_undef);
  }
  return obj;
}


SWIGINTERNINLINE SV * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static swig_type_info* info = 0;
  if (!info) {
    info = SWIG_TypeQuery("_p_char");
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(SV *obj, char** cptr, size_t* psize, int *alloc)
{
  if (SvMAGICAL(obj)) {
     SV *tmp = sv_newmortal();
     SvSetSV(tmp, obj);
     obj = tmp;
  }
  if (SvPOK(obj)) {
    STRLEN len = 0;
    char *cstr = SvPV(obj, len); 
    size_t size = len + 1;
    if (cptr) {
      if (alloc) {
	if (*alloc == SWIG_NEWOBJ) {
	  *cptr = (char *)memcpy(malloc((size)*sizeof(char)), cstr, sizeof(char)*(size));
	} else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      }
    }
    if (psize) *psize = size;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      char* vptr = 0; 
      if (SWIG_ConvertPtr(obj, (void**)&vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = vptr;
	if (psize) *psize = vptr ? (strlen(vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}




SWIGINTERN struct _WsXmlDoc *new__WsXmlDoc(char const *name,char const *ns){
    return ws_xml_create_doc(ns, name);
  }
SWIGINTERN void delete__WsXmlDoc(struct _WsXmlDoc *self){
    ws_xml_destroy_doc( self );
  }
SWIGINTERN char *_WsXmlDoc_string(struct _WsXmlDoc *self){
    int size;
    char *buf;
    /* force utf-8 encoding since e.g. winrm sends utf-16 */
    ws_xml_dump_memory_node_tree_enc( ws_xml_get_doc_root(self), &buf, &size, "UTF-8" );
    return buf;
  }
SWIGINTERN char *_WsXmlDoc_encode(struct _WsXmlDoc *self,char const *encoding){
    int size;
    char *buf;
    ws_xml_dump_memory_enc( self, &buf, &size, encoding );
    return buf;
  }
SWIGINTERN void _WsXmlDoc_dump_file(struct _WsXmlDoc *self,FILE *fp){
    ws_xml_dump_doc( fp, self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_root(struct _WsXmlDoc *self){
    return ws_xml_get_doc_root( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_envelope(struct _WsXmlDoc *self){
    return ws_xml_get_soap_envelope( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_header(struct _WsXmlDoc *self){
    return ws_xml_get_soap_header( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_body(struct _WsXmlDoc *self){
    return ws_xml_get_soap_body( self );
  }
SWIGINTERN WsXmlNodeH _WsXmlDoc_element(struct _WsXmlDoc *self,char const *name){
    return ws_xml_get_soap_element( self, name );
  }
SWIGINTERN char const *_WsXmlDoc_context(struct _WsXmlDoc *self){
    char *c = wsmc_get_enum_context( self );
    if (c) {
      if (*c)
        return c;
      u_free(c);
    }
    return NULL;
  }
SWIGINTERN WsXmlDocH _WsXmlDoc_generate_fault(struct _WsXmlDoc *self,WsmanStatus *s){
    return wsman_generate_fault( self, s->fault_code, s->fault_detail_code, s->fault_msg);
  }
SWIGINTERN int _WsXmlDoc_is_fault(struct _WsXmlDoc *self){
    return wsmc_check_for_fault( self );
  }
SWIGINTERN WsManFault *_WsXmlDoc_fault(struct _WsXmlDoc *self){
    WsManFault *f = NULL;
    if (wsmc_check_for_fault(self)) {
      f = (WsManFault *)calloc(1, sizeof(WsManFault));
      wsmc_get_fault_data(self, f);
    }
    return f;
  }
SWIGINTERN WsXmlDocH _WsXmlDoc_create_response_envelope(struct _WsXmlDoc *self,char const *action){
    return wsman_create_response_envelope(self, action);
  }
SWIGINTERN int _WsXmlDoc_is_end_of_sequence(struct _WsXmlDoc *self){
    return NULL != ws_xml_find_in_tree( ws_xml_get_soap_body( self ), XML_NS_ENUMERATION, WSENUM_END_OF_SEQUENCE, 1 );
  }
SWIGINTERN void delete___WsXmlNode(struct __WsXmlNode *self){
    ws_xml_unlink_node(self);
  }
SWIGINTERN char *__WsXmlNode_string(struct __WsXmlNode *self){
    int size;
    char *buf;
    ws_xml_dump_memory_node_tree( self, &buf, &size );
    return buf;
  }
SWIGINTERN void __WsXmlNode_dump_file(struct __WsXmlNode *self,FILE *fp){
    ws_xml_dump_node_tree( fp, self );
  }
SWIGINTERN int __WsXmlNode___eq__(struct __WsXmlNode *self,WsXmlNodeH n){ return self == n; }
SWIGINTERN char *__WsXmlNode_text(struct __WsXmlNode *self){
    return ws_xml_get_node_text( self );
  }
SWIGINTERN void __WsXmlNode_set_text(struct __WsXmlNode *self,char const *text){
    ws_xml_set_node_text( self, text );
  }
SWIGINTERN WsXmlDocH __WsXmlNode_doc(struct __WsXmlNode *self){
    return ws_xml_get_node_doc( self );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_parent(struct __WsXmlNode *self){
    return ws_xml_get_node_parent( self );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_child(struct __WsXmlNode *self){
    return xml_parser_get_first_child(self);
  }
SWIGINTERN char *__WsXmlNode_name(struct __WsXmlNode *self){
    return ws_xml_get_node_local_name( self );
  }
SWIGINTERN void __WsXmlNode_set_name(struct __WsXmlNode *self,char const *name){
    ws_xml_set_node_name( self, ws_xml_get_node_name_ns( self ), name );
  }
SWIGINTERN char *__WsXmlNode_ns(struct __WsXmlNode *self){
    return ws_xml_get_node_name_ns( self );
  }
SWIGINTERN void __WsXmlNode_set_ns(struct __WsXmlNode *self,char const *ns){
    ws_xml_set_ns( self, ns, ws_xml_get_node_name_ns_prefix(self) );
  }
SWIGINTERN char const *__WsXmlNode_prefix(struct __WsXmlNode *self){
    return ws_xml_get_node_name_ns_prefix(self);
  }
SWIGINTERN void __WsXmlNode_set_lang(struct __WsXmlNode *self,char const *lang){
    ws_xml_set_node_lang(self, lang);
  }

#include <limits.h>
#if !defined(SWIG_NO_LLONG_MAX)
# if !defined(LLONG_MAX) && defined(__GNUC__) && defined (__LONG_LONG_MAX__)
#   define LLONG_MAX __LONG_LONG_MAX__
#   define LLONG_MIN (-LLONG_MAX - 1LL)
#   define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
# endif
#endif


#include <errno.h>


#include <stdlib.h>
#ifdef _MSC_VER
# ifndef strtoull
#  define strtoull _strtoui64
# endif
# ifndef strtoll
#  define strtoll _strtoi64
# endif
#endif


SWIGINTERN int
SWIG_AsVal_double SWIG_PERL_DECL_ARGS_2(SV *obj, double *val)
{
  if (SvNIOK(obj)) {
    if (val) *val = SvNV(obj);
    return SWIG_OK;
  } else if (SvIOK(obj)) {
    if (val) *val = (double) SvIV(obj);
    return SWIG_AddCast(SWIG_OK);
  } else {
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      double v;
      errno = 0;
      v = strtod(nptr, &endptr);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
  }
  return SWIG_TypeError;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx, cx, rd;
   errno = 0;
   fx = floor(x);
   cx = ceil(x);
   rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_long SWIG_PERL_DECL_ARGS_2(SV *obj, long* val)
{
  if (SvUOK(obj)) {
    UV v = SvUV(obj);
    if (UVSIZE < sizeof(*val) || v <= LONG_MAX) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (IVSIZE <= sizeof(*val) || (v >= LONG_MIN && v <= LONG_MAX)) {
      if(val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      long v;
      errno = 0;
      v = strtol(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsVal_int SWIG_PERL_DECL_ARGS_2(SV * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long SWIG_PERL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (int)(v);
    }
  }  
  return res;
}

SWIGINTERN WsXmlNodeH __WsXmlNode_find(struct __WsXmlNode *self,char const *ns,char const *name,int recursive){
    return ws_xml_find_in_tree( self, ns, name, recursive );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_next(struct __WsXmlNode *self,int all){
    WsXmlNodeH next_node = xml_parser_get_next_child(self);
    if (next_node && !all) {
      const char *ns_uri = ws_xml_get_node_name_ns(self);
      const char *name = ws_xml_get_node_local_name(self);
      if (ws_xml_is_node_qname(next_node, ns_uri, name) == 0) {
        next_node = NULL;
      }
    }
    return next_node;
  }
SWIGINTERN int __WsXmlNode_size(struct __WsXmlNode *self,char const *name,char const *ns){
    return ws_xml_get_child_count_by_qname(self, ns, name);
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add__SWIG_0(struct __WsXmlNode *self,char const *ns,char const *name,char const *text){
    return ws_xml_add_child( self, ns, name, text );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add_before(struct __WsXmlNode *self,char const *ns,char const *name,char const *text){
    return ws_xml_add_prev_sibling( self, ns, name, text );
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_add__SWIG_1(struct __WsXmlNode *self,WsXmlNodeH node){
    ws_xml_duplicate_tree( self, node );
    return self;
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_get__SWIG_0(struct __WsXmlNode *self,int i,char const *name,char const *ns){
    if (i < 0 || i >= ws_xml_get_child_count_by_qname(self,ns,name))
      return NULL;
    return ws_xml_get_child(self, i, ns, name);
  }
SWIGINTERN WsXmlNodeH __WsXmlNode_get__SWIG_1(struct __WsXmlNode *self,char const *name,char const *ns){
    return ws_xml_get_child(self, 0, ns, name);
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr(struct __WsXmlNode *self,int index){
    return ws_xml_get_node_attr( self, index );
  }
SWIGINTERN int __WsXmlNode_attr_count(struct __WsXmlNode *self){
    return ws_xml_get_node_attr_count( self );
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr_find(struct __WsXmlNode *self,char const *ns,char const *name){
    return ws_xml_find_node_attr( self, ns, name );
  }
SWIGINTERN WsXmlAttrH __WsXmlNode_attr_add(struct __WsXmlNode *self,char const *ns,char const *name,char const *value){
    return ws_xml_add_node_attr( self, ns, name, value );
  }
SWIGINTERN epr_t *__WsXmlNode_epr(struct __WsXmlNode *self,char const *ns,char const *epr_node_name,int embedded){
    return epr_deserialize(self, ns, epr_node_name, embedded);
  }
SWIGINTERN char *__WsXmlAttr_name(struct __WsXmlAttr *self){
    return ws_xml_get_attr_name( self );
  }
SWIGINTERN char *__WsXmlAttr_ns(struct __WsXmlAttr *self){
    return ws_xml_get_attr_ns( self );
  }
SWIGINTERN char *__WsXmlAttr_value(struct __WsXmlAttr *self){
    return ws_xml_get_attr_value( self );
  }
SWIGINTERN void __WsXmlAttr_remove(struct __WsXmlAttr *self){
    ws_xml_remove_node_attr( self );
  }
SWIGINTERN epr_t *new_epr_t__SWIG_0(char const *uri,char const *address){
    return epr_create( uri, NULL, address);
  }
SWIGINTERN epr_t *new_epr_t__SWIG_1(char const *uri){
    return epr_from_string( uri );
  }
SWIGINTERN void delete_epr_t(epr_t *self){
    epr_destroy( self );
  }
SWIGINTERN void epr_t_add_selector(epr_t *self,char const *name,char const *text){
    epr_add_selector_text(self, name, text);
  }
SWIGINTERN int epr_t_serialize(epr_t *self,WsXmlNodeH node,char const *ns,char const *epr_node_name,int embedded){
    return epr_serialize(node, ns, epr_node_name, self, embedded);
  }
SWIGINTERN int epr_t_cmp(epr_t *self,epr_t const *epr2){
    return epr_cmp(self, epr2);
  }
SWIGINTERN char *epr_t_to_xml(epr_t *self,char const *ns,char const *epr_node_name){
    return epr_to_txt(self, ns?ns:"http://schemas.xmlsoap.org/ws/2004/08/addressing", epr_node_name?epr_node_name:"EndpointReference");
  }
SWIGINTERN char *epr_t_string(epr_t *self){
    return epr_to_string(self);
  }
SWIGINTERN int epr_t_selector_count(epr_t *self){
    return epr_selector_count(self);
  }
SWIGINTERN char *epr_t_resource_uri(epr_t *self){
    return epr_get_resource_uri(self);
  }
SWIGINTERN char *epr_t_selector(epr_t *self,char const *name){

    return wsman_epr_selector_by_name(self, name);
  }
SWIGINTERN AV *epr_t_selector_names(epr_t *self){

    int i;
    Target_Type ary = Target_SizedArray(self->refparams.selectorset.count);
    key_value_t *p = self->refparams.selectorset.selectors;
    for (i = 0; i < self->refparams.selectorset.count; i++) {
      Target_ListSet(ary, i, SWIG_FromCharPtr(p->key));
      ++p;
    }

    return (AV *)ary;



  }
SWIGINTERN char *epr_t_classname(epr_t *self){
    return uri_classname(self->refparams.uri);
  }
SWIGINTERN char *epr_t_namespace(epr_t *self){
    char *classname;
    int classnamelen, namespacelen;
    const char *prefix;
    const char *uri;

    /* try to get namespace from selectors (WS-Management style) */
    char *ns = get_cimnamespace_from_selectorset(&(self->refparams.selectorset));
    if (ns) {
      return strdup(ns);
    }
    /* WMI style? - extract namespace from uri */

    uri = self->refparams.uri;
    prefix = epr_prefix(uri);
    if (prefix == NULL) {
      return NULL; /* bad classname in URI */
    }
    classname = uri_classname(uri);
    if (classname == NULL)
      return NULL; /* bad URI */
    classnamelen = strlen(classname);
    free(classname);
    namespacelen = strlen(uri) - classnamelen - strlen(prefix) - 2; /* drop enclosing slashes */
    if (namespacelen <= 0)
      return strdup("");
    /* copy after prefix slash (+ 1) */
    return strndup(uri + strlen(prefix) + 1, namespacelen);
  }
SWIGINTERN char *epr_t_prefix(epr_t *self){
    return epr_prefix(self->refparams.uri);
  }
SWIGINTERN filter_t *new_filter_t(void){
    return filter_initialize();
  }
SWIGINTERN void delete_filter_t(filter_t *self){
    filter_destroy( self );
  }
SWIGINTERN int filter_t_associators(filter_t *self,epr_t *epr,char const *assocClass,char const *resultClass,char const *role,char const *resultRole,char **resultProp,int const propNum){
    return filter_set_assoc(self, epr, 0, assocClass, resultClass, role, resultRole, resultProp, propNum);
  }
SWIGINTERN int filter_t_references(filter_t *self,epr_t *epr,char const *assocClass,char const *resultClass,char const *role,char const *resultRole,char **resultProp,int const propNum){
    return filter_set_assoc(self, epr, 1, assocClass, resultClass, role, resultRole, resultProp, propNum);
  }
SWIGINTERN int filter_t_simple(filter_t *self,char const *dialect,char const *query){
    return filter_set_simple(self, dialect, query );
  }
SWIGINTERN int filter_t_xpath(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://www.w3.org/TR/1999/REC-xpath-19991116", query );
  }
SWIGINTERN int filter_t_cql(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://schemas.dmtf.org/wbem/cql/1/dsp0202.pdf", query );
  }
SWIGINTERN int filter_t_wql(filter_t *self,char const *query){
    return filter_set_simple(self,          "http://schemas.microsoft.com/wbem/wsman/1/WQL", query );
  }
SWIGINTERN void delete___WsEnumerateInfo(struct __WsEnumerateInfo *self){
  }
SWIGINTERN int __WsEnumerateInfo_max_items(struct __WsEnumerateInfo *self){ return self->maxItems; }
SWIGINTERN void __WsEnumerateInfo_set_max_items(struct __WsEnumerateInfo *self,int mi){ self->maxItems = mi; }
SWIGINTERN int __WsEnumerateInfo_flags(struct __WsEnumerateInfo *self){
    return self->flags;
  }
SWIGINTERN char const *__WsEnumerateInfo_epr_to(struct __WsEnumerateInfo *self){
    return self->epr_to;
  }
SWIGINTERN char const *__WsEnumerateInfo_epr_uri(struct __WsEnumerateInfo *self){
    return self->epr_uri;
  }
SWIGINTERN char const *__WsEnumerateInfo_encoding(struct __WsEnumerateInfo *self){
    return self->encoding;
  }
SWIGINTERN filter_t const *__WsEnumerateInfo_filter(struct __WsEnumerateInfo *self){
    return self->filter;
  }
SWIGINTERN int __WsEnumerateInfo_index(struct __WsEnumerateInfo *self){
    return self->index;
  }
SWIGINTERN void __WsEnumerateInfo_set_index(struct __WsEnumerateInfo *self,int i){
    self->index = i;
  }
SWIGINTERN int __WsEnumerateInfo_total_items(struct __WsEnumerateInfo *self){
    return self->totalItems;
  }
SWIGINTERN void __WsEnumerateInfo_set_total_items(struct __WsEnumerateInfo *self,int i){
    self->totalItems = i;
  }
SWIGINTERN WsXmlDocH __WsEnumerateInfo_pull_result(struct __WsEnumerateInfo *self){
    return (WsXmlDocH)self->pullResultPtr;
  }
SWIGINTERN void __WsEnumerateInfo_set_pull_result(struct __WsEnumerateInfo *self,WsXmlDocH result){
    self->pullResultPtr = (void *)result;
  }
SWIGINTERN WsXmlDocH __SoapOp_indoc(struct __SoapOp *self){
    return soap_get_op_doc(self, 1);
  }
SWIGINTERN void __SoapOp_set_indoc(struct __SoapOp *self,WsXmlDocH doc){
    soap_set_op_doc( self, doc, 1 );
  }
SWIGINTERN WsXmlDocH __SoapOp_outdoc(struct __SoapOp *self){
    return soap_get_op_doc(self, 0);
  }
SWIGINTERN void __SoapOp_set_outdoc(struct __SoapOp *self,WsXmlDocH doc){
    soap_set_op_doc( self, doc, 0 );
  }
SWIGINTERN struct __Soap *__SoapOp_soap(struct __SoapOp *self){
    return soap_get_op_soap(self);
  }
SWIGINTERN WsmanMessage *__SoapOp_msg(struct __SoapOp *self){
    return wsman_get_msg_from_op(self);
  }
SWIGINTERN unsigned long __SoapOp_maxsize(struct __SoapOp *self){
    return wsman_get_maxsize_from_op(self);
  }

SWIGINTERNINLINE SV *
SWIG_From_unsigned_SS_long  SWIG_PERL_DECL_ARGS_1(unsigned long value)
{
  SV *sv;
  if (UVSIZE >= sizeof(value) || value <= UV_MAX)
    sv = newSVuv(value);
  else
    sv = newSVpvf("%lu", value);
  return sv_2mortal(sv);
}

SWIGINTERN void delete___Soap(struct __Soap *self){
    soap_destroy(self);
  }
SWIGINTERN WsContextH __Soap_create_context(struct __Soap *self){
    return ws_create_context(self);
  }
SWIGINTERN WsContextH __Soap_context(struct __Soap *self){
    return ws_get_soap_context(self);
  }
SWIGINTERN WsContextH __Soap_create_ep_context(struct __Soap *self,WsXmlDocH doc){
    return ws_create_ep_context( self, doc );
  }
SWIGINTERN void delete__WS_CONTEXT(struct _WS_CONTEXT *self){
    ws_destroy_context(self);
  }
SWIGINTERN WsXmlDocH _WS_CONTEXT_indoc(struct _WS_CONTEXT *self){
    return self->indoc;
  }
SWIGINTERN struct __Soap *_WS_CONTEXT_runtime(struct _WS_CONTEXT *self){
    return ws_context_get_runtime(self);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_long SWIG_PERL_DECL_ARGS_2(SV *obj, unsigned long *val) 
{
  if (SvUOK(obj)) {
    UV v = SvUV(obj);
    if (UVSIZE <= sizeof(*val) || v <= ULONG_MAX) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (v >= 0 && (IVSIZE <= sizeof(*val) || v <= ULONG_MAX)) {
      if (val) *val = v;
      return SWIG_OK;
    }
    return SWIG_OverflowError;
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      unsigned long v;
      errno = 0;
      v = strtoul(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}

SWIGINTERN void _WS_CONTEXT_set_enumIdleTimeout(struct _WS_CONTEXT *self,unsigned long timeout){
    ws_set_context_enumIdleTimeout(self, timeout);
  }
SWIGINTERN char const *_WS_CONTEXT_classname(struct _WS_CONTEXT *self){
    return wsman_get_class_name(self);
  }
SWIGINTERN char const *_WS_CONTEXT_method(struct _WS_CONTEXT *self){
    return wsman_get_method_name(self);
  }
SWIGINTERN hash_t *_WS_CONTEXT_method_args(struct _WS_CONTEXT *self,char const *resource_uri){
    return wsman_get_method_args(self, resource_uri);
  }
SWIGINTERN int _WS_CONTEXT_max_elements(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_max_elements(self, doc);
  }
SWIGINTERN unsigned long _WS_CONTEXT_max_envelope_size(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_max_envelope_size(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_fragment_string(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_fragment_string(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_selector(struct _WS_CONTEXT *self,WsXmlDocH doc,char const *name,int index){
    return wsman_get_selector(self, doc, name, index);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors_from_epr(struct _WS_CONTEXT *self,WsXmlNodeH epr_node){
    return wsman_get_selectors_from_epr(self, epr_node);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_selector_list(self, doc);
  }
SWIGINTERN hash_t *_WS_CONTEXT_selectors_from_filter(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_selector_list_from_filter(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_action(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_action(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_resource_uri(struct _WS_CONTEXT *self,WsXmlDocH doc){
    return wsman_get_resource_uri(self, doc);
  }
SWIGINTERN char const *_WS_CONTEXT_option_set(struct _WS_CONTEXT *self,WsXmlDocH doc,char const *op){
    return wsman_get_option_set(self, doc, op);
  }
SWIGINTERN int _WS_CONTEXT_parse_enum_request(struct _WS_CONTEXT *self,WsEnumerateInfo *enumInfo){
    WsmanStatus status;
    wsman_status_init(&status);
    return wsman_parse_enum_request( self, enumInfo, &status);
    /* FIXME: return status */
  }
SWIGINTERN struct _WsmanStatus *new__WsmanStatus(int code,int detail,char const *msg){
    WsmanStatus *s = (WsmanStatus *)malloc(sizeof(WsmanStatus));
    wsman_status_init(s);
    if (code)
      s->fault_code = code;
    if (msg)
      s->fault_msg = strdup(msg);
    if (detail < 0
        || detail > OWSMAN_SYSTEM_ERROR) {
      SWIG_exception( SWIG_ValueError, "Bad fault detail" );
    }
    else {
      s->fault_detail_code = detail;
    }

    fail:

    return s;
  }
SWIGINTERN void delete__WsmanStatus(struct _WsmanStatus *self){
    if (self->fault_msg) free(self->fault_msg);
    free(self);
  }
SWIGINTERN char const *_WsmanStatus_to_s(struct _WsmanStatus *self){
    return self->fault_msg;
  }
SWIGINTERN void _WsmanStatus_set_code(struct _WsmanStatus *self,int code){ self->fault_code = code; }
SWIGINTERN int _WsmanStatus_code(struct _WsmanStatus *self){
    return self->fault_code;
  }
SWIGINTERN void _WsmanStatus_set_detail(struct _WsmanStatus *self,int detail){
    if (detail < 0
        || detail > OWSMAN_SYSTEM_ERROR) {
      SWIG_exception( SWIG_ValueError, "Bad fault detail" );
    }
    else {
      self->fault_detail_code = detail;
    }

    fail:
    return;

  }
SWIGINTERN int _WsmanStatus_detail(struct _WsmanStatus *self){
    return self->fault_detail_code;
  }
SWIGINTERN void _WsmanStatus_set_msg(struct _WsmanStatus *self,char const *msg){
    if (self->fault_msg)
      free(self->fault_msg);
    if (msg)
      self->fault_msg = strdup(msg);
    else
      self->fault_msg = NULL;
  }
SWIGINTERN char const *_WsmanStatus_msg(struct _WsmanStatus *self){
    return self->fault_msg;
  }
SWIGINTERN WsXmlDocH _WsmanStatus_generate_fault(struct _WsmanStatus *self,WsXmlDocH doc){
    return wsman_generate_fault( doc, self->fault_code, self->fault_detail_code, self->fault_msg);
  }
SWIGINTERN struct _WsManFault *new__WsManFault(WsXmlDocH doc){
    WsManFault *fault = wsmc_fault_new();
    wsmc_get_fault_data(doc, fault);
    return fault;
  }
SWIGINTERN void delete__WsManFault(struct _WsManFault *self){
    wsmc_fault_destroy(self);
  }
SWIGINTERN char const *_WsManFault_code(struct _WsManFault *self){
    return self->code;
  }
SWIGINTERN char const *_WsManFault_subcode(struct _WsManFault *self){
    return self->subcode;
  }
SWIGINTERN char const *_WsManFault_reason(struct _WsManFault *self){
    return self->reason;
  }
SWIGINTERN char const *_WsManFault_detail(struct _WsManFault *self){
    return self->fault_detail;
  }
SWIGINTERN int _WsManTransport_is_auth_method(struct _WsManTransport *self,int method){
    return wsman_is_auth_method((WsManClient *)self, method);
  }
SWIGINTERN void _WsManTransport_close(struct _WsManTransport *self){
    wsman_transport_close_transport((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_agent(struct _WsManTransport *self,char const *agent){
    wsman_transport_set_agent((WsManClient *)self, agent);
  }
SWIGINTERN char *_WsManTransport_agent(struct _WsManTransport *self){
    return wsman_transport_get_agent ((WsManClient *)self);
  }
SWIGINTERN char *_WsManTransport_get_username(struct _WsManTransport *self){
    return wsman_transport_get_userName((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_username(struct _WsManTransport *self,char *user_name){
    wsman_transport_set_userName((WsManClient *)self, user_name);
  }
SWIGINTERN char *_WsManTransport_get_password(struct _WsManTransport *self){
    return wsman_transport_get_password((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_password(struct _WsManTransport *self,char *password){
    wsman_transport_set_password((WsManClient *)self, password);
  }
SWIGINTERN char *_WsManTransport_get_proxy_username(struct _WsManTransport *self){
    return wsman_transport_get_proxy_username((WsManClient *)self );
  }
SWIGINTERN void _WsManTransport_set_proxy_username(struct _WsManTransport *self,char *proxy_username){
    wsman_transport_set_proxy_username((WsManClient *)self, proxy_username );
  }
SWIGINTERN char *_WsManTransport_get_proxy_password(struct _WsManTransport *self){
    return wsman_transport_get_proxy_password((WsManClient *)self );
  }
SWIGINTERN void _WsManTransport_set_proxy_password(struct _WsManTransport *self,char *proxy_password){
    wsman_transport_set_proxy_password((WsManClient *)self, proxy_password );
  }
SWIGINTERN void _WsManTransport_set_auth_method(struct _WsManTransport *self,char const *am){
    wsman_transport_set_auth_method((WsManClient *)self, am);
  }
SWIGINTERN char *_WsManTransport_auth_method(struct _WsManTransport *self){
    return wsman_transport_get_auth_method ((WsManClient *)self);
  }
SWIGINTERN char const *_WsManTransport_auth_name(int auth){
    return wsmc_transport_get_auth_name(auth);
  }
SWIGINTERN int _WsManTransport_auth_value(struct _WsManTransport *self){
    return wsmc_transport_get_auth_value((WsManClient *)self);
  }
SWIGINTERN char *_WsManTransport_error_string(int err){
    return wsman_transport_get_last_error_string(err);
  }
SWIGINTERN void _WsManTransport_set_timeout(struct _WsManTransport *self,unsigned long timeout){
    wsman_transport_set_timeout((WsManClient *)self, timeout);
  }
SWIGINTERN unsigned long _WsManTransport_timeout(struct _WsManTransport *self){
    return wsman_transport_get_timeout((WsManClient *)self);
  }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int SWIG_PERL_DECL_ARGS_2(SV * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (unsigned int)(v);
    }
  }  
  return res;
}

SWIGINTERN void _WsManTransport_set_verify_peer(struct _WsManTransport *self,unsigned int value){

    wsman_transport_set_verify_peer((WsManClient *)self, value);
  }
SWIGINTERN unsigned int _WsManTransport_verify_peer(struct _WsManTransport *self){
    return wsman_transport_get_verify_peer((WsManClient *)self);
  }

SWIGINTERNINLINE SV *
SWIG_From_unsigned_SS_int  SWIG_PERL_DECL_ARGS_1(unsigned int value)
{    
  return SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1(value);
}

SWIGINTERN void _WsManTransport_set_verify_host(struct _WsManTransport *self,unsigned int value){

    wsman_transport_set_verify_host((WsManClient *)self, value);
  }
SWIGINTERN unsigned int _WsManTransport_verify_host(struct _WsManTransport *self){
    return wsman_transport_get_verify_host((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_proxy(struct _WsManTransport *self,char const *proxy){
    wsman_transport_set_proxy((WsManClient *)self, proxy);
  }
SWIGINTERN char *_WsManTransport_proxy(struct _WsManTransport *self){
    return wsman_transport_get_proxy((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_proxyauth(struct _WsManTransport *self,char const *pauth){
    wsman_transport_set_proxyauth((WsManClient *)self, pauth);
  }
SWIGINTERN char *_WsManTransport_proxyauth(struct _WsManTransport *self){
    return wsman_transport_get_proxyauth((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_cainfo(struct _WsManTransport *self,char const *cainfo){
    wsman_transport_set_cainfo((WsManClient *)self, cainfo);
  }
SWIGINTERN char *_WsManTransport_cainfo(struct _WsManTransport *self){
    return wsman_transport_get_cainfo((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_certhumbprint(struct _WsManTransport *self,char const *arg){
    wsman_transport_set_certhumbprint((WsManClient *)self, arg);
  }
SWIGINTERN char *_WsManTransport_certhumbprint(struct _WsManTransport *self){
    return wsman_transport_get_certhumbprint((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_capath(struct _WsManTransport *self,char const *capath){
    wsman_transport_set_capath((WsManClient *)self, capath);
  }
SWIGINTERN char *_WsManTransport_capath(struct _WsManTransport *self){
    return wsman_transport_get_capath((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_caoid(struct _WsManTransport *self,char const *oid){
    wsman_transport_set_caoid((WsManClient *)self, oid);
  }
SWIGINTERN char *_WsManTransport_caoid(struct _WsManTransport *self){
    return wsman_transport_get_caoid((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_cert(struct _WsManTransport *self,char const *cert){
    wsman_transport_set_cert((WsManClient *)self, cert);
  }
SWIGINTERN char *_WsManTransport_cert(struct _WsManTransport *self){
    return wsman_transport_get_cert((WsManClient *)self);
  }
SWIGINTERN void _WsManTransport_set_key(struct _WsManTransport *self,char const *key){
    wsman_transport_set_key((WsManClient *)self, key);
  }
SWIGINTERN char *_WsManTransport_key(struct _WsManTransport *self){
    return wsman_transport_get_key((WsManClient *)self);
  }
SWIGINTERN client_opt_t *new_client_opt_t(void){
    client_opt_t *options = wsmc_options_init();
    return options;
  }
SWIGINTERN void delete_client_opt_t(client_opt_t *self){
    wsmc_options_destroy( self );
  }
SWIGINTERN void client_opt_t_set_dump_request(client_opt_t *self){
    wsmc_set_action_option(self, 0x0010 );
  }
SWIGINTERN void client_opt_t_clear_dump_request(client_opt_t *self){
    wsmc_clear_action_option(self, 0x0010 );
  }
SWIGINTERN void client_opt_t_set_flags(client_opt_t *self,unsigned long flags){
    wsmc_set_action_option(self, flags);
  }
SWIGINTERN unsigned long client_opt_t_get_flags(client_opt_t *self){
    return wsmc_get_action_option(self);
  }
SWIGINTERN void client_opt_t_clear_flags(client_opt_t *self,unsigned long flags){
    wsmc_clear_action_option(self, flags);
  }
SWIGINTERN void client_opt_t_reset_flags(client_opt_t *self){
    wsmc_clear_action_option(self, ~0x0000);
  }
SWIGINTERN void client_opt_t_set_max_envelope_size(client_opt_t *self,unsigned long size){
    self->max_envelope_size = size;
  }
SWIGINTERN unsigned long client_opt_t_get_max_envelope_size(client_opt_t *self){
    return self->max_envelope_size;
  }
SWIGINTERN void client_opt_t_set_max_elements(client_opt_t *self,int elements){
    self->max_elements = elements;
  }
SWIGINTERN int client_opt_t_get_max_elements(client_opt_t *self){
    return self->max_elements;
  }
SWIGINTERN void client_opt_t_set_timeout(client_opt_t *self,unsigned long timeout){
    self->timeout = timeout;
  }
SWIGINTERN unsigned long client_opt_t_get_timeout(client_opt_t *self){
    return self->timeout;
  }
SWIGINTERN void client_opt_t_set_fragment(client_opt_t *self,char *fragment){
    wsmc_set_fragment(fragment, self);
  }
SWIGINTERN char const *client_opt_t_get_fragment(client_opt_t *self){
    return self->fragment;
  }
SWIGINTERN void client_opt_t_set_cim_namespace(client_opt_t *self,char *cim_namespace){
    wsmc_set_cim_ns(cim_namespace, self);
  }
SWIGINTERN char const *client_opt_t_get_cim_namespace(client_opt_t *self){
    return self->cim_ns;
  }
SWIGINTERN void client_opt_t_set_reference(client_opt_t *self,char const *reference){
    wsmc_set_reference(reference, self);
  }
SWIGINTERN char const *client_opt_t_get_reference(client_opt_t *self){
    return self->reference;
  }
SWIGINTERN void client_opt_t_add_option(client_opt_t *self,char const *key,char const *value){

    wsmc_add_option(self, key, value);
  }
SWIGINTERN void client_opt_t_add_selector(client_opt_t *self,char const *key,char const *value){
    wsmc_add_selector(self, key, value);
  }
SWIGINTERN void client_opt_t_add_property__SWIG_0(client_opt_t *self,char const *key,char const *value){
    wsmc_add_property(self, key, value);
  }
SWIGINTERN void client_opt_t_add_property_array_item(client_opt_t *self,char const *key,char const *value){
    wsmc_add_property_array_item(self, key, value);
  }
SWIGINTERN void client_opt_t_add_property__SWIG_1(client_opt_t *self,char const *key,epr_t const *epr){
    wsmc_add_property_epr(self, key, epr);
  }
SWIGINTERN void client_opt_t_set_delivery_uri(client_opt_t *self,char const *delivery_uri){
    wsmc_set_delivery_uri(delivery_uri, self);
  }
SWIGINTERN char const *client_opt_t_delivery_uri(client_opt_t *self){
    return self->delivery_uri;
  }

/* Getting isfinite working pre C99 across multiple platforms is non-trivial. Users can provide SWIG_isfinite on older platforms. */
#ifndef SWIG_isfinite
/* isfinite() is a macro for C99 */
# if defined(isfinite)
#  define SWIG_isfinite(X) (isfinite(X))
# elif defined(__cplusplus) && __cplusplus >= 201103L
/* Use a template so that this works whether isfinite() is std::isfinite() or
 * in the global namespace.  The reality seems to vary between compiler
 * versions.
 *
 * Make sure namespace std exists to avoid compiler warnings.
 *
 * extern "C++" is required as this fragment can end up inside an extern "C" { } block
 */
namespace std { }
extern "C++" template<typename T>
inline int SWIG_isfinite_func(T x) {
  using namespace std;
  return isfinite(x);
}
#  define SWIG_isfinite(X) (SWIG_isfinite_func(X))
# elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))
#  define SWIG_isfinite(X) (__builtin_isfinite(X))
# elif defined(_MSC_VER)
#  define SWIG_isfinite(X) (_finite(X))
# elif defined(__sun) && defined(__SVR4)
#  include <ieeefp.h>
#  define SWIG_isfinite(X) (finite(X))
# endif
#endif


/* Accept infinite as a valid float value unless we are unable to check if a value is finite */
#ifdef SWIG_isfinite
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX) && SWIG_isfinite(X))
#else
# define SWIG_Float_Overflow_Check(X) ((X < -FLT_MAX || X > FLT_MAX))
#endif


SWIGINTERN int
SWIG_AsVal_float SWIG_PERL_DECL_ARGS_2(SV * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj, &v);
  if (SWIG_IsOK(res)) {
    if (SWIG_Float_Overflow_Check(v)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = (float)(v);
    }
  }  
  return res;
}

SWIGINTERN void client_opt_t_set_sub_expiry(client_opt_t *self,float event_subscription_expire){
	wsmc_set_sub_expiry(event_subscription_expire, self);
  }
SWIGINTERN float client_opt_t_sub_expiry(client_opt_t *self){
    return self->expires;
  }

SWIGINTERNINLINE SV *
SWIG_From_double  SWIG_PERL_DECL_ARGS_1(double value)
{
  return sv_2mortal(newSVnv(value));
}


SWIGINTERNINLINE SV *
SWIG_From_float  SWIG_PERL_DECL_ARGS_1(float value)
{    
  return SWIG_From_double  SWIG_PERL_CALL_ARGS_1(value);
}

SWIGINTERN void client_opt_t_set_heartbeat_interval(client_opt_t *self,float heartbeat_interval){
	wsmc_set_heartbeat_interval(heartbeat_interval, self);
  }
SWIGINTERN float client_opt_t_heartbeat_interval(client_opt_t *self){
    return self->heartbeat_interval;
  }
SWIGINTERN void client_opt_t_set_delivery_mode(client_opt_t *self,unsigned int delivery_mode){
    if (delivery_mode > 3)
      SWIG_exception( SWIG_ValueError, "Bad delivery mode" );
	
    wsmc_set_delivery_mode(delivery_mode, self);

    fail:
    return;

  }
SWIGINTERN int client_opt_t_delivery_mode(client_opt_t *self){
    return self->delivery_mode;
  }
SWIGINTERN void client_opt_t_set_delivery_security_mode(client_opt_t *self,unsigned int delivery_sec_mode){
    if (delivery_sec_mode > 10)
      SWIG_exception( SWIG_ValueError, "Bad delivery security mode" );
    wsmc_set_delivery_security_mode(delivery_sec_mode, self);

    fail:
    return;

  }
SWIGINTERN int client_opt_t_delivery_sec_mode(client_opt_t *self){
    return self->delivery_sec_mode;
  }
SWIGINTERN void client_opt_t_set_locale(client_opt_t *self,char const *locale){
	wsmc_set_locale(self, locale);
  }
SWIGINTERN char const *client_opt_t_locale(client_opt_t *self){
    return self->locale;
  }
SWIGINTERN struct _WsManClient *new__WsManClient__SWIG_0(char const *uri){
    struct _WsManClient *client = wsmc_create_from_uri( uri );
    if (client == NULL)
      SWIG_exception( SWIG_ValueError, "Can't create Openwsman::Client from given URI" );

    fail:

    return client;
  }
SWIGINTERN struct _WsManClient *new__WsManClient__SWIG_1(char const *hostname,int const port,char const *path,char const *scheme,char const *username,char const *password){
    struct _WsManClient *client = wsmc_create( hostname, port, path, scheme, username, password );
    if (client == NULL)
      SWIG_exception( SWIG_ValueError, "Can't create Openwsman::Client from given values" );

    fail:

    return client;
  }
SWIGINTERN void delete__WsManClient(struct _WsManClient *self){
    wsmc_release( self );
  }
SWIGINTERN void _WsManClient_set_dumpfile(struct _WsManClient *self,FILE *f){
    wsmc_set_dumpfile( self, f );
  }
SWIGINTERN long _WsManClient_response_code(struct _WsManClient *self){
    return wsmc_get_response_code( self );
  }
SWIGINTERN char *_WsManClient_scheme(struct _WsManClient *self){
    return wsmc_get_scheme( self );
  }
SWIGINTERN char *_WsManClient_host(struct _WsManClient *self){
    return wsmc_get_hostname( self );
  }
SWIGINTERN int _WsManClient_port(struct _WsManClient *self){
    return wsmc_get_port( self );
  }
SWIGINTERN char *_WsManClient_path(struct _WsManClient *self){
    return wsmc_get_path( self );
  }
SWIGINTERN char *_WsManClient_user(struct _WsManClient *self){
    return wsmc_get_user( self );
  }
SWIGINTERN char *_WsManClient_password(struct _WsManClient *self){
    return wsmc_get_password( self );
  }
SWIGINTERN WsManTransport *_WsManClient_transport(struct _WsManClient *self){
    wsmc_transport_init(self, NULL);
    wsmc_transport_set_auth_request_func( self, auth_request_callback );

    return (WsManTransport *)self;
  }
SWIGINTERN int _WsManClient_send_request(struct _WsManClient *self,WsXmlDocH request){
    return wsman_send_request(self, request);
  }
SWIGINTERN WsXmlDocH _WsManClient_build_envelope_from_response(struct _WsManClient *self){
    return wsmc_build_envelope_from_response(self);
  }
SWIGINTERN char *_WsManClient_encoding(struct _WsManClient *self){
    return wsmc_get_encoding(self);
  }
SWIGINTERN void _WsManClient_set_encoding(struct _WsManClient *self,char const *encoding){
    wsmc_set_encoding(self, encoding);
  }
SWIGINTERN WsXmlDocH _WsManClient_identify(struct _WsManClient *self,client_opt_t *options){










    return wsmc_action_identify( self, options );

  }
SWIGINTERN WsXmlDocH _WsManClient_get_from_epr(struct _WsManClient *self,client_opt_t *options,epr_t *epr){











    return wsmc_action_get_from_epr( self, epr, options);

  }
SWIGINTERN WsXmlDocH _WsManClient_delete_from_epr(struct _WsManClient *self,client_opt_t *options,epr_t *epr){











    return wsmc_action_delete_from_epr( self, epr, options);

  }
SWIGINTERN WsXmlDocH _WsManClient_enumerate(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char *resource_uri){












    return wsmc_action_enumerate( self, resource_uri, options, filter);

  }
SWIGINTERN WsXmlDocH _WsManClient_pull(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri,char const *context){













    return wsmc_action_pull( self, resource_uri, options, filter, context);

  }

#if defined(LLONG_MAX) && !defined(SWIG_LONG_LONG_AVAILABLE)
#  define SWIG_LONG_LONG_AVAILABLE
#endif


#ifdef SWIG_LONG_LONG_AVAILABLE
SWIGINTERN int
SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_DECL_ARGS_2(SV *obj, unsigned long long *val)
{
  if (SvUOK(obj)) {
    /* pretty sure this should be conditional on
     * (UVSIZE <= sizeof(*val) || v <= ULLONG_MAX) */
    if (val) *val = SvUV(obj);
    return SWIG_OK;
  } else if (SvIOK(obj)) {
    IV v = SvIV(obj);
    if (v >= 0 && (IVSIZE <= sizeof(*val) || v <= ULLONG_MAX)) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else {
    int dispatch = 0;
    const char *nptr = SvPV_nolen(obj);
    if (nptr) {
      char *endptr;
      unsigned long long v;
      errno = 0;
      v = strtoull(nptr, &endptr,0);
      if (errno == ERANGE) {
	errno = 0;
	return SWIG_OverflowError;
      } else {
	if (*endptr == '\0') {
	  if (val) *val = v;
	  return SWIG_Str2NumCast(SWIG_OK);
	}
      }
    }
    if (!dispatch) {
      const double mant_max = 1LL << DBL_MANT_DIG;
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double SWIG_PERL_CALL_ARGS_2(obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, mant_max)) {
	if (val) *val = (unsigned long long)(d);
	return res;
      }
    }
  }
  return SWIG_TypeError;
}
#endif


SWIGINTERNINLINE int
SWIG_AsVal_size_t SWIG_PERL_DECL_ARGS_2(SV * obj, size_t *val)
{
  int res = SWIG_TypeError;
#ifdef SWIG_LONG_LONG_AVAILABLE
  if (sizeof(size_t) <= sizeof(unsigned long)) {
#endif
    unsigned long v;
    res = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
#ifdef SWIG_LONG_LONG_AVAILABLE
  } else if (sizeof(size_t) <= sizeof(unsigned long long)) {
    unsigned long long v;
    res = SWIG_AsVal_unsigned_SS_long_SS_long SWIG_PERL_CALL_ARGS_2(obj, val ? &v : 0);
    if (SWIG_IsOK(res) && val) *val = (size_t)(v);
  }
#endif
  return res;
}

SWIGINTERN WsXmlDocH _WsManClient_create(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *data,size_t size,char const *encoding){














    return wsmc_action_create_fromtext( self, resource_uri, options, data, size, encoding);

  }
SWIGINTERN WsXmlDocH _WsManClient_put(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *data,size_t size,char const *encoding){














    return wsmc_action_put_fromtext( self, resource_uri, options, data, size, encoding);

  }
SWIGINTERN WsXmlDocH _WsManClient_release(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *context){












    return wsmc_action_release( self, resource_uri, options, context);

  }
SWIGINTERN WsXmlDocH _WsManClient_get(struct _WsManClient *self,client_opt_t *options,char const *resource_uri){











    return wsmc_action_get( self, resource_uri, options);

  }
SWIGINTERN WsXmlDocH _WsManClient_delete(struct _WsManClient *self,client_opt_t *options,char const *resource_uri){











    return wsmc_action_delete( self, resource_uri, options);

  }
SWIGINTERN WsXmlDocH _WsManClient_invoke(struct _WsManClient *self,client_opt_t *options,char const *resource_uri,char const *method,WsXmlDocH data){













    return wsmc_action_invoke( self, resource_uri, options, method, data);

  }
SWIGINTERN WsXmlDocH _WsManClient_subscribe(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri){












    return wsmc_action_subscribe(self,  resource_uri, options, filter);

  }
SWIGINTERN WsXmlDocH _WsManClient_unsubscribe(struct _WsManClient *self,client_opt_t *options,filter_t *filter,char const *resource_uri,char const *identifier){













    return wsmc_action_unsubscribe(self, resource_uri, options, identifier);

  }
SWIGINTERN WsXmlDocH _WsManClient_renew(struct _WsManClient *self,client_opt_t *options,char *resource_uri,char *identifier){












    return wsmc_action_renew(self, resource_uri, options, identifier);

  }
SWIGINTERN char *_WsManClient_fault_string(struct _WsManClient *self){
    return wsmc_get_fault_string(self);
  }
SWIGINTERN int _WsManClient_last_error(struct _WsManClient *self){
     return wsmc_get_last_error(self);
   }
#ifdef __cplusplus
extern "C" {
#endif

#define MAGIC_CLASS
SWIGCLASS_STATIC int swig_magic_readonly(pTHX_ SV *SWIGUNUSEDPARM(sv), MAGIC *SWIGUNUSEDPARM(mg)) {
    MAGIC_PPERL
    croak("Value is read-only.");
    return 0;
}

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif
XS(_wrap_new_XmlDoc) {
  {
    char *arg1 = 0 ;
    char *arg2 = (char *)NULL ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    struct _WsXmlDoc *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: new_XmlDoc(name,ns);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_XmlDoc" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_XmlDoc" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (struct _WsXmlDoc *)new__WsXmlDoc((char const *)arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_delete_XmlDoc) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_XmlDoc(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XmlDoc" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    delete__WsXmlDoc(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_string) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_string(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_string" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (char *)_WsXmlDoc_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_encode) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    char *arg2 = (char *)"utf-8" ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: XmlDoc_encode(self,encoding);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_encode" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlDoc_encode" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (char *)_WsXmlDoc_encode(arg1,(char const *)arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    free(result);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_dump_file) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlDoc_dump_file(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_dump_file" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    {
      arg2 = PerlIO_findFILE(IoIFP(sv_2io(ST(1))));
    }
    _WsXmlDoc_dump_file(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_root) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_root(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_root" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (WsXmlNodeH)_WsXmlDoc_root(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_envelope) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_envelope(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_envelope" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (WsXmlNodeH)_WsXmlDoc_envelope(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_header) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_header(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_header" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (WsXmlNodeH)_WsXmlDoc_header(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_body) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_body(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_body" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (WsXmlNodeH)_WsXmlDoc_body(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_element) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlDoc_element(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_element" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlDoc_element" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (WsXmlNodeH)_WsXmlDoc_element(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_context) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_context(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_context" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (char *)_WsXmlDoc_context(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_generate_fault) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    WsmanStatus *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlDoc_generate_fault(self,s);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_generate_fault" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlDoc_generate_fault" "', argument " "2"" of type '" "WsmanStatus *""'"); 
    }
    arg2 = (WsmanStatus *)(argp2);
    result = (WsXmlDocH)_WsXmlDoc_generate_fault(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_is_fault) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_is_fault(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_is_fault" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (int)_WsXmlDoc_is_fault(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_fault) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsManFault *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_fault(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_fault" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (WsManFault *)_WsXmlDoc_fault(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManFault, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_create_response_envelope) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    char *arg2 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: XmlDoc_create_response_envelope(self,action);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_create_response_envelope" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlDoc_create_response_envelope" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (WsXmlDocH)_WsXmlDoc_create_response_envelope(arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlDoc_is_end_of_sequence) {
  {
    struct _WsXmlDoc *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlDoc_is_end_of_sequence(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlDoc_is_end_of_sequence" "', argument " "1"" of type '" "struct _WsXmlDoc *""'"); 
    }
    arg1 = (struct _WsXmlDoc *)(argp1);
    result = (int)_WsXmlDoc_is_end_of_sequence(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_XmlNode) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_XmlNode(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_XmlNode" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    delete___WsXmlNode(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_string) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_string(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_string" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (char *)__WsXmlNode_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_dump_file) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_dump_file(self,fp);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_dump_file" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    {
      arg2 = PerlIO_findFILE(IoIFP(sv_2io(ST(1))));
    }
    __WsXmlNode_dump_file(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode___eq__) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    WsXmlNodeH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode___eq__(self,n);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode___eq__" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode___eq__" "', argument " "2"" of type '" "WsXmlNodeH""'"); 
    }
    arg2 = (WsXmlNodeH)(argp2);
    result = (int)__WsXmlNode___eq__(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_text) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_text(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_text" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (char *)__WsXmlNode_text(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_set_text) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_set_text(self,text);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_set_text" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_set_text" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    __WsXmlNode_set_text(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_doc) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_doc(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_doc" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (WsXmlDocH)__WsXmlNode_doc(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_parent) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_parent(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_parent" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (WsXmlNodeH)__WsXmlNode_parent(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_child) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_child(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_child" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (WsXmlNodeH)__WsXmlNode_child(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_name) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_name(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_name" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (char *)__WsXmlNode_name(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_set_name) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_set_name(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_set_name" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_set_name" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    __WsXmlNode_set_name(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_ns) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_ns(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_ns" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (char *)__WsXmlNode_ns(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_set_ns) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_set_ns(self,ns);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_set_ns" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_set_ns" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    __WsXmlNode_set_ns(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_prefix) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_prefix(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_prefix" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (char *)__WsXmlNode_prefix(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_set_lang) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_set_lang(self,lang);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_set_lang" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_set_lang" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    __WsXmlNode_set_lang(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_find) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    int arg4 = 1 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_find(self,ns,name,recursive);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_find" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_find" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_find" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
      if (!SWIG_IsOK(ecode4)) {
        SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XmlNode_find" "', argument " "4"" of type '" "int""'");
      } 
      arg4 = (int)(val4);
    }
    result = (WsXmlNodeH)__WsXmlNode_find(arg1,(char const *)arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_next) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_next(self,all);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_next" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XmlNode_next" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (WsXmlNodeH)__WsXmlNode_next(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_size) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = (char *)NULL ;
    char *arg3 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 3)) {
      SWIG_croak("Usage: XmlNode_size(self,name,ns);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_size" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_size" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_size" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (int)__WsXmlNode_size(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_add__SWIG_0) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_add(self,ns,name,text);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_add" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_add" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_add" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
      if (!SWIG_IsOK(res4)) {
        SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XmlNode_add" "', argument " "4"" of type '" "char const *""'");
      }
      arg4 = (char *)(buf4);
    }
    result = (WsXmlNodeH)__WsXmlNode_add__SWIG_0(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_add_before) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 3) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_add_before(self,ns,name,text);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_add_before" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_add_before" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_add_before" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    if (items > 3) {
      res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
      if (!SWIG_IsOK(res4)) {
        SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XmlNode_add_before" "', argument " "4"" of type '" "char const *""'");
      }
      arg4 = (char *)(buf4);
    }
    result = (WsXmlNodeH)__WsXmlNode_add_before(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_add__SWIG_1) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    WsXmlNodeH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_add(self,node);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_add" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_add" "', argument " "2"" of type '" "WsXmlNodeH""'"); 
    }
    arg2 = (WsXmlNodeH)(argp2);
    result = (WsXmlNodeH)__WsXmlNode_add__SWIG_1(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_add) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if (items == 2) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p___WsXmlNode, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(1), &vptr, SWIGTYPE_p___WsXmlNode, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if ((items >= 3) && (items <= 4)) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p___WsXmlNode, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(2), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (items > 3) {
        {
          int res = SWIG_AsCharPtrAndSize(ST(3), 0, NULL, 0);
          _v = SWIG_CheckState(res);
        }
        if (!_v) goto check_2;
        _ranki += _v*_pi;
        _rankm += _pi;
        _pi *= SWIG_MAXCASTRANK;
      }
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_XmlNode_add__SWIG_1); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_XmlNode_add__SWIG_0); return;
    }
  }
  
  croak("No matching function for overloaded 'XmlNode_add'");
  XSRETURN(0);
}


XS(_wrap_XmlNode_get__SWIG_0) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    int arg2 ;
    char *arg3 = (char *)NULL ;
    char *arg4 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 2) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_get(self,i,name,ns);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_get" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XmlNode_get" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_get" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    if (items > 3) {
      res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
      if (!SWIG_IsOK(res4)) {
        SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XmlNode_get" "', argument " "4"" of type '" "char const *""'");
      }
      arg4 = (char *)(buf4);
    }
    result = (WsXmlNodeH)__WsXmlNode_get__SWIG_0(arg1,arg2,(char const *)arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_get__SWIG_1) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    WsXmlNodeH result;
    dXSARGS;
    
    if ((items < 2) || (items > 3)) {
      SWIG_croak("Usage: XmlNode_get(self,name,ns);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_get" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_get" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_get" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (WsXmlNodeH)__WsXmlNode_get__SWIG_1(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlNode, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_get) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if ((items >= 2) && (items <= 4)) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p___WsXmlNode, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        {
          int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), NULL);
          _v = SWIG_CheckState(res);
        }
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (items > 2) {
        {
          int res = SWIG_AsCharPtrAndSize(ST(2), 0, NULL, 0);
          _v = SWIG_CheckState(res);
        }
        if (!_v) goto check_1;
        _ranki += _v*_pi;
        _rankm += _pi;
        _pi *= SWIG_MAXCASTRANK;
        if (items > 3) {
          {
            int res = SWIG_AsCharPtrAndSize(ST(3), 0, NULL, 0);
            _v = SWIG_CheckState(res);
          }
          if (!_v) goto check_1;
          _ranki += _v*_pi;
          _rankm += _pi;
          _pi *= SWIG_MAXCASTRANK;
        }
      }
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if ((items >= 2) && (items <= 3)) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p___WsXmlNode, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (items > 2) {
        {
          int res = SWIG_AsCharPtrAndSize(ST(2), 0, NULL, 0);
          _v = SWIG_CheckState(res);
        }
        if (!_v) goto check_2;
        _ranki += _v*_pi;
        _rankm += _pi;
        _pi *= SWIG_MAXCASTRANK;
      }
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_XmlNode_get__SWIG_0); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_XmlNode_get__SWIG_1); return;
    }
  }
  
  croak("No matching function for overloaded 'XmlNode_get'");
  XSRETURN(0);
}


XS(_wrap_XmlNode_attr) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    int arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    WsXmlAttrH result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: XmlNode_attr(self,index);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_attr" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "XmlNode_attr" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    result = (WsXmlAttrH)__WsXmlNode_attr(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_attr_count) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlNode_attr_count(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_attr_count" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    result = (int)__WsXmlNode_attr_count(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_attr_find) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    WsXmlAttrH result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: XmlNode_attr_find(self,ns,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_attr_find" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_attr_find" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_attr_find" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    result = (WsXmlAttrH)__WsXmlNode_attr_find(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_attr_add) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlAttrH result;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_attr_add(self,ns,name,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_attr_add" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_attr_add" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_attr_add" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "XmlNode_attr_add" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    result = (WsXmlAttrH)__WsXmlNode_attr_add(arg1,(char const *)arg2,(char const *)arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___WsXmlAttr, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_XmlNode_epr) {
  {
    struct __WsXmlNode *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    int arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    epr_t *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: XmlNode_epr(self,ns,epr_node_name,embedded);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlNode_epr" "', argument " "1"" of type '" "struct __WsXmlNode *""'"); 
    }
    arg1 = (struct __WsXmlNode *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "XmlNode_epr" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "XmlNode_epr" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "XmlNode_epr" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
    result = (epr_t *)__WsXmlNode_epr(arg1,(char const *)arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_epr_t, 0 | SWIG_SHADOW); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlAttr_name) {
  {
    struct __WsXmlAttr *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlAttr_name(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlAttr_name" "', argument " "1"" of type '" "struct __WsXmlAttr *""'"); 
    }
    arg1 = (struct __WsXmlAttr *)(argp1);
    result = (char *)__WsXmlAttr_name(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlAttr_ns) {
  {
    struct __WsXmlAttr *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlAttr_ns(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlAttr_ns" "', argument " "1"" of type '" "struct __WsXmlAttr *""'"); 
    }
    arg1 = (struct __WsXmlAttr *)(argp1);
    result = (char *)__WsXmlAttr_ns(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlAttr_value) {
  {
    struct __WsXmlAttr *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlAttr_value(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlAttr_value" "', argument " "1"" of type '" "struct __WsXmlAttr *""'"); 
    }
    arg1 = (struct __WsXmlAttr *)(argp1);
    result = (char *)__WsXmlAttr_value(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_XmlAttr_remove) {
  {
    struct __WsXmlAttr *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: XmlAttr_remove(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsXmlAttr, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "XmlAttr_remove" "', argument " "1"" of type '" "struct __WsXmlAttr *""'"); 
    }
    arg1 = (struct __WsXmlAttr *)(argp1);
    __WsXmlAttr_remove(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_EndPointReference__SWIG_0) {
  {
    char *arg1 = 0 ;
    char *arg2 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    epr_t *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: new_EndPointReference(uri,address);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_EndPointReference" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_EndPointReference" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (epr_t *)new_epr_t__SWIG_0((char const *)arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_epr_t, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_new_EndPointReference__SWIG_1) {
  {
    char *arg1 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int argvi = 0;
    epr_t *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_EndPointReference(uri);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_EndPointReference" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    result = (epr_t *)new_epr_t__SWIG_1((char const *)arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_epr_t, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  }
  SWIG_croak_null();
}


XS(_wrap_new_EndPointReference) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if (items == 1) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        int res = SWIG_AsCharPtrAndSize(ST(0), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if (items == 2) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        int res = SWIG_AsCharPtrAndSize(ST(0), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_EndPointReference__SWIG_1); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_EndPointReference__SWIG_0); return;
    }
  }
  
  croak("No matching function for overloaded 'new_EndPointReference'");
  XSRETURN(0);
}


XS(_wrap_delete_EndPointReference) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_EndPointReference(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EndPointReference" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    delete_epr_t(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_add_selector) {
  {
    epr_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: EndPointReference_add_selector(self,name,text);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_add_selector" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EndPointReference_add_selector" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "EndPointReference_add_selector" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    epr_t_add_selector(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_serialize) {
  {
    epr_t *arg1 = 0 ;
    WsXmlNodeH arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    int arg5 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int val5 ;
    int ecode5 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 5) || (items > 5)) {
      SWIG_croak("Usage: EndPointReference_serialize(self,node,ns,epr_node_name,embedded);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_serialize" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EndPointReference_serialize" "', argument " "2"" of type '" "WsXmlNodeH""'"); 
    }
    arg2 = (WsXmlNodeH)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "EndPointReference_serialize" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "EndPointReference_serialize" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    ecode5 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "EndPointReference_serialize" "', argument " "5"" of type '" "int""'");
    } 
    arg5 = (int)(val5);
    result = (int)epr_t_serialize(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_cmp) {
  {
    epr_t *arg1 = 0 ;
    epr_t *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EndPointReference_cmp(self,epr2);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_cmp" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EndPointReference_cmp" "', argument " "2"" of type '" "epr_t const *""'"); 
    }
    arg2 = (epr_t *)(argp2);
    result = (int)epr_t_cmp(arg1,(epr_t const *)arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_to_xml) {
  {
    epr_t *arg1 = 0 ;
    char *arg2 = (char *)NULL ;
    char *arg3 = (char *)NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 3)) {
      SWIG_croak("Usage: EndPointReference_to_xml(self,ns,epr_node_name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_to_xml" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EndPointReference_to_xml" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "EndPointReference_to_xml" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (char *)epr_t_to_xml(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_string) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_string(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_string" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (char *)epr_t_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_selector_count) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_selector_count(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_selector_count" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (int)epr_t_selector_count(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_resource_uri) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_resource_uri(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_resource_uri" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (char *)epr_t_resource_uri(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_selector) {
  {
    epr_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EndPointReference_selector(self,name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_selector" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EndPointReference_selector" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (char *)epr_t_selector(arg1,(char const *)arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_selector_names) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    AV *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_selector_names(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_selector_names" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (AV *)epr_t_selector_names(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_AV, 0 | 0); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_classname) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_classname(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_classname" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (char *)epr_t_classname(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_namespace) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_namespace(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_namespace" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (char *)epr_t_namespace(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EndPointReference_prefix) {
  {
    epr_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EndPointReference_prefix(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EndPointReference_prefix" "', argument " "1"" of type '" "epr_t *""'"); 
    }
    arg1 = (epr_t *)(argp1);
    result = (char *)epr_t_prefix(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_resultClass_set) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Filter_resultClass_set(self,resultClass);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_resultClass_set" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_resultClass_set" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    free((char*)arg1->resultClass);
    if (arg2) {
      size_t size = strlen((const char *)(arg2)) + 1;
      arg1->resultClass = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
    } else {
      arg1->resultClass = 0;
    }
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_resultClass_get) {
  {
    filter_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Filter_resultClass_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_resultClass_get" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    result = (char *) ((arg1)->resultClass);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_assocClass_set) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Filter_assocClass_set(self,assocClass);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_assocClass_set" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_assocClass_set" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    free((char*)arg1->assocClass);
    if (arg2) {
      size_t size = strlen((const char *)(arg2)) + 1;
      arg1->assocClass = (char *)(char *)memcpy(malloc((size)*sizeof(char)), (const char *)(arg2), sizeof(char)*(size));
    } else {
      arg1->assocClass = 0;
    }
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_assocClass_get) {
  {
    filter_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Filter_assocClass_get(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_assocClass_get" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    result = (char *) ((arg1)->assocClass);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Filter) {
  {
    int argvi = 0;
    filter_t *result = 0 ;
    dXSARGS;
    
    if ((items < 0) || (items > 0)) {
      SWIG_croak("Usage: new_Filter();");
    }
    result = (filter_t *)new_filter_t();
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_t, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Filter) {
  {
    filter_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Filter(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Filter" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    delete_filter_t(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_associators) {
  {
    filter_t *arg1 = 0 ;
    epr_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    char *arg5 = 0 ;
    char *arg6 = 0 ;
    char **arg7 = 0 ;
    int arg8 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int res6 ;
    char *buf6 = 0 ;
    int alloc6 = 0 ;
    void *argp7 = 0 ;
    int res7 = 0 ;
    int val8 ;
    int ecode8 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 8) || (items > 8)) {
      SWIG_croak("Usage: Filter_associators(self,epr,assocClass,resultClass,role,resultRole,resultProp,propNum);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_associators" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_associators" "', argument " "2"" of type '" "epr_t *""'"); 
    }
    arg2 = (epr_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Filter_associators" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Filter_associators" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Filter_associators" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    res6 = SWIG_AsCharPtrAndSize(ST(5), &buf6, NULL, &alloc6);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Filter_associators" "', argument " "6"" of type '" "char const *""'");
    }
    arg6 = (char *)(buf6);
    res7 = SWIG_ConvertPtr(ST(6), &argp7,SWIGTYPE_p_p_char, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Filter_associators" "', argument " "7"" of type '" "char **""'"); 
    }
    arg7 = (char **)(argp7);
    ecode8 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(7), &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Filter_associators" "', argument " "8"" of type '" "int""'");
    } 
    arg8 = (int)(val8);
    result = (int)filter_t_associators(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_references) {
  {
    filter_t *arg1 = 0 ;
    epr_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    char *arg5 = 0 ;
    char *arg6 = 0 ;
    char **arg7 = 0 ;
    int arg8 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int res6 ;
    char *buf6 = 0 ;
    int alloc6 = 0 ;
    void *argp7 = 0 ;
    int res7 = 0 ;
    int val8 ;
    int ecode8 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 8) || (items > 8)) {
      SWIG_croak("Usage: Filter_references(self,epr,assocClass,resultClass,role,resultRole,resultProp,propNum);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_references" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_references" "', argument " "2"" of type '" "epr_t *""'"); 
    }
    arg2 = (epr_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Filter_references" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Filter_references" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Filter_references" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    res6 = SWIG_AsCharPtrAndSize(ST(5), &buf6, NULL, &alloc6);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Filter_references" "', argument " "6"" of type '" "char const *""'");
    }
    arg6 = (char *)(buf6);
    res7 = SWIG_ConvertPtr(ST(6), &argp7,SWIGTYPE_p_p_char, 0 |  0 );
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "Filter_references" "', argument " "7"" of type '" "char **""'"); 
    }
    arg7 = (char **)(argp7);
    ecode8 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(7), &val8);
    if (!SWIG_IsOK(ecode8)) {
      SWIG_exception_fail(SWIG_ArgError(ecode8), "in method '" "Filter_references" "', argument " "8"" of type '" "int""'");
    } 
    arg8 = (int)(val8);
    result = (int)filter_t_references(arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6,arg7,arg8);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_simple) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Filter_simple(self,dialect,query);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_simple" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_simple" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Filter_simple" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    result = (int)filter_t_simple(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_xpath) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Filter_xpath(self,query);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_xpath" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_xpath" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (int)filter_t_xpath(arg1,(char const *)arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_cql) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Filter_cql(self,query);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_cql" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_cql" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (int)filter_t_cql(arg1,(char const *)arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Filter_wql) {
  {
    filter_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Filter_wql(self,query);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Filter_wql" "', argument " "1"" of type '" "filter_t *""'"); 
    }
    arg1 = (filter_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Filter_wql" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (int)filter_t_wql(arg1,(char const *)arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_delete_EnumerateInfo) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_EnumerateInfo(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_EnumerateInfo" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    delete___WsEnumerateInfo(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_max_items) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_max_items(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_max_items" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (int)__WsEnumerateInfo_max_items(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_set_max_items) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EnumerateInfo_set_max_items(self,mi);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_set_max_items" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EnumerateInfo_set_max_items" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    __WsEnumerateInfo_set_max_items(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_flags) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_flags(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_flags" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (int)__WsEnumerateInfo_flags(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_epr_to) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_epr_to(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_epr_to" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (char *)__WsEnumerateInfo_epr_to(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_epr_uri) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_epr_uri(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_epr_uri" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (char *)__WsEnumerateInfo_epr_uri(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_encoding) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_encoding(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_encoding" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (char *)__WsEnumerateInfo_encoding(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_filter) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    filter_t *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_filter(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_filter" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (filter_t *)__WsEnumerateInfo_filter(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_filter_t, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_index) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_index(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_index" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (int)__WsEnumerateInfo_index(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_set_index) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EnumerateInfo_set_index(self,i);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_set_index" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EnumerateInfo_set_index" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    __WsEnumerateInfo_set_index(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_total_items) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_total_items(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_total_items" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (int)__WsEnumerateInfo_total_items(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_set_total_items) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EnumerateInfo_set_total_items(self,i);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_set_total_items" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "EnumerateInfo_set_total_items" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    __WsEnumerateInfo_set_total_items(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_pull_result) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: EnumerateInfo_pull_result(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_pull_result" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    result = (WsXmlDocH)__WsEnumerateInfo_pull_result(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_EnumerateInfo_set_pull_result) {
  {
    struct __WsEnumerateInfo *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: EnumerateInfo_set_pull_result(self,result);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "EnumerateInfo_set_pull_result" "', argument " "1"" of type '" "struct __WsEnumerateInfo *""'"); 
    }
    arg1 = (struct __WsEnumerateInfo *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "EnumerateInfo_set_pull_result" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    __WsEnumerateInfo_set_pull_result(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_indoc) {
  {
    struct __SoapOp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SoapOp_indoc(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_indoc" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    result = (WsXmlDocH)__SoapOp_indoc(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_set_indoc) {
  {
    struct __SoapOp *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: SoapOp_set_indoc(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_set_indoc" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SoapOp_set_indoc" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    __SoapOp_set_indoc(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_outdoc) {
  {
    struct __SoapOp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SoapOp_outdoc(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_outdoc" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    result = (WsXmlDocH)__SoapOp_outdoc(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_set_outdoc) {
  {
    struct __SoapOp *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: SoapOp_set_outdoc(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_set_outdoc" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SoapOp_set_outdoc" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    __SoapOp_set_outdoc(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_soap) {
  {
    struct __SoapOp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    struct __Soap *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SoapOp_soap(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_soap" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    result = (struct __Soap *)__SoapOp_soap(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___Soap, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_msg) {
  {
    struct __SoapOp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsmanMessage *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SoapOp_msg(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_msg" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    result = (WsmanMessage *)__SoapOp_msg(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_WsmanMessage, 0 | 0); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_SoapOp_maxsize) {
  {
    struct __SoapOp *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: SoapOp_maxsize(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___SoapOp, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SoapOp_maxsize" "', argument " "1"" of type '" "struct __SoapOp *""'"); 
    }
    arg1 = (struct __SoapOp *)(argp1);
    result = (unsigned long)__SoapOp_maxsize(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Soap) {
  {
    struct __Soap *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Soap(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___Soap, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Soap" "', argument " "1"" of type '" "struct __Soap *""'"); 
    }
    arg1 = (struct __Soap *)(argp1);
    delete___Soap(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Soap_create_context) {
  {
    struct __Soap *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsContextH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Soap_create_context(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___Soap, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Soap_create_context" "', argument " "1"" of type '" "struct __Soap *""'"); 
    }
    arg1 = (struct __Soap *)(argp1);
    result = (WsContextH)__Soap_create_context(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Soap_context) {
  {
    struct __Soap *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsContextH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Soap_context(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___Soap, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Soap_context" "', argument " "1"" of type '" "struct __Soap *""'"); 
    }
    arg1 = (struct __Soap *)(argp1);
    result = (WsContextH)__Soap_context(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Soap_create_ep_context) {
  {
    struct __Soap *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    WsContextH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Soap_create_ep_context(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p___Soap, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Soap_create_ep_context" "', argument " "1"" of type '" "struct __Soap *""'"); 
    }
    arg1 = (struct __Soap *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Soap_create_ep_context" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    result = (WsContextH)__Soap_create_ep_context(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WS_CONTEXT, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Context) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Context(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Context" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    delete__WS_CONTEXT(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_indoc) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Context_indoc(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_indoc" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    result = (WsXmlDocH)_WS_CONTEXT_indoc(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_runtime) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    struct __Soap *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Context_runtime(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_runtime" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    result = (struct __Soap *)_WS_CONTEXT_runtime(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p___Soap, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_set_enumIdleTimeout) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Context_set_enumIdleTimeout(self,timeout);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_set_enumIdleTimeout" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Context_set_enumIdleTimeout" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    _WS_CONTEXT_set_enumIdleTimeout(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_classname) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Context_classname(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_classname" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    result = (char *)_WS_CONTEXT_classname(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_method) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Context_method(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_method" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    result = (char *)_WS_CONTEXT_method(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_method_args) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    hash_t *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Context_method_args(self,resource_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_method_args" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_method_args" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    result = (hash_t *)_WS_CONTEXT_method_args(arg1,(char const *)arg2);
    {
      ST(argvi) = hash2value(result);
      
    }
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Context_max_elements) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_max_elements(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_max_elements" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_max_elements" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (int)_WS_CONTEXT_max_elements(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_max_envelope_size) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_max_envelope_size(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_max_envelope_size" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_max_envelope_size" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (unsigned long)_WS_CONTEXT_max_envelope_size(arg1,arg2);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_fragment_string) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_fragment_string(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_fragment_string" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_fragment_string" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (char *)_WS_CONTEXT_fragment_string(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_selector) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    char *arg3 = 0 ;
    int arg4 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int val4 ;
    int ecode4 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: Context_selector(self,doc,name,index);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_selector" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_selector" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Context_selector" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    ecode4 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(3), &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Context_selector" "', argument " "4"" of type '" "int""'");
    } 
    arg4 = (int)(val4);
    result = (char *)_WS_CONTEXT_selector(arg1,arg2,(char const *)arg3,arg4);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_selectors_from_epr) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlNodeH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    hash_t *result = 0 ;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Context_selectors_from_epr(self,epr_node);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_selectors_from_epr" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p___WsXmlNode, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_selectors_from_epr" "', argument " "2"" of type '" "WsXmlNodeH""'"); 
    }
    arg2 = (WsXmlNodeH)(argp2);
    result = (hash_t *)_WS_CONTEXT_selectors_from_epr(arg1,arg2);
    {
      ST(argvi) = hash2value(result);
      
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_selectors) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    hash_t *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_selectors(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_selectors" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_selectors" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (hash_t *)_WS_CONTEXT_selectors(arg1,arg2);
    {
      ST(argvi) = hash2value(result);
      
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_selectors_from_filter) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    hash_t *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_selectors_from_filter(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_selectors_from_filter" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_selectors_from_filter" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (hash_t *)_WS_CONTEXT_selectors_from_filter(arg1,arg2);
    {
      ST(argvi) = hash2value(result);
      
    }
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_action) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_action(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_action" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_action" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (char *)_WS_CONTEXT_action(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_resource_uri) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: Context_resource_uri(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_resource_uri" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    if (items > 1) {
      res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_resource_uri" "', argument " "2"" of type '" "WsXmlDocH""'"); 
      }
      arg2 = (WsXmlDocH)(argp2);
    }
    result = (char *)_WS_CONTEXT_resource_uri(arg1,arg2);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Context_option_set) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Context_option_set(self,doc,op);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_option_set" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_option_set" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Context_option_set" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    result = (char *)_WS_CONTEXT_option_set(arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_Context_parse_enum_request) {
  {
    struct _WS_CONTEXT *arg1 = 0 ;
    WsEnumerateInfo *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Context_parse_enum_request(self,enumInfo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WS_CONTEXT, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Context_parse_enum_request" "', argument " "1"" of type '" "struct _WS_CONTEXT *""'"); 
    }
    arg1 = (struct _WS_CONTEXT *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_WsEnumerateInfo, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Context_parse_enum_request" "', argument " "2"" of type '" "WsEnumerateInfo *""'"); 
    }
    arg2 = (WsEnumerateInfo *)(argp2);
    result = (int)_WS_CONTEXT_parse_enum_request(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Status) {
  {
    int arg1 = 0 ;
    int arg2 = 0 ;
    char *arg3 = (char *)NULL ;
    int val1 ;
    int ecode1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    struct _WsmanStatus *result = 0 ;
    dXSARGS;
    
    if ((items < 0) || (items > 3)) {
      SWIG_croak("Usage: new_Status(code,detail,msg);");
    }
    if (items > 0) {
      ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
      if (!SWIG_IsOK(ecode1)) {
        SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Status" "', argument " "1"" of type '" "int""'");
      } 
      arg1 = (int)(val1);
    }
    if (items > 1) {
      ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
      if (!SWIG_IsOK(ecode2)) {
        SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Status" "', argument " "2"" of type '" "int""'");
      } 
      arg2 = (int)(val2);
    }
    if (items > 2) {
      res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Status" "', argument " "3"" of type '" "char const *""'");
      }
      arg3 = (char *)(buf3);
    }
    result = (struct _WsmanStatus *)new__WsmanStatus(arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsmanStatus, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Status) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Status(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Status" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    delete__WsmanStatus(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_to_s) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Status_to_s(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_to_s" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    result = (char *)_WsmanStatus_to_s(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_set_code) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Status_set_code(self,code);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_set_code" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Status_set_code" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    _WsmanStatus_set_code(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_code) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Status_code(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_code" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    result = (int)_WsmanStatus_code(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_set_detail) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Status_set_detail(self,detail);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_set_detail" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Status_set_detail" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    _WsmanStatus_set_detail(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_detail) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Status_detail(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_detail" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    result = (int)_WsmanStatus_detail(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_set_msg) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Status_set_msg(self,msg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_set_msg" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Status_set_msg" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsmanStatus_set_msg(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Status_msg) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Status_msg(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_msg" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    result = (char *)_WsmanStatus_msg(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Status_generate_fault) {
  {
    struct _WsmanStatus *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Status_generate_fault(self,doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsmanStatus, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Status_generate_fault" "', argument " "1"" of type '" "struct _WsmanStatus *""'"); 
    }
    arg1 = (struct _WsmanStatus *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Status_generate_fault" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    result = (WsXmlDocH)_WsmanStatus_generate_fault(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Fault) {
  {
    WsXmlDocH arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    struct _WsManFault *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Fault(doc);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Fault" "', argument " "1"" of type '" "WsXmlDocH""'"); 
    }
    arg1 = (WsXmlDocH)(argp1);
    result = (struct _WsManFault *)new__WsManFault(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManFault, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_delete_Fault) {
  {
    struct _WsManFault *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Fault(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManFault, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Fault" "', argument " "1"" of type '" "struct _WsManFault *""'"); 
    }
    arg1 = (struct _WsManFault *)(argp1);
    delete__WsManFault(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Fault_code) {
  {
    struct _WsManFault *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Fault_code(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Fault_code" "', argument " "1"" of type '" "struct _WsManFault *""'"); 
    }
    arg1 = (struct _WsManFault *)(argp1);
    result = (char *)_WsManFault_code(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Fault_subcode) {
  {
    struct _WsManFault *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Fault_subcode(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Fault_subcode" "', argument " "1"" of type '" "struct _WsManFault *""'"); 
    }
    arg1 = (struct _WsManFault *)(argp1);
    result = (char *)_WsManFault_subcode(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Fault_reason) {
  {
    struct _WsManFault *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Fault_reason(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Fault_reason" "', argument " "1"" of type '" "struct _WsManFault *""'"); 
    }
    arg1 = (struct _WsManFault *)(argp1);
    result = (char *)_WsManFault_reason(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Fault_detail) {
  {
    struct _WsManFault *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Fault_detail(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManFault, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Fault_detail" "', argument " "1"" of type '" "struct _WsManFault *""'"); 
    }
    arg1 = (struct _WsManFault *)(argp1);
    result = (char *)_WsManFault_detail(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_is_auth_method) {
  {
    struct _WsManTransport *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_is_auth_method(self,method);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_is_auth_method" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transport_is_auth_method" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (int)_WsManTransport_is_auth_method(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_close) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_close(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_close" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    _WsManTransport_close(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_agent) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_agent(self,agent);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_agent" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_agent" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_agent(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_agent) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_agent(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_agent" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_agent(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_get_username) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_get_username(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_get_username" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_get_username(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_username) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_username(self,user_name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_username" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_username" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_username(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_get_password) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_get_password(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_get_password" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_get_password(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_password) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_password(self,password);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_password" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_password" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_password(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_get_proxy_username) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_get_proxy_username(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_get_proxy_username" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_get_proxy_username(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_proxy_username) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_proxy_username(self,proxy_username);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_proxy_username" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_proxy_username" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_proxy_username(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_get_proxy_password) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_get_proxy_password(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_get_proxy_password" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_get_proxy_password(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_proxy_password) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_proxy_password(self,proxy_password);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_proxy_password" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_proxy_password" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_proxy_password(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_auth_method) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_auth_method(self,am);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_auth_method" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_auth_method" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_auth_method(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_auth_method) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_auth_method(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_auth_method" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_auth_method(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_auth_name) {
  {
    int arg1 ;
    int val1 ;
    int ecode1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_auth_name(auth);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Transport_auth_name" "', argument " "1"" of type '" "int""'");
    } 
    arg1 = (int)(val1);
    result = (char *)_WsManTransport_auth_name(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_auth_value) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_auth_value(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_auth_value" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (int)_WsManTransport_auth_value(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_error_string) {
  {
    int arg1 ;
    int val1 ;
    int ecode1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_error_string(err);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "Transport_error_string" "', argument " "1"" of type '" "int""'");
    } 
    arg1 = (int)(val1);
    result = (char *)_WsManTransport_error_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_timeout) {
  {
    struct _WsManTransport *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_timeout(self,timeout);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_timeout" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transport_set_timeout" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    _WsManTransport_set_timeout(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_timeout) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_timeout(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_timeout" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (unsigned long)_WsManTransport_timeout(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_verify_peer) {
  {
    struct _WsManTransport *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_verify_peer(self,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_verify_peer" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transport_set_verify_peer" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    _WsManTransport_set_verify_peer(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_verify_peer) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_verify_peer(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_verify_peer" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (unsigned int)_WsManTransport_verify_peer(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_verify_host) {
  {
    struct _WsManTransport *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_verify_host(self,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_verify_host" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Transport_set_verify_host" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    _WsManTransport_set_verify_host(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_verify_host) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_verify_host(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_verify_host" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (unsigned int)_WsManTransport_verify_host(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_proxy) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_proxy(self,proxy);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_proxy" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_proxy" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_proxy(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_proxy) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_proxy(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_proxy" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_proxy(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_proxyauth) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_proxyauth(self,pauth);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_proxyauth" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_proxyauth" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_proxyauth(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_proxyauth) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_proxyauth(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_proxyauth" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_proxyauth(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_cainfo) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_cainfo(self,cainfo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_cainfo" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_cainfo" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_cainfo(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_cainfo) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_cainfo(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_cainfo" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_cainfo(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_certhumbprint) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_certhumbprint(self,arg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_certhumbprint" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_certhumbprint" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_certhumbprint(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_certhumbprint) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_certhumbprint(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_certhumbprint" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_certhumbprint(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_capath) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_capath(self,capath);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_capath" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_capath" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_capath(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_capath) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_capath(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_capath" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_capath(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_caoid) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_caoid(self,oid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_caoid" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_caoid" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_caoid(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_caoid) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_caoid(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_caoid" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_caoid(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_cert) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_cert(self,cert);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_cert" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_cert" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_cert(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_cert) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_cert(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_cert" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_cert(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_set_key) {
  {
    struct _WsManTransport *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Transport_set_key(self,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_set_key" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Transport_set_key" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManTransport_set_key(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Transport_key) {
  {
    struct _WsManTransport *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Transport_key(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManTransport, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Transport_key" "', argument " "1"" of type '" "struct _WsManTransport *""'"); 
    }
    arg1 = (struct _WsManTransport *)(argp1);
    result = (char *)_WsManTransport_key(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_send_request) {
  {
    WsManClient *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_send_request(cl,request);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_send_request" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_send_request" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    result = (int)wsman_send_request(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsmc_transport_set_auth_request_func) {
  {
    WsManClient *arg1 = 0 ;
    wsman_auth_request_func_t arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 ;
    int res2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsmc_transport_set_auth_request_func(cl,f);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsmc_transport_set_auth_request_func" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    {
      res2 = SWIG_ConvertPtr(ST(1), &argp2, SWIGTYPE_p_wsman_auth_request_func_t,  0 );
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsmc_transport_set_auth_request_func" "', argument " "2"" of type '" "wsman_auth_request_func_t""'"); 
      }  
      if (!argp2) {
        SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "wsmc_transport_set_auth_request_func" "', argument " "2"" of type '" "wsman_auth_request_func_t""'");
      } else {
        arg2 = *((wsman_auth_request_func_t *)(argp2));
      }
    }
    wsmc_transport_set_auth_request_func(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_is_auth_method) {
  {
    WsManClient *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_is_auth_method(cl,method);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_is_auth_method" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "wsman_is_auth_method" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    result = (int)wsman_is_auth_method(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsmc_transport_init) {
  {
    WsManClient *arg1 = 0 ;
    void *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsmc_transport_init(cl,arg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsmc_transport_init" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1),SWIG_as_voidptrptr(&arg2), 0, 0);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsmc_transport_init" "', argument " "2"" of type '" "void *""'"); 
    }
    result = (int)wsmc_transport_init(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_close_transport) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_close_transport(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_close_transport" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    wsman_transport_close_transport(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsmc_transport_fini) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsmc_transport_fini(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsmc_transport_fini" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    wsmc_transport_fini(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_agent) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_agent(cl,agent);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_agent" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_agent" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_agent(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_agent) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_agent(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_agent" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_agent(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_userName) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_userName(cl,user_name);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_userName" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_userName" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_userName(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_userName) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_userName(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_userName" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_userName(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_password) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_password(cl,password);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_password" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_password" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_password(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_password) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_password(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_password" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_password(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_proxy_username) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_proxy_username(cl,proxy_username);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_proxy_username" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_proxy_username" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_proxy_username(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_proxy_username) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_proxy_username(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_proxy_username" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_proxy_username(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_proxy_password) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_proxy_password(cl,proxy_password);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_proxy_password" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_proxy_password" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_proxy_password(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_proxy_password) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_proxy_password(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_proxy_password" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_proxy_password(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_auth_method) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_auth_method(cl,am);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_auth_method" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_auth_method" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_auth_method(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_auth_method) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_auth_method(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_auth_method" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_auth_method(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsmc_transport_get_auth_name) {
  {
    wsman_auth_type_t arg1 ;
    void *argp1 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsmc_transport_get_auth_name(auth);");
    }
    {
      res1 = SWIG_ConvertPtr(ST(0), &argp1, SWIGTYPE_p_wsman_auth_type_t,  0 );
      if (!SWIG_IsOK(res1)) {
        SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsmc_transport_get_auth_name" "', argument " "1"" of type '" "wsman_auth_type_t""'"); 
      }  
      if (!argp1) {
        SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "wsmc_transport_get_auth_name" "', argument " "1"" of type '" "wsman_auth_type_t""'");
      } else {
        arg1 = *((wsman_auth_type_t *)(argp1));
      }
    }
    result = (char *)wsmc_transport_get_auth_name(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_wsmc_transport_get_auth_value) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    wsman_auth_type_t result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsmc_transport_get_auth_value(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsmc_transport_get_auth_value" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = wsmc_transport_get_auth_value(arg1);
    ST(argvi) = SWIG_NewPointerObj((wsman_auth_type_t *)memcpy((wsman_auth_type_t *)calloc(1,sizeof(wsman_auth_type_t)),&result,sizeof(wsman_auth_type_t)), SWIGTYPE_p_wsman_auth_type_t, SWIG_POINTER_OWN | 0); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_last_error_string) {
  {
    WS_LASTERR_Code arg1 ;
    void *argp1 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_last_error_string(err);");
    }
    {
      res1 = SWIG_ConvertPtr(ST(0), &argp1, SWIGTYPE_p_WS_LASTERR_Code,  0 );
      if (!SWIG_IsOK(res1)) {
        SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_last_error_string" "', argument " "1"" of type '" "WS_LASTERR_Code""'"); 
      }  
      if (!argp1) {
        SWIG_exception_fail(SWIG_NullReferenceError, "invalid null reference " "in method '" "wsman_transport_get_last_error_string" "', argument " "1"" of type '" "WS_LASTERR_Code""'");
      } else {
        arg1 = *((WS_LASTERR_Code *)(argp1));
      }
    }
    result = (char *)wsman_transport_get_last_error_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_timeout) {
  {
    WsManClient *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_timeout(cl,timeout);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_timeout" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "wsman_transport_set_timeout" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    wsman_transport_set_timeout(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_timeout) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_timeout(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_timeout" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (unsigned long)wsman_transport_get_timeout(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_verify_peer) {
  {
    WsManClient *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_verify_peer(cl,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_verify_peer" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "wsman_transport_set_verify_peer" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    wsman_transport_set_verify_peer(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_verify_peer) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_verify_peer(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_verify_peer" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (unsigned int)wsman_transport_get_verify_peer(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_verify_host) {
  {
    WsManClient *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_verify_host(cl,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_verify_host" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "wsman_transport_set_verify_host" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    wsman_transport_set_verify_host(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_verify_host) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_verify_host(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_verify_host" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (unsigned int)wsman_transport_get_verify_host(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_crlcheck) {
  {
    WsManClient *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_crlcheck(cl,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_crlcheck" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "wsman_transport_set_crlcheck" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    wsman_transport_set_crlcheck(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_crlcheck) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_crlcheck(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_crlcheck" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (unsigned int)wsman_transport_get_crlcheck(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_int  SWIG_PERL_CALL_ARGS_1((unsigned int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_crlfile) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_crlfile(cl,arg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_crlfile" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_crlfile" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_crlfile(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_crlfile) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_crlfile(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_crlfile" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_crlfile(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_proxy) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_proxy(cl,proxy);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_proxy" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_proxy" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_proxy(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_proxy) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_proxy(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_proxy" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_proxy(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_proxyauth) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_proxyauth(cl,pauth);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_proxyauth" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_proxyauth" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_proxyauth(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_proxyauth) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_proxyauth(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_proxyauth" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_proxyauth(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_cainfo) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_cainfo(cl,cainfo);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_cainfo" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_cainfo" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_cainfo(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_cainfo) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_cainfo(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_cainfo" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_cainfo(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_certhumbprint) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_certhumbprint(cl,arg);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_certhumbprint" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_certhumbprint" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_certhumbprint(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_certhumbprint) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_certhumbprint(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_certhumbprint" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_certhumbprint(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_capath) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_capath(cl,capath);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_capath" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_capath" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_capath(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_capath) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_capath(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_capath" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_capath(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_caoid) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_caoid(cl,oid);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_caoid" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_caoid" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_caoid(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_caoid) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_caoid(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_caoid" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_caoid(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_cert) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_cert(cl,cert);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_cert" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_cert" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_cert(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_cert) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_cert(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_cert" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_cert(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_set_key) {
  {
    WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: wsman_transport_set_key(cl,key);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_set_key" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "wsman_transport_set_key" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    wsman_transport_set_key(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_wsman_transport_get_key) {
  {
    WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: wsman_transport_get_key(cl);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "wsman_transport_get_key" "', argument " "1"" of type '" "WsManClient *""'"); 
    }
    arg1 = (WsManClient *)(argp1);
    result = (char *)wsman_transport_get_key(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_ClientOptions) {
  {
    int argvi = 0;
    client_opt_t *result = 0 ;
    dXSARGS;
    
    if ((items < 0) || (items > 0)) {
      SWIG_croak("Usage: new_ClientOptions();");
    }
    result = (client_opt_t *)new_client_opt_t();
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_client_opt_t, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_delete_ClientOptions) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_ClientOptions(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ClientOptions" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    delete_client_opt_t(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_dump_request) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_set_dump_request(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_dump_request" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    client_opt_t_set_dump_request(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_clear_dump_request) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_clear_dump_request(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_clear_dump_request" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    client_opt_t_clear_dump_request(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_flags) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_flags(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_flags" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_flags" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    client_opt_t_set_flags(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_flags) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_flags(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_flags" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (unsigned long)client_opt_t_get_flags(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_clear_flags) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_clear_flags(self,flags);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_clear_flags" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_clear_flags" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    client_opt_t_clear_flags(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_reset_flags) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_reset_flags(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_reset_flags" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    client_opt_t_reset_flags(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_max_envelope_size) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_max_envelope_size(self,size);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_max_envelope_size" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_max_envelope_size" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    client_opt_t_set_max_envelope_size(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_max_envelope_size) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_max_envelope_size(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_max_envelope_size" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (unsigned long)client_opt_t_get_max_envelope_size(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_max_elements) {
  {
    client_opt_t *arg1 = 0 ;
    int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_max_elements(self,elements);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_max_elements" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_max_elements" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    client_opt_t_set_max_elements(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_max_elements) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_max_elements(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_max_elements" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (int)client_opt_t_get_max_elements(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_timeout) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned long arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned long val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_timeout(self,timeout);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_timeout" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_long SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_timeout" "', argument " "2"" of type '" "unsigned long""'");
    } 
    arg2 = (unsigned long)(val2);
    client_opt_t_set_timeout(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_timeout) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    unsigned long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_timeout(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_timeout" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (unsigned long)client_opt_t_get_timeout(arg1);
    ST(argvi) = SWIG_From_unsigned_SS_long  SWIG_PERL_CALL_ARGS_1((unsigned long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_fragment) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_fragment(self,fragment);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_fragment" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_set_fragment" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    client_opt_t_set_fragment(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_fragment) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_fragment(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_fragment" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (char *)client_opt_t_get_fragment(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_cim_namespace) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_cim_namespace(self,cim_namespace);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_cim_namespace" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_set_cim_namespace" "', argument " "2"" of type '" "char *""'");
    }
    arg2 = (char *)(buf2);
    client_opt_t_set_cim_namespace(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_cim_namespace) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_cim_namespace(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_cim_namespace" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (char *)client_opt_t_get_cim_namespace(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_reference) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_reference(self,reference);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_reference" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_set_reference" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    client_opt_t_set_reference(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_get_reference) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_get_reference(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_get_reference" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (char *)client_opt_t_get_reference(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_option) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: ClientOptions_add_option(self,key,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_add_option" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_add_option" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ClientOptions_add_option" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    client_opt_t_add_option(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_selector) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: ClientOptions_add_selector(self,key,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_add_selector" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_add_selector" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ClientOptions_add_selector" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    client_opt_t_add_selector(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_property__SWIG_0) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: ClientOptions_add_property(self,key,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_add_property" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_add_property" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ClientOptions_add_property" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    client_opt_t_add_property__SWIG_0(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_property_array_item) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: ClientOptions_add_property_array_item(self,key,value);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_add_property_array_item" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_add_property_array_item" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ClientOptions_add_property_array_item" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    client_opt_t_add_property_array_item(arg1,(char const *)arg2,(char const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_property__SWIG_1) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    epr_t *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: ClientOptions_add_property(self,key,epr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_add_property" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_add_property" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "ClientOptions_add_property" "', argument " "3"" of type '" "epr_t const *""'"); 
    }
    arg3 = (epr_t *)(argp3);
    client_opt_t_add_property__SWIG_1(arg1,(char const *)arg2,(epr_t const *)arg3);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_add_property) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if (items == 3) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p_client_opt_t, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(2), &vptr, SWIGTYPE_p_epr_t, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if (items == 3) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(ST(0), &vptr, SWIGTYPE_p_client_opt_t, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(1), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(2), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_ClientOptions_add_property__SWIG_1); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_ClientOptions_add_property__SWIG_0); return;
    }
  }
  
  croak("No matching function for overloaded 'ClientOptions_add_property'");
  XSRETURN(0);
}


XS(_wrap_ClientOptions_set_delivery_uri) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_delivery_uri(self,delivery_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_delivery_uri" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_set_delivery_uri" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    client_opt_t_set_delivery_uri(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_delivery_uri) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_delivery_uri(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_delivery_uri" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (char *)client_opt_t_delivery_uri(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_sub_expiry) {
  {
    client_opt_t *arg1 = 0 ;
    float arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    float val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_sub_expiry(self,event_subscription_expire);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_sub_expiry" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_float SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_sub_expiry" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = (float)(val2);
    client_opt_t_set_sub_expiry(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_sub_expiry) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    float result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_sub_expiry(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_sub_expiry" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (float)client_opt_t_sub_expiry(arg1);
    ST(argvi) = SWIG_From_float  SWIG_PERL_CALL_ARGS_1((float)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_heartbeat_interval) {
  {
    client_opt_t *arg1 = 0 ;
    float arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    float val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_heartbeat_interval(self,heartbeat_interval);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_heartbeat_interval" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_float SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_heartbeat_interval" "', argument " "2"" of type '" "float""'");
    } 
    arg2 = (float)(val2);
    client_opt_t_set_heartbeat_interval(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_heartbeat_interval) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    float result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_heartbeat_interval(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_heartbeat_interval" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (float)client_opt_t_heartbeat_interval(arg1);
    ST(argvi) = SWIG_From_float  SWIG_PERL_CALL_ARGS_1((float)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_delivery_mode) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_delivery_mode(self,delivery_mode);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_delivery_mode" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_delivery_mode" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    client_opt_t_set_delivery_mode(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_delivery_mode) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_delivery_mode(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_delivery_mode" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (int)client_opt_t_delivery_mode(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_delivery_security_mode) {
  {
    client_opt_t *arg1 = 0 ;
    unsigned int arg2 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    unsigned int val2 ;
    int ecode2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_delivery_security_mode(self,delivery_sec_mode);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_delivery_security_mode" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    ecode2 = SWIG_AsVal_unsigned_SS_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "ClientOptions_set_delivery_security_mode" "', argument " "2"" of type '" "unsigned int""'");
    } 
    arg2 = (unsigned int)(val2);
    client_opt_t_set_delivery_security_mode(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_delivery_sec_mode) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_delivery_sec_mode(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_delivery_sec_mode" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (int)client_opt_t_delivery_sec_mode(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_set_locale) {
  {
    client_opt_t *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: ClientOptions_set_locale(self,locale);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_set_locale" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "ClientOptions_set_locale" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    client_opt_t_set_locale(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_ClientOptions_locale) {
  {
    client_opt_t *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: ClientOptions_locale(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "ClientOptions_locale" "', argument " "1"" of type '" "client_opt_t *""'"); 
    }
    arg1 = (client_opt_t *)(argp1);
    result = (char *)client_opt_t_locale(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_new_Client__SWIG_0) {
  {
    char *arg1 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int argvi = 0;
    struct _WsManClient *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: new_Client(uri);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Client" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    result = (struct _WsManClient *)new__WsManClient__SWIG_0((char const *)arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManClient, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  }
  SWIG_croak_null();
}


XS(_wrap_new_Client__SWIG_1) {
  {
    char *arg1 = 0 ;
    int arg2 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    char *arg5 = 0 ;
    char *arg6 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int val2 ;
    int ecode2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int res6 ;
    char *buf6 = 0 ;
    int alloc6 = 0 ;
    int argvi = 0;
    struct _WsManClient *result = 0 ;
    dXSARGS;
    
    if ((items < 6) || (items > 6)) {
      SWIG_croak("Usage: new_Client(hostname,port,path,scheme,username,password);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Client" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    ecode2 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), &val2);
    if (!SWIG_IsOK(ecode2)) {
      SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Client" "', argument " "2"" of type '" "int""'");
    } 
    arg2 = (int)(val2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Client" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_Client" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_Client" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    res6 = SWIG_AsCharPtrAndSize(ST(5), &buf6, NULL, &alloc6);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "new_Client" "', argument " "6"" of type '" "char const *""'");
    }
    arg6 = (char *)(buf6);
    result = (struct _WsManClient *)new__WsManClient__SWIG_1((char const *)arg1,arg2,(char const *)arg3,(char const *)arg4,(char const *)arg5,(char const *)arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManClient, SWIG_OWNER | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  }
  SWIG_croak_null();
}


XS(_wrap_new_Client) {
  dXSARGS;
  
  {
    unsigned long _index = 0;
    SWIG_TypeRank _rank = 0; 
    if (items == 1) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        int res = SWIG_AsCharPtrAndSize(ST(0), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_1;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 1;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_1:
    
    if (items == 6) {
      SWIG_TypeRank _ranki = 0;
      SWIG_TypeRank _rankm = 0;
      SWIG_TypeRank _pi = 1;
      int _v = 0;
      {
        int res = SWIG_AsCharPtrAndSize(ST(0), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        {
          int res = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(1), NULL);
          _v = SWIG_CheckState(res);
        }
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(2), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(3), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(4), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      {
        int res = SWIG_AsCharPtrAndSize(ST(5), 0, NULL, 0);
        _v = SWIG_CheckState(res);
      }
      if (!_v) goto check_2;
      _ranki += _v*_pi;
      _rankm += _pi;
      _pi *= SWIG_MAXCASTRANK;
      if (!_index || (_ranki < _rank)) {
        _rank = _ranki; _index = 2;
        if (_rank == _rankm) goto dispatch;
      }
    }
  check_2:
    
  dispatch:
    switch(_index) {
    case 1:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_Client__SWIG_0); return;
    case 2:
      PUSHMARK(MARK); SWIG_CALLXS(_wrap_new_Client__SWIG_1); return;
    }
  }
  
  croak("No matching function for overloaded 'new_Client'");
  XSRETURN(0);
}


XS(_wrap_delete_Client) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: delete_Client(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, SWIG_POINTER_DISOWN |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Client" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    delete__WsManClient(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_set_dumpfile) {
  {
    struct _WsManClient *arg1 = 0 ;
    FILE *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Client_set_dumpfile(self,f);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_set_dumpfile" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    {
      arg2 = PerlIO_findFILE(IoIFP(sv_2io(ST(1))));
    }
    _WsManClient_set_dumpfile(arg1,arg2);
    ST(argvi) = &PL_sv_undef;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_response_code) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    long result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_response_code(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_response_code" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (long)_WsManClient_response_code(arg1);
    ST(argvi) = SWIG_From_long  SWIG_PERL_CALL_ARGS_1((long)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_scheme) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_scheme(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_scheme" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_scheme(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_host) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_host(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_host" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_host(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_port) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_port(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_port" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (int)_WsManClient_port(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_path) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_path(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_path" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_path(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_user) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_user(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_user" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_user(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_password) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_password(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_password" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_password(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    free(result);
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_transport) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsManTransport *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_transport(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_transport" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (WsManTransport *)_WsManClient_transport(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsManTransport, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_send_request) {
  {
    struct _WsManClient *arg1 = 0 ;
    WsXmlDocH arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Client_send_request(self,request);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_send_request" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_send_request" "', argument " "2"" of type '" "WsXmlDocH""'"); 
    }
    arg2 = (WsXmlDocH)(argp2);
    result = (int)_WsManClient_send_request(arg1,arg2);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_build_envelope_from_response) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_build_envelope_from_response(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_build_envelope_from_response" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (WsXmlDocH)_WsManClient_build_envelope_from_response(arg1);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_encoding) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_encoding(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_encoding" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_encoding(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_set_encoding) {
  {
    struct _WsManClient *arg1 = 0 ;
    char *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Client_set_encoding(self,encoding);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_set_encoding" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_set_encoding" "', argument " "2"" of type '" "char const *""'");
    }
    arg2 = (char *)(buf2);
    _WsManClient_set_encoding(arg1,(char const *)arg2);
    ST(argvi) = &PL_sv_undef;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_identify) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 2) || (items > 2)) {
      SWIG_croak("Usage: Client_identify(self,options);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_identify" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_identify" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    result = (WsXmlDocH)_WsManClient_identify(arg1,arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    XSRETURN(argvi);
    fail:;
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_get_from_epr) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    epr_t *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Client_get_from_epr(self,options,epr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_get_from_epr" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_get_from_epr" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_get_from_epr" "', argument " "3"" of type '" "epr_t *""'"); 
    }
    arg3 = (epr_t *)(argp3);
    result = (WsXmlDocH)_WsManClient_get_from_epr(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_delete_from_epr) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    epr_t *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Client_delete_from_epr(self,options,epr);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_delete_from_epr" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_delete_from_epr" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_epr_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_delete_from_epr" "', argument " "3"" of type '" "epr_t *""'"); 
    }
    arg3 = (epr_t *)(argp3);
    result = (WsXmlDocH)_WsManClient_delete_from_epr(arg1,arg2,arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    XSRETURN(argvi);
    fail:;
    
    
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_enumerate) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    filter_t *arg3 = 0 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: Client_enumerate(self,options,filter,resource_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_enumerate" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_enumerate" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_enumerate" "', argument " "3"" of type '" "filter_t *""'"); 
    }
    arg3 = (filter_t *)(argp3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_enumerate" "', argument " "4"" of type '" "char *""'");
    }
    arg4 = (char *)(buf4);
    result = (WsXmlDocH)_WsManClient_enumerate(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_pull) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    filter_t *arg3 = 0 ;
    char *arg4 = 0 ;
    char *arg5 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 5) || (items > 5)) {
      SWIG_croak("Usage: Client_pull(self,options,filter,resource_uri,context);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_pull" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_pull" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_pull" "', argument " "3"" of type '" "filter_t *""'"); 
    }
    arg3 = (filter_t *)(argp3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_pull" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Client_pull" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    result = (WsXmlDocH)_WsManClient_pull(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_create) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    size_t arg5 ;
    char *arg6 = (char *)"utf-8" ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    size_t val5 ;
    int ecode5 = 0 ;
    int res6 ;
    char *buf6 = 0 ;
    int alloc6 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 5) || (items > 6)) {
      SWIG_croak("Usage: Client_create(self,options,resource_uri,data,size,encoding);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_create" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_create" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_create" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_create" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    ecode5 = SWIG_AsVal_size_t SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Client_create" "', argument " "5"" of type '" "size_t""'");
    } 
    arg5 = (size_t)(val5);
    if (items > 5) {
      res6 = SWIG_AsCharPtrAndSize(ST(5), &buf6, NULL, &alloc6);
      if (!SWIG_IsOK(res6)) {
        SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Client_create" "', argument " "6"" of type '" "char const *""'");
      }
      arg6 = (char *)(buf6);
    }
    result = (WsXmlDocH)_WsManClient_create(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_put) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    size_t arg5 ;
    char *arg6 = (char *)"utf-8" ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    size_t val5 ;
    int ecode5 = 0 ;
    int res6 ;
    char *buf6 = 0 ;
    int alloc6 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 5) || (items > 6)) {
      SWIG_croak("Usage: Client_put(self,options,resource_uri,data,size,encoding);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_put" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_put" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_put" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_put" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    ecode5 = SWIG_AsVal_size_t SWIG_PERL_CALL_ARGS_2(ST(4), &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Client_put" "', argument " "5"" of type '" "size_t""'");
    } 
    arg5 = (size_t)(val5);
    if (items > 5) {
      res6 = SWIG_AsCharPtrAndSize(ST(5), &buf6, NULL, &alloc6);
      if (!SWIG_IsOK(res6)) {
        SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "Client_put" "', argument " "6"" of type '" "char const *""'");
      }
      arg6 = (char *)(buf6);
    }
    result = (WsXmlDocH)_WsManClient_put(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5,(char const *)arg6);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    if (alloc6 == SWIG_NEWOBJ) free((char*)buf6);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_release) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: Client_release(self,options,resource_uri,context);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_release" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_release" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_release" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_release" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    result = (WsXmlDocH)_WsManClient_release(arg1,arg2,(char const *)arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_get) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Client_get(self,options,resource_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_get" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_get" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_get" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    result = (WsXmlDocH)_WsManClient_get(arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_delete) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 3) || (items > 3)) {
      SWIG_croak("Usage: Client_delete(self,options,resource_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_delete" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_delete" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_delete" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    result = (WsXmlDocH)_WsManClient_delete(arg1,arg2,(char const *)arg3);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_invoke) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    WsXmlDocH arg5 = NULL ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    void *argp5 = 0 ;
    int res5 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 4) || (items > 5)) {
      SWIG_croak("Usage: Client_invoke(self,options,resource_uri,method,data);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_invoke" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_invoke" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_invoke" "', argument " "3"" of type '" "char const *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_invoke" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    if (items > 4) {
      res5 = SWIG_ConvertPtr(ST(4), &argp5,SWIGTYPE_p__WsXmlDoc, 0 |  0 );
      if (!SWIG_IsOK(res5)) {
        SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Client_invoke" "', argument " "5"" of type '" "WsXmlDocH""'"); 
      }
      arg5 = (WsXmlDocH)(argp5);
    }
    result = (WsXmlDocH)_WsManClient_invoke(arg1,arg2,(char const *)arg3,(char const *)arg4,arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_subscribe) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    filter_t *arg3 = 0 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: Client_subscribe(self,options,filter,resource_uri);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_subscribe" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_subscribe" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_subscribe" "', argument " "3"" of type '" "filter_t *""'"); 
    }
    arg3 = (filter_t *)(argp3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_subscribe" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    result = (WsXmlDocH)_WsManClient_subscribe(arg1,arg2,arg3,(char const *)arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_unsubscribe) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    filter_t *arg3 = 0 ;
    char *arg4 = 0 ;
    char *arg5 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    void *argp3 = 0 ;
    int res3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int res5 ;
    char *buf5 = 0 ;
    int alloc5 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 5) || (items > 5)) {
      SWIG_croak("Usage: Client_unsubscribe(self,options,filter,resource_uri,identifier);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_unsubscribe" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_unsubscribe" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_ConvertPtr(ST(2), &argp3,SWIGTYPE_p_filter_t, 0 |  0 );
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_unsubscribe" "', argument " "3"" of type '" "filter_t *""'"); 
    }
    arg3 = (filter_t *)(argp3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_unsubscribe" "', argument " "4"" of type '" "char const *""'");
    }
    arg4 = (char *)(buf4);
    res5 = SWIG_AsCharPtrAndSize(ST(4), &buf5, NULL, &alloc5);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "Client_unsubscribe" "', argument " "5"" of type '" "char const *""'");
    }
    arg5 = (char *)(buf5);
    result = (WsXmlDocH)_WsManClient_unsubscribe(arg1,arg2,arg3,(char const *)arg4,(char const *)arg5);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
    XSRETURN(argvi);
    fail:;
    
    
    
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    if (alloc5 == SWIG_NEWOBJ) free((char*)buf5);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_renew) {
  {
    struct _WsManClient *arg1 = 0 ;
    client_opt_t *arg2 = 0 ;
    char *arg3 = 0 ;
    char *arg4 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    void *argp2 = 0 ;
    int res2 = 0 ;
    int res3 ;
    char *buf3 = 0 ;
    int alloc3 = 0 ;
    int res4 ;
    char *buf4 = 0 ;
    int alloc4 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 4) || (items > 4)) {
      SWIG_croak("Usage: Client_renew(self,options,resource_uri,identifier);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_renew" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    res2 = SWIG_ConvertPtr(ST(1), &argp2,SWIGTYPE_p_client_opt_t, 0 |  0 );
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Client_renew" "', argument " "2"" of type '" "client_opt_t *""'"); 
    }
    arg2 = (client_opt_t *)(argp2);
    res3 = SWIG_AsCharPtrAndSize(ST(2), &buf3, NULL, &alloc3);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "Client_renew" "', argument " "3"" of type '" "char *""'");
    }
    arg3 = (char *)(buf3);
    res4 = SWIG_AsCharPtrAndSize(ST(3), &buf4, NULL, &alloc4);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "Client_renew" "', argument " "4"" of type '" "char *""'");
    }
    arg4 = (char *)(buf4);
    result = (WsXmlDocH)_WsManClient_renew(arg1,arg2,arg3,arg4);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
    XSRETURN(argvi);
    fail:;
    
    
    if (alloc3 == SWIG_NEWOBJ) free((char*)buf3);
    if (alloc4 == SWIG_NEWOBJ) free((char*)buf4);
  }
  SWIG_croak_null();
}


XS(_wrap_Client_fault_string) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_fault_string(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_fault_string" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (char *)_WsManClient_fault_string(arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_Client_last_error) {
  {
    struct _WsManClient *arg1 = 0 ;
    void *argp1 = 0 ;
    int res1 = 0 ;
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: Client_last_error(self);");
    }
    res1 = SWIG_ConvertPtr(ST(0), &argp1,SWIGTYPE_p__WsManClient, 0 |  0 );
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Client_last_error" "', argument " "1"" of type '" "struct _WsManClient *""'"); 
    }
    arg1 = (struct _WsManClient *)(argp1);
    result = (int)_WsManClient_last_error(arg1);
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_set_debug) {
  {
    int arg1 ;
    int val1 ;
    int ecode1 = 0 ;
    int argvi = 0;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: set_debug(dbg);");
    }
    ecode1 = SWIG_AsVal_int SWIG_PERL_CALL_ARGS_2(ST(0), &val1);
    if (!SWIG_IsOK(ecode1)) {
      SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "set_debug" "', argument " "1"" of type '" "int""'");
    } 
    arg1 = (int)(val1);
    set_debug(arg1);
    ST(argvi) = &PL_sv_undef;
    
    XSRETURN(argvi);
    fail:;
    
  }
  SWIG_croak_null();
}


XS(_wrap_get_debug) {
  {
    int argvi = 0;
    int result;
    dXSARGS;
    
    if ((items < 0) || (items > 0)) {
      SWIG_croak("Usage: get_debug();");
    }
    result = (int)get_debug();
    ST(argvi) = SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(result)); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_create_soap_envelope) {
  {
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 0) || (items > 0)) {
      SWIG_croak("Usage: create_soap_envelope();");
    }
    result = (WsXmlDocH)create_soap_envelope();
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    XSRETURN(argvi);
    fail:;
  }
  SWIG_croak_null();
}


XS(_wrap_create_doc_from_file) {
  {
    char *arg1 = 0 ;
    char *arg2 = (char *)"UTF-8" ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: create_doc_from_file(filename,encoding);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "create_doc_from_file" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "create_doc_from_file" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (WsXmlDocH)create_doc_from_file((char const *)arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_create_doc_from_string) {
  {
    char *arg1 = 0 ;
    char *arg2 = (char *)"UTF-8" ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int res2 ;
    char *buf2 = 0 ;
    int alloc2 = 0 ;
    int argvi = 0;
    WsXmlDocH result;
    dXSARGS;
    
    if ((items < 1) || (items > 2)) {
      SWIG_croak("Usage: create_doc_from_string(buf,encoding);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "create_doc_from_string" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    if (items > 1) {
      res2 = SWIG_AsCharPtrAndSize(ST(1), &buf2, NULL, &alloc2);
      if (!SWIG_IsOK(res2)) {
        SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "create_doc_from_string" "', argument " "2"" of type '" "char const *""'");
      }
      arg2 = (char *)(buf2);
    }
    result = (WsXmlDocH)create_doc_from_string((char const *)arg1,(char const *)arg2);
    ST(argvi) = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p__WsXmlDoc, 0 | SWIG_SHADOW); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    if (alloc2 == SWIG_NEWOBJ) free((char*)buf2);
  }
  SWIG_croak_null();
}


XS(_wrap_uri_prefix) {
  {
    char *arg1 = 0 ;
    int res1 ;
    char *buf1 = 0 ;
    int alloc1 = 0 ;
    int argvi = 0;
    char *result = 0 ;
    dXSARGS;
    
    if ((items < 1) || (items > 1)) {
      SWIG_croak("Usage: uri_prefix(classname);");
    }
    res1 = SWIG_AsCharPtrAndSize(ST(0), &buf1, NULL, &alloc1);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "uri_prefix" "', argument " "1"" of type '" "char const *""'");
    }
    arg1 = (char *)(buf1);
    result = (char *)uri_prefix((char const *)arg1);
    ST(argvi) = SWIG_FromCharPtr((const char *)result); argvi++ ;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
    XSRETURN(argvi);
    fail:;
    if (alloc1 == SWIG_NEWOBJ) free((char*)buf1);
  }
  SWIG_croak_null();
}



/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static swig_type_info _swigt__p_AV = {"_p_AV", "AV *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_FILE = {"_p_FILE", "FILE *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WS_LASTERR_Code = {"_p_WS_LASTERR_Code", "WS_LASTERR_Code *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsEnumerateInfo = {"_p_WsEnumerateInfo", "WsEnumerateInfo *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsmanKnownStatusCode = {"_p_WsmanKnownStatusCode", "WsmanKnownStatusCode *|enum WsmanKnownStatusCode *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_WsmanMessage = {"_p_WsmanMessage", "WsmanMessage *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p__WS_CONTEXT = {"_p__WS_CONTEXT", "WsContextH|struct _WS_CONTEXT *|_WS_CONTEXT *", 0, 0, (void*)"openwsman::Context", 0};
static swig_type_info _swigt__p__WsManClient = {"_p__WsManClient", "WsManClient *|struct _WsManClient *|_WsManClient *", 0, 0, (void*)"openwsman::Client", 0};
static swig_type_info _swigt__p__WsManFault = {"_p__WsManFault", "WsManFault *|struct _WsManFault *|_WsManFault *", 0, 0, (void*)"openwsman::Fault", 0};
static swig_type_info _swigt__p__WsManTransport = {"_p__WsManTransport", "WsManTransport *|struct _WsManTransport *|_WsManTransport *", 0, 0, (void*)"openwsman::Transport", 0};
static swig_type_info _swigt__p__WsXmlDoc = {"_p__WsXmlDoc", "WsXmlDocH|struct _WsXmlDoc *|_WsXmlDoc *", 0, 0, (void*)"openwsman::XmlDoc", 0};
static swig_type_info _swigt__p__WsmanStatus = {"_p__WsmanStatus", "WsmanStatus *|struct _WsmanStatus *|_WsmanStatus *", 0, 0, (void*)"openwsman::Status", 0};
static swig_type_info _swigt__p___Soap = {"_p___Soap", "struct __Soap *|__Soap *", 0, 0, (void*)"openwsman::Soap", 0};
static swig_type_info _swigt__p___SoapOp = {"_p___SoapOp", "struct __SoapOp *|__SoapOp *", 0, 0, (void*)"openwsman::SoapOp", 0};
static swig_type_info _swigt__p___WsEnumerateInfo = {"_p___WsEnumerateInfo", "struct __WsEnumerateInfo *|__WsEnumerateInfo *", 0, 0, (void*)"openwsman::EnumerateInfo", 0};
static swig_type_info _swigt__p___WsXmlAttr = {"_p___WsXmlAttr", "WsXmlAttrH|struct __WsXmlAttr *|__WsXmlAttr *", 0, 0, (void*)"openwsman::XmlAttr", 0};
static swig_type_info _swigt__p___WsXmlNode = {"_p___WsXmlNode", "WsXmlNodeH|struct __WsXmlNode *|__WsXmlNode *", 0, 0, (void*)"openwsman::XmlNode", 0};
static swig_type_info _swigt__p___WsXmlNs = {"_p___WsXmlNs", "__WsXmlNs *", 0, 0, (void*)"openwsman::XmlNs", 0};
static swig_type_info _swigt__p___WsmanFaultCodeTable = {"_p___WsmanFaultCodeTable", "WsmanFaultCodeTable *|struct __WsmanFaultCodeTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultCodeType = {"_p___WsmanFaultCodeType", "WsmanFaultCodeType *|enum __WsmanFaultCodeType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultDetailTable = {"_p___WsmanFaultDetailTable", "WsmanFaultDetailTable *|struct __WsmanFaultDetailTable *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p___WsmanFaultDetailType = {"_p___WsmanFaultDetailType", "WsmanFaultDetailType *|enum __WsmanFaultDetailType *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_client_opt_t = {"_p_client_opt_t", "client_opt_t *", 0, 0, (void*)"openwsman::ClientOptions", 0};
static swig_type_info _swigt__p_epr_t = {"_p_epr_t", "epr_t *", 0, 0, (void*)"openwsman::EndPointReference", 0};
static swig_type_info _swigt__p_filter_t = {"_p_filter_t", "filter_t *", 0, 0, (void*)"openwsman::Filter", 0};
static swig_type_info _swigt__p_p_char = {"_p_p_char", "char **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wsman_auth_request_func_t = {"_p_wsman_auth_request_func_t", "wsman_auth_request_func_t *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_wsman_auth_type_t = {"_p_wsman_auth_type_t", "wsman_auth_type_t *", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_AV,
  &_swigt__p_FILE,
  &_swigt__p_WS_LASTERR_Code,
  &_swigt__p_WsEnumerateInfo,
  &_swigt__p_WsmanKnownStatusCode,
  &_swigt__p_WsmanMessage,
  &_swigt__p__WS_CONTEXT,
  &_swigt__p__WsManClient,
  &_swigt__p__WsManFault,
  &_swigt__p__WsManTransport,
  &_swigt__p__WsXmlDoc,
  &_swigt__p__WsmanStatus,
  &_swigt__p___Soap,
  &_swigt__p___SoapOp,
  &_swigt__p___WsEnumerateInfo,
  &_swigt__p___WsXmlAttr,
  &_swigt__p___WsXmlNode,
  &_swigt__p___WsXmlNs,
  &_swigt__p___WsmanFaultCodeTable,
  &_swigt__p___WsmanFaultCodeType,
  &_swigt__p___WsmanFaultDetailTable,
  &_swigt__p___WsmanFaultDetailType,
  &_swigt__p_char,
  &_swigt__p_client_opt_t,
  &_swigt__p_epr_t,
  &_swigt__p_filter_t,
  &_swigt__p_p_char,
  &_swigt__p_wsman_auth_request_func_t,
  &_swigt__p_wsman_auth_type_t,
};

static swig_cast_info _swigc__p_AV[] = {  {&_swigt__p_AV, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_FILE[] = {  {&_swigt__p_FILE, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WS_LASTERR_Code[] = {  {&_swigt__p_WS_LASTERR_Code, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsEnumerateInfo[] = {  {&_swigt__p_WsEnumerateInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsmanKnownStatusCode[] = {  {&_swigt__p_WsmanKnownStatusCode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_WsmanMessage[] = {  {&_swigt__p_WsmanMessage, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WS_CONTEXT[] = {  {&_swigt__p__WS_CONTEXT, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManClient[] = {  {&_swigt__p__WsManClient, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManFault[] = {  {&_swigt__p__WsManFault, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsManTransport[] = {  {&_swigt__p__WsManTransport, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsXmlDoc[] = {  {&_swigt__p__WsXmlDoc, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p__WsmanStatus[] = {  {&_swigt__p__WsmanStatus, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___Soap[] = {  {&_swigt__p___Soap, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___SoapOp[] = {  {&_swigt__p___SoapOp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsEnumerateInfo[] = {  {&_swigt__p___WsEnumerateInfo, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlAttr[] = {  {&_swigt__p___WsXmlAttr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlNode[] = {  {&_swigt__p___WsXmlNode, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsXmlNs[] = {  {&_swigt__p___WsXmlNs, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultCodeTable[] = {  {&_swigt__p___WsmanFaultCodeTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultCodeType[] = {  {&_swigt__p___WsmanFaultCodeType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultDetailTable[] = {  {&_swigt__p___WsmanFaultDetailTable, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p___WsmanFaultDetailType[] = {  {&_swigt__p___WsmanFaultDetailType, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_client_opt_t[] = {  {&_swigt__p_client_opt_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_epr_t[] = {  {&_swigt__p_epr_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_filter_t[] = {  {&_swigt__p_filter_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_char[] = {  {&_swigt__p_p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wsman_auth_request_func_t[] = {  {&_swigt__p_wsman_auth_request_func_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_wsman_auth_type_t[] = {  {&_swigt__p_wsman_auth_type_t, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_AV,
  _swigc__p_FILE,
  _swigc__p_WS_LASTERR_Code,
  _swigc__p_WsEnumerateInfo,
  _swigc__p_WsmanKnownStatusCode,
  _swigc__p_WsmanMessage,
  _swigc__p__WS_CONTEXT,
  _swigc__p__WsManClient,
  _swigc__p__WsManFault,
  _swigc__p__WsManTransport,
  _swigc__p__WsXmlDoc,
  _swigc__p__WsmanStatus,
  _swigc__p___Soap,
  _swigc__p___SoapOp,
  _swigc__p___WsEnumerateInfo,
  _swigc__p___WsXmlAttr,
  _swigc__p___WsXmlNode,
  _swigc__p___WsXmlNs,
  _swigc__p___WsmanFaultCodeTable,
  _swigc__p___WsmanFaultCodeType,
  _swigc__p___WsmanFaultDetailTable,
  _swigc__p___WsmanFaultDetailType,
  _swigc__p_char,
  _swigc__p_client_opt_t,
  _swigc__p_epr_t,
  _swigc__p_filter_t,
  _swigc__p_p_char,
  _swigc__p_wsman_auth_request_func_t,
  _swigc__p_wsman_auth_type_t,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_constant_info swig_constants[] = {
{0,0,0,0,0,0}
};
#ifdef __cplusplus
}
#endif
static swig_variable_info swig_variables[] = {
{0,0,0,0}
};
static swig_command_info swig_commands[] = {
{"openwsmanc::new_XmlDoc", _wrap_new_XmlDoc},
{"openwsmanc::delete_XmlDoc", _wrap_delete_XmlDoc},
{"openwsmanc::XmlDoc_string", _wrap_XmlDoc_string},
{"openwsmanc::XmlDoc_encode", _wrap_XmlDoc_encode},
{"openwsmanc::XmlDoc_dump_file", _wrap_XmlDoc_dump_file},
{"openwsmanc::XmlDoc_root", _wrap_XmlDoc_root},
{"openwsmanc::XmlDoc_envelope", _wrap_XmlDoc_envelope},
{"openwsmanc::XmlDoc_header", _wrap_XmlDoc_header},
{"openwsmanc::XmlDoc_body", _wrap_XmlDoc_body},
{"openwsmanc::XmlDoc_element", _wrap_XmlDoc_element},
{"openwsmanc::XmlDoc_context", _wrap_XmlDoc_context},
{"openwsmanc::XmlDoc_generate_fault", _wrap_XmlDoc_generate_fault},
{"openwsmanc::XmlDoc_is_fault", _wrap_XmlDoc_is_fault},
{"openwsmanc::XmlDoc_fault", _wrap_XmlDoc_fault},
{"openwsmanc::XmlDoc_create_response_envelope", _wrap_XmlDoc_create_response_envelope},
{"openwsmanc::XmlDoc_is_end_of_sequence", _wrap_XmlDoc_is_end_of_sequence},
{"openwsmanc::delete_XmlNode", _wrap_delete_XmlNode},
{"openwsmanc::XmlNode_string", _wrap_XmlNode_string},
{"openwsmanc::XmlNode_dump_file", _wrap_XmlNode_dump_file},
{"openwsmanc::XmlNode___eq__", _wrap_XmlNode___eq__},
{"openwsmanc::XmlNode_text", _wrap_XmlNode_text},
{"openwsmanc::XmlNode_set_text", _wrap_XmlNode_set_text},
{"openwsmanc::XmlNode_doc", _wrap_XmlNode_doc},
{"openwsmanc::XmlNode_parent", _wrap_XmlNode_parent},
{"openwsmanc::XmlNode_child", _wrap_XmlNode_child},
{"openwsmanc::XmlNode_name", _wrap_XmlNode_name},
{"openwsmanc::XmlNode_set_name", _wrap_XmlNode_set_name},
{"openwsmanc::XmlNode_ns", _wrap_XmlNode_ns},
{"openwsmanc::XmlNode_set_ns", _wrap_XmlNode_set_ns},
{"openwsmanc::XmlNode_prefix", _wrap_XmlNode_prefix},
{"openwsmanc::XmlNode_set_lang", _wrap_XmlNode_set_lang},
{"openwsmanc::XmlNode_find", _wrap_XmlNode_find},
{"openwsmanc::XmlNode_next", _wrap_XmlNode_next},
{"openwsmanc::XmlNode_size", _wrap_XmlNode_size},
{"openwsmanc::XmlNode_add_before", _wrap_XmlNode_add_before},
{"openwsmanc::XmlNode_add", _wrap_XmlNode_add},
{"openwsmanc::XmlNode_get", _wrap_XmlNode_get},
{"openwsmanc::XmlNode_attr", _wrap_XmlNode_attr},
{"openwsmanc::XmlNode_attr_count", _wrap_XmlNode_attr_count},
{"openwsmanc::XmlNode_attr_find", _wrap_XmlNode_attr_find},
{"openwsmanc::XmlNode_attr_add", _wrap_XmlNode_attr_add},
{"openwsmanc::XmlNode_epr", _wrap_XmlNode_epr},
{"openwsmanc::XmlAttr_name", _wrap_XmlAttr_name},
{"openwsmanc::XmlAttr_ns", _wrap_XmlAttr_ns},
{"openwsmanc::XmlAttr_value", _wrap_XmlAttr_value},
{"openwsmanc::XmlAttr_remove", _wrap_XmlAttr_remove},
{"openwsmanc::new_EndPointReference", _wrap_new_EndPointReference},
{"openwsmanc::delete_EndPointReference", _wrap_delete_EndPointReference},
{"openwsmanc::EndPointReference_add_selector", _wrap_EndPointReference_add_selector},
{"openwsmanc::EndPointReference_serialize", _wrap_EndPointReference_serialize},
{"openwsmanc::EndPointReference_cmp", _wrap_EndPointReference_cmp},
{"openwsmanc::EndPointReference_to_xml", _wrap_EndPointReference_to_xml},
{"openwsmanc::EndPointReference_string", _wrap_EndPointReference_string},
{"openwsmanc::EndPointReference_selector_count", _wrap_EndPointReference_selector_count},
{"openwsmanc::EndPointReference_resource_uri", _wrap_EndPointReference_resource_uri},
{"openwsmanc::EndPointReference_selector", _wrap_EndPointReference_selector},
{"openwsmanc::EndPointReference_selector_names", _wrap_EndPointReference_selector_names},
{"openwsmanc::EndPointReference_classname", _wrap_EndPointReference_classname},
{"openwsmanc::EndPointReference_namespace", _wrap_EndPointReference_namespace},
{"openwsmanc::EndPointReference_prefix", _wrap_EndPointReference_prefix},
{"openwsmanc::Filter_resultClass_set", _wrap_Filter_resultClass_set},
{"openwsmanc::Filter_resultClass_get", _wrap_Filter_resultClass_get},
{"openwsmanc::Filter_assocClass_set", _wrap_Filter_assocClass_set},
{"openwsmanc::Filter_assocClass_get", _wrap_Filter_assocClass_get},
{"openwsmanc::new_Filter", _wrap_new_Filter},
{"openwsmanc::delete_Filter", _wrap_delete_Filter},
{"openwsmanc::Filter_associators", _wrap_Filter_associators},
{"openwsmanc::Filter_references", _wrap_Filter_references},
{"openwsmanc::Filter_simple", _wrap_Filter_simple},
{"openwsmanc::Filter_xpath", _wrap_Filter_xpath},
{"openwsmanc::Filter_cql", _wrap_Filter_cql},
{"openwsmanc::Filter_wql", _wrap_Filter_wql},
{"openwsmanc::delete_EnumerateInfo", _wrap_delete_EnumerateInfo},
{"openwsmanc::EnumerateInfo_max_items", _wrap_EnumerateInfo_max_items},
{"openwsmanc::EnumerateInfo_set_max_items", _wrap_EnumerateInfo_set_max_items},
{"openwsmanc::EnumerateInfo_flags", _wrap_EnumerateInfo_flags},
{"openwsmanc::EnumerateInfo_epr_to", _wrap_EnumerateInfo_epr_to},
{"openwsmanc::EnumerateInfo_epr_uri", _wrap_EnumerateInfo_epr_uri},
{"openwsmanc::EnumerateInfo_encoding", _wrap_EnumerateInfo_encoding},
{"openwsmanc::EnumerateInfo_filter", _wrap_EnumerateInfo_filter},
{"openwsmanc::EnumerateInfo_index", _wrap_EnumerateInfo_index},
{"openwsmanc::EnumerateInfo_set_index", _wrap_EnumerateInfo_set_index},
{"openwsmanc::EnumerateInfo_total_items", _wrap_EnumerateInfo_total_items},
{"openwsmanc::EnumerateInfo_set_total_items", _wrap_EnumerateInfo_set_total_items},
{"openwsmanc::EnumerateInfo_pull_result", _wrap_EnumerateInfo_pull_result},
{"openwsmanc::EnumerateInfo_set_pull_result", _wrap_EnumerateInfo_set_pull_result},
{"openwsmanc::SoapOp_indoc", _wrap_SoapOp_indoc},
{"openwsmanc::SoapOp_set_indoc", _wrap_SoapOp_set_indoc},
{"openwsmanc::SoapOp_outdoc", _wrap_SoapOp_outdoc},
{"openwsmanc::SoapOp_set_outdoc", _wrap_SoapOp_set_outdoc},
{"openwsmanc::SoapOp_soap", _wrap_SoapOp_soap},
{"openwsmanc::SoapOp_msg", _wrap_SoapOp_msg},
{"openwsmanc::SoapOp_maxsize", _wrap_SoapOp_maxsize},
{"openwsmanc::delete_Soap", _wrap_delete_Soap},
{"openwsmanc::Soap_create_context", _wrap_Soap_create_context},
{"openwsmanc::Soap_context", _wrap_Soap_context},
{"openwsmanc::Soap_create_ep_context", _wrap_Soap_create_ep_context},
{"openwsmanc::delete_Context", _wrap_delete_Context},
{"openwsmanc::Context_indoc", _wrap_Context_indoc},
{"openwsmanc::Context_runtime", _wrap_Context_runtime},
{"openwsmanc::Context_set_enumIdleTimeout", _wrap_Context_set_enumIdleTimeout},
{"openwsmanc::Context_classname", _wrap_Context_classname},
{"openwsmanc::Context_method", _wrap_Context_method},
{"openwsmanc::Context_method_args", _wrap_Context_method_args},
{"openwsmanc::Context_max_elements", _wrap_Context_max_elements},
{"openwsmanc::Context_max_envelope_size", _wrap_Context_max_envelope_size},
{"openwsmanc::Context_fragment_string", _wrap_Context_fragment_string},
{"openwsmanc::Context_selector", _wrap_Context_selector},
{"openwsmanc::Context_selectors_from_epr", _wrap_Context_selectors_from_epr},
{"openwsmanc::Context_selectors", _wrap_Context_selectors},
{"openwsmanc::Context_selectors_from_filter", _wrap_Context_selectors_from_filter},
{"openwsmanc::Context_action", _wrap_Context_action},
{"openwsmanc::Context_resource_uri", _wrap_Context_resource_uri},
{"openwsmanc::Context_option_set", _wrap_Context_option_set},
{"openwsmanc::Context_parse_enum_request", _wrap_Context_parse_enum_request},
{"openwsmanc::new_Status", _wrap_new_Status},
{"openwsmanc::delete_Status", _wrap_delete_Status},
{"openwsmanc::Status_to_s", _wrap_Status_to_s},
{"openwsmanc::Status_set_code", _wrap_Status_set_code},
{"openwsmanc::Status_code", _wrap_Status_code},
{"openwsmanc::Status_set_detail", _wrap_Status_set_detail},
{"openwsmanc::Status_detail", _wrap_Status_detail},
{"openwsmanc::Status_set_msg", _wrap_Status_set_msg},
{"openwsmanc::Status_msg", _wrap_Status_msg},
{"openwsmanc::Status_generate_fault", _wrap_Status_generate_fault},
{"openwsmanc::new_Fault", _wrap_new_Fault},
{"openwsmanc::delete_Fault", _wrap_delete_Fault},
{"openwsmanc::Fault_code", _wrap_Fault_code},
{"openwsmanc::Fault_subcode", _wrap_Fault_subcode},
{"openwsmanc::Fault_reason", _wrap_Fault_reason},
{"openwsmanc::Fault_detail", _wrap_Fault_detail},
{"openwsmanc::Transport_is_auth_method", _wrap_Transport_is_auth_method},
{"openwsmanc::Transport_close", _wrap_Transport_close},
{"openwsmanc::Transport_set_agent", _wrap_Transport_set_agent},
{"openwsmanc::Transport_agent", _wrap_Transport_agent},
{"openwsmanc::Transport_get_username", _wrap_Transport_get_username},
{"openwsmanc::Transport_set_username", _wrap_Transport_set_username},
{"openwsmanc::Transport_get_password", _wrap_Transport_get_password},
{"openwsmanc::Transport_set_password", _wrap_Transport_set_password},
{"openwsmanc::Transport_get_proxy_username", _wrap_Transport_get_proxy_username},
{"openwsmanc::Transport_set_proxy_username", _wrap_Transport_set_proxy_username},
{"openwsmanc::Transport_get_proxy_password", _wrap_Transport_get_proxy_password},
{"openwsmanc::Transport_set_proxy_password", _wrap_Transport_set_proxy_password},
{"openwsmanc::Transport_set_auth_method", _wrap_Transport_set_auth_method},
{"openwsmanc::Transport_auth_method", _wrap_Transport_auth_method},
{"openwsmanc::Transport_auth_name", _wrap_Transport_auth_name},
{"openwsmanc::Transport_auth_value", _wrap_Transport_auth_value},
{"openwsmanc::Transport_error_string", _wrap_Transport_error_string},
{"openwsmanc::Transport_set_timeout", _wrap_Transport_set_timeout},
{"openwsmanc::Transport_timeout", _wrap_Transport_timeout},
{"openwsmanc::Transport_set_verify_peer", _wrap_Transport_set_verify_peer},
{"openwsmanc::Transport_verify_peer", _wrap_Transport_verify_peer},
{"openwsmanc::Transport_set_verify_host", _wrap_Transport_set_verify_host},
{"openwsmanc::Transport_verify_host", _wrap_Transport_verify_host},
{"openwsmanc::Transport_set_proxy", _wrap_Transport_set_proxy},
{"openwsmanc::Transport_proxy", _wrap_Transport_proxy},
{"openwsmanc::Transport_set_proxyauth", _wrap_Transport_set_proxyauth},
{"openwsmanc::Transport_proxyauth", _wrap_Transport_proxyauth},
{"openwsmanc::Transport_set_cainfo", _wrap_Transport_set_cainfo},
{"openwsmanc::Transport_cainfo", _wrap_Transport_cainfo},
{"openwsmanc::Transport_set_certhumbprint", _wrap_Transport_set_certhumbprint},
{"openwsmanc::Transport_certhumbprint", _wrap_Transport_certhumbprint},
{"openwsmanc::Transport_set_capath", _wrap_Transport_set_capath},
{"openwsmanc::Transport_capath", _wrap_Transport_capath},
{"openwsmanc::Transport_set_caoid", _wrap_Transport_set_caoid},
{"openwsmanc::Transport_caoid", _wrap_Transport_caoid},
{"openwsmanc::Transport_set_cert", _wrap_Transport_set_cert},
{"openwsmanc::Transport_cert", _wrap_Transport_cert},
{"openwsmanc::Transport_set_key", _wrap_Transport_set_key},
{"openwsmanc::Transport_key", _wrap_Transport_key},
{"openwsmanc::wsman_send_request", _wrap_wsman_send_request},
{"openwsmanc::wsmc_transport_set_auth_request_func", _wrap_wsmc_transport_set_auth_request_func},
{"openwsmanc::wsman_is_auth_method", _wrap_wsman_is_auth_method},
{"openwsmanc::wsmc_transport_init", _wrap_wsmc_transport_init},
{"openwsmanc::wsman_transport_close_transport", _wrap_wsman_transport_close_transport},
{"openwsmanc::wsmc_transport_fini", _wrap_wsmc_transport_fini},
{"openwsmanc::wsman_transport_set_agent", _wrap_wsman_transport_set_agent},
{"openwsmanc::wsman_transport_get_agent", _wrap_wsman_transport_get_agent},
{"openwsmanc::wsman_transport_set_userName", _wrap_wsman_transport_set_userName},
{"openwsmanc::wsman_transport_get_userName", _wrap_wsman_transport_get_userName},
{"openwsmanc::wsman_transport_set_password", _wrap_wsman_transport_set_password},
{"openwsmanc::wsman_transport_get_password", _wrap_wsman_transport_get_password},
{"openwsmanc::wsman_transport_set_proxy_username", _wrap_wsman_transport_set_proxy_username},
{"openwsmanc::wsman_transport_get_proxy_username", _wrap_wsman_transport_get_proxy_username},
{"openwsmanc::wsman_transport_set_proxy_password", _wrap_wsman_transport_set_proxy_password},
{"openwsmanc::wsman_transport_get_proxy_password", _wrap_wsman_transport_get_proxy_password},
{"openwsmanc::wsman_transport_set_auth_method", _wrap_wsman_transport_set_auth_method},
{"openwsmanc::wsman_transport_get_auth_method", _wrap_wsman_transport_get_auth_method},
{"openwsmanc::wsmc_transport_get_auth_name", _wrap_wsmc_transport_get_auth_name},
{"openwsmanc::wsmc_transport_get_auth_value", _wrap_wsmc_transport_get_auth_value},
{"openwsmanc::wsman_transport_get_last_error_string", _wrap_wsman_transport_get_last_error_string},
{"openwsmanc::wsman_transport_set_timeout", _wrap_wsman_transport_set_timeout},
{"openwsmanc::wsman_transport_get_timeout", _wrap_wsman_transport_get_timeout},
{"openwsmanc::wsman_transport_set_verify_peer", _wrap_wsman_transport_set_verify_peer},
{"openwsmanc::wsman_transport_get_verify_peer", _wrap_wsman_transport_get_verify_peer},
{"openwsmanc::wsman_transport_set_verify_host", _wrap_wsman_transport_set_verify_host},
{"openwsmanc::wsman_transport_get_verify_host", _wrap_wsman_transport_get_verify_host},
{"openwsmanc::wsman_transport_set_crlcheck", _wrap_wsman_transport_set_crlcheck},
{"openwsmanc::wsman_transport_get_crlcheck", _wrap_wsman_transport_get_crlcheck},
{"openwsmanc::wsman_transport_set_crlfile", _wrap_wsman_transport_set_crlfile},
{"openwsmanc::wsman_transport_get_crlfile", _wrap_wsman_transport_get_crlfile},
{"openwsmanc::wsman_transport_set_proxy", _wrap_wsman_transport_set_proxy},
{"openwsmanc::wsman_transport_get_proxy", _wrap_wsman_transport_get_proxy},
{"openwsmanc::wsman_transport_set_proxyauth", _wrap_wsman_transport_set_proxyauth},
{"openwsmanc::wsman_transport_get_proxyauth", _wrap_wsman_transport_get_proxyauth},
{"openwsmanc::wsman_transport_set_cainfo", _wrap_wsman_transport_set_cainfo},
{"openwsmanc::wsman_transport_get_cainfo", _wrap_wsman_transport_get_cainfo},
{"openwsmanc::wsman_transport_set_certhumbprint", _wrap_wsman_transport_set_certhumbprint},
{"openwsmanc::wsman_transport_get_certhumbprint", _wrap_wsman_transport_get_certhumbprint},
{"openwsmanc::wsman_transport_set_capath", _wrap_wsman_transport_set_capath},
{"openwsmanc::wsman_transport_get_capath", _wrap_wsman_transport_get_capath},
{"openwsmanc::wsman_transport_set_caoid", _wrap_wsman_transport_set_caoid},
{"openwsmanc::wsman_transport_get_caoid", _wrap_wsman_transport_get_caoid},
{"openwsmanc::wsman_transport_set_cert", _wrap_wsman_transport_set_cert},
{"openwsmanc::wsman_transport_get_cert", _wrap_wsman_transport_get_cert},
{"openwsmanc::wsman_transport_set_key", _wrap_wsman_transport_set_key},
{"openwsmanc::wsman_transport_get_key", _wrap_wsman_transport_get_key},
{"openwsmanc::new_ClientOptions", _wrap_new_ClientOptions},
{"openwsmanc::delete_ClientOptions", _wrap_delete_ClientOptions},
{"openwsmanc::ClientOptions_set_dump_request", _wrap_ClientOptions_set_dump_request},
{"openwsmanc::ClientOptions_clear_dump_request", _wrap_ClientOptions_clear_dump_request},
{"openwsmanc::ClientOptions_set_flags", _wrap_ClientOptions_set_flags},
{"openwsmanc::ClientOptions_get_flags", _wrap_ClientOptions_get_flags},
{"openwsmanc::ClientOptions_clear_flags", _wrap_ClientOptions_clear_flags},
{"openwsmanc::ClientOptions_reset_flags", _wrap_ClientOptions_reset_flags},
{"openwsmanc::ClientOptions_set_max_envelope_size", _wrap_ClientOptions_set_max_envelope_size},
{"openwsmanc::ClientOptions_get_max_envelope_size", _wrap_ClientOptions_get_max_envelope_size},
{"openwsmanc::ClientOptions_set_max_elements", _wrap_ClientOptions_set_max_elements},
{"openwsmanc::ClientOptions_get_max_elements", _wrap_ClientOptions_get_max_elements},
{"openwsmanc::ClientOptions_set_timeout", _wrap_ClientOptions_set_timeout},
{"openwsmanc::ClientOptions_get_timeout", _wrap_ClientOptions_get_timeout},
{"openwsmanc::ClientOptions_set_fragment", _wrap_ClientOptions_set_fragment},
{"openwsmanc::ClientOptions_get_fragment", _wrap_ClientOptions_get_fragment},
{"openwsmanc::ClientOptions_set_cim_namespace", _wrap_ClientOptions_set_cim_namespace},
{"openwsmanc::ClientOptions_get_cim_namespace", _wrap_ClientOptions_get_cim_namespace},
{"openwsmanc::ClientOptions_set_reference", _wrap_ClientOptions_set_reference},
{"openwsmanc::ClientOptions_get_reference", _wrap_ClientOptions_get_reference},
{"openwsmanc::ClientOptions_add_option", _wrap_ClientOptions_add_option},
{"openwsmanc::ClientOptions_add_selector", _wrap_ClientOptions_add_selector},
{"openwsmanc::ClientOptions_add_property_array_item", _wrap_ClientOptions_add_property_array_item},
{"openwsmanc::ClientOptions_add_property", _wrap_ClientOptions_add_property},
{"openwsmanc::ClientOptions_set_delivery_uri", _wrap_ClientOptions_set_delivery_uri},
{"openwsmanc::ClientOptions_delivery_uri", _wrap_ClientOptions_delivery_uri},
{"openwsmanc::ClientOptions_set_sub_expiry", _wrap_ClientOptions_set_sub_expiry},
{"openwsmanc::ClientOptions_sub_expiry", _wrap_ClientOptions_sub_expiry},
{"openwsmanc::ClientOptions_set_heartbeat_interval", _wrap_ClientOptions_set_heartbeat_interval},
{"openwsmanc::ClientOptions_heartbeat_interval", _wrap_ClientOptions_heartbeat_interval},
{"openwsmanc::ClientOptions_set_delivery_mode", _wrap_ClientOptions_set_delivery_mode},
{"openwsmanc::ClientOptions_delivery_mode", _wrap_ClientOptions_delivery_mode},
{"openwsmanc::ClientOptions_set_delivery_security_mode", _wrap_ClientOptions_set_delivery_security_mode},
{"openwsmanc::ClientOptions_delivery_sec_mode", _wrap_ClientOptions_delivery_sec_mode},
{"openwsmanc::ClientOptions_set_locale", _wrap_ClientOptions_set_locale},
{"openwsmanc::ClientOptions_locale", _wrap_ClientOptions_locale},
{"openwsmanc::new_Client", _wrap_new_Client},
{"openwsmanc::delete_Client", _wrap_delete_Client},
{"openwsmanc::Client_set_dumpfile", _wrap_Client_set_dumpfile},
{"openwsmanc::Client_response_code", _wrap_Client_response_code},
{"openwsmanc::Client_scheme", _wrap_Client_scheme},
{"openwsmanc::Client_host", _wrap_Client_host},
{"openwsmanc::Client_port", _wrap_Client_port},
{"openwsmanc::Client_path", _wrap_Client_path},
{"openwsmanc::Client_user", _wrap_Client_user},
{"openwsmanc::Client_password", _wrap_Client_password},
{"openwsmanc::Client_transport", _wrap_Client_transport},
{"openwsmanc::Client_send_request", _wrap_Client_send_request},
{"openwsmanc::Client_build_envelope_from_response", _wrap_Client_build_envelope_from_response},
{"openwsmanc::Client_encoding", _wrap_Client_encoding},
{"openwsmanc::Client_set_encoding", _wrap_Client_set_encoding},
{"openwsmanc::Client_identify", _wrap_Client_identify},
{"openwsmanc::Client_get_from_epr", _wrap_Client_get_from_epr},
{"openwsmanc::Client_delete_from_epr", _wrap_Client_delete_from_epr},
{"openwsmanc::Client_enumerate", _wrap_Client_enumerate},
{"openwsmanc::Client_pull", _wrap_Client_pull},
{"openwsmanc::Client_create", _wrap_Client_create},
{"openwsmanc::Client_put", _wrap_Client_put},
{"openwsmanc::Client_release", _wrap_Client_release},
{"openwsmanc::Client_get", _wrap_Client_get},
{"openwsmanc::Client_delete", _wrap_Client_delete},
{"openwsmanc::Client_invoke", _wrap_Client_invoke},
{"openwsmanc::Client_subscribe", _wrap_Client_subscribe},
{"openwsmanc::Client_unsubscribe", _wrap_Client_unsubscribe},
{"openwsmanc::Client_renew", _wrap_Client_renew},
{"openwsmanc::Client_fault_string", _wrap_Client_fault_string},
{"openwsmanc::Client_last_error", _wrap_Client_last_error},
{"openwsmanc::set_debug", _wrap_set_debug},
{"openwsmanc::get_debug", _wrap_get_debug},
{"openwsmanc::create_soap_envelope", _wrap_create_soap_envelope},
{"openwsmanc::create_doc_from_file", _wrap_create_doc_from_file},
{"openwsmanc::create_doc_from_string", _wrap_create_doc_from_string},
{"openwsmanc::uri_prefix", _wrap_uri_prefix},
{0,0}
};
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is made tough by the requirement that no dynamic
 * memory is used. Also, since swig_type_info structures store pointers to
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization.
 * The idea is that swig generates all the structures that are needed.
 * The runtime then collects these partially filled structures.
 * The SWIG_InitializeModule function takes these initial arrays out of
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned statically to an initial
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast dependencies. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns.
 *
 * We loop through the cast[] array associated with the type and mark casts
 * which have not been defined in previously loaded modules by assigning
 * cast pointer value to cast->next. We also hash cast->type->name string
 * and store the value in the cast->value field. If we encounter swig_cast_info
 * structure that represents a cast to self we move it to the beginning
 * of the cast array. One trick we need to do is making sure the type pointer
 * in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded.
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the list (because the cast->type) pointer will
 *     be correct.
 *
 * Once the cast array has been set up AND it does have new casts that need
 * to be added we sort non-self cast entries to move filtered out entries
 * to the end of the array and to arrange the rest in the increasing order
 * of their type pointer values. We store the index of the last added entry
 * in the cast->value field of the entry[0] (overwriting the name hash). Then
 * we sort <next|value> fields of the remaining entries to arrange hash values
 * in the increasing order. This way cast->next->type->name field matches
 * the cast->value hash.
 *
 * Example:
 * Array of casts for type stored at 0x5000, cast to type stored at 0x3000
 * has already been loaded
 *
 *      After sweep-and-hash:   After sort-by-type:    After sort-by-hash:
 *        ________________       ________________       ________________
 *       |                |     |                |     |                |
 * Entry |  type = 0x5000 |     |  type = 0x5000 |     |  type = 0x5000 |
 *   0   |                |     |                |     |                |
 *       |  next = Entry0 |     |  next = Entry0 |     |  next = Entry0 |
 *       | value = 1212   |     | value = 3      |     | value = 3      |
 *       |                |     |                |     |                |
 *       |================|     |================|     |================|
 *       |                |     |                |     |                |
 * Entry |  type = 0x2000 |     |  type = 0x1000 |     |  type = 0x1000 |
 *   1   |                |     |                |     |                |
 *       |  next = Entry1 |     |  next = Entry1 |     |  next = Entry3 |
 *       | value = 3434   |     | value = 4545   |     | value = 2323   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x3000 |     |  type = 0x2000 |     |  type = 0x2000 |
 *   2   |                |     |                |     |                |
 *       |  next = 0      |     |  next = Entry2 |     |  next = Entry2 |
 *       | value = 0      |     | value = 3434   |     | value = 3434   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x1000 |     |  type = 0x4000 |     |  type = 0x4000 |
 *   3   |                |     |                |     |                |
 *       |  next = Entry3 |     |  next = Entry3 |     |  next = Entry1 |
 *       | value = 4545   |     | value = 2323   |     | value = 4545   |
 *       |________________|     |________________|     |________________|
 *       |                |     |                |     |                |
 * Entry |  type = 0x4000 |     |  type = 0x3000 |     |  type = 0x3000 |
 *   4   |                |     |                |     |                |
 *       |  next = Entry4 |     |  next = 0      |     |  next = 0      |
 *       | value = 2323   |     | value = 0      |     | value = 0      |
 *       |________________|     |________________|     |________________|
 *
 * Once the cast array has been initialized, we use cast[0]->next field to link
 * it into the list of cast arrays for the type.
 *              ____    ____    ____
 * type->cast->|next|->|next|->|next|->0
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *             |----|  |----|  |----|
 *
 * Subsequent cast resolution works as follows:
 *
 * 1. Check whether the type matches the first entry in the current cast array.
 * 2. If not, then do a binary search over the (0:cast->value] interval using
 *    either type address or the hash value of the type name.
 * 3. If not found, then move over to the next cast array (cast[0]->next).
 *
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif

#ifndef SWIG_INIT_CLIENT_DATA_TYPE
#define SWIG_INIT_CLIENT_DATA_TYPE void *
#endif

/*
 * Sort function that puts cast entries with nonzero 'next' at the front
 * of the array while ordering them by addresses of their 'type' structs.
 */
SWIGINTERN int SWIG_CastCmpStruct(const void *pa, const void *pb) {
  swig_cast_info *pca = (swig_cast_info *)pa;
  swig_cast_info *pcb = (swig_cast_info *)pb;
  if (pca->type < pcb->type)
  return (pca->next || pcb->next == 0) ? -1 : 1;
  if (pca->type > pcb->type)
  return (pcb->next || pca->next == 0) ? 1 : -1;
  return 0;
}

/*
 * Shell-sort 'next' and 'value' field pairs to order them by 'value'.
 */
SWIGINTERN void SWIG_CastHashSort(swig_cast_info *cast, int size) {
  const int hmax = size/9;
  int h, i;
  for(h = 1; h <= hmax; h = 3*h+1);
  for(; h > 0; h /= 3)
  {
    for(i = h; i < size; ++i)
    {
      swig_cast_info *p = cast[i].next;
      unsigned int hash = cast[i].value;
      int j = i;
      while(j >= h && hash < cast[j-h].value)
      {
        cast[j].next = cast[j-h].next;
        cast[j].value = cast[j-h].value;
        j -= h;
      }
      cast[j].next = p;
      cast[j].value = hash;
    }
  }
}

SWIGRUNTIME void
SWIG_InitializeModule(SWIG_INIT_CLIENT_DATA_TYPE clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int init;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
    init = 1;
  } else {
    init = 0;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    iter=module_head;
    do {
      if (iter==&swig_module) {
        /* Our module is already in the list, so there's nothing more to do. */
        return;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* otherwise we must add our module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* When multiple interpreters are used, a module could have already been initialized in
       a different interpreter, but not yet have a pointer in this interpreter.
       In this case, we do not want to continue adding types... everything should be
       set up already */
  if (init == 0) return;
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %lu\n", (unsigned long)swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *target_type;
    swig_cast_info *cast, *first;
    int num_mapped = 0;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = first = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      target_type = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        target_type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
        if (target_type) {
          /* Target type already defined in another module */
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: found cast %s\n", target_type->name);
#endif
          if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
            printf("SWIG_InitializeModule: skip old type %s\n", target_type->name);
#endif
            cast->type = target_type;
            target_type = 0;
          } else {
            /* Check if this cast is already in the list */
            swig_cast_info *ocast = SWIG_TypeCheck(target_type->name, type);
#ifdef SWIGRUNTIME_DEBUG
            if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", target_type->name);
#endif
            if (!ocast) target_type = 0;
          }
        }
      }
      
      if (!target_type) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        /* Set inclusion mark for sorting */
        cast->next = cast;
        num_mapped++;
        
        if (type == cast->type) {
#ifdef SWIGRUNTIME_DEBUG
          printf("%s : self cast at pos [%li]\n", type->name, cast - first);
#endif
          if (cast - first) {
            /* Move cast to itself to the first entry in the array */
            
            swig_cast_info tmp = *cast;
            *cast = *first;
            *first = tmp;
          }
          first++;
          
        } else {
          cast->value = SWIG_Hash(cast->type->name, (unsigned int)strlen(cast->type->name));
        }
      }
      cast++;
    }
    
    if (num_mapped) {
      if (cast - first) {
        swig_cast_info *tmp;
        
        /* Sort casts by type address for binary search in SWIG_TypeCheckStruct */
        qsort(first, cast - first, sizeof(swig_cast_info), SWIG_CastCmpStruct);
        
        /* Remap back links for added entries */
        cast = swig_module.cast_initial[i] + num_mapped;
        for (tmp = first; tmp < cast; tmp++) {
          tmp->next = tmp;
        }
      }
      
      /* Set the value field of the first entry to the index of the last added entry */
      cast = swig_module.cast_initial[i];
      cast->value = num_mapped - 1;
      
      num_mapped -= (int)(first - cast);
      if (num_mapped > 1) {
        /* Sort <'next','value'> pairs by 'value' for binary search in SWIG_TypeCheck */
        
        SWIG_CastHashSort(first, num_mapped);
      }
      
      first = type->cast;
      if (first) {
        /* Link the current set into the list of cast arrays */
        cast->next = first->next;
        first->next = cast;
      } else {
        cast->next = 0;
        type->cast = cast;
      }
    }
    
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %lu %s\n", (unsigned long)i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      swig_cast_info *head, *cast;
      head = swig_module.types[i]->cast;
      while (head) {
        for (cast = head; (cast - head) <= head->value; cast++) {
          if (!cast->converter) {
            if (cast->type && !cast->type->clientdata)
            SWIG_TypeClientData(cast->type, swig_module.types[i]->clientdata);
          }
        }
        head = head->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#if defined(__cplusplus) && ! defined(XSPROTO)
extern "C"
#endif

XS(SWIG_init) {
  dXSARGS;
  int i;
  (void)items;
  
  SWIG_InitializeModule(0);
  
  /* Install commands */
  for (i = 0; swig_commands[i].name; i++) {
    /* Casts only needed for Perl < 5.10. */
#ifdef __cplusplus
    newXS(const_cast<char*>(swig_commands[i].name), swig_commands[i].wrapper, const_cast<char*>(__FILE__));
#else
    newXS((char*)swig_commands[i].name, swig_commands[i].wrapper, (char*)__FILE__);
#endif
  }
  
  /* Install variables */
  for (i = 0; swig_variables[i].name; i++) {
    SV *sv;
    sv = get_sv(swig_variables[i].name, TRUE | 0x2 | GV_ADDMULTI);
    if (swig_variables[i].type) {
      SWIG_MakePtr(sv,(void *)1, *swig_variables[i].type,0);
    } else {
      sv_setiv(sv,(IV) 0);
    }
    swig_create_magic(sv, swig_variables[i].name, swig_variables[i].set, swig_variables[i].get); 
  }
  
  /* Install constant */
  for (i = 0; swig_constants[i].type; i++) {
    SV *sv;
    sv = get_sv(swig_constants[i].name, TRUE | 0x2 | GV_ADDMULTI);
    switch(swig_constants[i].type) {
    case SWIG_INT:
      sv_setiv(sv, (IV) swig_constants[i].lvalue);
      break;
    case SWIG_FLOAT:
      sv_setnv(sv, (double) swig_constants[i].dvalue);
      break;
    case SWIG_STRING:
      sv_setpv(sv, (const char *) swig_constants[i].pvalue);
      break;
    case SWIG_POINTER:
      SWIG_MakePtr(sv, swig_constants[i].pvalue, *(swig_constants[i].ptype),0);
      break;
    case SWIG_BINARY:
      SWIG_MakePackedObj(sv, swig_constants[i].pvalue, swig_constants[i].lvalue, *(swig_constants[i].ptype));
      break;
    default:
      break;
    }
    SvREADONLY_on(sv);
  }
  
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OPENWSMAN_MAJOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(2)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OPENWSMAN_MINOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(8)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OPENWSMAN_PATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OPENWSMAN_VERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("2.8.1"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlNs, (void*) "openwsman::XmlNs");
  SWIG_TypeClientData(SWIGTYPE_p__WsXmlDoc, (void*) "openwsman::XmlDoc");
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlNode, (void*) "openwsman::XmlNode");
  SWIG_TypeClientData(SWIGTYPE_p___WsXmlAttr, (void*) "openwsman::XmlAttr");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SOAP_1_1", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/soap/envelope"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SOAP_1_2", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://www.w3.org/2003/05/soap-envelope"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_XML_NAMESPACES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://www.w3.org/XML/1998/namespace"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_ADDRESSING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_DISCOVERY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/10/discovery"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_EVENTING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_ENUMERATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_TRANSFER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_XML_SCHEMA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://www.w3.org/2001/XMLSchema"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SCHEMA_INSTANCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://www.w3.org/2001/XMLSchema-instance"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_POLICY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/policy"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_TRUST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2005/02/trust"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SCHEMA_INSTANCE_PREFIX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("xsi"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_SCHEMA_INSTANCE_NIL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("nil"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_OPENWSMAN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schema.openwsman.org/2006/openwsman"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_CIM_SCHEMA", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/common"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_CIM_CLASS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/cim-schema/2"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_CIM_BINDING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_CIM_INTRINSIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.openwsman.org/wbem/wscim/1/intrinsic"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_WS_MAN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_WSMAN_FAULT_DETAIL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/faultDetail"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_WS_MAN_CAT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2005/06/wsmancat"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_WSMAN_ID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/identity/1/wsmanidentity.xsd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP1_2_CONTENT_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("application/soap+xml;charset=UTF-8"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_CONTENT_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("application/soap+xml"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIMXML_CONTENT_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("application/xml"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_SKIP_DEF_FILTERS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x01)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_ACTION_PREFIX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x02)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_ONE_WAY_OP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x04)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_NO_RESP_OP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x08)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_DONT_KEEP_INDOC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x10)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_CLIENT_RESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x20)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_CUSTOM_DISPATCHER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x40)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_IDENTIFY_DISPATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x80)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMID_IDENTIFY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Identify"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMID_IDENTIFY_RESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("IdentifyResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMID_PROTOCOL_VERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ProtocolVersion"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMID_PRODUCT_VENDOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ProductVendor"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMID_PRODUCT_VERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ProductVersion"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_SCHEMA_NIL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("nil"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_TO_ANONYMOUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_MESSAGE_ID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MessageID"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_ADDRESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Address"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EndpointReference"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_ACTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Action"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_RELATES_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RelatesTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("To"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_REPLY_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ReplyTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_FROM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("From"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_FAULT_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("FaultTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_REFERENCE_PROPERTIES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ReferenceProperties"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_REFERENCE_PARAMETERS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ReferenceParameters"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/addressing/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_ENVELOPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Envelope"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_HEADER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Header"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_BODY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Body"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_CODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Code"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_VALUE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Value"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_SUBCODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Subcode"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_REASON", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Reason"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_TEXT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Text"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_LANG", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("lang"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_DETAIL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Detail"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_FAULT_DETAIL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("FaultDetail"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_MUST_UNDERSTAND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("mustUnderstand"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_VERSION_MISMATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("VersionMismatch"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_UPGRADE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Upgrade"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_SUPPORTED_ENVELOPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SupportedEnvelope"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_GET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Get"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_GETRESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/GetResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_GET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Get"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_GET_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("GetResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_PUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Put"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_PUTRESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/PutResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_PUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Put"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_PUT_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("PutResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_CREATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Create"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_CREATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Create"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_DELETE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_DELETE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Delete"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_ACTION_DELETERESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/DeleteResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRANSFER_DELETE_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("DeleteResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSXF_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/transfer/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_ENUMERATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_ENUMERATERESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/EnumerateResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ENUMERATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Enumerate"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ENUMERATE_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_RELEASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Release"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_RELEASE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Release"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_RELEASE_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ReleaseResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_PULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_PULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Pull"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_PULL_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("PullResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_RENEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/Renew"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_GETSTATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/GetStatus"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "ENUM_ACTION_ENUMEND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/EnumerationEnd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_SUBSCRIBE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Subscribe"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_SUBSCRIBE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Subscribe"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_SUBSCRIBE_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SubscribeResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_PULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Pull"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_GETSTATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/GetStaus"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_UNSUBSCRIBE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Unsubscribe"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_UNSUBSCRIBE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Unsubscribe"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_UNSUBSCRIBE_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("UnsubscribeResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_SUBEND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/SubscriptionEnd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "EVT_ACTION_RENEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/Renew"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_RENEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Renew"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_RENEW_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RenewResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_EVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Events"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_HEARTBEAT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Heartbeat"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_DROPPEDEVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/DroppedEvents"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_ACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Ack"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_EVENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/wbmem/wsman/1/wsman/Event"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/09/enumeration/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_RENEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Renew"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_RENEW_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RenewResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_GET_STATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("GetStatus"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_GET_STATUS_RESP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("GetStatusResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_END_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EndTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_EXPIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Expires"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Filter"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Dialect"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ENUMERATION_CONTEXT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerationContext"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_MAX_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MaxTime"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_MAX_ELEMENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MaxElements"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_MAX_CHARACTERS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MaxCharacters"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ITEMS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Items"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_END_OF_SEQUENCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EndOfSequence"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_ENUMERATION_END", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerationEnd"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_REASON", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Reason"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_CODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Code"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_SOURCE_SHUTTING_DOWN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SourceShuttingDown"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSENUM_SOURCE_CANCELING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SourceCanceling"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Delivery"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_NOTIFY_TO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("NotifyTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_ENDTO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EndTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_EXPIRES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Expires"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Mode"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_SUBSCRIPTION_MANAGER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SubscriptionManager"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_IDENTIFIER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Identifier"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Filter"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Dialect"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY_MODE_PUSH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.xmlsoap.org/ws/2004/08/eventing/DeliveryModes/Push"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY_MODE_PUSHWITHACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/PushWithAck"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY_MODE_EVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Events"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEVENT_DELIVERY_MODE_PULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/Pull"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_ASSOCIATED_INSTANCES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("AssociatedInstances"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_ASSOCIATION_INSTANCES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("AssociationInstances"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_OBJECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Object"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_ASSOCIATION_CLASS_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("AssociationClassName"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_RESULT_CLASS_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ResultClassName"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_ROLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Role"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_RESULT_ROLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ResultRole"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_INCLUDE_RESULT_PROPERTY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("IncludeResultProperty"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_RESOURCE_NS_PREFIX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("p"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_SYSTEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("System"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_LOCALE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Locale"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_RESOURCE_URI", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ResourceURI"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_SELECTOR_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SelectorSet"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_SELECTOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Selector"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Name"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_REQUEST_TOTAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RequestTotalItemsCountEstimate"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_TOTAL_ESTIMATE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("TotalItemsCountEstimate"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_OPTIMIZE_ENUM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("OptimizeEnumeration"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_MAX_ELEMENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MaxElements"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ENUM_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateEPR"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ENUM_OBJ_AND_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateObjectAndEPR"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ENUM_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerationMode"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ITEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Item"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_FRAGMENT_TRANSFER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("FragmentTransfer"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_XML_FRAGMENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("XmlFragment"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_OPTION_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("OptionSet"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_OPTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Option"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_TOTAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Total"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_HEARTBEATS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Heartbeats"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_EVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Events"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ACTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Action"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_EVENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Event"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ACKREQUESTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("AckRequested"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_MAX_ENVELOPE_SIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MaxEnvelopeSize"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_OPERATION_TIMEOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("OperationTimeout"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_FAULT_SUBCODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("FaultSubCode"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Filter"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Dialect"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_CONTENTCODING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ContentEncoding"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_CONNECTIONRETRY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ConnectionRetry"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_SENDBOOKMARKS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SendBookmarks"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_BOOKMARK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Bookmark"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_DROPPEDEVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("DroppedEvents"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Auth"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_PROFILE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Profile"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_CERTIFICATETHUMBPRINT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("CertificateThumbprint"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_DEFAULTBOOKMARK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/bookmark/earliest"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WXF_RESOURCE_CREATED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ResourceCreated"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_POLYMORPHISM_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("PolymorphismMode"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_INCLUDE_SUBCLASS_PROP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("IncludeSubClassProperties"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_EXCLUDE_SUBCLASS_PROP", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ExcludeSubClassProperties"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_NONE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("None"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_DERIVED_REPRESENTATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("DerivedRepresentation"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_ACTION_FAULT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding/fault"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_SHOW_EXTENSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ShowExtensions"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_EXCLUDE_NIL_PROPS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ExcludeNilProperties"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_RESOURCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Resource"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_RESOURCE_URI", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("ResourceUri"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_VERSION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Version"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_NOTES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Notes"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_VENDOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Vendor"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_DISPLAY_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("DisplayName"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_KEYWORDS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Keywords"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_ACCESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Access"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_RELATIONSHIPS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Relationsships"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_COMPLIANCE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Compliance"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_OPERATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Operation"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_SELECTOR_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SelectorSet"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_SELECTOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Selector"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_OPTION_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("OptionSet"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_ACTION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Action"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_SELECTOR_SET_REF", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("SelectorSetRef"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_LOCATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Location"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Name"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMANCAT_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Type"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_SELECTOR_FILTER_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/SelectorFilter"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_WQL_FILTER_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.microsoft.com/wbem/wsman/1/WQL"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_XPATH_FILTER_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://www.w3.org/TR/1999/REC-xpath-19991116"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_CQL_FILTER_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/cql/1/dsp0202.pdf"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_ASSOCIATION_FILTER_DIALECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/cimbinding/associationFilter"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSM_XPATH_EVENTROOT_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/filter/eventRootXPath"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSFW_RESPONSE_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Response"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSFW_INDOC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("indoc"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSFW_ENUM_PREFIX", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("_en."));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_NAMESPACE_SELECTOR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("__cimnamespace"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ALL_AVAILABLE_CLASSES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1/*"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "XML_NS_CIM_ALL_CLASS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wscim/1"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ACTION_ENUMERATE_CLASS_NAMES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateClassNames"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ACTION_ENUMERATE_INSTANCE_NAMES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateInstanceNames"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ACTION_ENUMERATE_CLASSES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("EnumerateClasses"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ACTION_GET_CLASS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("GetClass"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "CIM_ACTION_DELETE_CLASS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("DeleteClass"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_ISSUEDTOKENS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("IssuedTokens"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_REQUESTSECURITYTOKENRESPONSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RequestSecurityTokenResponse"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_TOKENTYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("TokenType"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_REQUESTEDSECURITYTOKEN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("RequestedSecurityToken"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSSE_USERNAMETOKEN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("UsernameToken"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSSE_USERNAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Username"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSSE_PASSWORD", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Password"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSP_APPLIESTO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("AppliesTo"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_USERNAMETOKEN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#UsernameToken"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_CERTIFICATETHUMBPRINT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/token/certificateThumbprint"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTP_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/basic"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTP_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/digest"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/basic"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/digest"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/basic"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/digest"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/spnego-kerberos"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTPS_MUTUAL_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual/spnego-kerberos"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SECURITY_PROFILE_HTTP_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/http/spnego-kerberos"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_NONE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_ENUMERATION_COUNT_ESTIMATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0001)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_ENUMERATION_OPTIMIZATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0002)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_ENUMERATION_ENUM_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0004)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_ENUMERATION_ENUM_OBJ_AND_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0008)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_DUMP_REQUEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0010)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_INCLUDESUBCLASSPROPERTIES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0020)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_EXCLUDESUBCLASSPROPERTIES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0040)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_POLYMORPHISM_NONE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0080)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_MUND_MAX_ESIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0100)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_MUND_LOCALE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0200)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_MUND_OPTIONSET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0400)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_MUND_FRAGMENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x0800)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_CIM_EXTENSIONS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x1000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_CIM_REFERENCES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x2000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_CIM_ASSOCIATORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0x4000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_EVENT_SENDBOOKMARK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0X8000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FLAG_CIM_SCHEMA_OPT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0X10000)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_PUSH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_PUSHWITHACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_EVENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(2)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_PULL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(3)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_NO_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_BASIC_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_DIGEST_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(2)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_PASS_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(3)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_NTLM_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(4)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_GSSNEGOTIATE_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(5)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WS_MAX_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(6)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_AUTO", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTP_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTP_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(2)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(3)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(4)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(5)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_BASIC", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(6)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_DIGEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(7)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(8)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTPS_MUTUAL_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(9)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_SEC_HTTP_SPNEGO_KERBEROS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(10)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_epr_t, (void*) "openwsman::EndPointReference");
  SWIG_TypeClientData(SWIGTYPE_p_filter_t, (void*) "openwsman::Filter");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_SENDER_CODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Sender"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_MUSTUNDERSTAND_CODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("MustUnderstand"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_RECEIVER_CODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Receiver"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_SENDER_CODE_NS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("s:Sender"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_RECEIVER_CODE_NS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("s:Receiver"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FAULT_MUSTUNDERSTAND_CODE_NS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("s:MustUnderstand"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_RC_OK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_RC_OK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ACCESS_DENIED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_ACCESS_DENIED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_ACTION_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_ACTION_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ALREADY_EXISTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_ALREADY_EXISTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_CANNOT_PROCESS_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_CANNOT_PROCESS_FILTER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_CANNOT_PROCESS_FILTER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_CANNOT_PROCESS_FILTER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_CONCURRENCY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_CONCURRENCY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_DELIVERY_MODE_REQUESTED_UNAVAILABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_DELIVERY_MODE_REQUESTED_UNAVAILABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DELIVERY_REFUSED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DELIVERY_REFUSED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_DESTINATION_UNREACHABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_DESTINATION_UNREACHABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_ENCODING_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_ENCODING_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_ENDPOINT_UNAVAILABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_ENDPOINT_UNAVAILABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_EVENT_DELIVER_TO_UNUSABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_EVENT_DELIVER_TO_UNUSABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_EVENT_SOURCE_UNABLE_TO_PROCESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_EVENT_SOURCE_UNABLE_TO_PROCESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_FILTER_DIALECT_REQUESTED_UNAVAILABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_FILTER_DIALECT_REQUESTED_UNAVAILABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_FILTERING_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_FILTERING_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_FILTERING_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_FILTERING_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_FILTERING_REQUESTED_UNAVAILABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_FILTERING_REQUESTED_UNAVAILABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_FRAGMENT_DIALECT_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_FRAGMENT_DIALECT_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_INTERNAL_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_INTERNAL_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_INVALID_BOOKMARK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_INVALID_BOOKMARK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_INVALID_ENUMERATION_CONTEXT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_INVALID_ENUMERATION_CONTEXT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_INVALID_EXPIRATION_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_INVALID_EXPIRATION_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_INVALID_EXPIRATION_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_INVALID_EXPIRATION_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_INVALID_MESSAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_INVALID_MESSAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_INVALID_MESSAGE_INFORMATION_HEADER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_INVALID_MESSAGE_INFORMATION_HEADER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_INVALID_OPTIONS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_INVALID_OPTIONS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_INVALID_PARAMETER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_INVALID_PARAMETER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WXF_INVALID_REPRESENTATION", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WXF_INVALID_REPRESENTATION)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_INVALID_SELECTORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_INVALID_SELECTORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_MESSAGE_INFORMATION_HEADER_REQUIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_MESSAGE_INFORMATION_HEADER_REQUIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_NO_ACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_NO_ACK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_QUOTA_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_QUOTA_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_SCHEMA_VALIDATION_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_SCHEMA_VALIDATION_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_TIMED_OUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_TIMED_OUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_TIMED_OUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_TIMED_OUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_UNABLE_TO_RENEW", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_UNABLE_TO_RENEW)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSE_UNSUPPORTED_EXPIRATION_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSE_UNSUPPORTED_EXPIRATION_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSEN_UNSUPPORTED_EXPIRATION_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSEN_UNSUPPORTED_EXPIRATION_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_UNSUPPORTED_FEATURE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_UNSUPPORTED_FEATURE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_FAULT_VERSION_MISMATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOAP_FAULT_VERSION_MISMATCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_FAULT_MUSTUNDERSTAND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOAP_FAULT_MUSTUNDERSTAND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMB_POLYMORPHISM_MODE_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMB_POLYMORPHISM_MODE_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_UNKNOWN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_UNKNOWN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_OK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_OK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ACK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ACK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ACTION_MISMATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ACTION_MISMATCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ALREADY_EXISTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ALREADY_EXISTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_AMBIGUOUS_SELECTORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_AMBIGUOUS_SELECTORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ASYNCHRONOUS_REQUEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ASYNCHRONOUS_REQUEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ADDRESSING_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ADDRESSING_MODE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_AUTHERIZATION_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_AUTHERIZATION_MODE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_BOOKMARKS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_BOOKMARKS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_CHARECHTER_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_CHARECHTER_SET)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_DELIVERY_RETRIES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_DELIVERY_RETRIES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_DUPLICATE_SELECTORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_DUPLICATE_SELECTORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ENCODING_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ENCODING_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_ENUMERATION_MODE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_ENUMERATION_MODE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_EXPIRATION_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_EXPIRATION_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_EXPIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_EXPIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_FILTERING_REQUIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_FILTERING_REQUIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_FORMAT_MISMATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_FORMAT_MISMATCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_FORMAT_SECURITY_TOKEN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_FORMAT_SECURITY_TOKEN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_FRAGMENT_LEVEL_ACCESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_FRAGMENT_LEVEL_ACCESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_HEARTBEATS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_HEARTBEATS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INSECURE_ADDRESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INSECURE_ADDRESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INSUFFICIENT_SELECTORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INSUFFICIENT_SELECTORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_ADDRESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_ADDRESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_FORMAT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_FORMAT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_FRAGMENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_FRAGMENT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_NAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_NAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_NAMESPACE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_NAMESPACE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_RESOURCEURI", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_RESOURCEURI)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_SELECTOR_ASSIGNMENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_SELECTOR_ASSIGNMENT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_SYSTEM", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_SYSTEM)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_TIMEOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_TIMEOUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_VALUE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_VALUE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_INVALID_VALUES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_INVALID_VALUES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_LOCALE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_LOCALE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MAX_ELEMENTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MAX_ELEMENTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MAX_ENVELOPE_POLICY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MAX_ENVELOPE_POLICY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MAX_ENVELOPE_SIZE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MAX_ENVELOPE_SIZE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MAX_TIME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MAX_TIME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MINIMUM_ENVELOPE_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MINIMUM_ENVELOPE_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_MISSING_VALUES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_MISSING_VALUES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_OPERATION_TIMEOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_OPERATION_TIMEOUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_OPTION_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_OPTION_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_OPTION_SET", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_OPTION_SET)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_READ_ONLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_READ_ONLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_RESOURCE_OFFLINE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_RESOURCE_OFFLINE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_RENAME", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_RENAME)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_SELECTOR_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_SELECTOR_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_SERVICE_ENVELOPE_LIMIT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_SERVICE_ENVELOPE_LIMIT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_TARGET_ALREADY_EXISTS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_TARGET_ALREADY_EXISTS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_TYPE_MISMATCH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_TYPE_MISMATCH)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_UNEXPECTED_SELECTORS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_UNEXPECTED_SELECTORS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_UNREPORTABLE_SUCCESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_UNREPORTABLE_SUCCESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_UNUSABLE_ADDRESS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_UNUSABLE_ADDRESS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_URI_LIMIT_EXCEEDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_URI_LIMIT_EXCEEDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_DETAIL_WHITESPACE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_DETAIL_WHITESPACE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSA_DETAIL_DUPLICATE_MESSAGE_ID", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSA_DETAIL_DUPLICATE_MESSAGE_ID)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "SOAP_DETAIL_HEADER_NOT_UNDERSTOOD", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(SOAP_DETAIL_HEADER_NOT_UNDERSTOOD)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WST_DETAIL_UNSUPPORTED_TOKENTYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WST_DETAIL_UNSUPPORTED_TOKENTYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSP_DETAIL_INVALID_EPR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSP_DETAIL_INVALID_EPR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OWSMAN_DETAIL_ENDPOINT_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(OWSMAN_DETAIL_ENDPOINT_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OWSMAN_NO_DETAILS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(OWSMAN_NO_DETAILS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "OWSMAN_SYSTEM_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(OWSMAN_SYSTEM_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NONE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NONE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CANCELLED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CANCELLED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CANT_RESOLVE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CANT_RESOLVE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CANT_RESOLVE_PROXY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CANT_RESOLVE_PROXY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CANT_CONNECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CANT_CONNECT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CANT_CONNECT_PROXY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CANT_CONNECT_PROXY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_SSL_FAILED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_SSL_FAILED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_IO_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_IO_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_MALFORMED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_MALFORMED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_TRY_AGAIN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_TRY_AGAIN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CONTINUE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CONTINUE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_SWITCHING_PROTOCOLS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_SWITCHING_PROTOCOLS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PROCESSING", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PROCESSING)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_OK", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_OK)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CREATED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CREATED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_ACCEPTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_ACCEPTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NON_AUTHORITATIVE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NON_AUTHORITATIVE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NO_CONTENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NO_CONTENT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_RESET_CONTENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_RESET_CONTENT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PARTIAL_CONTENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PARTIAL_CONTENT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_MULTI_STATUS", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_MULTI_STATUS)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_MULTIPLE_CHOICES", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_MULTIPLE_CHOICES)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_MOVED_PERMANENTLY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_MOVED_PERMANENTLY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_FOUND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_FOUND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_MOVED_TEMPORARILY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_MOVED_TEMPORARILY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_SEE_OTHER", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_SEE_OTHER)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_MODIFIED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_MODIFIED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_USE_PROXY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_USE_PROXY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_APPEARING_IN_THIS_PROTOCOL)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_TEMPORARY_REDIRECT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_TEMPORARY_REDIRECT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_BAD_REQUEST", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_BAD_REQUEST)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_UNAUTHORIZED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_UNAUTHORIZED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PAYMENT_REQUIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PAYMENT_REQUIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_FORBIDDEN", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_FORBIDDEN)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_FOUND", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_FOUND)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_METHOD_NOT_ALLOWED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_METHOD_NOT_ALLOWED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_ACCEPTABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_ACCEPTABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PROXY_AUTHENTICATION_REQUIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PROXY_AUTHENTICATION_REQUIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PROXY_UNAUTHORIZED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PROXY_UNAUTHORIZED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_REQUEST_TIMEOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_REQUEST_TIMEOUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_CONFLICT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_CONFLICT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_GONE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_GONE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_LENGTH_REQUIRED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_LENGTH_REQUIRED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_PRECONDITION_FAILED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_PRECONDITION_FAILED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_REQUEST_ENTITY_TOO_LARGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_REQUEST_ENTITY_TOO_LARGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_REQUEST_URI_TOO_LONG", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_REQUEST_URI_TOO_LONG)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_UNSUPPORTED_MEDIA_TYPE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_UNSUPPORTED_MEDIA_TYPE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_REQUESTED_RANGE_NOT_SATISFIABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_INVALID_RANGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_INVALID_RANGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_EXPECTATION_FAILED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_EXPECTATION_FAILED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_UNPROCESSABLE_ENTITY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_UNPROCESSABLE_ENTITY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_LOCKED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_LOCKED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_FAILED_DEPENDENCY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_FAILED_DEPENDENCY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_INTERNAL_SERVER_ERROR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_INTERNAL_SERVER_ERROR)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_IMPLEMENTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_IMPLEMENTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_BAD_GATEWAY", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_BAD_GATEWAY)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_SERVICE_UNAVAILABLE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_SERVICE_UNAVAILABLE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_GATEWAY_TIMEOUT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_GATEWAY_TIMEOUT)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_HTTP_VERSION_NOT_SUPPORTED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_HTTP_VERSION_NOT_SUPPORTED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_INSUFFICIENT_STORAGE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_INSUFFICIENT_STORAGE)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "WSMAN_STATUS_NOT_EXTENDED", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(WSMAN_STATUS_NOT_EXTENDED)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p___WsEnumerateInfo, (void*) "openwsman::EnumerateInfo");
  SWIG_TypeClientData(SWIGTYPE_p___SoapOp, (void*) "openwsman::SoapOp");
  SWIG_TypeClientData(SWIGTYPE_p___Soap, (void*) "openwsman::Soap");
  SWIG_TypeClientData(SWIGTYPE_p__WS_CONTEXT, (void*) "openwsman::Context");
  SWIG_TypeClientData(SWIGTYPE_p__WsmanStatus, (void*) "openwsman::Status");
  SWIG_TypeClientData(SWIGTYPE_p__WsManFault, (void*) "openwsman::Fault");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NO_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "BASIC_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DIGEST_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(2)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PASS_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(3)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NTLM_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(4)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "GSSNEGOTIATE_AUTH", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(5)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NO_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("No Auth"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "BASIC_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Basic"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DIGEST_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Digest"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "PASS_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Passport"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "NTLM_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("NTLM"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "GSSNEGOTIATE_AUTH_STR", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("GSS-Negotiate"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p__WsManTransport, (void*) "openwsman::Transport");
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "TRUE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(1)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "FALSE", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_From_int  SWIG_PERL_CALL_ARGS_1((int)(0)));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  /*@SWIG:/usr/share/swig/4.4.1/perl5/perltypemaps.swg,67,%set_constant@*/ do {
    SV *sv = get_sv((char*) SWIG_prefix "DEFAULT_USER_AGENT", TRUE | 0x2 | GV_ADDMULTI);
    sv_setsv(sv, SWIG_FromCharPtr("Openwsman"));
    SvREADONLY_on(sv);
  } while(0) /*@SWIG@*/;
  SWIG_TypeClientData(SWIGTYPE_p_client_opt_t, (void*) "openwsman::ClientOptions");
  SWIG_TypeClientData(SWIGTYPE_p__WsManClient, (void*) "openwsman::Client");
  ST(0) = &PL_sv_yes;
  XSRETURN(1);
}

